/*

  disasmdump
  copyright (C) bytepointer.com
  disassembler routines copyright (c) 2007-2013 Oleh Yuschuk
  
  View changelist history is globals.h

  LICENSE NOTE: Excluding the source files (c) Oleh Yuschuk (part of the disassembler
  engine), I wrote the routines for these source files either specifically for this
  edition of the project or pulled them from one of my personal code libraries
   whose source code is not currently public. All routines present however form an Open-
  Source project in their entirety.  Disasmdump is free software: you can redistribute it
  and/or modify it under the terms of the GNU General Public License
  version 3 as published by the Free Software Foundation. This program is distributed
  as is, in the hope that it will be useful, but WITHOUT ANY WARRANTY; Glitch or no
  glitch, there is not even an implied warranty of MERCHANTABILITY or FITNESS for any
  endeavor or particular purpose.

  See the GNU General Public License for more details.
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
  
  ---
  
  If you have a question, feedback or would like to submit a bug, please visit
  bytepointer.com, or send a direct e-mail to admin@.
  
  Thank you for your interest!

*/
#include "cmdline.h"
#include "olly_disasm_engine/disasm_interface.h"


//
// outputOpCodeValues() - simple helper to output the number of hex digits designated by uOpCodeByteSize
//      returns the total number of characters output
//
ulong outputOpCodeValues(const char* pszOpCode, ulong uOpCodeByteSize)
{
    const char* p = pszOpCode;
    ulong uOpCodeLen = uOpCodeByteSize*2; //it takes two hex characters to make one byte
    ulong uOpCodeCurCharIdx = 0;
    ulong uOutChars = 0;

    //loop through hex codes outputting each opcode , and ignoring spaces
    while(*p && uOpCodeCurCharIdx<uOpCodeLen)
    {
        //output the character
        _fputchar(*p);
        
        //if this is part of the numeric hex opcode (and not a space or other symbol, increment our character index)
        if
        (
            (*p>='0' && *p<='9')
            ||
            (*p>='a' && *p<='f')
            ||
            (*p>='A' && *p<='F')
        )
        {
            ++uOpCodeCurCharIdx;
        }
        
        ++uOutChars;
        ++p;

   } //while
   
   return(uOutChars);

} //outputOpCodeValues()


//
// disasm() - this wrapper is solely responsible for interfacing with the Olly Disassembling Engine
//
//  pData           - pointer to buffer of data to disassemble
//  uDataSize       - maximum amount to disassemble before end of buffer (pData may not be pointing at beginning of buffer)
//  uTotalDataSize  - total size of the data buffer allocated (only used for messages)
//  uDisasmToSize   - if zero, disassembly is attempted on all of uDataSize; if the end
//                    of the data results in a partial instruction, the engine simply displays "???"
//                    for the last instruction line; to avoid this situation, a size less than or
//                    equal to uDataSize should be passed as this argument to stop disassembly prior;
//                    this allows us to pass more data than we need and have enough surrounding
//                    data to disassemble a potential partial instruction at the end of the buffer.
//
//                    For example, lets say we want to disassemble a total of 4 bytes, but the 4th
//                    byte begins a 5-byte CALL instruction; disassembling a buffer of 4 bytes only
//                    would result in the last instruction being a "???"; lets say we happen pass in
//                    an extra 6 bytes just to be safe (because in this example, we have them), so we set
//                    uDisasmToSize to 4 and uDataSize at 10 (as our input data). The disassembler will
//                    have enough information to decode the full CALL instruction because the data exists
//                    in the input buffer, but we can exit our disassembler routine on the "next" instruction
//                    boundary.
//
int disasmWrapper(const byte* pData, dword uDataSize, dword uDisasmToSize, size_t uInitAddr, uint uOpCodeColSize, uint uDisasmColSize, uint uIndentSize, bool bVerbose, bool bShowHexValues, bool bUppercase, bool bDefaultSeg, bool bWantColAddr, bool bWantColOpCodes, bool bWantColChars, ColorScheme* pColors, size_t uTotalDataSize, bool bConstrainToEndOfPeSection)
{
    bool bStdOutIsConsole = true;
    CONSOLE_SCREEN_BUFFER_INFO csbi;
    TCHAR szIndent[MAX_INDENT_LEN+1];
    int iRetCode = CODE_SUCCESS;
    HANDLE hConsole = INVALID_HANDLE_VALUE;
    WORD wOrigText = 0;

    //if zero passed for uDisasmToSize, default to buffer size
    if (!uDisasmToSize) uDisasmToSize = uDataSize;

    //init variables needed to change the color
    if (pColors)
    {
        //get handle to current console
        hConsole = GetStdHandle(STD_OUTPUT_HANDLE);
        if (INVALID_HANDLE_VALUE == hConsole)
        {
            _tprintf(TEXT("ERROR: GetStdHandle() failed with code=%u\n"),GetLastError());
            iRetCode = CODE_ERROR_API;
        }
        else
        {
            //is this a handle to a display console?
            bStdOutIsConsole = FILE_TYPE_CHAR == GetFileType(hConsole);
            if (!bStdOutIsConsole)
            {
                //console is being redirected - disable colors
                pColors = NULL;
            }
            else if (!GetConsoleScreenBufferInfo(hConsole,&csbi)) //fails with code 6 on a redirected console
            {
                _tprintf(TEXT("ERROR: GetConsoleScreenBufferInfo() failed with code=%u\n"),GetLastError());
                iRetCode = CODE_ERROR_API;
            }
            else
            {
                //save off original text color
                wOrigText = csbi.wAttributes;
                
                //retain original background attributes
                word uOrigBgAttr = wOrigText&0xFFF0;

                //apply original background to color scheme
                pColors->wDefault       |= uOrigBgAttr;
                pColors->wAddrCol       |= uOrigBgAttr;
                pColors->wOpCodeCol     |= uOrigBgAttr;
                pColors->wOpCodePartial |= uOrigBgAttr;
                pColors->wDisDefault    |= uOrigBgAttr;
                pColors->wDisPlain      |= uOrigBgAttr;
                pColors->wDisOpImm      |= uOrigBgAttr;
                pColors->wDisOpMemStack |= uOrigBgAttr;
                pColors->wDisOpMem      |= uOrigBgAttr;
                pColors->wDisOpReg      |= uOrigBgAttr;
                pColors->wDisOpRegFpu   |= uOrigBgAttr;
                pColors->wDisOpRegSys   |= uOrigBgAttr;
                pColors->wDisInsCall    |= uOrigBgAttr;
                pColors->wDisInsRet     |= uOrigBgAttr;
                pColors->wDisInsBranchC |= uOrigBgAttr;
                pColors->wDisInsBranchU |= uOrigBgAttr;
                pColors->wDisInsStack   |= uOrigBgAttr;
                pColors->wDisInsFpu     |= uOrigBgAttr;
                pColors->wDisInsSys     |= uOrigBgAttr;
                pColors->wDisBrackets   |= uOrigBgAttr;
                pColors->wDisOp         |= uOrigBgAttr;

            } //GetConsoleScreenBufferInfo() successfully retrieved current color attributes

        } //GetStdHandle() succeeded

    } //we have colors
    
    //validate args:
    //      uDisasmToSize must be less than or equal to uDataSize - both must be nonzero
    //      passed indent size cannot exceed buffer we allocated for it
    if (CODE_SUCCEEDED(iRetCode))
    {
        if ( !(pData && uDisasmToSize && uDataSize && uDisasmToSize<=uDataSize && uIndentSize <= MAXLEN(szIndent)-1) )
        {
            _tprintf(TEXT("ERROR: invalid arguments passed to disasm() wrapper\n"));
            iRetCode = CODE_ERROR_INTERNAL;
        }
        else
        {
            //init szIndent string with spaces, followed by null termination
            TCHAR* p = szIndent;
            TCHAR* pEnd = szIndent+uIndentSize;
            while (p<pEnd) *p++ = TEXT(' ');
            *p = 0;
            
            //init disassembler
            Preparedisasm();
        
            //init disassembler structures and configuration
            t_disasm da;
            t_config daCfg;
            ZEROMEM(daCfg);
            
            if (!bShowHexValues)
            {
                daCfg.memmode           = NUM_STD|NUM_DECIMAL; // Constant part of address, NUM_xxx
                daCfg.jmpmode           = NUM_STD|NUM_LONG;    // Jump/call destination, NUM_xxx
                daCfg.binconstmode      = NUM_STD|NUM_LONG;    // Binary constants, NUM_xxx
                daCfg.constmode         = NUM_STD|NUM_DECIMAL; // Numeric constants, NUM_xxx
            }
            else
            {
                daCfg.memmode           = NUM_STD|NUM_X;    // Constant part of address, NUM_xxx
                daCfg.jmpmode           = NUM_STD|NUM_X;    // Jump/call destination, NUM_xxx
                daCfg.binconstmode      = NUM_STD|NUM_X;    // Binary constants, NUM_xxx
                daCfg.constmode         = NUM_STD|NUM_X;    // Numeric constants, NUM_xxx
            }

            daCfg.disasmmode        = DAMODE_MASM;      // Main style, one of DAMODE_xxx
            daCfg.lowercase         = bUppercase?0:1;   // Force lowercase display
            daCfg.tabarguments      = 0;                // Tab between mnemonic and arguments
            daCfg.extraspace        = 1;                // Extra space between arguments
            daCfg.useretform        = 0;                // Use RET instead of RETN
            daCfg.shortstringcmds   = 1;                // Use short form of string commands
            daCfg.putdefseg         = bDefaultSeg?1:0;  // Display default segments in listing
            daCfg.showmemsize       = 1;                // Always show memory size
            daCfg.shownear          = 0;                // Show NEAR modifiers
            daCfg.ssesizemode       = 1;                // How to decode size of SSE operands
            daCfg.jumphintmode      = 0;                // How to decode jump hints
            daCfg.sizesens          = 0;                // How to decode size-sensitive mnemonics
            daCfg.simplifiedst      = 0;                // How to decode top of FPU stack
            daCfg.hiliteoperands    = 1;                // Highlight operands
            
            //disassembler flags - might as well only add the features we are using keep disassembly fast,
            //  although nothing breaks by unconditionally having the disassembler produce everything
            //  whether or not it is used
            int damode = DA_TEXT;
            if (bWantColOpCodes) damode |= DA_DUMP;
            if (pColors) damode |= DA_HILITE;

            //send bytes to disassembler
            ulong uCurIp = (ulong)uInitAddr;
            ulong uCurPos=0;
            ulong uDisasmBytes = 0;
            
            do
            {
                //this call invokes the OllyDbg Disassembling engine at current position in buffer
                ulong uOpCodeSize = Disasm((byte*)(pData+uCurPos),(ulong)uDataSize-uCurPos,uCurIp,&da,damode,&daCfg,Decodeaddress);
                if (0 == uOpCodeSize)
                {
                    _tprintf(TEXT("ERROR: fatal error encountered during disassembly\n"));
                    break;
                }
                else
                {
                    //disassembly line
                    if (pColors) SetConsoleTextAttribute(hConsole,pColors->wDefault);
                    if (uIndentSize) _fputts(szIndent,stdout);
                    
                    //addr column
                    if (pColors) SetConsoleTextAttribute(hConsole,pColors->wAddrCol);
                    if (bWantColAddr) _tprintf(TEXT("%08X "),da.ip);
                    
                    //opcode column
                    if (bWantColOpCodes)
                    {
                        if (pColors) SetConsoleTextAttribute(hConsole,pColors->wOpCodeCol);
                        
                        //uppercase the hex bytes
                        TCHAR* p = da.dump;
                        while(*p)
                        {
                            *p++ = TO_UPPER(*p);
                        }

                        //when not using colors or current instruction length is within amount we are disassembling
                        //  dump the entire instruction
                        ulong uOpcodeLen = (ulong)_tcslen(da.dump);
                        if (!pColors || uCurPos+uOpCodeSize<uDisasmToSize)
                        {
                            _fputts(da.dump,stdout);
                        }
                        else
                        {
                            //output the included portion of the opcode(s) in the normal color, the remaining portion in the partial color
                            //  we need to track uOutChars so we know where we are in string (bytes may be grouped together with surrounding spaces)
                            ulong uOutChars = outputOpCodeValues(da.dump,(ulong)uDisasmToSize-uCurPos);
                            SetConsoleTextAttribute(hConsole,pColors->wOpCodePartial);
                            outputOpCodeValues(da.dump + uOutChars,uCurPos+uOpCodeSize-uDisasmToSize);
                        }

                        //pad the remainder of the opcode column length
                        if (uOpcodeLen<uOpCodeColSize)
                        {
                            for (ulong i=uOpcodeLen; i<uOpCodeColSize; ++i) _fputchar(TEXT(' '));
                        }

                    } //opcode column
                                        
                    if (!pColors)
                    {
                        //no color, just dump disassembly
                        _fputts(da.result,stdout);
                    }
                    else
                    {
                        //use the mask to determine which color to emit per character
                        TCHAR* p = da.result;
                        byte* pMask = da.mask;
                        
                        //initialize last color and set it
                        word wColor = pColors->wDisDefault;
                        word wLastColor = wColor;
                        SetConsoleTextAttribute(hConsole,wLastColor);
                        
                        //loop through each disasm char (and corresponding mask char)
                        while(*p)
                        {
                            //override the colors of certain characters before consulting Olly's hilighting mask
                            if (TEXT('[') == *p || TEXT(']') == *p)
                            {
                                //color brackets specifically
                                wColor = pColors->wDisBrackets;
                            }
                            else if (TEXT(',') == *p || TEXT(':') == *p || TEXT('+') == *p || TEXT('*') == *p)
                            {
                                //other operators and symbols
                                wColor = pColors->wDisOp;
                            }
                            else
                            {
                                //set corresponding color based on mask code character
                                switch (*pMask)
                                {
                                    case DRAW_CONST:    wColor = pColors->wDisOpImm;      break;  // C
                                    case DRAW_STKMEM:   wColor = pColors->wDisOpMemStack; break;  // K
                                    case DRAW_MEM:      wColor = pColors->wDisOpMem;      break;  // M
                                    case DRAW_IREG:     wColor = pColors->wDisOpReg;      break;  // R
                                    case DRAW_FREG:     wColor = pColors->wDisOpRegFpu;   break;  // F
                                    case DRAW_SYSREG:   wColor = pColors->wDisOpRegSys;   break;  // S
                                    case DRAW_CALL:     wColor = pColors->wDisInsCall;    break;  // @
                                    case DRAW_RET:      wColor = pColors->wDisInsRet;     break;  // <
                                    case DRAW_CJMP:     wColor = pColors->wDisInsBranchC; break;  // ?
                                    case DRAW_JUMP:     wColor = pColors->wDisInsBranchU; break;  // >
                                    case DRAW_PUSHPOP:  wColor = pColors->wDisInsStack;   break;  // =
                                    case DRAW_FPU:      wColor = pColors->wDisInsFpu;     break;  // 1
                                    case DRAW_SUSPECT:  wColor = pColors->wDisInsSys;     break;  // !
                                    case DRAW_PLAIN:    wColor = pColors->wDisPlain;      break;  // .
                                    default:
                                        wColor = pColors->wDisDefault;
                                } //switch mask code
                            }
                            
                            //change the color if different from last character
                            if (wColor != wLastColor) SetConsoleTextAttribute(hConsole,wColor);
                            wLastColor = wColor;
                            
                            //output current disassembly character
                            _fputchar(*p);
                            
                            //next disasm char
                            ++p; ++pMask;

                        } //while
                        
                        //reset back to normal text color
                        SetConsoleTextAttribute(hConsole, wOrigText);

                    } //color syntax hilighting enabled
                    

                    /* //UNCOMMENT TO OUTPUT SYNTAX HILITING MASK FOR DEBUGGING * / 
                    if (da.masksize>0)
                    {
                        int i = 0;
                        CONSOLE_SCREEN_BUFFER_INFO csbi;
                        if (GetConsoleScreenBufferInfo(hConsole,&csbi))
                        {
                            //pad current cursor position up to this column
                            for (i = csbi.dwCursorPosition.X; i<80; ++i)
                            {
                                _fputchar(TEXT(' '));
                            }
                        }
                        
                        for (i=0; i<da.masksize; i++)
                        {
                            _fputchar(da.mask[i]);
                        }
                    };
                    /**/

                    //done - restore color
                    if (pColors) SetConsoleTextAttribute(hConsole,wOrigText);

                    //character column, display text representation of opcodes (when possible)
                    if (bWantColChars)
                    {
                        const byte* pCur = pData+uCurPos;
                        const byte* pCurOpCodeEnd = pCur + uOpCodeSize;
                        ulong uDisasmLen = (ulong)_tcslen(da.result);
                        
                        //pad the remainder up to the char column length
                        _fputts(TEXT(" "),stdout); ++uDisasmLen;
                        for (ulong i=uDisasmLen; i<uDisasmColSize; ++i) _fputchar(TEXT(' '));

                        //output character data                        
                        while (pCur < pCurOpCodeEnd)
                        {
                            byte chCur = *pCur++;
                            if (chCur<32 || chCur>126) chCur = '.';
                            _tprintf(TEXT("%c"),chCur);
                        }
                    }

                    //line complete
                    _fputts(TEXT("\n"),stdout);
                    
                    //did this result in a complete disassembly?
                    if (0 != _tcscmp(da.result,TEXT("???")))
                    {
                        uDisasmBytes += uOpCodeSize;
                    }
                    
                    uCurPos += uOpCodeSize;
                    uCurIp += uOpCodeSize;
                }
            }    
            while (uCurPos<uDisasmToSize);
            
            //output a verbose summary line
            if (bVerbose)
            {
                _fputts(TEXT("\n"),stdout);
                
                //how much did we disassemble? show amount disassembled relative to input data size
                if (uDisasmBytes == uTotalDataSize)
                {
                    //we disassembled the whole file
                    _tprintf(TEXT("disassembled%s %u byte%s"),1 != uTotalDataSize?TEXT(" all"):TEXT(""),uTotalDataSize,1 != uTotalDataSize?TEXT("s"):TEXT(""));
                }
                else
                {
                    _tprintf(TEXT("disassembled %u of %u bytes"),uDisasmBytes,uTotalDataSize);
                }

                //was disassembly size auto-determined by PE section size?
                if (bConstrainToEndOfPeSection)
                {
                    _tprintf(TEXT(" (to end of PE section)"));
                }

                //add some info if we displayed partial instructions
                if (uDisasmToSize < uDataSize)
                {
                    _tprintf(TEXT(" / %u opcode bytes %s"),uDisasmToSize,pColors?TEXT("hilighted"):TEXT("requested"));
                }
                                
                _fputts(TEXT("\n"),stdout);
            }
            
            //cleanup
            Finishdisasm();
            
        } //we have a buffer with data

    } //color init didn't fail
    
    return(iRetCode);

} //disasm()

//
// isAddressWithinPeSection() - the calling function calls this function for a given PE section to see if address is within range
//  the return values are as follows:
//
//      <0  - an error occurred
//      0   - not found
//      1   - found
//
ResultCode isAddressWithinPeSection(dword dwAddr, dword* puRetFileOffset, size_t uFileSize, const TCHAR* pszSection,
    dword dwMemStart, dword dwMemSize, dword dwMemDataSize,
    dword dwSectFileStart, dword dwSectFileSize, dword dwSectFileDataSize)
{
    ResultCode iRetCode = CODE_SUCCESS;
    assert(puRetFileOffset);
    dword dwMemEnd = dwMemStart+dwMemSize;
    
    //if address within the mapped memory range - calculate the file offset
    if (dwAddr>=dwMemStart && dwAddr<dwMemEnd)
    {
        //section offset out of range of total section size within file - this can be in sections with memory size
        //  initialized by loader to zeros, but which do not take up space in the file
        //
        //  i.e. ensure address is not out of bounds of section
        dword dwSectionOffset = dwAddr-dwMemStart;
        if (dwSectionOffset>=dwSectFileSize)
        {
            _tprintf(TEXT("ERROR: the pointer 0x%X specifies memory not mapped in file (sect=%s, ptroffset=0x%X, sectfsize=0x%X)\n"),
                dwAddr,
                pszSection,
                dwSectionOffset,
                dwSectFileSize
            );
        }
        else if (!puRetFileOffset)
        {
            iRetCode = CODE_ERROR_INTERNAL;
        }
        else
        {
            //found - convert section address to file offset and ensure its not bigger than file
            *puRetFileOffset = dwSectFileStart+dwSectionOffset;
            if (*puRetFileOffset >= uFileSize)
            {
                _tprintf(TEXT("ERROR: invalid PE data; pointer 0x%X maps outside image (sect=%s, ptroffset=0x%X, file_size=%u\n"),
                    dwAddr,
                    pszSection,
                    dwSectionOffset,
                    uFileSize
                );
            }
            else
            {
                iRetCode = 1;

                /*
                _tprintf(TEXT("address 0x%X maps to \"%s\" section, file offset 0x%X\n"),
                    dwAddr,
                    pszSection,
                    *puRetFileOffset
                );
                */
            }
        }
    } //section within range
    
    return(iRetCode);
    
} //isAddressWithinPeSection()


//
// getFileOffsetFromPePointer() - this routine converts a PE pointer specified by pdwStartOffset into a file offset; the offset result
//  is returned back through pdwStartOffset; if the input address is NULL, the module's code entry point is used
//
//      pdwAddress          [in/out]    input pointer_address; upon function success, the file offset is returned here
//      pdwDisasmStartAddr  [out]       corresponding pointer address for returned file offset; always the same as input *pdwStartOffset
//                                      unless NULL is specified and this will return the address of the module's code entry point
//      pdwMaxSizeLeft      [out]       since address will correspond to a offset into a specific section, this function returns
//                                      the maximum bytes from that offset to the end of the section
//      pszFilename         [in]        input filename that corresponds to pData for output
//      pData               [in]        pe file
//      uDataSize           [in]        size of pe file
//      bVerbose            [in]        debugging verbosity
//
ResultCode getFileOffsetFromPePointer(dword* pdwAddress, dword* pdwDisasmStartAddr, dword* pdwMaxSizeLeft, const TCHAR* pszFilename, const byte* pData, size_t uDataSize, bool bVerbose)
{
    ResultCode iRetCode = CODE_SUCCESS;
    const TCHAR* pszPrefix = TEXT("    ");
    
    //sanity check function args
    if ( !(pdwAddress && pdwDisasmStartAddr && pData && uDataSize) )
    {
        iRetCode = CODE_ERROR_ARGUMENTS;
    }
    else
    {
        IMAGE_NT_HEADERS*       pNtHeaders      = NULL;
        IMAGE_OPTIONAL_HEADER*  pOptionalHeader = NULL; //IMAGE_OPTIONAL_HEADER aka. PE Header
        IMAGE_DOS_HEADER*       pDosHeader      = (IMAGE_DOS_HEADER*)pData;
        bool                    bPe32Plus       = false;
        
        //verify valid DOS header signature    
        if (IMAGE_DOS_SIGNATURE != pDosHeader->e_magic)
        {
            _tprintf(TEXT("ERROR: invalid IMAGE_DOS_SIGNATURE\n"));
            iRetCode = CODE_ERROR_INVALID;
        }
        else
        {
            //verify valid NT header signature
            pNtHeaders = (IMAGE_NT_HEADERS*)((byte*)pData + pDosHeader->e_lfanew);
            if (IMAGE_NT_SIGNATURE != pNtHeaders->Signature)
            {
                _tprintf(TEXT("ERROR: invalid IMAGE_NT_SIGNATURE\n"));
                iRetCode = CODE_ERROR_INVALID;
            }
            else
            {
                //verify that we're dealing with a PE32 or PE32+
                pOptionalHeader = &pNtHeaders->OptionalHeader;
                if (pOptionalHeader->Magic == IMAGE_NT_OPTIONAL_HDR32_MAGIC)
                {
                    bPe32Plus = false;
                }
                else if(pOptionalHeader->Magic == IMAGE_NT_OPTIONAL_HDR64_MAGIC)
                {
                    bPe32Plus = true;
                }
                else
                {
                    _tprintf(TEXT("ERROR: PE32 or PE32+ signature not found (magic=0x%X)\n"),pOptionalHeader->Magic);
                    iRetCode = CODE_ERROR_INVALID;
                } 

                //signature OK?
                if (CODE_SUCCEEDED(iRetCode))
                {
                    if (bVerbose)
                    {
                        _tprintf
                        (
                            TEXT("%s%s signature found\n")
                            TEXT("%smachine cpu type is %s\n")
                            TEXT("%simage checksum is 0x%08X\n")
                            TEXT("%sfile alignment is 0x%X\n")
                            TEXT("%ssection alignment is 0x%X\n")
                            ,
                            pszPrefix,!bPe32Plus?TEXT("PE32"):TEXT("PE32+ (64-bit OS)"),
                            pszPrefix,peGetMachineString(pNtHeaders->FileHeader.Machine),
                            pszPrefix,pOptionalHeader->CheckSum,
                            pszPrefix,pOptionalHeader->FileAlignment,
                            pszPrefix,pOptionalHeader->SectionAlignment
                        );
                    }                

                    if (bPe32Plus)
	                {
		                IMAGE_OPTIONAL_HEADER64* pOptionalHeader64 = (IMAGE_OPTIONAL_HEADER64*)pOptionalHeader; 
		                qword qwLoaderBase = pOptionalHeader64->ImageBase;
		                if (bVerbose) _tprintf(TEXT("load base 0x%08I64X\n"),qwLoaderBase);
		                
		                //PE32+ images unsupported
		                _tprintf(TEXT("ABORT: PE32+ images not currently supported\n"));
		                iRetCode = CODE_ERROR_UNSUPPORTED;
                    }
                    else
                    {
                        //get base pointer address
		                IMAGE_OPTIONAL_HEADER32* pOptionalHeader32 = (IMAGE_OPTIONAL_HEADER32*)pOptionalHeader; 
		                dword dwLoaderBase = pOptionalHeader32->ImageBase;
		                if (bVerbose) _tprintf(TEXT("%sload base 0x%08X\n"),pszPrefix,dwLoaderBase);
                            	            
	                    //output entry point - same DWORD location for both PE and PE+ regardless of the Optional header we use, so just use default
	                    dword dwEntryPoint = dwLoaderBase + pNtHeaders->OptionalHeader.AddressOfEntryPoint;
	                    
	                    if (bVerbose)
	                    {
	                        _tprintf(TEXT("%s%scode entry point address 0x%X\n"),pszPrefix,0 == *pdwAddress?TEXT("looking up "):TEXT(""),dwEntryPoint);
	                    }
	                    
	                    //if zero passes as the start pointer, this means to use the code entry point address
	                    if (0 == *pdwAddress)
	                    {
	                        *pdwAddress = dwEntryPoint;
	                    }
    	                
	                    //at this point we know the the reference address is in *pdwAddress and will be the address we reference
	                    //  to find the file_offset
	                    *pdwDisasmStartAddr = *pdwAddress;


                        //
                        // Check pointer against headers section
                        //

                    #ifdef _MSC_VER
                        #pragma warning (push)
                        #pragma warning (disable:4311) //warning C4311: 'type cast' : pointer truncation from 'IMAGE_SECTION_HEADER *' to 'dword'
                    #endif
                        const TCHAR* pszHeadersSection = TEXT("HEADERS");
                        const TCHAR* pszFoundSection = NULL;

                        //section table immediately follows NT headers
                        IMAGE_SECTION_HEADER*   pCurSection = (IMAGE_SECTION_HEADER*)((byte*)pNtHeaders + sizeof(IMAGE_NT_HEADERS));
                        size_t                  uSectionCount = pNtHeaders->FileHeader.NumberOfSections;

                        if (!uSectionCount)
                        {
                            _tprintf(TEXT("WARNING: no PE sections found\n"));
                        }
                        
                        //current section range is from DOS header and up to and including sections
                        dword uSectionStart = dwLoaderBase;
                        dword uSectionDataSize = (dword)((byte*)pCurSection-pData) + uSectionCount*sizeof(IMAGE_SECTION_HEADER);
                        dword uSectionDataEnd = uSectionStart + uSectionDataSize;
                        
                        //calculate loaded section size by rounding section up to nearest SectionAlignment
                        dword uLoadedSectionEnd = peRoundUpToSectionAlignment(pOptionalHeader->SectionAlignment,uSectionDataEnd);
                        dword uLoadedSectionSize = uLoadedSectionEnd - uSectionStart;
                        
                        dword uFileOffsetStart = 0;
                        dword uFileOffsetSize = pOptionalHeader->SizeOfHeaders;//SizeOfHeaders is rounded to file alignment
                        dword uFileOffsetDataSize = uSectionDataSize;

                    #ifdef _MSC_VER
                        #pragma warning (pop)
                    #endif
                    
                        //is address within range of headers?
                        dword uCalcFileOffset = 0;
                        bool bFound = false;
                        iRetCode = isAddressWithinPeSection
                        (
                            *pdwAddress,&uCalcFileOffset,uDataSize,pszHeadersSection,
                            uSectionStart,uLoadedSectionSize,uSectionDataSize,
                            uFileOffsetStart,uFileOffsetSize,uFileOffsetDataSize
                        );
                        if (iRetCode>=1)
                        {
                            //FOUND in headers section
                            pszFoundSection = pszHeadersSection;
                            bFound = true;
                            iRetCode = CODE_SUCCESS;
                        }
                        else if (uSectionCount)
                        {
                            //
                            // look through each section for address
                            //
                            
                            //output header section info
                            if (bVerbose) _tprintf(TEXT("%s%u PE section%s found\n"),pszPrefix,uSectionCount,1!=uSectionCount?TEXT("s"):TEXT(""));

                            TCHAR szSection[9]; //section names can be up to 8 characters, and account for NULL terminator
                            for (size_t uSecIdx=0; CODE_SUCCEEDED(iRetCode) && !bFound && uSecIdx<uSectionCount; ++uSecIdx, ++pCurSection)
                            {
                                //may be 8 chars w/o null terminator - may use "%.8s" to avoid using a separate buffer
                                szSection[MAXLEN(szSection)-1] = 0;
                            #ifndef WIDE
                                stringCopy(szSection,MAXLEN(szSection),(const char*)pCurSection->Name,MAXLEN(szSection)-1);
                            #else
                                stringAnsiToWide((char*)pCurSection->Name,szSection,MAXLEN(szSection),MAXLEN(szSection)-1);
                            #endif
                                
                                uSectionStart = dwLoaderBase + pCurSection->VirtualAddress;
                                
                                //Here, we need to calculate the size of the section as if loaded into memory. The PE
                                //spec describes VirtualSize as:
                                //
                                //  "The total size of the section when loaded into memory. If this value is greater
                                //   than SizeOfRawData, the section is zero-padded."
                                //
                                //This would seem to be the size as if loaded in memory, but it is actually the raw size
                                //  of the section without alignment padding.  To determine loaded section size, we need
                                //  to manually take the size and round it up to the nearest SectionAlignment.
                                uSectionDataSize = pCurSection->Misc.VirtualSize;

                                //manually calculate loaded section size by rounding to nearest section alignment
                                dword uSectionDataEnd = uSectionStart + uSectionDataSize;
                                dword uLoadedSectionEnd = peRoundUpToSectionAlignment(pOptionalHeader->SectionAlignment,uSectionDataEnd);
                                uLoadedSectionSize = uLoadedSectionEnd - uSectionStart; //in practice, this value is usually identical to pCurSection->Misc.VirtualSize, which we store in uFileOffsetDataSize
                                
                                uFileOffsetStart = pCurSection->PointerToRawData;
                                uFileOffsetSize = pCurSection->SizeOfRawData; //size of initialized data on disk, rounded up to nearest FileAlignment
                                uFileOffsetDataSize = pCurSection->Misc.VirtualSize;
                                
                                iRetCode = isAddressWithinPeSection
                                (
                                    *pdwAddress,&uCalcFileOffset,uDataSize,szSection,
                                    uSectionStart,uLoadedSectionSize,uSectionDataSize,
                                    uFileOffsetStart,uFileOffsetSize,uFileOffsetDataSize
                                );
                                
                                //was it found?
                                if (iRetCode >= 1)
                                {
                                    pszFoundSection = szSection;
                                    bFound = true;
                                    iRetCode = CODE_SUCCESS;
                                }
           
                            } //for each section
                            
                        } //we have PE sections 
                        
                        //if no errors above, we either found address or we didn't
                        if (CODE_SUCCEEDED(iRetCode))
                        {
                            if (!bFound)
                            {
                                iRetCode = CODE_ERROR_NOT_FOUND;
                                _tprintf(TEXT("ERROR: address 0x%x not found\n"),*pdwAddress);
                            }
                            else
                            {
                                if (bVerbose) _tprintf(TEXT("%saddress 0x%X found in \"%s\" section at file offset 0x%X\n"),pszPrefix,*pdwAddress,pszFoundSection,uCalcFileOffset);
                                
                                //return calculated file offset and maximum data bytes left in section
                                *pdwAddress = uCalcFileOffset;
                                
                                //BUGFIX -- uFileOffsetDataSize represents size in memory, not file, so swap with uFileOffsetSize
                                //*pdwMaxSizeLeft = uFileOffsetDataSize - (uCalcFileOffset - uFileOffsetStart);
                                *pdwMaxSizeLeft = uFileOffsetSize - (uCalcFileOffset - uFileOffsetStart);
                                
                            } //file offset found

                        } //no errors occurred
                        
                    } //32-bit PE -- (!bPe32Plus)
                    
                } //optional header (PE header) magic signature OK

            } //valid NT signature

        } //valid DOS signature

    } //function arguments sanity check OK

    return(iRetCode);
    
} //getFileOffsetFromPePointer()

//
// main() - console app entry point
//
int CDECL _tmain(int argc, TCHAR** argv)
{
    //initialize CRT debugging in debug mode for memory leak detection
#if defined(DEBUG) && defined(_CRTDBG_MAP_ALLOC) && defined(WINDOWS)
    _CrtSetDbgFlag(DEFAULT_CRTDBG_FLAGS);
#endif
    
    //parse the command line options
    Options options;
    ResultCode iRetCode = options.processCommandLine(argc,argv);
    if (CODE_SUCCEEDED(iRetCode))
    {
        //do we need to translate a PE pointer address?
        bool bConstrainToEndOfPeSection = false;
        dword uMaxPeSectionSizeLeft = 0;
        if (options.m_bPeAddr)
        {
            if (options.m_bVerbose) _tprintf(TEXT("analyzing \"%s\"...\n"),options.m_pszFilename);
            iRetCode = getFileOffsetFromPePointer(&options.m_uStartOffset,&options.m_uInitAddr,&uMaxPeSectionSizeLeft,options.m_pszFilename,options.m_pDataBuf,options.m_uDataSize,options.m_bVerbose);
            
            //if successful, constrain amount of 
            if (CODE_SUCCEEDED(iRetCode))
            {
                if (options.m_uDisasmToSize && options.m_uDisasmToSize > uMaxPeSectionSizeLeft)
                {
                    _tprintf(TEXT("ERROR: cannot disassemble from 0x%X past end of current section (max_size=0x%X)\n"),options.m_uStartOffset,uMaxPeSectionSizeLeft);
                    iRetCode = CODE_ERROR_RANGE;
                }
                else if (!options.m_uDisasmToSize)
                {
                    options.m_uDisasmToSize = uMaxPeSectionSizeLeft;
                    bConstrainToEndOfPeSection = true;
                }
            }
        }
        
        if (CODE_SUCCEEDED(iRetCode))
        {
            //run the disassembler
            const byte* pStart = options.m_pDataBuf + options.m_uStartOffset;
            dword uRemainSize = options.m_uDataSize - options.m_uStartOffset;
            iRetCode = disasmWrapper
            (
                pStart,
                uRemainSize,
                options.m_uDisasmToSize,
                options.m_uInitAddr,
                options.m_uOpCodeColSize,
                options.m_uDisasmColSize,
                options.m_uIndentSize,
                options.m_bVerbose,
                options.m_bShowHexValues,
                options.m_bUppercase,
                options.m_bDefaultSeg,
                options.m_bWantColAddr,
                options.m_bWantColOpCodes,
                options.m_bWantColChars,
                options.m_bColor?&options.m_cs:NULL,
                options.m_uDataSize,
                bConstrainToEndOfPeSection
            );
        }
        
    } //command line successfully parsed  

    return(resultCodeToExitCode(iRetCode));

} //main()
