/*

[bytepointer.com] a minimal amount of definitions were moved into this file from
                  from the original disasm.h (from the Olly Disassembler 2.01 public
                  source code); just enough to declare and use the required types,
                  constants and functions so they could be called externally without
                  worry about interfering with my own framework declarations.

*/
#ifndef _DISASMFUNCS_H_
#define _DISASMFUNCS_H_

typedef unsigned char  uchar;          // Unsigned character (byte)
typedef unsigned short ushort;         // Unsigned short
typedef unsigned long  ulong;          // Unsigned long

#ifdef UNICODE
  #define tchar        wchar_t         // Or replace it with TCHAR
#else
  #define tchar        char
#endif

#define NREG           8               // Number of registers (of any type)
#define TEXTLEN        256             // Max length of text string
#define NOPERAND       4               // Maximal allowed number of operands

// Disassembling options.
#define DA_TEXT        0x00000001      // Decode command to text and comment
#define   DA_HILITE    0x00000002      // Use syntax highlighting
#define   DA_JZ        0x00000004      // JZ, JNZ instead of JE, JNE
#define   DA_JC        0x00000008      // JC, JNC instead of JAE, JB
#define DA_DUMP        0x00000020      // Dump command to hexadecimal text
#define DA_PSEUDO      0x00000400      // List pseudooperands

#define DAMODE_MASM    0               // MASM assembling/disassembling style
#define DAMODE_IDEAL   1               // IDEAL assembling/disassembling style
#define DAMODE_HLA     2               // HLA assembling/disassembling style
#define DAMODE_ATT     3               // AT&T disassembling style

#define NUM_STYLE     0x0003           // Mask to extract hex style
#define   NUM_STD     0x0000           // 123, 12345678h, 0ABCD1234h
#define   NUM_X       0x0001           // 123, 0x12345678, 0xABCD1234
#define   NUM_OLLY    0x0002           // 123., 12345678, 0ABCD1234
#define NUM_LONG      0x0010           // 00001234h instead of 1234h
#define NUM_DECIMAL   0x0020           // 123 instead of 7Bh if under DECLIMIT

// Command highlighting.
#define DRAW_PLAIN     '.' //0x0000000C      // Plain commands
#define DRAW_JUMP      '>' //0x0000000D      // Unconditional jump commands
#define DRAW_CJMP      '?' //0x0000000E      // Conditional jump commands
#define DRAW_PUSHPOP   '=' //0x0000000F      // PUSH/POP commands
#define DRAW_CALL      '@' //0x00000010      // CALL commands
#define DRAW_RET       '<' //0x00000011      // RET commands
#define DRAW_FPU       '1' //0x00000012      // FPU, MMX, 3DNow! and SSE commands
#define DRAW_SUSPECT   '!' //0x00000013      // Bad, system and privileged commands
// Operand highlighting.
#define DRAW_IREG      'R' //0x00000018      // General purpose registers
#define DRAW_FREG      'F' //0x00000019      // FPU, MMX and SSE registers
#define DRAW_SYSREG    'S' //0x0000001A      // Segment and system registers
#define DRAW_STKMEM    'K' //0x0000001B      // Memory accessed over ESP or EBP
#define DRAW_MEM       'M' //0x0000001C      // Any other memory
#define DRAW_CONST     'C' //0x0000001E      // Constant

// Description of disassembled operand
typedef struct t_operand
{
    // Description of operand.
    ulong          features;             // Operand features, set of OP_xxx
    ulong          arg;                  // Operand type, set of B_xxx
    //  int            optype;               // DEC_INT, DEC_FLOAT or DEC_UNKNOWN
    int            opsize;               // Total size of data, bytes
    int            granularity;          // Size of element (opsize exc. MMX/SSE)
    int            reg;                  // REG_xxx (also ESP in POP) or REG_UNDEF
    ulong          uses;                 // List of used regs (not in address!)
    ulong          modifies;             // List of modified regs (not in addr!)
    // Description of memory address.
    int            seg;                  // Selector (SEG_xxx)
    uchar          scale[NREG];          // Scales of registers in memory address
    ulong          aregs;                // List of registers used in address
    ulong          opconst;              // Constant or const part of address
    ulong          selector;             // Immediate selector in far jump/call
    // Textual decoding.
    tchar          text[TEXTLEN];        // Operand, decoded to text
} t_operand;

// Note that used registers are those which contents is necessary to create
// result. Modified registers are those which value is changed. For example,
// command MOV EAX,[EBX+ECX] uses EBX and ECX and modifies EAX. Command
// ADD ESI,EDI uses ESI and EDI and modifies ESI.

//Disassembled command
typedef struct t_disasm 
{             
    ulong          ip;                   // Address of first command byte
    ulong          size;                 // Full length of command, bytes
    ulong          cmdtype;              // Type of command, D_xxx
    ulong          exttype;              // More features, set of DX_xxx
    ulong          prefixes;             // List of prefixes, set of PF_xxx
    ulong          nprefix;              // Number of prefixes, including SSE2
    ulong          memfixup;             // Offset of first 4-byte fixup or -1
    ulong          immfixup;             // Offset of second 4-byte fixup or -1
    int            errors;               // Set of DAE_xxx
    int            warnings;             // Set of DAW_xxx
    ulong          uses;                 // List of used registers
    ulong          modifies;             // List of modified registers
    ulong          memconst;             // Constant in memory address or 0
    ulong          stackinc;             // Data size in ENTER/RETN/RETF
    t_operand      op[NOPERAND];         // Operands
    tchar          dump[TEXTLEN];        // Hex dump of the command
    tchar          result[TEXTLEN];      // Fully decoded command as text
    uchar          mask[TEXTLEN];        // Mask to highlight result
    int            masksize;             // Length of mask corresponding to result
} t_disasm;

//Disassembler configuration
typedef struct t_config
{
    int            disasmmode;           // Main style, one of DAMODE_xxx
    int            memmode;              // Constant part of address, NUM_xxx
    int            jmpmode;              // Jump/call destination, NUM_xxx
    int            binconstmode;         // Binary constants, NUM_xxx
    int            constmode;            // Numeric constants, NUM_xxx
    int            lowercase;            // Force lowercase display
    int            tabarguments;         // Tab between mnemonic and arguments
    int            extraspace;           // Extra space between arguments
    int            useretform;           // Use RET instead of RETN
    int            shortstringcmds;      // Use short form of string commands
    int            putdefseg;            // Display default segments in listing
    int            showmemsize;          // Always show memory size
    int            shownear;             // Show NEAR modifiers
    int            ssesizemode;          // How to decode size of SSE operands
    int            jumphintmode;         // How to decode jump hints
    int            sizesens;             // How to decode size-sensitive mnemonics
    int            simplifiedst;         // How to decode top of FPU stack
    int            hiliteoperands;       // Highlight operands
} t_config;

int         Preparedisasm       (void);
void        Finishdisasm        (void);
ulong       Disasm              (uchar* cmd, ulong cmdsize, ulong cmdip, t_disasm* da, int cmdmode, t_config* cmdconfig, int (*decodeaddress)(tchar* s, ulong addr));
tchar*      Geterrwarnmessage   (ulong errors, ulong warnings);
int         Decodeaddress       (tchar *s, ulong addr);

#endif //_DISASMFUNCS_H_