/*

  disasmdump
  copyright (C) bytepointer.com
  disassembler routines copyright (c) 2007-2013 Oleh Yuschuk
  
  View changelist history is globals.h

  LICENSE NOTE: Excluding the source files (c) Oleh Yuschuk (part of the disassembler
  engine), I wrote the routines for these source files either specifically for this
  edition of the project or pulled them from one of my personal code libraries
   whose source code is not currently public. All routines present however form an Open-
  Source project in their entirety.  Disasmdump is free software: you can redistribute it
  and/or modify it under the terms of the GNU General Public License
  version 3 as published by the Free Software Foundation. This program is distributed
  as is, in the hope that it will be useful, but WITHOUT ANY WARRANTY; Glitch or no
  glitch, there is not even an implied warranty of MERCHANTABILITY or FITNESS for any
  endeavor or particular purpose.

  See the GNU General Public License for more details.
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
  
  ---
  
  If you have a question, feedback or would like to submit a bug, please visit
  bytepointer.com, or send a direct e-mail to admin@.
  
  Thank you for your interest!

*/
#include "stdinc.h"

//
// resultCodeToExitCode() - converts a result code to a compliant ERRORLEVEL exit code
//      Basically, any success code is converted to 0, all failure codes are converted
//      to a positive failure code
//
int resultCodeToExitCode(int code)
{
    return(CODE_SUCCEEDED(code)?0:-code);
}

//
// getFilenamePtr() - returns the pointer in pszFilePath to the filename
//
const TCHAR* getFilenamePtr(const TCHAR* pszFilePath)
{
    const TCHAR* pszFilename = NULL;

    if (pszFilePath && *pszFilePath)
    {
        const TCHAR* p = pszFilePath;

        //go to end of String
        while (*p) p++;

        //go backwards looking for first path slash or ':'
        for (p--; (p!=pszFilePath) && (*p!=PATHSLASH) && (*p!=':'); p--) ;

        //if we hit slash or colon, we obviously don't want that part of filename
        //  so advance just past it
        pszFilename = p + ((*p==PATHSLASH) || (*p==':'));
    }

    return(pszFilename);

} //getFilenamePtr()

//
// stringToNum32_internal() - internal function that performs the meat of converting a string to a 32 bit value.
//
// ARGUMENTS:
//      pszNum                      - string data of contiguous memory to be converted
//      puRet                       - result of conversion
//      pbRetNegative               - optional pointer to let caller know if uRetNum has a negative sign
//      bForceDefaultBase           - if false, 0x and 0b and 0 (hex, binary and octal notation) will override uDefaultBase,
//                                    otherwise the base designated by uDefaultBase is the basis for conversion.
//      uDefaultBase                - specifies default base of converted number; must be between 2 and 36
//                                    using character ranges 0-9,A-Z or a-z
//      uDataLen                    - if nonzero, only process n characters of passed string, otherwise position of NULL is used
//                                    to determine end of digits.
//
// RETURN VALUES:
//      CODE_SUCCESS                - string processed without errors
//      CODE_ERROR_ARGUMENTS        - bad arguments passed
//      CODE_ERROR_LIMIT_REACHED    - edge case where last addressable memory location falls within the data which is unsupported
//      CODE_ERROR_SYNTAX           - string is not a valid number recognized by function or negative disallowed by unsigned type
//      CODE_ERROR_OVERFLOW         - string represents a number too large for the underlying type
//
// NOTES: This function operates under the assumption that the ranges 'A'-'Z' and 'a'-'z'
// are each contigous in their character table mappings (e.g. ASCII) for the target platform.
// The ANSI standard does not guarantee an order for these ranges.
//
static int stringToNum32_internal(const TCHAR* pszNum, dword* puRet, bool* pbRetNegative, bool bForceDefaultBase, dword uDefaultBase, size_t uDataLen, bool bFailIfNegative)
{
    int iRetCode = CODE_SUCCESS;
    
    //arg check - must be at least 1 character
    if ( !(pszNum && puRet && uDefaultBase>=2 && uDefaultBase<=36 && (uDataLen || *pszNum)) )
    {
        iRetCode = CODE_ERROR_ARGUMENTS;
    }
    else
    {
        const TCHAR*     p           = pszNum;
        const TCHAR*     pEnd        = NULL;
        bool            bNegative   = false;
        
        //find the end of the string - either by custom size passed, or by searching for NULL
        if (uDataLen)
        {
            //size passed in - overrides NULL terminator search; just jump to end of string - guard against pointer overflow
            pEnd = p + uDataLen;
            if (pEnd<=p) pEnd = NULL;
        }
        else
        {
            //adjust pEnd until its pointing at last character in string - guard against pointer overflow
            pEnd = pszNum;
            while(pEnd && *pEnd) ++pEnd;
        }

        //an unlikely edge case, but possible within theoretical limits (pointer bits wrap to zero/NULL)
        if (!pEnd)
        {
            iRetCode = CODE_ERROR_LIMIT_REACHED;
        }
        else
        {
            //
            // Look for sign and base designations from front of string
            //

            //skip past leading whitespace (if any)
            while(p<pEnd && IS_WS(*p)) ++p;

            //look for positive or negative signs
            if (p<pEnd)
            {
                if ((*p==TEXT('-')) || (*p==TEXT('+')))
                {
                    bNegative = (*p==TEXT('-'));
                    ++p;

                    //skip past any additional whitespace (if any)
                    while(p<pEnd && IS_WS(*p)) ++p;
                }
            }

            //see if we are looking at a base declaration that isn't part of number
            //  (i.e. 0x800 (hex), 0b001 (binary), 0100 (octal))
            if (!bForceDefaultBase)
            {
                if (*p==TEXT('0') && p+1<pEnd)
                {
                    const TCHAR* pNextChar = p+1;

                    //hex base
                    if (*pNextChar==TEXT('x') || *pNextChar==TEXT('X'))
                    {
                        //only change base if bForceDefaultBase is false
                        uDefaultBase = 16;
                        ++p; ++p;
                    }
                    //binary base
                    else if (*pNextChar==TEXT('b') || *pNextChar==TEXT('B'))
                    {
                        //only change base if bForceDefaultBase is false
                        uDefaultBase = 2;
                        ++p; ++p;
                    }
                    //octal base - assumption is that this nothing matched above is an octal based number
                    else
                    {
                        uDefaultBase = 8;
                        ++p;
                    }
                }
            }

            //
            // Parse the number backwards
            //

            //if we are pointing at EOS at this point, no additional data was supplied
            if (p>=pEnd)
            {
                iRetCode = CODE_ERROR_SYNTAX;
            }
            //if fail if negative flag is set and a negative was encountered, return appropriate error
            else if (bFailIfNegative && bNegative)
            {
                iRetCode = CODE_ERROR_SYNTAX;
            }
            else
            {
                //keep track of where the number starts (most significant digit)
                const TCHAR* pNumStart = p;

                //go to last character in string
                p = pEnd-1;
                assert(p<=pEnd);
                
                //go backwards until non-whitespace is hit
                while (p>pNumStart && IS_WS(*p)) --p;
                
                //at loop entry, p is pointing at least significant digit
                dword uDigitVal      = 0;
                dword uDigitPower    = 1;
                *puRet = 0;
                dword uLastNum = 0;
                dword uLastDigitPower = 0;
                do
                {
                    //if in the ASCII range from 0 - 9, just convert to binary digit
                    if ((*p>=TEXT('0')) && (*p<=TEXT('9')))
                    {
                        uDigitVal = *p - TEXT('0');
                    }
                    //otherwise, if it's in the alphabet somewhere, the digit is an offset from letter 'a' as 10
                    //  (i.e. letter 'F' would be equal to 15 as it is in hex)
                    else if ((*p>=TEXT('A')) && (*p<=TEXT('Z')))
                    {
                        uDigitVal = *p - TEXT('A') + 10;
                    }
                    else if ((*p>=TEXT('a')) && (*p<=TEXT('z')))
                    {
                        uDigitVal = *p - TEXT('a') + 10;
                    }
                    else iRetCode = CODE_ERROR_SYNTAX;

                    //is digit valid for base (within range)
                    if (uDigitVal>uDefaultBase-1)
                    {
                        iRetCode = CODE_ERROR_SYNTAX;
                    }
                    else
                    {
                        *puRet += uDigitVal * uDigitPower;
                        uDigitPower *= uDefaultBase;
                        
                        //if for some reason the number prior to processing this digit is bigger,
                        //  we know we have overflowed the number because we are always adding
                        //  also check to ensure current power has not overflowed unless we are at final
                        //  digit
                        if (*puRet < uLastNum || (uDigitPower < uLastDigitPower && p>pNumStart) )
                        {
                            iRetCode = CODE_ERROR_OVERFLOW;
                        }
                        else
                        {                            
                            //keep track of the last number and power
                            uLastNum = *puRet;
                            uLastDigitPower = uDigitPower;
                        }

                    } //base is valid

                //until we processed last digit or a fatal error has occurred
                } while (p-- > pNumStart && CODE_SUCCEEDED(iRetCode));

            } //at least 1 character of numeric data exists

            //pass back negative flag if caller desires this information
            if (pbRetNegative)
            {
                *pbRetNegative = bNegative;
            }

        } //pEnd validated

    } //args OK

    return(iRetCode);

} //stringToNum32_internal()

//
// stringToInt32()
//
// Converts a string to a 32-bit-representation of the number in signed and unsigned formats (31 bits max, reserving the sign bit).
//
// ARGUMENTS:
//      pszNum                      - string data of contiguous memory to be converted
//      RetNum                      - result of conversion
//      bForceDefaultBase           - if false, 0x and 0b and 0 (hex, binary and octal notation) will override uDefaultBase,
//                                    otherwise the base designated by uDefaultBase is the basis for conversion.
//      uDefaultBase                - specifies default base of converted number; must be between 2 and 36
//                                    using character ranges 0-9,A-Z or a-z
//      uDataLen                    - if nonzero, only process n characters of passed string, otherwise position of NULL is used
//                                    to determine end of digits.
//      bFailIfNegative             - if a negative sign is found, this function call will fail with a syntax error; this
//                                    is used to enforce and validate unsigned numbers.
//
// SYNTAX:
//      (WS)*[-+](WS)*(0[xb])*num(WS)*
//
//      (...or more human readable...)
//
//      sign        base prefix     numerical data (alphanumerical data too)
//      ['+','-']   ['0x','0b']     0..9[,A...]
//
//
ResultCode stringToInt32(const TCHAR* pszNum, int32* puRet, bool bForceDefaultBase, dword uDefaultBase, size_t uDataLen)
{
    //do an unsigned conversion first, noting if we got a negative sign
    dword       uNumWrapper = 0;
    bool        bNegative   = false;
    int         iRetCode    = stringToNum32_internal(pszNum,&uNumWrapper,&bNegative,bForceDefaultBase,uDefaultBase,uDataLen,false);
    if (CODE_SUCCEEDED(iRetCode))
    {
        //We now have an unsigned representation of a number, with or without our negative flag set;
        //  to safely place this number into an int type we need to check the sign bit, even if the
        //  the negative flag is not set. If the unsigned representation is nonzero, we need to 
        //  subtract a 1 prior to looking at the sign bit; this will account for a negative number
        //  being able to store a 1 higher magnitude value than the unsigned counterpart.
        //  (e.g. a signed byte ranges from -128 to +127) 
        dword uNumCompare = uNumWrapper;
        if (uNumCompare) --uNumCompare;

        //if sign bit set, we would overflow the int type
        if (HAS_BITS(uNumCompare,(dword)1 << (sizeof(dword)*8-1) ))
        {
            iRetCode = CODE_ERROR_OVERFLOW;
        }
        else
        {
            //if a negative sign was parsed, convert to negative number using two's compliment
            if (bNegative)
            {
                uNumWrapper = ~(uNumWrapper)+1;
            }

            //assign int value by pointer casting - avoiding implicit bit changes by avoiding direct type conversions
            *puRet = (int32)uNumWrapper;

        } //result fits in an int

    } //number data parsed OK

    return(iRetCode);

} //stringToInt32()

//
// stringToDword() - same as stringToInt32(), but the negative sign is disallowed and a full 32 bits can be used
//
ResultCode stringToDword(const TCHAR* pszNum, dword* puRet, bool bForceDefaultBase, dword uDefaultBase, size_t uDataLen)
{
    return(stringToNum32_internal(pszNum,puRet,NULL,bForceDefaultBase,uDefaultBase,uDataLen,true));
    
} //stringToDword()

//
// stringCopy() - copies source String into destination buffer, and restricts
//      the amount copied to the destination buffer by the passed buffer size.
//      the return value is 0 if an error occurred or the final String could not
//      fit entirely into buffer.  The return value is the length of
//      the destination String when uCharCopyCount is ZERO, or uCharCopyCount.
//
//      uCharCopyCount is the number of characters to copy.  If ZERO, the function
//      will copy entire String and a NULL termination character.  If there is not enough
//      space, the String will be truncated.
//
//      If uCharCopyCount is nonzero, only those amount of characters will be copied
//      to buffer within destination limit.  Unless NULL terminator is included within
//      range being copied, the final String WILL NOT be NULL terminated.
//
size_t stringCopy(TCHAR* pszDest, size_t uMaxLenDest, const TCHAR* pszSource, size_t uCopyChars)
{
    size_t uRetCount = 0;
    
    //check args
    if (pszDest && pszSource && uMaxLenDest)
    {
        //configure source and destination pointers
        const TCHAR*     p   = pszSource;
        TCHAR*           d   = pszDest;

        if (uCopyChars)
        {
            //copy by explicit count without regard to source String (like strncpy)
            size_t          uActualCopyCount    = SMALLEST(uMaxLenDest,uCopyChars);
            const char*     pLastChar           = pszSource+uActualCopyCount-1;

            //copy characters loop
            while (p<=pLastChar)
            {
                *d++ = *p++;
            }
            
            //determine return value
            uRetCount = uActualCopyCount;
            if (uActualCopyCount < uCopyChars)
            {
                //truncation occurred
                uRetCount = 0;
            }
        }
        else
        {
            //copy entire String (always make room for NULL) - like strcpy()
            while (*p && (uRetCount<uMaxLenDest-1))
            {
                *d++ = *p++;
                ++uRetCount;
            }
            
            //NULL terminate
            *d = 0;

            //if we truncated String, return 0
            if (*p)
            {   
                assert((size_t)uRetCount == uMaxLenDest-1);
                uRetCount = 0;
            }
        }
    }

    return(uRetCount);

} //stringCopy()

//
// stringAnsiToWide() - transforms an ANSI String pszSource to a WIDE String
//      whose buffer is indicated by pszDest.  If pszSource is NULL or the same as dest,
//      the conversion will be done in place.
//
ResultCode stringAnsiToWide(const char* pszSource, WCHAR* pszDest, size_t uMaxLenDest, size_t uCopyLen)
{
    ResultCode retCode = CODE_ERROR_ARGUMENTS;
    if (pszDest && uMaxLenDest)
    {
        //are we converting the buffer in-place?
        if (!pszSource || (void*)pszSource == (void*)pszDest)
        {
            retCode = stringAnsiToWideInPlace((char*)pszDest,uMaxLenDest,uCopyLen);
        }
        else
        {
            const char*     p           = pszSource;
            WCHAR*          d           = pszDest;
            retCode = CODE_SUCCESS;

            if (!uCopyLen)
            {
                //
                // do conversion by NULL terminator in source
                //
                const WCHAR* pLastDest = pszDest + uMaxLenDest-1;

                //copy until NULL terminator is hit in source
                while (*p && d<pLastDest)
                {
                    *d++ = (WCHAR)(*p++);
                }

                //NULL terminate destination
                *d = 0;

                //if we didn't run out of room in destination, source string will be at NULL terminator
                if (0 != *p)
                {
                    retCode = CODE_ERROR_BUFFER_SIZE;
                }
            }
            else
            {
                //
                // do conversion by count
                //

                //ensure we have enough space in wide buffer
                if (uCopyLen > uMaxLenDest)
                {
                    retCode = CODE_ERROR_BUFFER_SIZE;
                }
                else
                {
                    //copy just the number of characters specified and nothing more
                    while (uCopyLen--)
                    {
                        *d++ = (WCHAR)(*p++);
                    }
                }
            }
        }
    }

    return(retCode);

} //stringAnsiToWide()

//
// stringAnsiToWideInPlace() - if uStringLength is 0, the size will be determined by where the NULL terminator lies
//
ResultCode stringAnsiToWideInPlace(char* pszDest, size_t uMaxSizeDest, size_t uCopyLen)
{
    ResultCode retCode = CODE_ERROR_ARGUMENTS;
    
    //validate passed arguments
    if (pszDest && uMaxSizeDest)
    {
        //set up ANSI string variables
        const char* pszAnsi     = pszDest;
        const char* pszAnsiEnd  = pszAnsi;
        if (!uCopyLen)
        {
            //locate end of string
            while(*pszAnsiEnd) ++pszAnsiEnd;
            
            //go past NULL terminator
            ++pszAnsiEnd;
            
            //...so NULL is included within count and data we will copy
            uCopyLen = pszAnsiEnd-pszAnsi;
            assert(uCopyLen);
        }
        else
        {
            //string length already specified
            pszAnsiEnd += uCopyLen;
        }
        //at this point, pszAnsiEnd points 1 past last ansi source character
        //  (this is where the NULL terminator may or may not be)
        
        //calculate required byte size for wide buffer
        size_t uReqAllocSize = uCopyLen * sizeof(WCHAR);
        assert(uReqAllocSize);
        
        //ensure we have enough space
        if (uReqAllocSize>uMaxSizeDest)
        {
            retCode = CODE_ERROR_BUFFER_SIZE;
        }
        else
        {
            //set up destination pointer
            WCHAR*  d = ((WCHAR*)pszDest) + uCopyLen - 1;
            
            //duplicate the buffer backwards (so we can do the conversion in place)
            const char* s = pszAnsiEnd-1; //init source to point at last character in buffer
            while (s >= pszAnsi)
            {
                *d-- = (WCHAR)( *s-- );
            }

            //success!
            retCode = CODE_SUCCESS;
        }
    }

    return(retCode);

} //stringAnsiToWideInPlace()

//
// peGetMachineString()
//
const TCHAR* peGetMachineString(WORD uMachine)
{
    const TCHAR* pszRet = NULL;
    switch(uMachine)
    {
        case IMAGE_FILE_MACHINE_I386:           pszRet = TEXT("i386"); break;
        case IMAGE_FILE_MACHINE_AMD64:          pszRet = TEXT("x64"); break;
        case IMAGE_FILE_MACHINE_IA64:           pszRet = TEXT("Itanium"); break; //Intel 64
        
        case IMAGE_FILE_MACHINE_POWERPC:        pszRet = TEXT("PowerPC Little-Endian"); break;
        case IMAGE_FILE_MACHINE_POWERPCFP:      pszRet = TEXT("PowerPC with Floating Point"); break;

        case IMAGE_FILE_MACHINE_ARM:            pszRet = TEXT("ARM Little-Endian"); break;
        case IMAGE_FILE_MACHINE_ARMNT:          pszRet = TEXT("ARMv7+ Thumb Mode"); break;
        case IMAGE_FILE_MACHINE_ARM64:          pszRet = TEXT("ARMv8 64-bit Mode"); break;
        case IMAGE_FILE_MACHINE_THUMB:          pszRet = TEXT("ARM or Thumb (Interworking)"); break;
        
        case IMAGE_FILE_MACHINE_R3000:          pszRet = TEXT("MIPS Little-Endian, 0x160 Big-Endian"); break;
        case IMAGE_FILE_MACHINE_R4000:          pszRet = TEXT("MIPS Little-Endian"); break;
        case IMAGE_FILE_MACHINE_R10000:         pszRet = TEXT("MIPS Little-Endian"); break;
        case IMAGE_FILE_MACHINE_WCEMIPSV2:      pszRet = TEXT("MIPS Little-Endian WCE v2"); break;
        case IMAGE_FILE_MACHINE_MIPS16:         pszRet = TEXT("MIPS16"); break;        
        case IMAGE_FILE_MACHINE_MIPSFPU:        pszRet = TEXT("MIPS with FPU"); break;
        case IMAGE_FILE_MACHINE_MIPSFPU16:      pszRet = TEXT("MIPS16 with FPU"); break;

        case IMAGE_FILE_MACHINE_SH3:            pszRet = TEXT("Hitachi SH3"); break;
        case IMAGE_FILE_MACHINE_SH3DSP:         pszRet = TEXT("Hitachi SH3 DSP"); break;
        case IMAGE_FILE_MACHINE_SH3E:           pszRet = TEXT("Hitachi SH3E Little-Endian"); break;
        case IMAGE_FILE_MACHINE_SH4:            pszRet = TEXT("Hitachi SH4"); break;
        case IMAGE_FILE_MACHINE_SH5:            pszRet = TEXT("Hitachi SH5"); break;

        case IMAGE_FILE_MACHINE_ALPHA:          pszRet = TEXT("Alpha_AXP"); break;
        case IMAGE_FILE_MACHINE_ALPHA64:        pszRet = TEXT("ALPHA64"); break; //a.k.a. IMAGE_FILE_MACHINE_AXP64

        case IMAGE_FILE_MACHINE_AM33:           pszRet = TEXT("Matsushita AM33"); break;        
        case IMAGE_FILE_MACHINE_TRICORE:        pszRet = TEXT("Infineon"); break;
        case IMAGE_FILE_MACHINE_EBC:            pszRet = TEXT("EFI Byte Code"); break;
        case IMAGE_FILE_MACHINE_M32R:           pszRet = TEXT("Mitsubishi M32R Little-Endian"); break;
        case IMAGE_FILE_MACHINE_CEF:            pszRet = TEXT("CEF"); break; //?
        case IMAGE_FILE_MACHINE_CEE:            pszRet = TEXT("CEE"); break; //?
        
        case IMAGE_FILE_MACHINE_UNKNOWN:        pszRet = TEXT("unknown"); break;
            
        default:
            pszRet = TEXT("unrecognized");
    }
    
    return(pszRet);

} //peGetMachineString()

//
// peRoundUpToSectionAlignment() - rounds value up to nearest dwAlign
//
DWORD peRoundUpToSectionAlignment(DWORD dwAlign, DWORD dwVal)
{
    if (dwAlign)
    {
        //do the rounding with bitwise operations...

        //create bit mask of bits to keep
        //  e.g. if section alignment is 0x1000                        1000000000000
        //       we want the following bitmask      11111111111111111111000000000000
        DWORD dwMask = ~(dwAlign-1);
        
        //round up by adding full alignment (dwAlign-1 since if already aligned we don't want anything to change),
        //  then mask off any lower bits
        dwVal = (dwVal + dwAlign-1) & dwMask;
    }

    return(dwVal);

} //peRoundUpToSectionAlignment()
