////////////////////////////////////////////////////////////////////////////////
//                                                                            //
// OllyDbg Disassembling Engine v2.01                                         //
//                                                                            //
// Copyright (c) 2007-2013 Oleh Yuschuk, ollydbg@t-online.de                  //
//                                                                            //
// This code is part of the OllyDbg Disassembler v2.01                        //
//                                                                            //
// Disassembling engine is free software; you can redistribute it and/or      //
// modify it under the terms of the GNU General Public License as published   //
// by the Free Software Foundation; either version 3 of the License, or (at   //
// your option) any later version.                                            //
//                                                                            //
// This code is distributed in the hope that it will be useful, but WITHOUT   //
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or      //
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for   //
// more details.                                                              //
//                                                                            //
// You should have received a copy of the GNU General Public License along    //
// with this program. If not, see <http://www.gnu.org/licenses/>.             //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
//                                                                            //
// This is a fast disassembler that can be used to determine the length of    //
// the binary 80x86 32-bit command and its attributes, to convert it to the   //
// human-readable text form, highlight its operands, and create hexadecimal   //
// dump of the binary command.                                                //
//                                                                            //
// It is a stripped down version of the disassembler used by OllyDbg 2.01.    //
// It can't analyse and comment the contents of the operands, or predict the  //
// results of the command execution. Analysis-dependent features are not      //
// included, too. Most other features are kept.                               //
//                                                                            //
// Disassembler supports integer, FPU, MMX, 3DNow, SSE1-SSE4.1 and AVX        //
// instructions. 64-bit mode, AVX2, FMA and XOP are not (yet) supported.      //
//                                                                            //
// This code can be compiled either in ASCII or UNICODE mode. It is reentrant //
// (thread-safe, feature not available in the original OllyDbg code).         //
//                                                                            //
// Typical operation speed on 3-GHz Phenom II in MASM mode is:                //
//                                                                            //
//   Command length and info:          130 ns/command (7,700,000 commands/s)  //
//   Disassembly:                      290 ns/command (3,400,000 commands/s)  //
//   Disassembly, dump, highlighting:  350 ns/command (2,800,000 commands/s)  //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////

//[bytepointer.com]: disabled some warnings as the original code was c
#pragma warning (disable:4244) //warning C4244: '+=' : conversion from 'int' to 'uchar', possible loss of data
#pragma warning (disable:4244) //warning C4244: '+=' : conversion from 'int' to 'uchar', possible loss of data
#pragma warning (disable:4100) //warning C4100: 'XXX' : unreferenced formal parameter

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#pragma hdrstop

#include "disasm_interface.h" //[bytepointer.com]: broke out some definitions from disasm.h here for inclusion into rest of project
#include "disasm.h"           //                   remainder of disasm.h is only used by this source file


//
// [bytepointer.com] the contents of the original cmdlist.c pasted here to keep engine in one compilation unit; disclaimers where identical
//

//instruction definitions were originally in a file cmdlist.c / just copied them in here so we only have one compilation unit
const t_bincmd bincmd[] = {

  { T("PAUSE"),
        D_SSE|D_MUSTF3, 0,
        1, 0x000000FF, 0x00000090, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("NOP"),
        D_CMD, DX_NOP,
        1, 0x000000FF, 0x00000090, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("NOP"),
        D_CMD|D_UNDOC, DX_NOP,
        2, 0x0000FFFF, 0x0000190F, 0x00,
        B_INT,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("NOP"),
        D_CMD|D_UNDOC, DX_NOP,
        2, 0x0000FFFF, 0x00001A0F, 0x00,
        B_INT,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("NOP"),
        D_CMD|D_UNDOC, DX_NOP,
        2, 0x0000FFFF, 0x00001B0F, 0x00,
        B_INT,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("NOP"),
        D_CMD|D_UNDOC, DX_NOP,
        2, 0x0000FFFF, 0x00001C0F, 0x00,
        B_INT,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("NOP"),
        D_CMD|D_UNDOC, DX_NOP,
        2, 0x0000FFFF, 0x00001D0F, 0x00,
        B_INT,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("NOP"),
        D_CMD|D_UNDOC, DX_NOP,
        2, 0x0000FFFF, 0x00001E0F, 0x00,
        B_INT,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("NOP"),
        D_CMD, DX_NOP,
        2, 0x0000FFFF, 0x00001F0F, 0x00,
        B_INT,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("MONITOR"),
        D_SYS|D_RARE, 0,
        3, 0x00FFFFFF, 0x00C8010F, 0x00,
        B_EAXMEM|B_PSEUDO,
        B_ECX|B_BINARY|B_PSEUDO,
        B_EDX|B_BINARY|B_PSEUDO,
        B_NONE },

  { T("MWAIT"),
        D_SYS|D_RARE, 0,
        3, 0x00FFFFFF, 0x00C9010F, 0x00,
        B_EAX|B_BINARY|B_PSEUDO,
        B_ECX|B_BINARY|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("CLAC"),
        D_SYS|D_RARE, 0,
        3, 0x00FFFFFF, 0x00CA010F, 0x00,
        B_EAX|B_BINARY|B_PSEUDO,
        B_ECX|B_BINARY|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("STAC"),
        D_SYS|D_RARE, 0,
        3, 0x00FFFFFF, 0x00CB010F, 0x00,
        B_EAX|B_BINARY|B_PSEUDO,
        B_ECX|B_BINARY|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("AAA"),
        D_CMD|D_ALLFLAGS|D_RARE, 0,
        1, 0x000000FF, 0x00000037, 0x00,
        B_AL|B_UPD|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("AAD"),
        D_CMD|D_ALLFLAGS|D_RARE, 0,
        2, 0x0000FFFF, 0x00000AD5, 0x00,
        B_AX|B_UPD|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("AAD"),
        D_CMD|D_ALLFLAGS|D_RARE, 0,
        1, 0x000000FF, 0x000000D5, 0x00,
        B_AX|B_UPD|B_PSEUDO,
        B_CONST8|B_UNSIGNED,
        B_NONE,
        B_NONE },

  { T("AAM"),
        D_CMD|D_ALLFLAGS|D_RARE, 0,
        2, 0x0000FFFF, 0x00000AD4, 0x00,
        B_AX|B_UPD|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("AAM"),
        D_CMD|D_ALLFLAGS|D_RARE, 0,
        1, 0x000000FF, 0x000000D4, 0x00,
        B_AX|B_UPD|B_PSEUDO,
        B_CONST8|B_UNSIGNED,
        B_NONE,
        B_NONE },

  { T("AAS"),
        D_CMD|D_ALLFLAGS|D_RARE, 0,
        1, 0x000000FF, 0x0000003F, 0x00,
        B_AL|B_UPD|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("ADC"),
        D_CMD|D_SIZE01|D_ALLFLAGS|D_USESCARRY, DX_JZ|DX_JB,
        1, 0x000000FE, 0x00000014, 0x00,
        B_ACC|B_UPD,
        B_CONST|B_NOADDR,
        B_NONE,
        B_NONE },

  { T("ADC"),
        D_CMD|D_SIZE01|D_LOCKABLE|D_ALLFLAGS|D_USESCARRY, DX_JZ|DX_JB,
        1, 0x000038FE, 0x00001080, 0x00,
        B_INT|B_SHOWSIZE|B_UPD,
        B_CONST|B_NOADDR,
        B_NONE,
        B_NONE },

  { T("ADC"),
        D_CMD|D_SIZE01|D_LOCKABLE|D_ALLFLAGS|D_USESCARRY, DX_JZ|DX_JB,
        1, 0x000038FE, 0x00001082, 0x00,
        B_INT|B_SHOWSIZE|B_UPD,
        B_SXTCONST,
        B_NONE,
        B_NONE },

  { T("ADC"),
        D_CMD|D_SIZE01|D_LOCKABLE|D_ALLFLAGS|D_USESCARRY, DX_JZ|DX_JB,
        1, 0x000000FE, 0x00000010, 0x00,
        B_INT|B_UPD,
        B_REG,
        B_NONE,
        B_NONE },

  { T("ADC"),
        D_CMD|D_SIZE01|D_ALLFLAGS|D_USESCARRY, DX_JZ|DX_JB,
        1, 0x000000FE, 0x00000012, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("ADD"),
        D_CMD|D_SIZE01|D_ALLFLAGS, DX_JZ|DX_JB|DX_ADD,
        1, 0x000000FE, 0x00000004, 0x00,
        B_ACC|B_UPD,
        B_CONST,
        B_NONE,
        B_NONE },

  { T("ADD"),
        D_CMD|D_SIZE01|D_LOCKABLE|D_ALLFLAGS, DX_JZ|DX_JB|DX_ADD,
        1, 0x000038FE, 0x00000080, 0x00,
        B_INT|B_SHOWSIZE|B_UPD,
        B_CONST,
        B_NONE,
        B_NONE },

  { T("ADD"),
        D_CMD|D_SIZE01|D_LOCKABLE|D_ALLFLAGS, DX_JZ|DX_JB|DX_ADD,
        1, 0x000038FE, 0x00000082, 0x00,
        B_INT|B_SHOWSIZE|B_UPD,
        B_SXTCONST,
        B_NONE,
        B_NONE },

  { T("ADD"),
        D_CMD|D_SIZE01|D_LOCKABLE|D_ALLFLAGS, DX_JZ|DX_JB|DX_ADD,
        1, 0x000000FE, 0x00000000, 0x00,
        B_INT|B_UPD,
        B_REG,
        B_NONE,
        B_NONE },

  { T("ADD"),
        D_CMD|D_SIZE01|D_ALLFLAGS, DX_JZ|DX_JB|DX_ADD,
        1, 0x000000FE, 0x00000002, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("AND"),
        D_TEST|D_SIZE01|D_ALLFLAGS, DX_JZ,
        1, 0x000000FE, 0x00000024, 0x00,
        B_ACC|B_BINARY|B_UPD,
        B_CONST|B_BINARY,
        B_NONE,
        B_NONE },

  { T("AND"),
        D_TEST|D_SIZE01|D_LOCKABLE|D_ALLFLAGS, DX_JZ,
        1, 0x000038FE, 0x00002080, 0x00,
        B_INT|B_BINARY|B_SHOWSIZE|B_UPD,
        B_CONST|B_BINARY,
        B_NONE,
        B_NONE },

  { T("AND"),
        D_TEST|D_SIZE01|D_LOCKABLE|D_ALLFLAGS, DX_JZ,
        1, 0x000038FE, 0x00002082, 0x00,
        B_INT|B_BINARY|B_SHOWSIZE|B_UPD,
        B_SXTCONST|B_BINARY,
        B_NONE,
        B_NONE },

  { T("AND"),
        D_TEST|D_SIZE01|D_LOCKABLE|D_ALLFLAGS, DX_JZ,
        1, 0x000000FE, 0x00000020, 0x00,
        B_INT|B_BINARY|B_UPD,
        B_REG|B_BINARY,
        B_NONE,
        B_NONE },

  { T("AND"),
        D_TEST|D_SIZE01|D_ALLFLAGS, DX_JZ,
        1, 0x000000FE, 0x00000022, 0x00,
        B_REG|B_BINARY|B_UPD,
        B_INT|B_BINARY,
        B_NONE,
        B_NONE },

  { T("ARPL"),
        D_SYS|D_FLAGZ|D_RARE, 0,
        1, 0x000000FF, 0x00000063, 0x00,
        B_INT16|B_UPD,
        B_REG16,
        B_NONE,
        B_NONE },

  { T("BOUND"),
        D_CMD|D_RARE, 0,
        1, 0x000000FF, 0x00000062, 0x00,
        B_REG|B_SIGNED,
        B_INTPAIR|B_MEMONLY,
        B_NONE,
        B_NONE },

  { T("BSF"),
        D_CMD|D_ALLFLAGS, DX_JZ,
        2, 0x0000FFFF, 0x0000BC0F, 0x00,
        B_REG|B_CHG,
        B_INT|B_BINARY,
        B_NONE,
        B_NONE },

  { T("BSR"),
        D_CMD|D_NOREP|D_ALLFLAGS, DX_JZ,
        2, 0x0000FFFF, 0x0000BD0F, 0x00,
        B_REG|B_CHG,
        B_INT|B_BINARY,
        B_NONE,
        B_NONE },

  { T("BSWAP"),
        D_CMD, 0,
        2, 0x0000F8FF, 0x0000C80F, 0x00,
        B_REGCMD|B_32BITONLY|B_NOESP|B_UPD,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("BT"),
        D_TEST|D_ALLFLAGS, DX_JC,
        2, 0x0000FFFF, 0x0000A30F, 0x00,
        B_INT|B_BINARY,
        B_REG|B_BITCNT,
        B_NONE,
        B_NONE },

  { T("BT"),
        D_TEST|D_ALLFLAGS, DX_JC,
        2, 0x0038FFFF, 0x0020BA0F, 0x00,
        B_INT|B_BINARY|B_SHOWSIZE,
        B_CONST8|B_BITCNT,
        B_NONE,
        B_NONE },

  { T("BTC"),
        D_CMD|D_LOCKABLE|D_ALLFLAGS, DX_JC,
        2, 0x0000FFFF, 0x0000BB0F, 0x00,
        B_INT|B_BINARY|B_NOESP|B_UPD,
        B_REG|B_BITCNT,
        B_NONE,
        B_NONE },

  { T("BTC"),
        D_CMD|D_LOCKABLE|D_ALLFLAGS, DX_JC,
        2, 0x0038FFFF, 0x0038BA0F, 0x00,
        B_INT|B_BINARY|B_NOESP|B_SHOWSIZE|B_UPD,
        B_CONST8|B_BITCNT,
        B_NONE,
        B_NONE },

  { T("BTR"),
        D_CMD|D_LOCKABLE|D_ALLFLAGS, DX_JC,
        2, 0x0000FFFF, 0x0000B30F, 0x00,
        B_INT|B_BINARY|B_NOESP|B_UPD,
        B_REG|B_BITCNT,
        B_NONE,
        B_NONE },

  { T("BTR"),
        D_CMD|D_LOCKABLE|D_ALLFLAGS, DX_JC,
        2, 0x0038FFFF, 0x0030BA0F, 0x00,
        B_INT|B_BINARY|B_NOESP|B_SHOWSIZE|B_UPD,
        B_CONST8|B_BITCNT,
        B_NONE,
        B_NONE },

  { T("BTS"),
        D_CMD|D_LOCKABLE|D_ALLFLAGS, DX_JC,
        2, 0x0000FFFF, 0x0000AB0F, 0x00,
        B_INT|B_BINARY|B_NOESP|B_UPD,
        B_REG|B_BITCNT,
        B_NONE,
        B_NONE },

  { T("BTS"),
        D_CMD|D_LOCKABLE|D_ALLFLAGS, DX_JC,
        2, 0x0038FFFF, 0x0028BA0F, 0x00,
        B_INT|B_BINARY|B_NOESP|B_SHOWSIZE|B_UPD,
        B_CONST8|B_BITCNT,
        B_NONE,
        B_NONE },

  { T("CALL"),
        D_CALL|D_CHGESP, 0,
        1, 0x000000FF, 0x000000E8, 0x00,
        B_OFFSET|B_JMPCALL,
        B_PUSHRET|B_CHG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("CALL"),
        D_CALL|D_CHGESP, 0,
        1, 0x000038FF, 0x000010FF, 0x00,
        B_INT|B_JMPCALL,
        B_PUSHRET|B_CHG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("CALL"),
        D_CALLFAR|D_CHGESP|D_RARE, 0,
        1, 0x000000FF, 0x0000009A, 0x00,
        B_FARCONST|B_JMPCALLFAR,
        B_PUSHRETF|B_CHG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("CALL"),
        D_CALLFAR|D_CHGESP|D_RARE, 0,
        1, 0x000038FF, 0x000018FF, 0x00,
        B_SEGOFFS|B_JMPCALLFAR|B_MEMONLY,
        B_PUSHRETF|B_CHG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("CBW"),
        D_CMD|D_DATA16, 0,
        1, 0x000000FF, 0x00000098, 0x00,
        B_AX|B_UPD|B_PSEUDO,
        B_AL|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("CBW"),
        D_CMD|D_DATA16, 0,
        1, 0x000000FF, 0x00000098, 0x00,
        B_AX|B_UPD|B_PSEUDO,
        B_AL,
        B_NONE,
        B_NONE },

  { T("CDQ"),
        D_CMD|D_DATA32, 0,
        1, 0x000000FF, 0x00000099, 0x00,
        B_EDX|B_CHG|B_PSEUDO,
        B_EAX|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("CDQ"),
        D_CMD|D_DATA32, 0,
        1, 0x000000FF, 0x00000099, 0x00,
        B_EDX|B_CHG|B_PSEUDO,
        B_EAX,
        B_NONE,
        B_NONE },

  { T("CLC"),
        D_CMD|D_FLAGC, 0,
        1, 0x000000FF, 0x000000F8, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("CLD"),
        D_CMD|D_FLAGD, 0,
        1, 0x000000FF, 0x000000FC, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("CLFLUSH"),
        D_CMD|D_MEMORY|D_RARE, 0,
        2, 0x0038FFFF, 0x0038AE0F, 0x00,
        B_ANYMEM|B_MEMONLY,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("CLI"),
        D_CMD|D_RARE, 0,
        1, 0x000000FF, 0x000000FA, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("CLTS"),
        D_PRIVILEGED|D_RARE, 0,
        2, 0x0000FFFF, 0x0000060F, 0x00,
        B_CR0|B_UPD|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("CMC"),
        D_CMD|D_FLAGC, 0,
        1, 0x000000FF, 0x000000F5, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("CMOVO"),
        D_MOVC|D_COND, 0,
        2, 0x0000FFFF, 0x0000400F, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("CMOVNO"),
        D_MOVC|D_COND, 0,
        2, 0x0000FFFF, 0x0000410F, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("CMOVB"),
        D_MOVC|D_COND|D_USESCARRY, DX_JB,
        2, 0x0000FFFF, 0x0000420F, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("CMOVC"),
        D_MOVC|D_COND|D_USESCARRY, DX_JC,
        2, 0x0000FFFF, 0x0000420F, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("CMOVNAE"),
        D_MOVC|D_COND|D_USESCARRY, DX_JB,
        2, 0x0000FFFF, 0x0000420F, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("CMOVAE"),
        D_MOVC|D_COND|D_USESCARRY, DX_JB,
        2, 0x0000FFFF, 0x0000430F, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("CMOVNB"),
        D_MOVC|D_COND|D_USESCARRY, DX_JB,
        2, 0x0000FFFF, 0x0000430F, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("CMOVNC"),
        D_MOVC|D_COND|D_USESCARRY, DX_JC,
        2, 0x0000FFFF, 0x0000430F, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("CMOVE"),
        D_MOVC|D_COND, DX_JE,
        2, 0x0000FFFF, 0x0000440F, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("CMOVZ"),
        D_MOVC|D_COND, DX_JZ,
        2, 0x0000FFFF, 0x0000440F, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("CMOVNE"),
        D_MOVC|D_COND, DX_JE,
        2, 0x0000FFFF, 0x0000450F, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("CMOVNZ"),
        D_MOVC|D_COND, DX_JZ,
        2, 0x0000FFFF, 0x0000450F, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("CMOVBE"),
        D_MOVC|D_COND|D_USESCARRY, 0,
        2, 0x0000FFFF, 0x0000460F, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("CMOVNA"),
        D_MOVC|D_COND|D_USESCARRY, 0,
        2, 0x0000FFFF, 0x0000460F, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("CMOVA"),
        D_MOVC|D_COND|D_USESCARRY, 0,
        2, 0x0000FFFF, 0x0000470F, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("CMOVNBE"),
        D_MOVC|D_COND|D_USESCARRY, 0,
        2, 0x0000FFFF, 0x0000470F, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("CMOVS"),
        D_MOVC|D_COND, 0,
        2, 0x0000FFFF, 0x0000480F, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("CMOVNS"),
        D_MOVC|D_COND, 0,
        2, 0x0000FFFF, 0x0000490F, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("CMOVPE"),
        D_MOVC|D_COND, 0,
        2, 0x0000FFFF, 0x00004A0F, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("CMOVP"),
        D_MOVC|D_COND, 0,
        2, 0x0000FFFF, 0x00004A0F, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("CMOVPO"),
        D_MOVC|D_COND, 0,
        2, 0x0000FFFF, 0x00004B0F, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("CMOVNP"),
        D_MOVC|D_COND, 0,
        2, 0x0000FFFF, 0x00004B0F, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("CMOVL"),
        D_MOVC|D_COND, 0,
        2, 0x0000FFFF, 0x00004C0F, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("CMOVNGE"),
        D_MOVC|D_COND, 0,
        2, 0x0000FFFF, 0x00004C0F, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("CMOVGE"),
        D_MOVC|D_COND, 0,
        2, 0x0000FFFF, 0x00004D0F, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("CMOVNL"),
        D_MOVC|D_COND, 0,
        2, 0x0000FFFF, 0x00004D0F, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("CMOVLE"),
        D_MOVC|D_COND, 0,
        2, 0x0000FFFF, 0x00004E0F, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("CMOVNG"),
        D_MOVC|D_COND, 0,
        2, 0x0000FFFF, 0x00004E0F, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("CMOVG"),
        D_MOVC|D_COND, 0,
        2, 0x0000FFFF, 0x00004F0F, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("CMOVNLE"),
        D_MOVC|D_COND, 0,
        2, 0x0000FFFF, 0x00004F0F, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("CMP"),
        D_TEST|D_SIZE01|D_ALLFLAGS|D_HLADIR, DX_JE|DX_JB,
        1, 0x000000FE, 0x0000003C, 0x00,
        B_ACC,
        B_CONST,
        B_NONE,
        B_NONE },

  { T("CMP"),
        D_TEST|D_SIZE01|D_ALLFLAGS|D_HLADIR, DX_JE|DX_JB,
        1, 0x000038FE, 0x00003880, 0x00,
        B_INT|B_SHOWSIZE,
        B_CONST,
        B_NONE,
        B_NONE },

  { T("CMP"),
        D_TEST|D_SIZE01|D_ALLFLAGS|D_HLADIR, DX_JE|DX_JB,
        1, 0x000038FE, 0x00003882, 0x00,
        B_INT|B_SHOWSIZE,
        B_SXTCONST,
        B_NONE,
        B_NONE },

  { T("CMP"),
        D_TEST|D_SIZE01|D_ALLFLAGS|D_HLADIR, DX_JE|DX_JB,
        1, 0x000000FE, 0x00000038, 0x00,
        B_INT,
        B_REG,
        B_NONE,
        B_NONE },

  { T("CMP"),
        D_TEST|D_SIZE01|D_ALLFLAGS|D_HLADIR, DX_JE|DX_JB,
        1, 0x000000FE, 0x0000003A, 0x00,
        B_REG,
        B_INT,
        B_NONE,
        B_NONE },

  { T("CMPXCHG"),
        D_CMD|D_SIZE01|D_LOCKABLE|D_ALLFLAGS|D_HLADIR, DX_JE|DX_JB,
        2, 0x0000FEFF, 0x0000B00F, 0x00,
        B_INT|B_UPD,
        B_REG,
        B_ACC|B_UPD|B_PSEUDO,
        B_NONE },

  { T("CMPXCHG8B"),
        D_CMD|D_LOCKABLE|D_MEMORY|D_ALLFLAGS, DX_JE|DX_JB,
        2, 0x0038FFFF, 0x0008C70F, 0x00,
        B_INT64|B_MEMONLY|B_UPD,
        B_EAX|B_UPD|B_PSEUDO,
        B_EDX|B_UPD|B_PSEUDO,
        B_EBX|B_PSEUDO },

  { T("CPUID"),
        D_CMD, 0,
        2, 0x0000FFFF, 0x0000A20F, 0x00,
        B_EAX|B_CHG|B_PSEUDO,
        B_EBX|B_CHG|B_PSEUDO,
        B_ECX|B_CHG|B_PSEUDO,
        B_EDX|B_CHG|B_PSEUDO },

  { T("CWD"),
        D_CMD|D_DATA16, 0,
        1, 0x000000FF, 0x00000099, 0x00,
        B_DX|B_CHG|B_PSEUDO,
        B_AX|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("CWD"),
        D_CMD|D_DATA16, 0,
        1, 0x000000FF, 0x00000099, 0x00,
        B_DX|B_CHG|B_PSEUDO,
        B_AX,
        B_NONE,
        B_NONE },

  { T("CWDE"),
        D_CMD|D_DATA32, 0,
        1, 0x000000FF, 0x00000098, 0x00,
        B_EAX|B_UPD|B_PSEUDO,
        B_AX|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("CWDE"),
        D_CMD|D_DATA32, 0,
        1, 0x000000FF, 0x00000098, 0x00,
        B_EAX|B_UPD|B_PSEUDO,
        B_AX,
        B_NONE,
        B_NONE },

  { T("DAA"),
        D_CMD|D_ALLFLAGS|D_USESCARRY|D_RARE, DX_JC,
        1, 0x000000FF, 0x00000027, 0x00,
        B_AL|B_UPD|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("DAS"),
        D_CMD|D_ALLFLAGS|D_USESCARRY|D_RARE, DX_JC,
        1, 0x000000FF, 0x0000002F, 0x00,
        B_AL|B_UPD|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("DEC"),
        D_CMD|D_SIZE01|D_LOCKABLE|D_NOCFLAG, DX_JZ,
        1, 0x000038FE, 0x000008FE, 0x00,
        B_INT|B_SHOWSIZE|B_UPD,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("DEC"),
        D_CMD|D_NOCFLAG, DX_JZ,
        1, 0x000000F8, 0x00000048, 0x00,
        B_REGCMD|B_UPD,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("DIV"),
        D_CMD|D_ALLFLAGS, 0,
        1, 0x000038FF, 0x000030F6, 0x00,
        B_INT8|B_SHOWSIZE,
        B_AX|B_UPD|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("DIV"),
        D_CMD|D_ALLFLAGS, 0,
        1, 0x000038FF, 0x000030F7, 0x00,
        B_INT1632|B_UNSIGNED|B_NOESP|B_SHOWSIZE,
        B_DXEDX|B_UPD|B_PSEUDO,
        B_ACC|B_UPD|B_PSEUDO,
        B_NONE },

  { T("EMMS"),
        D_CMD, 0,
        2, 0x0000FFFF, 0x0000770F, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("ENTER"),
        D_CMD|D_CHGESP, 0,
        1, 0x000000FF, 0x000000C8, 0x00,
        B_CONST16|B_STACKINC,
        B_CONST8_2|B_UNSIGNED,
        B_PUSH|B_CHG|B_PSEUDO,
        B_BPEBP|B_CHG|B_PSEUDO },

  { T("WAIT"),
        D_CMD, 0,
        1, 0x000000FF, 0x0000009B, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FWAIT"),
        D_CMD, 0,
        1, 0x000000FF, 0x0000009B, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("HLT"),
        D_PRIVILEGED|D_RARE, 0,
        1, 0x000000FF, 0x000000F4, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("IDIV"),
        D_CMD|D_ALLFLAGS, 0,
        1, 0x000038FF, 0x000038F6, 0x00,
        B_INT8|B_SIGNED|B_SHOWSIZE,
        B_AX|B_UPD|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("IDIV"),
        D_CMD|D_ALLFLAGS, 0,
        1, 0x000038FF, 0x000038F7, 0x00,
        B_INT1632|B_SIGNED|B_NOESP|B_SHOWSIZE,
        B_DXEDX|B_UPD|B_PSEUDO,
        B_ACC|B_UPD|B_PSEUDO,
        B_NONE },

  { T("IMUL"),
        D_CMD|D_ALLFLAGS, 0,
        1, 0x000038FF, 0x000028F6, 0x00,
        B_AX|B_UPD|B_PSEUDO,
        B_INT8|B_SIGNED|B_SHOWSIZE,
        B_AL|B_SIGNED|B_PSEUDO,
        B_NONE },

  { T("IMUL"),
        D_CMD|D_ALLFLAGS, 0,
        1, 0x000038FF, 0x000028F7, 0x00,
        B_DXEDX|B_CHG|B_PSEUDO,
        B_ACC|B_UPD|B_PSEUDO,
        B_INT1632|B_SIGNED|B_NOESP|B_SHOWSIZE,
        B_NONE },

  { T("IMUL"),
        D_CMD|D_ALLFLAGS, 0,
        2, 0x0000FFFF, 0x0000AF0F, 0x00,
        B_REG|B_UPD,
        B_INT|B_NOESP,
        B_NONE,
        B_NONE },

  { T("IMUL"),
        D_CMD|D_ALLFLAGS, 0,
        1, 0x000000FF, 0x0000006B, 0x00,
        B_REG|B_CHG,
        B_INT|B_NOESP,
        B_SXTCONST,
        B_NONE },

  { T("IMUL"),
        D_CMD|D_ALLFLAGS, 0,
        1, 0x000000FF, 0x00000069, 0x00,
        B_REG|B_CHG,
        B_INT|B_NOESP,
        B_CONST|B_SIGNED,
        B_NONE },

  { T("IN"),
        D_IO|D_SIZE01|D_RARE, 0,
        1, 0x000000FE, 0x000000E4, 0x00,
        B_ACC|B_CHG,
        B_CONST8|B_PORT,
        B_NONE,
        B_NONE },

  { T("IN"),
        D_IO|D_SIZE01|D_RARE, 0,
        1, 0x000000FE, 0x000000EC, 0x00,
        B_ACC|B_CHG,
        B_DXPORT|B_PORT,
        B_NONE,
        B_NONE },

  { T("INC"),
        D_CMD|D_SIZE01|D_LOCKABLE|D_NOCFLAG, DX_JZ,
        1, 0x000038FE, 0x000000FE, 0x00,
        B_INT|B_SHOWSIZE|B_UPD,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("INC"),
        D_CMD|D_NOCFLAG, DX_JZ,
        1, 0x000000F8, 0x00000040, 0x00,
        B_REGCMD|B_UPD,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("INT"),
        D_INT, 0,
        1, 0x000000FF, 0x000000CD, 0x00,
        B_CONST8,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("INT3"),
        D_INT|D_RARE, 0,
        1, 0x000000FF, 0x000000CC, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("INTO"),
        D_INT|D_RARE, 0,
        1, 0x000000FF, 0x000000CE, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("INT1"),
        D_INT|D_UNDOC, 0,
        1, 0x000000FF, 0x000000F1, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("ICEBP"),
        D_INT|D_UNDOC, 0,
        1, 0x000000FF, 0x000000F1, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("INVD"),
        D_PRIVILEGED|D_RARE, 0,
        2, 0x0000FFFF, 0x0000080F, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("INVLPG"),
        D_PRIVILEGED|D_MEMORY|D_RARE, 0,
        2, 0x0038FFFF, 0x0038010F, 0x00,
        B_ANYMEM|B_MEMONLY,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("IRET*"),
        D_RETFAR|D_ALLFLAGS|D_CHGESP|D_WILDCARD|D_RARE, 0,
        1, 0x000000FF, 0x000000CF, 0x00,
        B_STKTOPFAR|B_JMPCALLFAR|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JO"),
        D_JMC|D_BHINT|D_COND, 0,
        1, 0x000000FF, 0x00000070, 0x00,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JO"),
        D_JMC|D_BHINT|D_COND, 0,
        2, 0x0000FFFF, 0x0000800F, 0x00,
        B_OFFSET|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JNO"),
        D_JMC|D_BHINT|D_COND, 0,
        1, 0x000000FF, 0x00000071, 0x00,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JNO"),
        D_JMC|D_BHINT|D_COND, 0,
        2, 0x0000FFFF, 0x0000810F, 0x00,
        B_OFFSET|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JB"),
        D_JMC|D_BHINT|D_COND|D_USESCARRY, DX_JB,
        1, 0x000000FF, 0x00000072, 0x00,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JC"),
        D_JMC|D_BHINT|D_COND|D_USESCARRY, DX_JC,
        1, 0x000000FF, 0x00000072, 0x00,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JNAE"),
        D_JMC|D_BHINT|D_COND|D_USESCARRY, DX_JB,
        1, 0x000000FF, 0x00000072, 0x00,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JB"),
        D_JMC|D_BHINT|D_COND|D_USESCARRY, DX_JB,
        2, 0x0000FFFF, 0x0000820F, 0x00,
        B_OFFSET|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JC"),
        D_JMC|D_BHINT|D_COND|D_USESCARRY, DX_JC,
        2, 0x0000FFFF, 0x0000820F, 0x00,
        B_OFFSET|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JNAE"),
        D_JMC|D_BHINT|D_COND|D_USESCARRY, DX_JB,
        2, 0x0000FFFF, 0x0000820F, 0x00,
        B_OFFSET|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JAE"),
        D_JMC|D_BHINT|D_COND|D_USESCARRY, DX_JB,
        1, 0x000000FF, 0x00000073, 0x00,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JNB"),
        D_JMC|D_BHINT|D_COND|D_USESCARRY, DX_JB,
        1, 0x000000FF, 0x00000073, 0x00,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JNC"),
        D_JMC|D_BHINT|D_COND|D_USESCARRY, DX_JC,
        1, 0x000000FF, 0x00000073, 0x00,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JAE"),
        D_JMC|D_BHINT|D_COND|D_USESCARRY, DX_JB,
        2, 0x0000FFFF, 0x0000830F, 0x00,
        B_OFFSET|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JNB"),
        D_JMC|D_BHINT|D_COND|D_USESCARRY, DX_JB,
        2, 0x0000FFFF, 0x0000830F, 0x00,
        B_OFFSET|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JNC"),
        D_JMC|D_BHINT|D_COND|D_USESCARRY, DX_JC,
        2, 0x0000FFFF, 0x0000830F, 0x00,
        B_OFFSET|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JE"),
        D_JMC|D_BHINT|D_COND, DX_JE,
        1, 0x000000FF, 0x00000074, 0x00,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JZ"),
        D_JMC|D_BHINT|D_COND, DX_JZ,
        1, 0x000000FF, 0x00000074, 0x00,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JE"),
        D_JMC|D_BHINT|D_COND, DX_JE,
        2, 0x0000FFFF, 0x0000840F, 0x00,
        B_OFFSET|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JZ"),
        D_JMC|D_BHINT|D_COND, DX_JZ,
        2, 0x0000FFFF, 0x0000840F, 0x00,
        B_OFFSET|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JNE"),
        D_JMC|D_BHINT|D_COND, DX_JE,
        1, 0x000000FF, 0x00000075, 0x00,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JNZ"),
        D_JMC|D_BHINT|D_COND, DX_JZ,
        1, 0x000000FF, 0x00000075, 0x00,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JNE"),
        D_JMC|D_BHINT|D_COND, DX_JE,
        2, 0x0000FFFF, 0x0000850F, 0x00,
        B_OFFSET|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JNZ"),
        D_JMC|D_BHINT|D_COND, DX_JZ,
        2, 0x0000FFFF, 0x0000850F, 0x00,
        B_OFFSET|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JBE"),
        D_JMC|D_BHINT|D_COND|D_USESCARRY, 0,
        1, 0x000000FF, 0x00000076, 0x00,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JNA"),
        D_JMC|D_BHINT|D_COND|D_USESCARRY, 0,
        1, 0x000000FF, 0x00000076, 0x00,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JBE"),
        D_JMC|D_BHINT|D_COND|D_USESCARRY, 0,
        2, 0x0000FFFF, 0x0000860F, 0x00,
        B_OFFSET|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JNA"),
        D_JMC|D_BHINT|D_COND|D_USESCARRY, 0,
        2, 0x0000FFFF, 0x0000860F, 0x00,
        B_OFFSET|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JA"),
        D_JMC|D_BHINT|D_COND|D_USESCARRY, 0,
        1, 0x000000FF, 0x00000077, 0x00,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JNBE"),
        D_JMC|D_BHINT|D_COND|D_USESCARRY, 0,
        1, 0x000000FF, 0x00000077, 0x00,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JA"),
        D_JMC|D_BHINT|D_COND|D_USESCARRY, 0,
        2, 0x0000FFFF, 0x0000870F, 0x00,
        B_OFFSET|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JNBE"),
        D_JMC|D_BHINT|D_COND|D_USESCARRY, 0,
        2, 0x0000FFFF, 0x0000870F, 0x00,
        B_OFFSET|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JS"),
        D_JMC|D_BHINT|D_COND, 0,
        1, 0x000000FF, 0x00000078, 0x00,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JS"),
        D_JMC|D_BHINT|D_COND, 0,
        2, 0x0000FFFF, 0x0000880F, 0x00,
        B_OFFSET|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JNS"),
        D_JMC|D_BHINT|D_COND, 0,
        1, 0x000000FF, 0x00000079, 0x00,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JNS"),
        D_JMC|D_BHINT|D_COND, 0,
        2, 0x0000FFFF, 0x0000890F, 0x00,
        B_OFFSET|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JPE"),
        D_JMC|D_BHINT|D_COND|D_RARE, 0,
        1, 0x000000FF, 0x0000007A, 0x00,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JP"),
        D_JMC|D_BHINT|D_COND|D_RARE, 0,
        1, 0x000000FF, 0x0000007A, 0x00,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JPE"),
        D_JMC|D_BHINT|D_COND|D_RARE, 0,
        2, 0x0000FFFF, 0x00008A0F, 0x00,
        B_OFFSET|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JP"),
        D_JMC|D_BHINT|D_COND|D_RARE, 0,
        2, 0x0000FFFF, 0x00008A0F, 0x00,
        B_OFFSET|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JPO"),
        D_JMC|D_BHINT|D_COND|D_RARE, 0,
        1, 0x000000FF, 0x0000007B, 0x00,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JNP"),
        D_JMC|D_BHINT|D_COND|D_RARE, 0,
        1, 0x000000FF, 0x0000007B, 0x00,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JPO"),
        D_JMC|D_BHINT|D_COND|D_RARE, 0,
        2, 0x0000FFFF, 0x00008B0F, 0x00,
        B_OFFSET|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JNP"),
        D_JMC|D_BHINT|D_COND|D_RARE, 0,
        2, 0x0000FFFF, 0x00008B0F, 0x00,
        B_OFFSET|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JL"),
        D_JMC|D_BHINT|D_COND, 0,
        1, 0x000000FF, 0x0000007C, 0x00,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JNGE"),
        D_JMC|D_BHINT|D_COND, 0,
        1, 0x000000FF, 0x0000007C, 0x00,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JL"),
        D_JMC|D_BHINT|D_COND, 0,
        2, 0x0000FFFF, 0x00008C0F, 0x00,
        B_OFFSET|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JNGE"),
        D_JMC|D_BHINT|D_COND, 0,
        2, 0x0000FFFF, 0x00008C0F, 0x00,
        B_OFFSET|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JGE"),
        D_JMC|D_BHINT|D_COND, 0,
        1, 0x000000FF, 0x0000007D, 0x00,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JNL"),
        D_JMC|D_BHINT|D_COND, 0,
        1, 0x000000FF, 0x0000007D, 0x00,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JGE"),
        D_JMC|D_BHINT|D_COND, 0,
        2, 0x0000FFFF, 0x00008D0F, 0x00,
        B_OFFSET|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JNL"),
        D_JMC|D_BHINT|D_COND, 0,
        2, 0x0000FFFF, 0x00008D0F, 0x00,
        B_OFFSET|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JLE"),
        D_JMC|D_BHINT|D_COND, 0,
        1, 0x000000FF, 0x0000007E, 0x00,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JNG"),
        D_JMC|D_BHINT|D_COND, 0,
        1, 0x000000FF, 0x0000007E, 0x00,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JLE"),
        D_JMC|D_BHINT|D_COND, 0,
        2, 0x0000FFFF, 0x00008E0F, 0x00,
        B_OFFSET|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JNG"),
        D_JMC|D_BHINT|D_COND, 0,
        2, 0x0000FFFF, 0x00008E0F, 0x00,
        B_OFFSET|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JG"),
        D_JMC|D_BHINT|D_COND, 0,
        1, 0x000000FF, 0x0000007F, 0x00,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JNLE"),
        D_JMC|D_BHINT|D_COND, 0,
        1, 0x000000FF, 0x0000007F, 0x00,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JG"),
        D_JMC|D_BHINT|D_COND, 0,
        2, 0x0000FFFF, 0x00008F0F, 0x00,
        B_OFFSET|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JNLE"),
        D_JMC|D_BHINT|D_COND, 0,
        2, 0x0000FFFF, 0x00008F0F, 0x00,
        B_OFFSET|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JCXZ"),
        D_JMCX|D_ADDR16|D_BHINT, 0,
        1, 0x000000FF, 0x000000E3, 0x00,
        B_CX|B_PSEUDO,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE },

  { T("JECXZ"),
        D_JMCX|D_ADDR32|D_BHINT, 0,
        1, 0x000000FF, 0x000000E3, 0x00,
        B_ECX|B_PSEUDO,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE },

  { T("JMP"),
        D_JMP, 0,
        1, 0x000000FF, 0x000000EB, 0x00,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JMP"),
        D_JMP, 0,
        1, 0x000000FF, 0x000000E9, 0x00,
        B_OFFSET|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JMP"),
        D_JMP, 0,
        1, 0x000038FF, 0x000020FF, 0x00,
        B_INT|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JMP"),
        D_JMPFAR|D_SUSPICIOUS, 0,
        1, 0x000000FF, 0x000000EA, 0x00,
        B_FARCONST|B_JMPCALLFAR,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JMP"),
        D_JMPFAR|D_RARE, 0,
        1, 0x000038FF, 0x000028FF, 0x00,
        B_SEGOFFS|B_JMPCALLFAR|B_MEMONLY|B_SHOWSIZE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("LAHF"),
        D_CMD, 0,
        1, 0x000000FF, 0x0000009F, 0x00,
        B_AH|B_CHG|B_PSEUDO,
        B_FLAGS8|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("LAR"),
        D_CMD|D_FLAGZ|D_RARE, DX_JZ,
        2, 0x0000FFFF, 0x0000020F, 0x00,
        B_REG|B_BINARY|B_NOESP|B_CHG,
        B_INT|B_BINARY|B_NOESP,
        B_NONE,
        B_NONE },

  { T("LDS"),
        D_CMD|D_RARE, 0,
        1, 0x000000FF, 0x000000C5, 0x00,
        B_SEGDS|B_CHG|B_PSEUDO,
        B_REG|B_BINARY|B_CHG,
        B_SEGOFFS|B_MEMONLY,
        B_NONE },

  { T("LES"),
        D_CMD|D_RARE, 0,
        1, 0x000000FF, 0x000000C4, 0x00,
        B_SEGES|B_CHG|B_PSEUDO,
        B_REG|B_BINARY|B_CHG,
        B_SEGOFFS|B_MEMONLY,
        B_NONE },

  { T("LFS"),
        D_CMD|D_RARE, 0,
        2, 0x0000FFFF, 0x0000B40F, 0x00,
        B_SEGFS|B_CHG|B_PSEUDO,
        B_REG|B_BINARY|B_CHG,
        B_SEGOFFS|B_MEMONLY,
        B_NONE },

  { T("LGS"),
        D_CMD|D_RARE, 0,
        2, 0x0000FFFF, 0x0000B50F, 0x00,
        B_SEGGS|B_CHG|B_PSEUDO,
        B_REG|B_BINARY|B_CHG,
        B_SEGOFFS|B_MEMONLY,
        B_NONE },

  { T("LSS"),
        D_CMD|D_RARE, 0,
        2, 0x0000FFFF, 0x0000B20F, 0x00,
        B_SEGSS|B_CHG|B_PSEUDO,
        B_REG|B_BINARY|B_CHG,
        B_SEGOFFS|B_MEMONLY,
        B_NONE },

  { T("LEA"),
        D_CMD|D_HLADIR, DX_LEA,
        1, 0x000000FF, 0x0000008D, 0x00,
        B_REG|B_BINARY|B_CHG,
        B_ANYMEM|B_MEMONLY|B_NOSEG,
        B_NONE,
        B_NONE },

  { T("LEAVE"),
        D_CMD|D_CHGESP, 0,
        1, 0x000000FF, 0x000000C9, 0x00,
        B_BPEBP|B_CHG|B_PSEUDO,
        B_EBPMEM|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("LGDT"),
        D_PRIVILEGED|D_MEMORY|D_RARE, 0,
        2, 0x0038FFFF, 0x0010010F, 0x00,
        B_DESCR|B_MEMONLY,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("LIDT"),
        D_PRIVILEGED|D_MEMORY|D_RARE, 0,
        2, 0x0038FFFF, 0x0018010F, 0x00,
        B_DESCR|B_MEMONLY,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("LLDT"),
        D_PRIVILEGED|D_RARE, 0,
        2, 0x0038FFFF, 0x0010000F, 0x00,
        B_INT16|B_NOESP,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("LMSW"),
        D_PRIVILEGED|D_RARE, 0,
        2, 0x0038FFFF, 0x0030010F, 0x00,
        B_CR0|B_UPD|B_PSEUDO,
        B_INT16|B_NOESP,
        B_NONE,
        B_NONE },

  { T("LOOP"),
        D_JMCX|D_ADDR32, 0,
        1, 0x000000FF, 0x000000E2, 0x00,
        B_ECX|B_UPD|B_PSEUDO,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE },

  { T("LOOPD"),
        D_JMCX|D_ADDR32, 0,
        1, 0x000000FF, 0x000000E2, 0x00,
        B_ECX|B_UPD|B_PSEUDO,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE },

  { T("LOOPW"),
        D_JMCX|D_ADDR16, 0,
        1, 0x000000FF, 0x000000E2, 0x00,
        B_CX|B_UPD|B_PSEUDO,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE },

  { T("LOOPZ"),
        D_JMCX|D_ADDR32|D_COND, 0,
        1, 0x000000FF, 0x000000E1, 0x00,
        B_ECX|B_UPD|B_PSEUDO,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE },

  { T("LOOPDZ"),
        D_JMCX|D_ADDR32|D_COND, 0,
        1, 0x000000FF, 0x000000E1, 0x00,
        B_ECX|B_UPD|B_PSEUDO,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE },

  { T("LOOPWZ"),
        D_JMCX|D_ADDR16|D_COND, 0,
        1, 0x000000FF, 0x000000E1, 0x00,
        B_CX|B_UPD|B_PSEUDO,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE },

  { T("LOOPE"),
        D_JMCX|D_ADDR32|D_COND, 0,
        1, 0x000000FF, 0x000000E1, 0x00,
        B_ECX|B_UPD|B_PSEUDO,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE },

  { T("LOOPDE"),
        D_JMCX|D_ADDR32|D_COND, 0,
        1, 0x000000FF, 0x000000E1, 0x00,
        B_ECX|B_UPD|B_PSEUDO,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE },

  { T("LOOPWE"),
        D_JMCX|D_ADDR16|D_COND, 0,
        1, 0x000000FF, 0x000000E1, 0x00,
        B_CX|B_UPD|B_PSEUDO,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE },

  { T("LOOPNZ"),
        D_JMCX|D_ADDR32|D_COND, 0,
        1, 0x000000FF, 0x000000E0, 0x00,
        B_ECX|B_UPD|B_PSEUDO,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE },

  { T("LOOPDNZ"),
        D_JMCX|D_ADDR32|D_COND, 0,
        1, 0x000000FF, 0x000000E0, 0x00,
        B_ECX|B_UPD|B_PSEUDO,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE },

  { T("LOOPWNZ"),
        D_JMCX|D_ADDR16|D_COND, 0,
        1, 0x000000FF, 0x000000E0, 0x00,
        B_CX|B_UPD|B_PSEUDO,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE },

  { T("LOOPNE"),
        D_JMCX|D_ADDR32|D_COND, 0,
        1, 0x000000FF, 0x000000E0, 0x00,
        B_ECX|B_UPD|B_PSEUDO,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE },

  { T("LOOPDNE"),
        D_JMCX|D_ADDR32|D_COND, 0,
        1, 0x000000FF, 0x000000E0, 0x00,
        B_ECX|B_UPD|B_PSEUDO,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE },

  { T("LOOPWNE"),
        D_JMCX|D_ADDR16|D_COND, 0,
        1, 0x000000FF, 0x000000E0, 0x00,
        B_CX|B_UPD|B_PSEUDO,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE },

  { T("LSL"),
        D_SYS|D_FLAGZ|D_RARE, 0,
        2, 0x0000FFFF, 0x0000030F, 0x00,
        B_REG|B_NOESP|B_CHG,
        B_INT|B_BINARY|B_NOESP,
        B_NONE,
        B_NONE },

  { T("LTR"),
        D_PRIVILEGED|D_RARE, 0,
        2, 0x0038FFFF, 0x0018000F, 0x00,
        B_INT16|B_NOESP,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("MOV"),
        D_MOV|D_SIZE01, 0,
        1, 0x000000FE, 0x00000088, 0x00,
        B_INT|B_CHG,
        B_REG,
        B_NONE,
        B_NONE },

  { T("MOV"),
        D_MOV|D_SIZE01, 0,
        1, 0x000000FE, 0x0000008A, 0x00,
        B_REG|B_CHG,
        B_INT,
        B_NONE,
        B_NONE },

  { T("MOV"),
        D_CMD|D_REGISTER|D_RARE, 0,
        1, 0x0000C0FF, 0x0000C08C, 0x00,
        B_INT|B_REGISTER|B_NOESP|B_CHG,
        B_SEG,
        B_NONE,
        B_NONE },

  { T("MOV"),
        D_CMD|D_MEMORY|D_RARE, 0,
        1, 0x000000FF, 0x0000008C, 0x00,
        B_INT16|B_MEMORY|B_CHG,
        B_SEG,
        B_NONE,
        B_NONE },

  { T("MOV"),
        D_CMD|D_RARE, 0,
        1, 0x000000FF, 0x0000008E, 0x00,
        B_SEGNOCS|B_CHG,
        B_INT|B_REGISTER|B_NOESP,
        B_NONE,
        B_NONE },

  { T("MOV"),
        D_CMD|D_RARE, 0,
        1, 0x000000FF, 0x0000008E, 0x00,
        B_SEGNOCS|B_CHG,
        B_INT16|B_MEMORY|B_NOESP,
        B_NONE,
        B_NONE },

  { T("MOV"),
        D_MOV|D_SIZE01, 0,
        1, 0x000000FE, 0x000000A0, 0x00,
        B_ACC|B_CHG,
        B_IMMINT,
        B_NONE,
        B_NONE },

  { T("MOV"),
        D_MOV|D_SIZE01, 0,
        1, 0x000000FE, 0x000000A2, 0x00,
        B_IMMINT|B_CHG,
        B_ACC,
        B_NONE,
        B_NONE },

  { T("MOV"),
        D_MOV, 0,
        1, 0x000000F8, 0x000000B0, 0x00,
        B_REGCMD8|B_CHG,
        B_CONST8,
        B_NONE,
        B_NONE },

  { T("MOV"),
        D_MOV, 0,
        1, 0x000000F8, 0x000000B8, 0x00,
        B_REGCMD|B_NOESP|B_CHG,
        B_CONST,
        B_NONE,
        B_NONE },

  { T("MOV"),
        D_MOV|D_SIZE01, 0,
        1, 0x000038FE, 0x000000C6, 0x00,
        B_INT|B_NOESP|B_SHOWSIZE|B_CHG,
        B_CONST,
        B_NONE,
        B_NONE },

  { T("MOV"),
        D_PRIVILEGED|D_RARE, 0,
        2, 0x0000FFFF, 0x0000220F, 0x00,
        B_CR|B_CHG,
        B_INT32|B_BINARY|B_REGONLY|B_NOESP,
        B_NONE,
        B_NONE },

  { T("MOV"),
        D_PRIVILEGED|D_RARE, 0,
        2, 0x0000FFFF, 0x0000200F, 0x00,
        B_INT32|B_BINARY|B_REGONLY|B_NOESP|B_CHG,
        B_CR,
        B_NONE,
        B_NONE },

  { T("MOV"),
        D_PRIVILEGED|D_RARE, 0,
        2, 0x0000FFFF, 0x0000230F, 0x00,
        B_DR|B_CHG,
        B_INT32|B_BINARY|B_REGONLY|B_NOESP,
        B_NONE,
        B_NONE },

  { T("MOV"),
        D_PRIVILEGED|D_RARE, 0,
        2, 0x0000FFFF, 0x0000210F, 0x00,
        B_INT32|B_BINARY|B_REGONLY|B_NOESP|B_CHG,
        B_DR,
        B_NONE,
        B_NONE },

  { T("MOVSX"),
        D_MOV, 0,
        2, 0x0000FFFF, 0x0000BE0F, 0x00,
        B_REG|B_NOESP|B_CHG,
        B_INT8|B_SIGNED|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("MOVSX"),
        D_MOV, 0,
        2, 0x0000FFFF, 0x0000BF0F, 0x00,
        B_REG32|B_NOESP|B_CHG,
        B_INT16|B_SIGNED|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("MOVZX"),
        D_MOV, 0,
        2, 0x0000FFFF, 0x0000B60F, 0x00,
        B_REG|B_NOESP|B_CHG,
        B_INT8|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("MOVZX"),
        D_MOV, 0,
        2, 0x0000FFFF, 0x0000B70F, 0x00,
        B_REG32|B_NOESP|B_CHG,
        B_INT16|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("MUL"),
        D_CMD|D_ALLFLAGS, 0,
        1, 0x000038FF, 0x000020F6, 0x00,
        B_AX|B_UPD|B_PSEUDO,
        B_AL|B_PSEUDO,
        B_INT8|B_UNSIGNED|B_SHOWSIZE,
        B_NONE },

  { T("MUL"),
        D_CMD|D_ALLFLAGS, 0,
        1, 0x000038FF, 0x000020F7, 0x00,
        B_DXEDX|B_CHG|B_PSEUDO,
        B_ACC|B_UPD|B_PSEUDO,
        B_INT1632|B_UNSIGNED|B_NOESP|B_SHOWSIZE,
        B_NONE },

  { T("NEG"),
        D_CMD|D_SIZE01|D_LOCKABLE|D_ALLFLAGS, DX_JZ|DX_JC,
        1, 0x000038FE, 0x000018F6, 0x00,
        B_INT|B_NOESP|B_SHOWSIZE|B_UPD,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("NOT"),
        D_CMD|D_SIZE01|D_LOCKABLE, 0,
        1, 0x000038FE, 0x000010F6, 0x00,
        B_INT|B_NOESP|B_SHOWSIZE|B_UPD,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("OR"),
        D_CMD|D_SIZE01|D_ALLFLAGS, DX_JZ|DX_JB,
        1, 0x000000FE, 0x0000000C, 0x00,
        B_ACC|B_BINARY|B_UPD,
        B_CONST|B_BINARY,
        B_NONE,
        B_NONE },

  { T("OR"),
        D_CMD|D_SIZE01|D_LOCKABLE|D_ALLFLAGS, DX_JZ|DX_JB,
        1, 0x000038FE, 0x00000880, 0x00,
        B_INT|B_BINARY|B_NOESP|B_SHOWSIZE|B_UPD,
        B_CONST|B_BINARY,
        B_NONE,
        B_NONE },

  { T("OR"),
        D_CMD|D_SIZE01|D_LOCKABLE|D_ALLFLAGS, DX_JZ|DX_JB,
        1, 0x000038FE, 0x00000882, 0x00,
        B_INT|B_BINARY|B_NOESP|B_SHOWSIZE|B_UPD,
        B_SXTCONST|B_BINARY,
        B_NONE,
        B_NONE },

  { T("OR"),
        D_CMD|D_SIZE01|D_LOCKABLE|D_ALLFLAGS, DX_JZ|DX_JB,
        1, 0x000000FE, 0x00000008, 0x00,
        B_INT|B_BINARY|B_NOESP|B_UPD,
        B_REG|B_BINARY,
        B_NONE,
        B_NONE },

  { T("OR"),
        D_CMD|D_SIZE01|D_ALLFLAGS, DX_JZ|DX_JB,
        1, 0x000000FE, 0x0000000A, 0x00,
        B_REG|B_BINARY|B_NOESP|B_UPD,
        B_INT|B_BINARY,
        B_NONE,
        B_NONE },

  { T("OUT"),
        D_IO|D_SIZE01|D_RARE, 0,
        1, 0x000000FE, 0x000000E6, 0x00,
        B_CONST8|B_PORT,
        B_ACC,
        B_NONE,
        B_NONE },

  { T("OUT"),
        D_IO|D_SIZE01|D_RARE, 0,
        1, 0x000000FE, 0x000000EE, 0x00,
        B_DXPORT|B_PORT,
        B_ACC,
        B_NONE,
        B_NONE },

  { T("POP"),
        D_POP|D_CHGESP, 0,
        1, 0x000038FF, 0x0000008F, 0x00,
        B_INT|B_SHOWSIZE|B_CHG,
        B_STKTOP|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("POP"),
        D_POP|D_CHGESP, 0,
        1, 0x000000F8, 0x00000058, 0x00,
        B_REGCMD|B_CHG,
        B_STKTOP|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("POP"),
        D_POP|D_CHGESP|D_RARE, 0,
        1, 0x000000FF, 0x0000001F, 0x00,
        B_SEGDS|B_CHG,
        B_STKTOP|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("POP"),
        D_POP|D_CHGESP|D_RARE, 0,
        1, 0x000000FF, 0x00000007, 0x00,
        B_SEGES|B_CHG,
        B_STKTOP|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("POP"),
        D_POP|D_CHGESP|D_RARE, DX_JB,
        1, 0x000000FF, 0x00000017, 0x00,
        B_SEGSS|B_CHG,
        B_STKTOP|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("POP"),
        D_POP|D_CHGESP|D_RARE, 0,
        2, 0x0000FFFF, 0x0000A10F, 0x00,
        B_SEGFS|B_CHG,
        B_STKTOP|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("POP"),
        D_POP|D_CHGESP|D_RARE, 0,
        2, 0x0000FFFF, 0x0000A90F, 0x00,
        B_SEGGS|B_CHG,
        B_STKTOP|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("POPA*"),
        D_CMD|D_CHGESP|D_WILDCARD, 0,
        1, 0x000000FF, 0x00000061, 0x00,
        B_STKTOPA|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("POPF*"),
        D_POP|D_ALLFLAGS|D_CHGESP|D_WILDCARD, 0,
        1, 0x000000FF, 0x0000009D, 0x00,
        B_EFL|B_CHG|B_PSEUDO,
        B_STKTOPEFL|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("PUSH"),
        D_PUSH|D_CHGESP, 0,
        1, 0x000038FF, 0x000030FF, 0x00,
        B_INT|B_SHOWSIZE,
        B_PUSH|B_CHG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("PUSH"),
        D_PUSH|D_CHGESP, 0,
        1, 0x000000F8, 0x00000050, 0x00,
        B_REGCMD,
        B_PUSH|B_CHG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("PUSH"),
        D_PUSH|D_CHGESP, 0,
        1, 0x000000FF, 0x0000006A, 0x00,
        B_SXTCONST|B_SHOWSIZE,
        B_PUSH|B_CHG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("PUSH"),
        D_PUSH|D_CHGESP, 0,
        1, 0x000000FF, 0x00000068, 0x00,
        B_CONSTL|B_SHOWSIZE,
        B_PUSH|B_CHG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("PUSH"),
        D_PUSH|D_CHGESP|D_RARE, 0,
        1, 0x000000FF, 0x0000000E, 0x00,
        B_SEGCS,
        B_PUSH|B_CHG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("PUSH"),
        D_PUSH|D_CHGESP|D_RARE, 0,
        1, 0x000000FF, 0x00000016, 0x00,
        B_SEGSS,
        B_PUSH|B_CHG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("PUSH"),
        D_PUSH|D_CHGESP|D_RARE, 0,
        1, 0x000000FF, 0x0000001E, 0x00,
        B_SEGDS,
        B_PUSH|B_CHG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("PUSH"),
        D_PUSH|D_CHGESP|D_RARE, 0,
        1, 0x000000FF, 0x00000006, 0x00,
        B_SEGES,
        B_PUSH|B_CHG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("PUSH"),
        D_PUSH|D_CHGESP|D_RARE, 0,
        2, 0x0000FFFF, 0x0000A00F, 0x00,
        B_SEGFS,
        B_PUSH|B_CHG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("PUSH"),
        D_PUSH|D_CHGESP|D_RARE, 0,
        2, 0x0000FFFF, 0x0000A80F, 0x00,
        B_SEGGS,
        B_PUSH|B_CHG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("PUSHA*"),
        D_CMD|D_CHGESP|D_WILDCARD, 0,
        1, 0x000000FF, 0x00000060, 0x00,
        B_PUSHA|B_CHG|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("PUSHF*"),
        D_PUSH|D_CHGESP|D_WILDCARD, DX_JB,
        1, 0x000000FF, 0x0000009C, 0x00,
        B_EFL|B_PSEUDO,
        B_PUSH|B_CHG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("RCL"),
        D_CMD|D_SIZE01|D_FLAGSCO|D_USESCARRY, DX_JC,
        1, 0x000038FE, 0x000010D0, 0x00,
        B_INT|B_NOESP|B_SHOWSIZE|B_UPD,
        B_1|B_SHIFTCNT,
        B_NONE,
        B_NONE },

  { T("RCL"),
        D_CMD|D_SIZE01|D_FLAGSCO|D_USESCARRY, DX_JC,
        1, 0x000038FE, 0x000010D2, 0x00,
        B_INT|B_NOESP|B_SHOWSIZE|B_UPD,
        B_CL|B_SHIFTCNT,
        B_NONE,
        B_NONE },

  { T("RCL"),
        D_CMD|D_SIZE01|D_FLAGSCO|D_USESCARRY, DX_JC,
        1, 0x000038FE, 0x000010C0, 0x00,
        B_INT|B_NOESP|B_SHOWSIZE|B_UPD,
        B_CONST8|B_SHIFTCNT,
        B_NONE,
        B_NONE },

  { T("RCR"),
        D_CMD|D_SIZE01|D_FLAGSCO|D_USESCARRY, DX_JC,
        1, 0x000038FE, 0x000018D0, 0x00,
        B_INT|B_NOESP|B_SHOWSIZE|B_UPD,
        B_1|B_SHIFTCNT,
        B_NONE,
        B_NONE },

  { T("RCR"),
        D_CMD|D_SIZE01|D_FLAGSCO|D_USESCARRY, DX_JC,
        1, 0x000038FE, 0x000018D2, 0x00,
        B_INT|B_NOESP|B_SHOWSIZE|B_UPD,
        B_CL|B_SHIFTCNT,
        B_NONE,
        B_NONE },

  { T("RCR"),
        D_CMD|D_SIZE01|D_FLAGSCO|D_USESCARRY, DX_JC,
        1, 0x000038FE, 0x000018C0, 0x00,
        B_INT|B_NOESP|B_SHOWSIZE|B_UPD,
        B_CONST8|B_SHIFTCNT,
        B_NONE,
        B_NONE },

  { T("ROL"),
        D_CMD|D_SIZE01|D_FLAGSCO, DX_JC,
        1, 0x000038FE, 0x000000D0, 0x00,
        B_INT|B_NOESP|B_SHOWSIZE|B_UPD,
        B_1|B_SHIFTCNT,
        B_NONE,
        B_NONE },

  { T("ROL"),
        D_CMD|D_SIZE01|D_FLAGSCO, DX_JC,
        1, 0x000038FE, 0x000000D2, 0x00,
        B_INT|B_NOESP|B_SHOWSIZE|B_UPD,
        B_CL|B_SHIFTCNT,
        B_NONE,
        B_NONE },

  { T("ROL"),
        D_CMD|D_SIZE01|D_FLAGSCO, DX_JC,
        1, 0x000038FE, 0x000000C0, 0x00,
        B_INT|B_NOESP|B_SHOWSIZE|B_UPD,
        B_CONST8|B_SHIFTCNT,
        B_NONE,
        B_NONE },

  { T("ROR"),
        D_CMD|D_SIZE01|D_FLAGSCO, DX_JC,
        1, 0x000038FE, 0x000008D0, 0x00,
        B_INT|B_NOESP|B_SHOWSIZE|B_UPD,
        B_1|B_SHIFTCNT,
        B_NONE,
        B_NONE },

  { T("ROR"),
        D_CMD|D_SIZE01|D_FLAGSCO, DX_JC,
        1, 0x000038FE, 0x000008D2, 0x00,
        B_INT|B_NOESP|B_SHOWSIZE|B_UPD,
        B_CL|B_SHIFTCNT,
        B_NONE,
        B_NONE },

  { T("ROR"),
        D_CMD|D_SIZE01|D_FLAGSCO, DX_JC,
        1, 0x000038FE, 0x000008C0, 0x00,
        B_INT|B_NOESP|B_SHOWSIZE|B_UPD,
        B_CONST8|B_SHIFTCNT,
        B_NONE,
        B_NONE },

  { T("RDMSR"),
        D_PRIVILEGED|D_RARE, 0,
        2, 0x0000FFFF, 0x0000320F, 0x00,
        B_EDX|B_BINARY|B_CHG|B_PSEUDO,
        B_EAX|B_BINARY|B_CHG|B_PSEUDO,
        B_ECX|B_PSEUDO,
        B_NONE },

  { T("RDPMC"),
        D_SYS|D_RARE, 0,
        2, 0x0000FFFF, 0x0000330F, 0x00,
        B_EDX|B_BINARY|B_CHG|B_PSEUDO,
        B_EAX|B_BINARY|B_CHG|B_PSEUDO,
        B_ECX|B_PSEUDO,
        B_NONE },

  { T("RDTSC"),
        D_SYS|D_RARE, 0,
        2, 0x0000FFFF, 0x0000310F, 0x00,
        B_EDX|B_BINARY|B_CHG|B_PSEUDO,
        B_EAX|B_BINARY|B_CHG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("RDTSCP"),
        D_SYS|D_RARE, 0,
        3, 0x00FFFFFF, 0x00F9010F, 0x00,
        B_EDX|B_BINARY|B_CHG|B_PSEUDO,
        B_EAX|B_BINARY|B_CHG|B_PSEUDO,
        B_ECX|B_BINARY|B_CHG|B_PSEUDO,
        B_NONE },

  { T("RETN"),
        D_RET|D_NOREP|D_CHGESP, DX_RETN,
        1, 0x000000FF, 0x000000C3, 0x00,
        B_STKTOP|B_JMPCALL|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("RET"),
        D_RET|D_NOREP|D_CHGESP, 0,
        1, 0x000000FF, 0x000000C3, 0x00,
        B_STKTOP|B_JMPCALL|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("RETN"),
        D_RET|D_MUSTREP|D_CHGESP, DX_RETN,
        1, 0x000000FF, 0x000000C3, 0x00,
        B_STKTOP|B_JMPCALL|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("RET"),
        D_RET|D_MUSTREP|D_CHGESP, 0,
        1, 0x000000FF, 0x000000C3, 0x00,
        B_STKTOP|B_JMPCALL|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("RETN"),
        D_RET|D_CHGESP, DX_RETN,
        1, 0x000000FF, 0x000000C2, 0x00,
        B_STKTOP|B_JMPCALL|B_PSEUDO,
        B_CONST16|B_STACKINC,
        B_NONE,
        B_NONE },

  { T("RET"),
        D_RET|D_CHGESP, 0,
        1, 0x000000FF, 0x000000C2, 0x00,
        B_STKTOP|B_JMPCALL|B_PSEUDO,
        B_CONST16|B_STACKINC,
        B_NONE,
        B_NONE },

  { T("RETF"),
        D_RETFAR|D_CHGESP|D_RARE, 0,
        1, 0x000000FF, 0x000000CB, 0x00,
        B_STKTOPFAR|B_JMPCALLFAR|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("RETF"),
        D_RETFAR|D_CHGESP|D_RARE, 0,
        1, 0x000000FF, 0x000000CA, 0x00,
        B_STKTOPFAR|B_JMPCALLFAR|B_PSEUDO,
        B_CONST16|B_STACKINC,
        B_NONE,
        B_NONE },

  { T("RSM"),
        D_PRIVILEGED|D_RARE, 0,
        2, 0x0000FFFF, 0x0000AA0F, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("SAHF"),
        D_CMD|D_ALLFLAGS, 0,
        1, 0x000000FF, 0x0000009E, 0x00,
        B_AH|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("SHL"),
        D_CMD|D_SIZE01|D_ALLFLAGS, DX_JZ|DX_JC,
        1, 0x000038FE, 0x000020D0, 0x00,
        B_INT|B_NOESP|B_SHOWSIZE|B_UPD,
        B_1|B_SHIFTCNT,
        B_NONE,
        B_NONE },

  { T("SHL"),
        D_CMD|D_SIZE01|D_ALLFLAGS, DX_JZ|DX_JC,
        1, 0x000038FE, 0x000020D2, 0x00,
        B_INT|B_NOESP|B_SHOWSIZE|B_UPD,
        B_CL|B_SHIFTCNT,
        B_NONE,
        B_NONE },

  { T("SHL"),
        D_CMD|D_SIZE01|D_ALLFLAGS, DX_JZ|DX_JC,
        1, 0x000038FE, 0x000020C0, 0x00,
        B_INT|B_NOESP|B_SHOWSIZE|B_UPD,
        B_CONST8|B_SHIFTCNT,
        B_NONE,
        B_NONE },

  { T("SAL"),
        D_CMD|D_SIZE01|D_ALLFLAGS, DX_JZ|DX_JC,
        1, 0x000038FE, 0x000020D0, 0x00,
        B_INT|B_NOESP|B_SHOWSIZE|B_UPD,
        B_1|B_SHIFTCNT,
        B_NONE,
        B_NONE },

  { T("SAL"),
        D_CMD|D_SIZE01|D_ALLFLAGS, DX_JZ|DX_JC,
        1, 0x000038FE, 0x000020D2, 0x00,
        B_INT|B_NOESP|B_SHOWSIZE|B_UPD,
        B_CL|B_SHIFTCNT,
        B_NONE,
        B_NONE },

  { T("SAL"),
        D_CMD|D_SIZE01|D_ALLFLAGS, DX_JZ|DX_JC,
        1, 0x000038FE, 0x000020C0, 0x00,
        B_INT|B_NOESP|B_SHOWSIZE|B_UPD,
        B_CONST8|B_SHIFTCNT,
        B_NONE,
        B_NONE },

  { T("SAL"),
        D_CMD|D_SIZE01|D_ALLFLAGS|D_UNDOC, DX_JZ|DX_JC,
        1, 0x000038FE, 0x000030D0, 0x00,
        B_INT|B_NOESP|B_SHOWSIZE|B_UPD,
        B_1|B_SHIFTCNT,
        B_NONE,
        B_NONE },

  { T("SAL"),
        D_CMD|D_SIZE01|D_ALLFLAGS|D_UNDOC, DX_JZ|DX_JC,
        1, 0x000038FE, 0x000030D2, 0x00,
        B_INT|B_NOESP|B_SHOWSIZE|B_UPD,
        B_CL|B_SHIFTCNT,
        B_NONE,
        B_NONE },

  { T("SAL"),
        D_CMD|D_SIZE01|D_ALLFLAGS|D_UNDOC, DX_JZ|DX_JC,
        1, 0x000038FE, 0x000030C0, 0x00,
        B_INT|B_NOESP|B_SHOWSIZE|B_UPD,
        B_CONST8|B_SHIFTCNT,
        B_NONE,
        B_NONE },

  { T("SALC"),
        D_CMD|D_ALLFLAGS|D_UNDOC, DX_JZ|DX_JC,
        1, 0x000000FF, 0x000000D6, 0x00,
        B_AL|B_UPD|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("SHR"),
        D_CMD|D_SIZE01|D_ALLFLAGS, DX_JZ|DX_JC,
        1, 0x000038FE, 0x000028D0, 0x00,
        B_INT|B_NOESP|B_SHOWSIZE|B_UPD,
        B_1|B_SHIFTCNT,
        B_NONE,
        B_NONE },

  { T("SHR"),
        D_CMD|D_SIZE01|D_ALLFLAGS, DX_JZ|DX_JC,
        1, 0x000038FE, 0x000028D2, 0x00,
        B_INT|B_NOESP|B_SHOWSIZE|B_UPD,
        B_CL|B_SHIFTCNT,
        B_NONE,
        B_NONE },

  { T("SHR"),
        D_CMD|D_SIZE01|D_ALLFLAGS, DX_JZ|DX_JC,
        1, 0x000038FE, 0x000028C0, 0x00,
        B_INT|B_NOESP|B_SHOWSIZE|B_UPD,
        B_CONST8|B_SHIFTCNT,
        B_NONE,
        B_NONE },

  { T("SAR"),
        D_CMD|D_SIZE01|D_ALLFLAGS, DX_JZ|DX_JC,
        1, 0x000038FE, 0x000038D0, 0x00,
        B_INT|B_NOESP|B_SHOWSIZE|B_UPD,
        B_1|B_SHIFTCNT,
        B_NONE,
        B_NONE },

  { T("SAR"),
        D_CMD|D_SIZE01|D_ALLFLAGS, DX_JZ|DX_JC,
        1, 0x000038FE, 0x000038D2, 0x00,
        B_INT|B_NOESP|B_SHOWSIZE|B_UPD,
        B_CL|B_SHIFTCNT,
        B_NONE,
        B_NONE },

  { T("SAR"),
        D_CMD|D_SIZE01|D_ALLFLAGS, DX_JZ|DX_JC,
        1, 0x000038FE, 0x000038C0, 0x00,
        B_INT|B_NOESP|B_SHOWSIZE|B_UPD,
        B_CONST8|B_SHIFTCNT,
        B_NONE,
        B_NONE },

  { T("SBB"),
        D_CMD|D_SIZE01|D_ALLFLAGS|D_USESCARRY, DX_JZ|DX_JB,
        1, 0x000000FE, 0x0000001C, 0x00,
        B_ACC|B_UPD,
        B_CONST,
        B_NONE,
        B_NONE },

  { T("SBB"),
        D_CMD|D_SIZE01|D_LOCKABLE|D_ALLFLAGS|D_USESCARRY, DX_JZ|DX_JB,
        1, 0x000038FE, 0x00001880, 0x00,
        B_INT|B_SHOWSIZE|B_UPD,
        B_CONST,
        B_NONE,
        B_NONE },

  { T("SBB"),
        D_CMD|D_SIZE01|D_LOCKABLE|D_ALLFLAGS|D_USESCARRY, DX_JZ|DX_JB,
        1, 0x000038FE, 0x00001882, 0x00,
        B_INT|B_SHOWSIZE|B_UPD,
        B_SXTCONST,
        B_NONE,
        B_NONE },

  { T("SBB"),
        D_CMD|D_SIZE01|D_LOCKABLE|D_ALLFLAGS|D_USESCARRY, DX_JZ|DX_JB,
        1, 0x000000FE, 0x00000018, 0x00,
        B_INT|B_UPD,
        B_REG,
        B_NONE,
        B_NONE },

  { T("SBB"),
        D_CMD|D_SIZE01|D_ALLFLAGS|D_USESCARRY, DX_JZ|DX_JB,
        1, 0x000000FE, 0x0000001A, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("SETO"),
        D_SETC|D_COND, 0,
        2, 0x0000FFFF, 0x0000900F, 0x00,
        B_INT8|B_CHG,
        B_ANYREG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SETNO"),
        D_SETC|D_COND, 0,
        2, 0x0000FFFF, 0x0000910F, 0x00,
        B_INT8|B_CHG,
        B_ANYREG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SETB"),
        D_SETC|D_COND|D_USESCARRY, DX_JB,
        2, 0x0000FFFF, 0x0000920F, 0x00,
        B_INT8|B_CHG,
        B_ANYREG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SETC"),
        D_SETC|D_COND|D_USESCARRY, DX_JC,
        2, 0x0000FFFF, 0x0000920F, 0x00,
        B_INT8|B_CHG,
        B_ANYREG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SETNAE"),
        D_SETC|D_COND|D_USESCARRY, DX_JB,
        2, 0x0000FFFF, 0x0000920F, 0x00,
        B_INT8|B_CHG,
        B_ANYREG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SETAE"),
        D_SETC|D_COND|D_USESCARRY, DX_JB,
        2, 0x0000FFFF, 0x0000930F, 0x00,
        B_INT8|B_CHG,
        B_ANYREG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SETNB"),
        D_SETC|D_COND|D_USESCARRY, DX_JB,
        2, 0x0000FFFF, 0x0000930F, 0x00,
        B_INT8|B_CHG,
        B_ANYREG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SETNC"),
        D_SETC|D_COND|D_USESCARRY, DX_JC,
        2, 0x0000FFFF, 0x0000930F, 0x00,
        B_INT8|B_CHG,
        B_ANYREG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SETE"),
        D_SETC|D_COND, DX_JE,
        2, 0x0000FFFF, 0x0000940F, 0x00,
        B_INT8|B_CHG,
        B_ANYREG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SETZ"),
        D_SETC|D_COND, DX_JZ,
        2, 0x0000FFFF, 0x0000940F, 0x00,
        B_INT8|B_CHG,
        B_ANYREG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SETNE"),
        D_SETC|D_COND, DX_JE,
        2, 0x0000FFFF, 0x0000950F, 0x00,
        B_INT8|B_CHG,
        B_ANYREG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SETNZ"),
        D_SETC|D_COND, DX_JZ,
        2, 0x0000FFFF, 0x0000950F, 0x00,
        B_INT8|B_CHG,
        B_ANYREG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SETBE"),
        D_SETC|D_COND|D_USESCARRY, 0,
        2, 0x0000FFFF, 0x0000960F, 0x00,
        B_INT8|B_CHG,
        B_ANYREG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SETNA"),
        D_SETC|D_COND|D_USESCARRY, 0,
        2, 0x0000FFFF, 0x0000960F, 0x00,
        B_INT8|B_CHG,
        B_ANYREG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SETA"),
        D_SETC|D_COND|D_USESCARRY, 0,
        2, 0x0000FFFF, 0x0000970F, 0x00,
        B_INT8|B_CHG,
        B_ANYREG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SETNBE"),
        D_SETC|D_COND|D_USESCARRY, 0,
        2, 0x0000FFFF, 0x0000970F, 0x00,
        B_INT8|B_CHG,
        B_ANYREG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SETS"),
        D_SETC|D_COND, 0,
        2, 0x0000FFFF, 0x0000980F, 0x00,
        B_INT8|B_CHG,
        B_ANYREG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SETNS"),
        D_SETC|D_COND, 0,
        2, 0x0000FFFF, 0x0000990F, 0x00,
        B_INT8|B_CHG,
        B_ANYREG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SETPE"),
        D_SETC|D_COND|D_RARE, 0,
        2, 0x0000FFFF, 0x00009A0F, 0x00,
        B_INT8|B_CHG,
        B_ANYREG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SETP"),
        D_SETC|D_COND|D_RARE, 0,
        2, 0x0000FFFF, 0x00009A0F, 0x00,
        B_INT8|B_CHG,
        B_ANYREG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SETPO"),
        D_SETC|D_COND|D_RARE, 0,
        2, 0x0000FFFF, 0x00009B0F, 0x00,
        B_INT8|B_CHG,
        B_ANYREG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SETNP"),
        D_SETC|D_COND|D_RARE, 0,
        2, 0x0000FFFF, 0x00009B0F, 0x00,
        B_INT8|B_CHG,
        B_ANYREG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SETL"),
        D_SETC|D_COND, 0,
        2, 0x0000FFFF, 0x00009C0F, 0x00,
        B_INT8|B_CHG,
        B_ANYREG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SETNGE"),
        D_SETC|D_COND, 0,
        2, 0x0000FFFF, 0x00009C0F, 0x00,
        B_INT8|B_CHG,
        B_ANYREG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SETGE"),
        D_SETC|D_COND, 0,
        2, 0x0000FFFF, 0x00009D0F, 0x00,
        B_INT8|B_CHG,
        B_ANYREG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SETNL"),
        D_SETC|D_COND, 0,
        2, 0x0000FFFF, 0x00009D0F, 0x00,
        B_INT8|B_CHG,
        B_ANYREG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SETLE"),
        D_SETC|D_COND, 0,
        2, 0x0000FFFF, 0x00009E0F, 0x00,
        B_INT8|B_CHG,
        B_ANYREG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SETNG"),
        D_SETC|D_COND, 0,
        2, 0x0000FFFF, 0x00009E0F, 0x00,
        B_INT8|B_CHG,
        B_ANYREG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SETG"),
        D_SETC|D_COND, 0,
        2, 0x0000FFFF, 0x00009F0F, 0x00,
        B_INT8|B_CHG,
        B_ANYREG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SETNLE"),
        D_SETC|D_COND, 0,
        2, 0x0000FFFF, 0x00009F0F, 0x00,
        B_INT8|B_CHG,
        B_ANYREG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SGDT"),
        D_SYS|D_MEMORY|D_RARE, 0,
        2, 0x0038FFFF, 0x0000010F, 0x00,
        B_DESCR|B_MEMONLY|B_CHG,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("SIDT"),
        D_SYS|D_MEMORY|D_RARE, 0,
        2, 0x0038FFFF, 0x0008010F, 0x00,
        B_DESCR|B_MEMONLY|B_CHG,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("SHLD"),
        D_CMD|D_ALLFLAGS, DX_JZ|DX_JC,
        2, 0x0000FFFF, 0x0000A40F, 0x00,
        B_INT|B_NOESP|B_UPD,
        B_REG,
        B_CONST8|B_SHIFTCNT,
        B_NONE },

  { T("SHLD"),
        D_CMD|D_ALLFLAGS, DX_JZ|DX_JC,
        2, 0x0000FFFF, 0x0000A50F, 0x00,
        B_INT|B_NOESP|B_UPD,
        B_REG,
        B_CL|B_SHIFTCNT,
        B_NONE },

  { T("SHRD"),
        D_CMD|D_ALLFLAGS, DX_JZ|DX_JC,
        2, 0x0000FFFF, 0x0000AC0F, 0x00,
        B_INT|B_NOESP|B_UPD,
        B_REG,
        B_CONST8|B_SHIFTCNT,
        B_NONE },

  { T("SHRD"),
        D_CMD|D_ALLFLAGS, DX_JZ|DX_JC,
        2, 0x0000FFFF, 0x0000AD0F, 0x00,
        B_INT|B_NOESP|B_UPD,
        B_REG,
        B_CL|B_SHIFTCNT,
        B_NONE },

  { T("SLDT"),
        D_SYS|D_RARE, 0,
        2, 0x0038FFFF, 0x0000000F, 0x00,
        B_INT|B_NOESP|B_CHG,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("SMSW"),
        D_SYS|D_MEMORY|D_RARE, 0,
        2, 0x0038FFFF, 0x0020010F, 0x00,
        B_INT16|B_MEMONLY|B_CHG,
        B_CR0|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SMSW"),
        D_SYS|D_REGISTER|D_RARE, 0,
        2, 0x0038FFFF, 0x0020010F, 0x00,
        B_INT|B_REGONLY|B_NOESP|B_CHG,
        B_CR0|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("STC"),
        D_CMD|D_FLAGC, 0,
        1, 0x000000FF, 0x000000F9, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("STD"),
        D_CMD|D_FLAGD, 0,
        1, 0x000000FF, 0x000000FD, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("STI"),
        D_CMD|D_RARE, 0,
        1, 0x000000FF, 0x000000FB, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("STMXCSR"),
        D_CMD|D_MEMORY, 0,
        2, 0x0038FFFF, 0x0018AE0F, 0x00,
        B_INT32|B_BINARY|B_MEMONLY|B_NOESP|B_SHOWSIZE|B_CHG,
        B_MXCSR|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("STR"),
        D_SYS|D_MEMORY|D_RARE, 0,
        2, 0x0038FFFF, 0x0008000F, 0x00,
        B_INT16|B_MEMONLY|B_CHG,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("STR"),
        D_SYS|D_REGISTER|D_RARE, 0,
        2, 0x0038FFFF, 0x0008000F, 0x00,
        B_INT|B_REGONLY|B_NOESP|B_CHG,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("SUB"),
        D_CMD|D_SIZE01|D_ALLFLAGS, DX_JZ|DX_JB|DX_SUB,
        1, 0x000000FE, 0x0000002C, 0x00,
        B_ACC|B_UPD,
        B_CONST,
        B_NONE,
        B_NONE },

  { T("SUB"),
        D_CMD|D_SIZE01|D_LOCKABLE|D_ALLFLAGS, DX_JZ|DX_JB|DX_SUB,
        1, 0x000038FE, 0x00002880, 0x00,
        B_INT|B_SHOWSIZE|B_UPD,
        B_CONST,
        B_NONE,
        B_NONE },

  { T("SUB"),
        D_CMD|D_SIZE01|D_LOCKABLE|D_ALLFLAGS, DX_JZ|DX_JB|DX_SUB,
        1, 0x000038FE, 0x00002882, 0x00,
        B_INT|B_SHOWSIZE|B_UPD,
        B_SXTCONST,
        B_NONE,
        B_NONE },

  { T("SUB"),
        D_CMD|D_SIZE01|D_LOCKABLE|D_ALLFLAGS, DX_JZ|DX_JB|DX_SUB,
        1, 0x000000FE, 0x00000028, 0x00,
        B_INT|B_UPD,
        B_REG,
        B_NONE,
        B_NONE },

  { T("SUB"),
        D_CMD|D_SIZE01|D_ALLFLAGS, DX_JZ|DX_JB|DX_SUB,
        1, 0x000000FE, 0x0000002A, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("SYSENTER"),
        D_SYS|D_RARE, 0,
        2, 0x0000FFFF, 0x0000340F, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("SYSEXIT"),
        D_SYS|D_ALLFLAGS|D_SUSPICIOUS, 0,
        2, 0x0000FFFF, 0x0000350F, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("TEST"),
        D_TEST|D_SIZE01|D_ALLFLAGS, DX_JZ,
        1, 0x000000FE, 0x000000A8, 0x00,
        B_ACC|B_BINARY,
        B_CONST|B_BINARY,
        B_NONE,
        B_NONE },

  { T("TEST"),
        D_TEST|D_SIZE01|D_ALLFLAGS, DX_JZ,
        1, 0x000038FE, 0x000000F6, 0x00,
        B_INT|B_BINARY|B_SHOWSIZE,
        B_CONST|B_BINARY,
        B_NONE,
        B_NONE },

  { T("TEST"),
        D_TEST|D_SIZE01|D_ALLFLAGS|D_UNDOC, DX_JZ,
        1, 0x000038FE, 0x000008F6, 0x00,
        B_INT|B_BINARY|B_SHOWSIZE,
        B_CONST|B_BINARY,
        B_NONE,
        B_NONE },

  { T("TEST"),
        D_TEST|D_SIZE01|D_ALLFLAGS, DX_JZ,
        1, 0x000000FE, 0x00000084, 0x00,
        B_INT|B_BINARY,
        B_REG|B_BINARY,
        B_NONE,
        B_NONE },

  { T("TEST"),
        D_TEST|D_SIZE01|D_ALLFLAGS, DX_JZ,
        1, 0x000000FE, 0x00000084, 0x00,
        B_REG|B_BINARY,
        B_INT|B_BINARY,
        B_NONE,
        B_NONE },

  { T("UD1"),
        D_CMD|D_UNDOC, 0,
        2, 0x0000FFFF, 0x0000B90F, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("UD2"),
        D_CMD, 0,
        2, 0x0000FFFF, 0x00000B0F, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("VERR"),
        D_CMD|D_FLAGZ|D_RARE, 0,
        2, 0x0038FFFF, 0x0020000F, 0x00,
        B_INT16|B_NOESP,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("VERW"),
        D_CMD|D_FLAGZ|D_RARE, 0,
        2, 0x0038FFFF, 0x0028000F, 0x00,
        B_INT16|B_NOESP,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("WBINVD"),
        D_PRIVILEGED|D_RARE, 0,
        2, 0x0000FFFF, 0x0000090F, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("WRMSR"),
        D_PRIVILEGED|D_RARE, 0,
        2, 0x0000FFFF, 0x0000300F, 0x00,
        B_EDX|B_BINARY|B_PSEUDO,
        B_EAX|B_BINARY|B_PSEUDO,
        B_ECX|B_PSEUDO,
        B_NONE },

  { T("XADD"),
        D_CMD|D_SIZE01|D_LOCKABLE|D_ALLFLAGS, DX_JE|DX_JB,
        2, 0x0000FEFF, 0x0000C00F, 0x00,
        B_INT|B_UPD,
        B_REG|B_CHG,
        B_NONE,
        B_NONE },

  { T("XCHG"),
        D_MOV|D_LOCKABLE, 0,
        1, 0x000000F8, 0x00000090, 0x00,
        B_ACC|B_CHG,
        B_REGCMD|B_CHG,
        B_NONE,
        B_NONE },

  { T("XCHG"),
        D_MOV, 0,
        1, 0x000000F8, 0x00000090, 0x00,
        B_REGCMD|B_CHG,
        B_ACC|B_CHG,
        B_NONE,
        B_NONE },

  { T("XCHG"),
        D_MOV|D_SIZE01|D_LOCKABLE, 0,
        1, 0x000000FE, 0x00000086, 0x00,
        B_INT|B_CHG,
        B_REG|B_CHG,
        B_NONE,
        B_NONE },

  { T("XCHG"),
        D_MOV|D_SIZE01|D_LOCKABLE, 0,
        1, 0x000000FE, 0x00000086, 0x00,
        B_REG|B_CHG,
        B_INT|B_CHG,
        B_NONE,
        B_NONE },

  { T("XLAT"),
        D_CMD, 0,
        1, 0x000000FF, 0x000000D7, 0x00,
        B_AL|B_CHG|B_PSEUDO,
        B_XLATMEM,
        B_NONE,
        B_NONE },

  { T("XLATB"),
        D_CMD, 0,
        1, 0x000000FF, 0x000000D7, 0x00,
        B_AL|B_UPD|B_PSEUDO,
        B_XLATMEM|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("XOR"),
        D_CMD|D_SIZE01|D_ALLFLAGS, DX_JZ,
        1, 0x000000FE, 0x00000034, 0x00,
        B_ACC|B_BINARY|B_UPD,
        B_CONST|B_BINARY,
        B_NONE,
        B_NONE },

  { T("XOR"),
        D_CMD|D_SIZE01|D_LOCKABLE|D_ALLFLAGS, DX_JZ,
        1, 0x000038FE, 0x00003080, 0x00,
        B_INT|B_BINARY|B_NOESP|B_SHOWSIZE|B_UPD,
        B_CONST|B_BINARY,
        B_NONE,
        B_NONE },

  { T("XOR"),
        D_CMD|D_SIZE01|D_LOCKABLE|D_ALLFLAGS, DX_JZ,
        1, 0x000038FE, 0x00003082, 0x00,
        B_INT|B_BINARY|B_NOESP|B_SHOWSIZE|B_UPD,
        B_SXTCONST|B_BINARY,
        B_NONE,
        B_NONE },

  { T("XOR"),
        D_CMD|D_SIZE01|D_LOCKABLE|D_ALLFLAGS, DX_JZ,
        1, 0x000000FE, 0x00000030, 0x00,
        B_INT|B_BINARY|B_UPD,
        B_REG|B_BINARY,
        B_NONE,
        B_NONE },

  { T("XOR"),
        D_CMD|D_SIZE01|D_ALLFLAGS, DX_JZ,
        1, 0x000000FE, 0x00000032, 0x00,
        B_REG|B_BINARY|B_UPD,
        B_INT|B_BINARY,
        B_NONE,
        B_NONE },

  { T("CMPS"),
        D_CMD|D_SIZE01|D_LONGFORM|D_NOREP|D_ALLFLAGS|D_HLADIR, DX_JE|DX_JB,
        1, 0x000000FE, 0x000000A6, 0x00,
        B_STRSRC|B_SHOWSIZE,
        B_STRDEST|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("CMPSB"),
        D_CMD|D_NOREP|D_ALLFLAGS, DX_JE|DX_JB,
        1, 0x000000FF, 0x000000A6, 0x00,
        B_STRSRC8|B_PSEUDO,
        B_STRDEST8|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("CMPS*"),
        D_CMD|D_NOREP|D_ALLFLAGS|D_WILDCARD, DX_JE|DX_JB,
        1, 0x000000FF, 0x000000A7, 0x00,
        B_STRSRC|B_PSEUDO,
        B_STRDEST|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("CMPS"),
        D_STRING|D_SIZE01|D_LONGFORM|D_MUSTREPE|D_ALLFLAGS|D_HLADIR, 0,
        1, 0x000000FE, 0x000000A6, 0x00,
        B_STRSRC|B_SHOWSIZE,
        B_STRDEST|B_SHOWSIZE,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("CMPSB"),
        D_STRING|D_MUSTREPE|D_ALLFLAGS, 0,
        1, 0x000000FF, 0x000000A6, 0x00,
        B_STRSRC8|B_PSEUDO,
        B_STRDEST8|B_PSEUDO,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("CMPS*"),
        D_STRING|D_MUSTREPE|D_ALLFLAGS|D_WILDCARD, 0,
        1, 0x000000FF, 0x000000A7, 0x00,
        B_STRSRC|B_PSEUDO,
        B_STRDEST|B_PSEUDO,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("CMPS"),
        D_STRING|D_SIZE01|D_LONGFORM|D_MUSTREPNE|D_ALLFLAGS|D_HLADIR, 0,
        1, 0x000000FE, 0x000000A6, 0x00,
        B_STRSRC|B_SHOWSIZE,
        B_STRDEST|B_SHOWSIZE,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("CMPSB"),
        D_STRING|D_MUSTREPNE|D_ALLFLAGS, 0,
        1, 0x000000FF, 0x000000A6, 0x00,
        B_STRSRC8|B_PSEUDO,
        B_STRDEST8|B_PSEUDO,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("CMPS*"),
        D_STRING|D_MUSTREPNE|D_ALLFLAGS|D_WILDCARD, 0,
        1, 0x000000FF, 0x000000A7, 0x00,
        B_STRSRC|B_PSEUDO,
        B_STRDEST|B_PSEUDO,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("LODS"),
        D_CMD|D_SIZE01|D_LONGFORM|D_NOREP, 0,
        1, 0x000000FE, 0x000000AC, 0x00,
        B_ACC|B_CHG|B_PSEUDO,
        B_STRSRC|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("LODSB"),
        D_CMD|D_NOREP, 0,
        1, 0x000000FF, 0x000000AC, 0x00,
        B_AL|B_CHG|B_PSEUDO,
        B_STRSRC8|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("LODS*"),
        D_CMD|D_NOREP|D_WILDCARD, 0,
        1, 0x000000FF, 0x000000AD, 0x00,
        B_ACC|B_CHG|B_PSEUDO,
        B_STRSRC|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("LODS"),
        D_STRING|D_SIZE01|D_LONGFORM|D_MUSTREP|D_RARE, 0,
        1, 0x000000FE, 0x000000AC, 0x00,
        B_ACC|B_CHG|B_PSEUDO,
        B_STRSRC|B_SHOWSIZE,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("LODSB"),
        D_STRING|D_MUSTREP|D_RARE, 0,
        1, 0x000000FF, 0x000000AC, 0x00,
        B_AL|B_CHG|B_PSEUDO,
        B_STRSRC8|B_PSEUDO,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("LODS*"),
        D_STRING|D_MUSTREP|D_WILDCARD|D_RARE, 0,
        1, 0x000000FF, 0x000000AD, 0x00,
        B_ACC|B_CHG|B_PSEUDO,
        B_STRSRC|B_PSEUDO,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("LODS"),
        D_STRING|D_SIZE01|D_LONGFORM|D_MUSTREPNE|D_UNDOC, 0,
        1, 0x000000FE, 0x000000AC, 0x00,
        B_ACC|B_CHG|B_PSEUDO,
        B_STRSRC|B_SHOWSIZE,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("LODSB"),
        D_STRING|D_MUSTREPNE|D_UNDOC, 0,
        1, 0x000000FF, 0x000000AC, 0x00,
        B_AL|B_CHG|B_PSEUDO,
        B_STRSRC8|B_PSEUDO,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("LODS*"),
        D_STRING|D_MUSTREPNE|D_WILDCARD|D_UNDOC, 0,
        1, 0x000000FF, 0x000000AD, 0x00,
        B_ACC|B_CHG|B_PSEUDO,
        B_STRSRC|B_PSEUDO,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("MOVS"),
        D_CMD|D_SIZE01|D_LONGFORM|D_NOREP, 0,
        1, 0x000000FE, 0x000000A4, 0x00,
        B_STRDEST|B_SHOWSIZE|B_CHG,
        B_STRSRC|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("MOVSB"),
        D_CMD|D_NOREP, 0,
        1, 0x000000FF, 0x000000A4, 0x00,
        B_STRDEST8|B_CHG|B_PSEUDO,
        B_STRSRC8|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("MOVS*"),
        D_CMD|D_NOREP|D_WILDCARD, 0,
        1, 0x000000FF, 0x000000A5, 0x00,
        B_STRDEST|B_CHG|B_PSEUDO,
        B_STRSRC|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("MOVS"),
        D_STRING|D_SIZE01|D_LONGFORM|D_MUSTREP, 0,
        1, 0x000000FE, 0x000000A4, 0x00,
        B_STRDEST|B_SHOWSIZE|B_CHG,
        B_STRSRC|B_SHOWSIZE,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("MOVSB"),
        D_STRING|D_MUSTREP, 0,
        1, 0x000000FF, 0x000000A4, 0x00,
        B_STRDEST8|B_CHG|B_PSEUDO,
        B_STRSRC8|B_PSEUDO,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("MOVS*"),
        D_STRING|D_MUSTREP|D_WILDCARD, 0,
        1, 0x000000FF, 0x000000A5, 0x00,
        B_STRDEST|B_CHG|B_PSEUDO,
        B_STRSRC|B_PSEUDO,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("MOVS"),
        D_STRING|D_SIZE01|D_LONGFORM|D_MUSTREPNE|D_UNDOC, 0,
        1, 0x000000FE, 0x000000A4, 0x00,
        B_STRDEST|B_SHOWSIZE|B_CHG,
        B_STRSRC|B_SHOWSIZE,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("MOVSB"),
        D_STRING|D_MUSTREPNE|D_UNDOC, 0,
        1, 0x000000FF, 0x000000A4, 0x00,
        B_STRDEST8|B_CHG|B_PSEUDO,
        B_STRSRC8|B_PSEUDO,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("MOVS*"),
        D_STRING|D_MUSTREPNE|D_WILDCARD|D_UNDOC, 0,
        1, 0x000000FF, 0x000000A5, 0x00,
        B_STRDEST|B_CHG|B_PSEUDO,
        B_STRSRC|B_PSEUDO,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("SCAS"),
        D_CMD|D_SIZE01|D_LONGFORM|D_NOREP|D_ALLFLAGS, DX_JE|DX_JB,
        1, 0x000000FE, 0x000000AE, 0x00,
        B_STRDEST|B_SHOWSIZE,
        B_ACC|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SCASB"),
        D_CMD|D_NOREP|D_ALLFLAGS, DX_JE|DX_JB,
        1, 0x000000FF, 0x000000AE, 0x00,
        B_STRDEST8|B_PSEUDO,
        B_AL|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SCAS*"),
        D_CMD|D_NOREP|D_ALLFLAGS|D_WILDCARD, DX_JE|DX_JB,
        1, 0x000000FF, 0x000000AF, 0x00,
        B_STRDEST|B_PSEUDO,
        B_ACC|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SCAS"),
        D_STRING|D_SIZE01|D_LONGFORM|D_MUSTREPE|D_ALLFLAGS, 0,
        1, 0x000000FE, 0x000000AE, 0x00,
        B_STRDEST|B_SHOWSIZE,
        B_ACC|B_PSEUDO,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("SCASB"),
        D_STRING|D_MUSTREPE|D_ALLFLAGS, 0,
        1, 0x000000FF, 0x000000AE, 0x00,
        B_STRDEST8|B_PSEUDO,
        B_AL|B_PSEUDO,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("SCAS*"),
        D_STRING|D_MUSTREPE|D_ALLFLAGS|D_WILDCARD, 0,
        1, 0x000000FF, 0x000000AF, 0x00,
        B_STRDEST|B_PSEUDO,
        B_ACC|B_PSEUDO,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("SCAS"),
        D_STRING|D_SIZE01|D_LONGFORM|D_MUSTREPNE|D_ALLFLAGS, 0,
        1, 0x000000FE, 0x000000AE, 0x00,
        B_STRDEST|B_SHOWSIZE,
        B_ACC|B_PSEUDO,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("SCASB"),
        D_STRING|D_MUSTREPNE|D_ALLFLAGS, 0,
        1, 0x000000FF, 0x000000AE, 0x00,
        B_STRDEST8|B_PSEUDO,
        B_AL|B_PSEUDO,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("SCAS*"),
        D_STRING|D_MUSTREPNE|D_ALLFLAGS|D_WILDCARD, 0,
        1, 0x000000FF, 0x000000AF, 0x00,
        B_STRDEST|B_PSEUDO,
        B_ACC|B_PSEUDO,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("STOS"),
        D_CMD|D_SIZE01|D_LONGFORM|D_NOREP, 0,
        1, 0x000000FE, 0x000000AA, 0x00,
        B_STRDEST|B_SHOWSIZE|B_CHG,
        B_ACC|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("STOSB"),
        D_CMD|D_NOREP, 0,
        1, 0x000000FF, 0x000000AA, 0x00,
        B_STRDEST8|B_CHG|B_PSEUDO,
        B_AL|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("STOS*"),
        D_CMD|D_NOREP|D_WILDCARD, 0,
        1, 0x000000FF, 0x000000AB, 0x00,
        B_STRDEST|B_CHG|B_PSEUDO,
        B_ACC|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("STOS"),
        D_STRING|D_SIZE01|D_LONGFORM|D_MUSTREP, 0,
        1, 0x000000FE, 0x000000AA, 0x00,
        B_STRDEST|B_SHOWSIZE|B_CHG,
        B_ACC|B_PSEUDO,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("STOSB"),
        D_STRING|D_MUSTREP, 0,
        1, 0x000000FF, 0x000000AA, 0x00,
        B_STRDEST8|B_CHG|B_PSEUDO,
        B_AL|B_PSEUDO,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("STOS*"),
        D_STRING|D_MUSTREP|D_WILDCARD, 0,
        1, 0x000000FF, 0x000000AB, 0x00,
        B_STRDEST|B_CHG|B_PSEUDO,
        B_ACC|B_PSEUDO,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("STOS"),
        D_STRING|D_SIZE01|D_LONGFORM|D_MUSTREPNE|D_UNDOC, 0,
        1, 0x000000FE, 0x000000AA, 0x00,
        B_STRDEST|B_SHOWSIZE|B_CHG,
        B_ACC|B_PSEUDO,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("STOSB"),
        D_STRING|D_MUSTREPNE|D_UNDOC, 0,
        1, 0x000000FF, 0x000000AA, 0x00,
        B_STRDEST8|B_CHG|B_PSEUDO,
        B_AL|B_PSEUDO,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("STOS*"),
        D_STRING|D_MUSTREPNE|D_WILDCARD|D_UNDOC, 0,
        1, 0x000000FF, 0x000000AB, 0x00,
        B_STRDEST|B_CHG|B_PSEUDO,
        B_ACC|B_PSEUDO,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("INS"),
        D_IO|D_SIZE01|D_LONGFORM|D_NOREP|D_RARE, 0,
        1, 0x000000FE, 0x0000006C, 0x00,
        B_STRDEST|B_SHOWSIZE|B_CHG,
        B_DXPORT|B_PORT,
        B_NONE,
        B_NONE },

  { T("INSB"),
        D_IO|D_NOREP|D_RARE, 0,
        1, 0x000000FF, 0x0000006C, 0x00,
        B_STRDEST8|B_CHG|B_PSEUDO,
        B_DXPORT|B_PORT|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("INS*"),
        D_IO|D_NOREP|D_WILDCARD|D_RARE, 0,
        1, 0x000000FF, 0x0000006D, 0x00,
        B_STRDEST|B_CHG|B_PSEUDO,
        B_DXPORT|B_PORT|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("INS"),
        D_IO|D_SIZE01|D_LONGFORM|D_MUSTREP|D_RARE, 0,
        1, 0x000000FE, 0x0000006C, 0x00,
        B_STRDEST|B_SHOWSIZE|B_CHG,
        B_DXPORT|B_PORT,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("INSB"),
        D_IO|D_MUSTREP|D_RARE, 0,
        1, 0x000000FF, 0x0000006C, 0x00,
        B_STRDEST8|B_CHG|B_PSEUDO,
        B_DXPORT|B_PORT|B_PSEUDO,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("INS*"),
        D_IO|D_MUSTREP|D_WILDCARD|D_RARE, 0,
        1, 0x000000FF, 0x0000006D, 0x00,
        B_STRDEST|B_CHG|B_PSEUDO,
        B_DXPORT|B_PORT|B_PSEUDO,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("INS"),
        D_IO|D_SIZE01|D_LONGFORM|D_MUSTREPNE|D_UNDOC, 0,
        1, 0x000000FE, 0x0000006C, 0x00,
        B_STRDEST|B_SHOWSIZE|B_CHG,
        B_DXPORT|B_PORT,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("INSB"),
        D_IO|D_MUSTREPNE|D_UNDOC, 0,
        1, 0x000000FF, 0x0000006C, 0x00,
        B_STRDEST8|B_CHG|B_PSEUDO,
        B_DXPORT|B_PORT|B_PSEUDO,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("INS*"),
        D_IO|D_MUSTREPNE|D_WILDCARD|D_UNDOC, 0,
        1, 0x000000FF, 0x0000006D, 0x00,
        B_STRDEST|B_CHG|B_PSEUDO,
        B_DXPORT|B_PORT|B_PSEUDO,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("OUTS"),
        D_IO|D_SIZE01|D_LONGFORM|D_NOREP|D_RARE, 0,
        1, 0x000000FE, 0x0000006E, 0x00,
        B_DXPORT|B_PORT,
        B_STRSRC|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("OUTSB"),
        D_IO|D_NOREP|D_RARE, 0,
        1, 0x000000FF, 0x0000006E, 0x00,
        B_DXPORT|B_PORT|B_PSEUDO,
        B_STRSRC8|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("OUTS*"),
        D_IO|D_NOREP|D_WILDCARD|D_RARE, 0,
        1, 0x000000FF, 0x0000006F, 0x00,
        B_DXPORT|B_PORT|B_PSEUDO,
        B_STRSRC|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("OUTS"),
        D_IO|D_SIZE01|D_LONGFORM|D_MUSTREP|D_RARE, 0,
        1, 0x000000FE, 0x0000006E, 0x00,
        B_DXPORT|B_PORT,
        B_STRSRC|B_SHOWSIZE,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("OUTSB"),
        D_IO|D_MUSTREP|D_RARE, 0,
        1, 0x000000FF, 0x0000006E, 0x00,
        B_DXPORT|B_PORT|B_PSEUDO,
        B_STRSRC8|B_PSEUDO,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("OUTS*"),
        D_IO|D_MUSTREP|D_WILDCARD|D_RARE, 0,
        1, 0x000000FF, 0x0000006F, 0x00,
        B_DXPORT|B_PORT|B_PSEUDO,
        B_STRSRC|B_PSEUDO,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("OUTS"),
        D_IO|D_SIZE01|D_LONGFORM|D_MUSTREPNE|D_UNDOC, 0,
        1, 0x000000FE, 0x0000006E, 0x00,
        B_DXPORT|B_PORT,
        B_STRSRC|B_SHOWSIZE,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("OUTSB"),
        D_IO|D_MUSTREPNE|D_UNDOC, 0,
        1, 0x000000FF, 0x0000006E, 0x00,
        B_DXPORT|B_PORT|B_PSEUDO,
        B_STRSRC8|B_PSEUDO,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("OUTS*"),
        D_IO|D_MUSTREPNE|D_WILDCARD|D_UNDOC, 0,
        1, 0x000000FF, 0x0000006F, 0x00,
        B_DXPORT|B_PORT|B_PSEUDO,
        B_STRSRC|B_PSEUDO,
        B_STRCNT|B_UPD|B_PSEUDO,
        B_NONE },

  { T("MOVBE"),
        D_CMD|D_NOREP, 0,
        3, 0x00FFFFFF, 0x00F0380F, 0x00,
        B_REG|B_CHG,
        B_INT|B_MEMONLY,
        B_NONE,
        B_NONE },

  { T("MOVBE"),
        D_CMD|D_NOREP, 0,
        3, 0x00FFFFFF, 0x00F1380F, 0x00,
        B_INT|B_MEMONLY|B_CHG,
        B_REG,
        B_NONE,
        B_NONE },

  { T("XGETBV"),
        D_SYS|D_MUSTNONE|D_RARE, 0,
        3, 0x00FFFFFF, 0x00D0010F, 0x00,
        B_EAX|B_CHG|B_PSEUDO,
        B_EDX|B_CHG|B_PSEUDO,
        B_ECX|B_PSEUDO,
        B_NONE },

  { T("XSETBV"),
        D_PRIVILEGED|D_MUSTNONE|D_RARE, 0,
        3, 0x00FFFFFF, 0x00D1010F, 0x00,
        B_EAX|B_PSEUDO,
        B_EDX|B_PSEUDO,
        B_ECX|B_PSEUDO,
        B_NONE },

  { T("XRSTOR"),
        D_SYS|D_MUSTNONE|D_MEMORY|D_RARE, 0,
        2, 0x0038FFFF, 0x0028AE0F, 0x00,
        B_ANYMEM|B_MEMONLY,
        B_EAX|B_PSEUDO,
        B_EDX|B_PSEUDO,
        B_NONE },

  { T("XSAVE"),
        D_SYS|D_MUSTNONE|D_MEMORY|D_RARE, 0,
        2, 0x0038FFFF, 0x0020AE0F, 0x00,
        B_ANYMEM|B_MEMONLY|B_CHG,
        B_EAX|B_PSEUDO,
        B_EDX|B_PSEUDO,
        B_NONE },

  { T("F2XM1"),
        D_FPU, 0,
        2, 0x0000FFFF, 0x0000F0D9, 0x00,
        B_ST0|B_CHG|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FABS"),
        D_FPU, 0,
        2, 0x0000FFFF, 0x0000E1D9, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FCHS"),
        D_FPU, 0,
        2, 0x0000FFFF, 0x0000E0D9, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FCLEX"),
        D_FPU, 0,
        2, 0x0000FFFF, 0x0000E2DB, 0x00,
        B_FST|B_UPD|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FCOMPP"),
        D_FPU|D_FPUPOP2, 0,
        2, 0x0000FFFF, 0x0000D9DE, 0x00,
        B_ST0|B_PSEUDO,
        B_ST1|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FCOS"),
        D_FPU, 0,
        2, 0x0000FFFF, 0x0000FFD9, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FDECSTP"),
        D_FPU|D_FPUPUSH, 0,
        2, 0x0000FFFF, 0x0000F6D9, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FINCSTP"),
        D_FPU|D_FPUPOP, 0,
        2, 0x0000FFFF, 0x0000F7D9, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FINIT"),
        D_FPU, 0,
        2, 0x0000FFFF, 0x0000E3DB, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FLD1"),
        D_FPU|D_FPUPUSH, 0,
        2, 0x0000FFFF, 0x0000E8D9, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FLDL2T"),
        D_FPU|D_FPUPUSH, 0,
        2, 0x0000FFFF, 0x0000E9D9, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FLDL2E"),
        D_FPU|D_FPUPUSH, 0,
        2, 0x0000FFFF, 0x0000EAD9, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FLDPI"),
        D_FPU|D_FPUPUSH, 0,
        2, 0x0000FFFF, 0x0000EBD9, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FLDLG2"),
        D_FPU|D_FPUPUSH, 0,
        2, 0x0000FFFF, 0x0000ECD9, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FLDLN2"),
        D_FPU|D_FPUPUSH, 0,
        2, 0x0000FFFF, 0x0000EDD9, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FLDZ"),
        D_FPU|D_FPUPUSH, 0,
        2, 0x0000FFFF, 0x0000EED9, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FNOP"),
        D_FPU, 0,
        2, 0x0000FFFF, 0x0000D0D9, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FPATAN"),
        D_FPU|D_FPUPOP, 0,
        2, 0x0000FFFF, 0x0000F3D9, 0x00,
        B_ST1|B_UPD|B_PSEUDO,
        B_ST0|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FPREM"),
        D_FPU, 0,
        2, 0x0000FFFF, 0x0000F8D9, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_ST1|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FPREM1"),
        D_FPU, 0,
        2, 0x0000FFFF, 0x0000F5D9, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_ST1|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FPTAN"),
        D_FPU|D_FPUPUSH, 0,
        2, 0x0000FFFF, 0x0000F2D9, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FRNDINT"),
        D_FPU, 0,
        2, 0x0000FFFF, 0x0000FCD9, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FSCALE"),
        D_FPU, 0,
        2, 0x0000FFFF, 0x0000FDD9, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_ST1|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FSIN"),
        D_FPU, 0,
        2, 0x0000FFFF, 0x0000FED9, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FSINCOS"),
        D_FPU|D_FPUPUSH, 0,
        2, 0x0000FFFF, 0x0000FBD9, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FSQRT"),
        D_FPU, 0,
        2, 0x0000FFFF, 0x0000FAD9, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FSTSW"),
        D_FPU, 0,
        2, 0x0000FFFF, 0x0000E0DF, 0x00,
        B_AX|B_CHG,
        B_FST|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FTST"),
        D_FPU, 0,
        2, 0x0000FFFF, 0x0000E4D9, 0x00,
        B_ST0|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FUCOMPP"),
        D_FPU|D_FPUPOP2, 0,
        2, 0x0000FFFF, 0x0000E9DA, 0x00,
        B_ST0|B_PSEUDO,
        B_ST1|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FXAM"),
        D_FPU, 0,
        2, 0x0000FFFF, 0x0000E5D9, 0x00,
        B_ST0|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FXTRACT"),
        D_FPU|D_FPUPUSH, 0,
        2, 0x0000FFFF, 0x0000F4D9, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FYL2X"),
        D_FPU|D_FPUPOP, 0,
        2, 0x0000FFFF, 0x0000F1D9, 0x00,
        B_ST1|B_UPD|B_PSEUDO,
        B_ST0|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FYL2XP1"),
        D_FPU|D_FPUPOP, 0,
        2, 0x0000FFFF, 0x0000F9D9, 0x00,
        B_ST1|B_UPD|B_PSEUDO,
        B_ST0|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FENI"),
        D_FPU|D_RARE, 0,
        2, 0x0000FFFF, 0x0000E0DB, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FDISI"),
        D_FPU|D_RARE, 0,
        2, 0x0000FFFF, 0x0000E1DB, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FADD"),
        D_FPU, 0,
        2, 0x0000F8FF, 0x0000C0D8, 0x00,
        B_ST0|B_UPD,
        B_ST,
        B_NONE,
        B_NONE },

  { T("FADD"),
        D_FPU, 0,
        2, 0x0000F8FF, 0x0000C0DC, 0x00,
        B_ST|B_UPD,
        B_ST0,
        B_NONE,
        B_NONE },

  { T("FADDP"),
        D_FPU|D_FPUPOP, 0,
        2, 0x0000F8FF, 0x0000C0DE, 0x00,
        B_ST|B_UPD,
        B_ST0,
        B_NONE,
        B_NONE },

  { T("FADDP"),
        D_FPU|D_FPUPOP, 0,
        2, 0x0000FFFF, 0x0000C1DE, 0x00,
        B_ST1|B_UPD|B_PSEUDO,
        B_ST0|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FCMOVB"),
        D_FPU|D_COND, 0,
        2, 0x0000F8FF, 0x0000C0DA, 0x00,
        B_ST0|B_CHG,
        B_ST,
        B_NONE,
        B_NONE },

  { T("FCMOVE"),
        D_FPU|D_COND, 0,
        2, 0x0000F8FF, 0x0000C8DA, 0x00,
        B_ST0|B_CHG,
        B_ST,
        B_NONE,
        B_NONE },

  { T("FCMOVBE"),
        D_FPU|D_COND, 0,
        2, 0x0000F8FF, 0x0000D0DA, 0x00,
        B_ST0|B_CHG,
        B_ST,
        B_NONE,
        B_NONE },

  { T("FCMOVU"),
        D_FPU|D_COND, 0,
        2, 0x0000F8FF, 0x0000D8DA, 0x00,
        B_ST0|B_CHG,
        B_ST,
        B_NONE,
        B_NONE },

  { T("FCMOVNB"),
        D_FPU|D_COND, 0,
        2, 0x0000F8FF, 0x0000C0DB, 0x00,
        B_ST0|B_CHG,
        B_ST,
        B_NONE,
        B_NONE },

  { T("FCMOVNE"),
        D_FPU|D_COND, 0,
        2, 0x0000F8FF, 0x0000C8DB, 0x00,
        B_ST0|B_CHG,
        B_ST,
        B_NONE,
        B_NONE },

  { T("FCMOVNBE"),
        D_FPU|D_COND, 0,
        2, 0x0000F8FF, 0x0000D0DB, 0x00,
        B_ST0|B_CHG,
        B_ST,
        B_NONE,
        B_NONE },

  { T("FCMOVNU"),
        D_FPU|D_COND, 0,
        2, 0x0000F8FF, 0x0000D8DB, 0x00,
        B_ST0|B_CHG,
        B_ST,
        B_NONE,
        B_NONE },

  { T("FCOM"),
        D_FPU, 0,
        2, 0x0000F8FF, 0x0000D0D8, 0x00,
        B_ST0|B_PSEUDO,
        B_ST,
        B_NONE,
        B_NONE },

  { T("FCOM"),
        D_FPU, 0,
        2, 0x0000FFFF, 0x0000D1D8, 0x00,
        B_ST0|B_PSEUDO,
        B_ST1|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FCOMP"),
        D_FPU|D_FPUPOP, 0,
        2, 0x0000F8FF, 0x0000D8D8, 0x00,
        B_ST0|B_PSEUDO,
        B_ST,
        B_NONE,
        B_NONE },

  { T("FCOMP"),
        D_FPU|D_FPUPOP, 0,
        2, 0x0000FFFF, 0x0000D9D8, 0x00,
        B_ST0|B_PSEUDO,
        B_ST1|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FCOMI"),
        D_FPU|D_FLAGSZPC, 0,
        2, 0x0000F8FF, 0x0000F0DB, 0x00,
        B_ST0,
        B_ST,
        B_NONE,
        B_NONE },

  { T("FCOMIP"),
        D_FPU|D_FLAGSZPC|D_FPUPOP, 0,
        2, 0x0000F8FF, 0x0000F0DF, 0x00,
        B_ST0,
        B_ST,
        B_NONE,
        B_NONE },

  { T("FUCOMI"),
        D_FPU|D_FLAGSZPC, 0,
        2, 0x0000F8FF, 0x0000E8DB, 0x00,
        B_ST0,
        B_ST,
        B_NONE,
        B_NONE },

  { T("FUCOMIP"),
        D_FPU|D_FLAGSZPC|D_FPUPOP, 0,
        2, 0x0000F8FF, 0x0000E8DF, 0x00,
        B_ST0,
        B_ST,
        B_NONE,
        B_NONE },

  { T("FDIV"),
        D_FPU, 0,
        2, 0x0000F8FF, 0x0000F0D8, 0x00,
        B_ST0|B_UPD,
        B_ST,
        B_NONE,
        B_NONE },

  { T("FDIV"),
        D_FPU, 0,
        2, 0x0000F8FF, 0x0000F8DC, 0x00,
        B_ST|B_UPD,
        B_ST0,
        B_NONE,
        B_NONE },

  { T("FDIVP"),
        D_FPU|D_FPUPOP, 0,
        2, 0x0000F8FF, 0x0000F8DE, 0x00,
        B_ST|B_UPD,
        B_ST0,
        B_NONE,
        B_NONE },

  { T("FDIVP"),
        D_FPU|D_FPUPOP, 0,
        2, 0x0000FFFF, 0x0000F9DE, 0x00,
        B_ST1|B_UPD|B_PSEUDO,
        B_ST0|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FDIVR"),
        D_FPU, 0,
        2, 0x0000F8FF, 0x0000F8D8, 0x00,
        B_ST0|B_UPD,
        B_ST,
        B_NONE,
        B_NONE },

  { T("FDIVR"),
        D_FPU, 0,
        2, 0x0000F8FF, 0x0000F0DC, 0x00,
        B_ST|B_UPD,
        B_ST0,
        B_NONE,
        B_NONE },

  { T("FDIVRP"),
        D_FPU|D_FPUPOP, 0,
        2, 0x0000F8FF, 0x0000F0DE, 0x00,
        B_ST|B_UPD,
        B_ST0,
        B_NONE,
        B_NONE },

  { T("FDIVRP"),
        D_FPU|D_FPUPOP, 0,
        2, 0x0000FFFF, 0x0000F1DE, 0x00,
        B_ST1|B_UPD|B_PSEUDO,
        B_ST0|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FFREE"),
        D_FPU, 0,
        2, 0x0000F8FF, 0x0000C0DD, 0x00,
        B_ST,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FFREEP"),
        D_FPU|D_FPUPOP|D_UNDOC, 0,
        2, 0x0000F8FF, 0x0000C0DF, 0x00,
        B_ST,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FLD"),
        D_FPU|D_FPUPUSH, 0,
        2, 0x0000F8FF, 0x0000C0D9, 0x00,
        B_ST,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FMUL"),
        D_FPU, 0,
        2, 0x0000F8FF, 0x0000C8D8, 0x00,
        B_ST0|B_UPD,
        B_ST,
        B_NONE,
        B_NONE },

  { T("FMUL"),
        D_FPU, 0,
        2, 0x0000F8FF, 0x0000C8DC, 0x00,
        B_ST|B_UPD,
        B_ST0,
        B_NONE,
        B_NONE },

  { T("FMULP"),
        D_FPU|D_FPUPOP, 0,
        2, 0x0000F8FF, 0x0000C8DE, 0x00,
        B_ST|B_UPD,
        B_ST0,
        B_NONE,
        B_NONE },

  { T("FMULP"),
        D_FPU|D_FPUPOP, 0,
        2, 0x0000FFFF, 0x0000C9DE, 0x00,
        B_ST1|B_UPD|B_PSEUDO,
        B_ST0|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FST"),
        D_FPU, 0,
        2, 0x0000F8FF, 0x0000D0DD, 0x00,
        B_ST|B_CHG,
        B_ST0|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FSTP"),
        D_FPU|D_FPUPOP, 0,
        2, 0x0000F8FF, 0x0000D8DD, 0x00,
        B_ST|B_CHG,
        B_ST0|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FSUB"),
        D_FPU, 0,
        2, 0x0000F8FF, 0x0000E0D8, 0x00,
        B_ST0|B_UPD,
        B_ST,
        B_NONE,
        B_NONE },

  { T("FSUB"),
        D_FPU, 0,
        2, 0x0000F8FF, 0x0000E8DC, 0x00,
        B_ST|B_UPD,
        B_ST0,
        B_NONE,
        B_NONE },

  { T("FSUBP"),
        D_FPU|D_FPUPOP, 0,
        2, 0x0000F8FF, 0x0000E8DE, 0x00,
        B_ST|B_UPD,
        B_ST0,
        B_NONE,
        B_NONE },

  { T("FSUBP"),
        D_FPU|D_FPUPOP, 0,
        2, 0x0000FFFF, 0x0000E9DE, 0x00,
        B_ST1|B_UPD|B_PSEUDO,
        B_ST0|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FSUBR"),
        D_FPU, 0,
        2, 0x0000F8FF, 0x0000E8D8, 0x00,
        B_ST0|B_UPD,
        B_ST,
        B_NONE,
        B_NONE },

  { T("FSUBR"),
        D_FPU, 0,
        2, 0x0000F8FF, 0x0000E0DC, 0x00,
        B_ST|B_UPD,
        B_ST0,
        B_NONE,
        B_NONE },

  { T("FSUBRP"),
        D_FPU|D_FPUPOP, 0,
        2, 0x0000F8FF, 0x0000E0DE, 0x00,
        B_ST|B_UPD,
        B_ST0,
        B_NONE,
        B_NONE },

  { T("FSUBRP"),
        D_FPU|D_FPUPOP, 0,
        2, 0x0000FFFF, 0x0000E1DE, 0x00,
        B_ST1|B_UPD|B_PSEUDO,
        B_ST0|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FUCOM"),
        D_FPU, 0,
        2, 0x0000F8FF, 0x0000E0DD, 0x00,
        B_ST0|B_PSEUDO,
        B_ST,
        B_NONE,
        B_NONE },

  { T("FUCOM"),
        D_FPU, 0,
        2, 0x0000FFFF, 0x0000E1DD, 0x00,
        B_ST0|B_PSEUDO,
        B_ST1|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FUCOMP"),
        D_FPU|D_FPUPOP, 0,
        2, 0x0000F8FF, 0x0000E8DD, 0x00,
        B_ST0|B_PSEUDO,
        B_ST,
        B_NONE,
        B_NONE },

  { T("FUCOMP"),
        D_FPU|D_FPUPOP, 0,
        2, 0x0000FFFF, 0x0000E9DD, 0x00,
        B_ST0|B_PSEUDO,
        B_ST1|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FXCH"),
        D_FPU, 0,
        2, 0x0000F8FF, 0x0000C8D9, 0x00,
        B_ST0|B_CHG|B_PSEUDO,
        B_ST|B_CHG,
        B_NONE,
        B_NONE },

  { T("FXCH"),
        D_FPU, 0,
        2, 0x0000FFFF, 0x0000C9D9, 0x00,
        B_ST0|B_CHG|B_PSEUDO,
        B_ST1|B_CHG|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FADD"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000000D8, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_FLOAT32|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("FADD"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000000DC, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_FLOAT64|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("FIADD"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000000DA, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_INT32|B_SIGNED|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("FIADD"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000000DE, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_INT16|B_SIGNED|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("FBLD"),
        D_FPU|D_MEMORY|D_FPUPUSH|D_RARE, 0,
        1, 0x000038FF, 0x000020DF, 0x00,
        B_BCD|B_MEMORY,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FBSTP"),
        D_FPU|D_MEMORY|D_FPUPOP|D_RARE, 0,
        1, 0x000038FF, 0x000030DF, 0x00,
        B_BCD|B_MEMORY|B_CHG,
        B_ST0|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FCOM"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000010D8, 0x00,
        B_ST0|B_PSEUDO,
        B_FLOAT32|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("FCOM"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000010DC, 0x00,
        B_ST0|B_PSEUDO,
        B_FLOAT64|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("FCOMP"),
        D_FPU|D_MEMORY|D_FPUPOP, 0,
        1, 0x000038FF, 0x000018D8, 0x00,
        B_ST0|B_PSEUDO,
        B_FLOAT32|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("FCOMP"),
        D_FPU|D_MEMORY|D_FPUPOP, 0,
        1, 0x000038FF, 0x000018DC, 0x00,
        B_ST0|B_PSEUDO,
        B_FLOAT64|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("FDIV"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000030D8, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_FLOAT32|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("FDIV"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000030DC, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_FLOAT64|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("FIDIV"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000030DA, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_INT32|B_SIGNED|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("FIDIV"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000030DE, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_INT16|B_SIGNED|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("FDIVR"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000038D8, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_FLOAT32|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("FDIVR"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000038DC, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_FLOAT64|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("FIDIVR"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000038DA, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_INT32|B_SIGNED|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("FIDIVR"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000038DE, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_INT16|B_SIGNED|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("FICOM"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000010DE, 0x00,
        B_ST0|B_PSEUDO,
        B_INT16|B_SIGNED|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("FICOM"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000010DA, 0x00,
        B_ST0|B_PSEUDO,
        B_INT32|B_SIGNED|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("FICOMP"),
        D_FPU|D_MEMORY|D_FPUPOP, 0,
        1, 0x000038FF, 0x000018DE, 0x00,
        B_ST0|B_PSEUDO,
        B_INT16|B_SIGNED|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("FICOMP"),
        D_FPU|D_MEMORY|D_FPUPOP, 0,
        1, 0x000038FF, 0x000018DA, 0x00,
        B_ST0|B_PSEUDO,
        B_INT32|B_SIGNED|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("FILD"),
        D_FPU|D_MEMORY|D_FPUPUSH, 0,
        1, 0x000038FF, 0x000000DF, 0x00,
        B_INT16|B_SIGNED|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FILD"),
        D_FPU|D_MEMORY|D_FPUPUSH, 0,
        1, 0x000038FF, 0x000000DB, 0x00,
        B_INT32|B_SIGNED|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FILD"),
        D_FPU|D_MEMORY|D_FPUPUSH, 0,
        1, 0x000038FF, 0x000028DF, 0x00,
        B_INT64|B_SIGNED|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FIST"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000010DF, 0x00,
        B_INT16|B_SIGNED|B_MEMORY|B_SHOWSIZE|B_CHG,
        B_ST0|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FIST"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000010DB, 0x00,
        B_INT32|B_SIGNED|B_MEMORY|B_SHOWSIZE|B_CHG,
        B_ST0|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FISTP"),
        D_FPU|D_MEMORY|D_FPUPOP, 0,
        1, 0x000038FF, 0x000018DF, 0x00,
        B_INT16|B_SIGNED|B_MEMORY|B_SHOWSIZE|B_CHG,
        B_ST0|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FISTP"),
        D_FPU|D_MEMORY|D_FPUPOP, 0,
        1, 0x000038FF, 0x000018DB, 0x00,
        B_INT32|B_SIGNED|B_MEMORY|B_SHOWSIZE|B_CHG,
        B_ST0|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FISTP"),
        D_FPU|D_MEMORY|D_FPUPOP, 0,
        1, 0x000038FF, 0x000038DF, 0x00,
        B_INT64|B_SIGNED|B_MEMORY|B_SHOWSIZE|B_CHG,
        B_ST0|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FISTTP"),
        D_FPU|D_MEMORY|D_FPUPOP, 0,
        1, 0x000038FF, 0x000008DF, 0x00,
        B_INT16|B_SIGNED|B_MEMORY|B_SHOWSIZE|B_CHG,
        B_ST0|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FISTTP"),
        D_FPU|D_MEMORY|D_FPUPOP, 0,
        1, 0x000038FF, 0x000008DB, 0x00,
        B_INT32|B_SIGNED|B_MEMORY|B_SHOWSIZE|B_CHG,
        B_ST0|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FISTTP"),
        D_FPU|D_MEMORY|D_FPUPOP, 0,
        1, 0x000038FF, 0x000008DD, 0x00,
        B_INT64|B_SIGNED|B_MEMORY|B_SHOWSIZE|B_CHG,
        B_ST0|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FLD"),
        D_FPU|D_MEMORY|D_FPUPUSH, 0,
        1, 0x000038FF, 0x000000D9, 0x00,
        B_FLOAT32|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FLD"),
        D_FPU|D_MEMORY|D_FPUPUSH, 0,
        1, 0x000038FF, 0x000000DD, 0x00,
        B_FLOAT64|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FLD"),
        D_FPU|D_MEMORY|D_FPUPUSH, 0,
        1, 0x000038FF, 0x000028DB, 0x00,
        B_FLOAT80|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FLDCW"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000028D9, 0x00,
        B_FCW|B_CHG|B_PSEUDO,
        B_INT16|B_BINARY|B_MEMORY,
        B_NONE,
        B_NONE },

  { T("FLDENV"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000020D9, 0x00,
        B_LONGDATA|B_MEMORY,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FMUL"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000008D8, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_FLOAT32|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("FMUL"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000008DC, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_FLOAT64|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("FIMUL"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000008DA, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_INT32|B_SIGNED|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("FIMUL"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000008DE, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_INT16|B_SIGNED|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("FRSTOR"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000020DD, 0x00,
        B_LONGDATA|B_MEMORY,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FSAVE"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000030DD, 0x00,
        B_LONGDATA|B_MEMORY|B_CHG,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FST"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000010D9, 0x00,
        B_FLOAT32|B_MEMORY|B_SHOWSIZE|B_CHG,
        B_ST0|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FST"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000010DD, 0x00,
        B_FLOAT64|B_MEMORY|B_SHOWSIZE|B_CHG,
        B_ST0|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FSTP"),
        D_FPU|D_MEMORY|D_FPUPOP, 0,
        1, 0x000038FF, 0x000018D9, 0x00,
        B_FLOAT32|B_MEMORY|B_SHOWSIZE|B_CHG,
        B_ST0|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FSTP"),
        D_FPU|D_MEMORY|D_FPUPOP, 0,
        1, 0x000038FF, 0x000018DD, 0x00,
        B_FLOAT64|B_MEMORY|B_SHOWSIZE|B_CHG,
        B_ST0|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FSTP"),
        D_FPU|D_MEMORY|D_FPUPOP, 0,
        1, 0x000038FF, 0x000038DB, 0x00,
        B_FLOAT80|B_MEMORY|B_SHOWSIZE|B_CHG,
        B_ST0|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FSTCW"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000038D9, 0x00,
        B_INT16|B_BINARY|B_MEMORY|B_CHG,
        B_FCW|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FSTENV"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000030D9, 0x00,
        B_LONGDATA|B_MEMORY|B_CHG,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FSTSW"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000038DD, 0x00,
        B_INT16|B_BINARY|B_MEMORY|B_CHG,
        B_FST|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("FSUB"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000020D8, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_FLOAT32|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("FSUB"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000020DC, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_FLOAT64|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("FISUB"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000020DA, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_INT32|B_SIGNED|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("FISUB"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000020DE, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_INT16|B_SIGNED|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("FSUBR"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000028D8, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_FLOAT32|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("FSUBR"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000028DC, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_FLOAT64|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("FISUBR"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000028DA, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_INT32|B_SIGNED|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("FISUBR"),
        D_FPU|D_MEMORY, 0,
        1, 0x000038FF, 0x000028DE, 0x00,
        B_ST0|B_UPD|B_PSEUDO,
        B_INT16|B_SIGNED|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("FSETPM"),
        D_FPU|D_UNDOC, 0,
        2, 0x0000FFFF, 0x0000E4DB, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("ADDPD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000580F, 0x00,
        B_SREGF64x2|B_UPD,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("VADDPD"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x00000058, 0x00,
        B_SREGF64x2|B_CHG,
        B_SVEXF64x2,
        B_SSEF64x2,
        B_NONE },

  { T("ADDPS"),
        D_SSE|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000580F, 0x00,
        B_SREGF32x4|B_UPD,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("VADDPS"),
        D_AVX|D_MUSTNONE, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x00000058, 0x00,
        B_SREGF32x4|B_CHG,
        B_SVEXF32x4,
        B_SSEF32x4,
        B_NONE },

  { T("ADDSD"),
        D_SSE|D_MUSTF2, 0,
        2, 0x0000FFFF, 0x0000580F, 0x00,
        B_SREGF64L|B_UPD,
        B_SSEF64L,
        B_NONE,
        B_NONE },

  { T("VADDSD"),
        D_AVX|D_MUSTF2, DX_VEX|DX_IGNOREL,
        1, 0x000000FF, 0x00000058, 0x00,
        B_SREGF64L|B_CHG,
        B_SVEXF64L,
        B_SSEF64L,
        B_NONE },

  { T("ADDSS"),
        D_SSE|D_MUSTF3, 0,
        2, 0x0000FFFF, 0x0000580F, 0x00,
        B_SREGF32L|B_UPD,
        B_SSEF32L,
        B_NONE,
        B_NONE },

  { T("VADDSS"),
        D_AVX|D_MUSTF3, DX_VEX|DX_IGNOREL,
        1, 0x000000FF, 0x00000058, 0x00,
        B_SREGF32L|B_CHG,
        B_SVEXF32L,
        B_SSEF32L,
        B_NONE },

  { T("ADDSUBPD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000D00F, 0x00,
        B_SREGF64x2|B_UPD,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("VADDSUBPD"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x000000D0, 0x00,
        B_SREGF64x2|B_CHG,
        B_SVEXF64x2,
        B_SSEF64x2,
        B_NONE },

  { T("ADDSUBPS"),
        D_SSE|D_MUSTF2, 0,
        2, 0x0000FFFF, 0x0000D00F, 0x00,
        B_SREGF32x4|B_UPD,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("VADDSUBPS"),
        D_AVX|D_MUSTF2, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x000000D0, 0x00,
        B_SREGF32x4|B_CHG,
        B_SVEXF32x4,
        B_SSEF32x4,
        B_NONE },

  { T("ANDPD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000540F, 0x00,
        B_SREGF64x2|B_UPD,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("VANDPD"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x00000054, 0x00,
        B_SREGF64x2|B_CHG,
        B_SVEXF64x2,
        B_SSEF64x2,
        B_NONE },

  { T("ANDPS"),
        D_SSE|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000540F, 0x00,
        B_SREGF32x4|B_UPD,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("VANDPS"),
        D_AVX|D_MUSTNONE, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x00000054, 0x00,
        B_SREGF32x4|B_CHG,
        B_SVEXF32x4,
        B_SSEF32x4,
        B_NONE },

  { T("ANDNPD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000550F, 0x00,
        B_SREGF64x2|B_UPD,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("VANDNPD"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x00000055, 0x00,
        B_SREGF64x2|B_CHG,
        B_SVEXF64x2,
        B_SSEF64x2,
        B_NONE },

  { T("ANDNPS"),
        D_SSE|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000550F, 0x00,
        B_SREGF32x4|B_UPD,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("VANDNPS"),
        D_AVX|D_MUSTNONE, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x00000055, 0x00,
        B_SREGF32x4|B_CHG,
        B_SVEXF32x4,
        B_SSEF32x4,
        B_NONE },

  { T("CMP*PD"),
        D_SSE|D_POSTBYTE|D_MUST66|D_WILDCARD, 0,
        2, 0x0000FFFF, 0x0000C20F, 0x00,
        B_SREGF64x2|B_UPD,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("VCMP*PD"),
        D_AVX|D_POSTBYTE|D_MUST66|D_WILDCARD, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x000000C2, 0x00,
        B_SREGF64x2|B_CHG,
        B_SVEXF64x2,
        B_SSEF64x2,
        B_NONE },

  { T("CMPPD"),
        D_SSE|D_MUST66|D_SUSPICIOUS, 0,
        2, 0x0000FFFF, 0x0000C20F, 0x00,
        B_SREGF64x2|B_UPD,
        B_SSEF64x2,
        B_CONST8,
        B_NONE },

  { T("VCMPPD"),
        D_AVX|D_MUST66|D_SUSPICIOUS, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x000000C2, 0x00,
        B_SREGF64x2|B_CHG,
        B_SVEXF64x2,
        B_SSEF64x2,
        B_CONST8 },

  { T("CMP*PS"),
        D_SSE|D_POSTBYTE|D_MUSTNONE|D_WILDCARD, 0,
        2, 0x0000FFFF, 0x0000C20F, 0x00,
        B_SREGF32x4|B_UPD,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("VCMP*PS"),
        D_AVX|D_POSTBYTE|D_MUSTNONE|D_WILDCARD, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x000000C2, 0x00,
        B_SREGF32x4|B_CHG,
        B_SVEXF32x4,
        B_SSEF32x4,
        B_NONE },

  { T("CMPPS"),
        D_SSE|D_MUSTNONE|D_SUSPICIOUS, 0,
        2, 0x0000FFFF, 0x0000C20F, 0x00,
        B_SREGF32x4|B_UPD,
        B_SSEF32x4,
        B_CONST8,
        B_NONE },

  { T("VCMPPS"),
        D_AVX|D_MUSTNONE|D_SUSPICIOUS, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x000000C2, 0x00,
        B_SREGF32x4|B_CHG,
        B_SVEXF32x4,
        B_SSEF32x4,
        B_CONST8 },

  { T("CMP*SD"),
        D_SSE|D_POSTBYTE|D_MUSTF2|D_WILDCARD, 0,
        2, 0x0000FFFF, 0x0000C20F, 0x00,
        B_SREGF64L|B_UPD,
        B_SSEF64L,
        B_NONE,
        B_NONE },

  { T("VCMP*SD"),
        D_AVX|D_POSTBYTE|D_MUSTF2|D_WILDCARD, DX_VEX|DX_IGNOREL,
        1, 0x000000FF, 0x000000C2, 0x00,
        B_SREGF64L|B_CHG,
        B_SVEXF64L,
        B_SSEF64L,
        B_NONE },

  { T("CMPSD"),
        D_SSE|D_MUSTF2|D_SUSPICIOUS, 0,
        2, 0x0000FFFF, 0x0000C20F, 0x00,
        B_SREGF64L|B_UPD,
        B_SVEXF64L,
        B_CONST8,
        B_NONE },

  { T("VCMPSD"),
        D_AVX|D_MUSTF2|D_SUSPICIOUS, DX_VEX|DX_IGNOREL,
        1, 0x000000FF, 0x000000C2, 0x00,
        B_SREGF64L|B_CHG,
        B_SVEXF64L,
        B_SSEF64L,
        B_CONST8 },

  { T("CMP*SS"),
        D_SSE|D_POSTBYTE|D_MUSTF3|D_WILDCARD, 0,
        2, 0x0000FFFF, 0x0000C20F, 0x00,
        B_SREGF32L|B_UPD,
        B_SSEF32L,
        B_NONE,
        B_NONE },

  { T("VCMP*SS"),
        D_AVX|D_POSTBYTE|D_MUSTF3|D_WILDCARD, DX_VEX|DX_IGNOREL,
        1, 0x000000FF, 0x000000C2, 0x00,
        B_SREGF32L|B_CHG,
        B_SVEXF32L,
        B_SSEF32L,
        B_NONE },

  { T("CMPSS"),
        D_SSE|D_MUSTF3|D_SUSPICIOUS, 0,
        2, 0x0000FFFF, 0x0000C20F, 0x00,
        B_SREGF32L|B_UPD,
        B_SSEF32L,
        B_CONST8,
        B_NONE },

  { T("VCMPSS"),
        D_AVX|D_MUSTF3|D_SUSPICIOUS, DX_VEX|DX_IGNOREL,
        1, 0x000000FF, 0x000000C2, 0x00,
        B_SREGF32L|B_CHG,
        B_SVEXF32L,
        B_SSEF32L,
        B_CONST8 },

  { T("COMISD"),
        D_SSE|D_MUST66|D_ALLFLAGS, 0,
        2, 0x0000FFFF, 0x00002F0F, 0x00,
        B_SREGF64L,
        B_SSEF64L,
        B_NONE,
        B_NONE },

  { T("VCOMISD"),
        D_AVX|D_MUST66|D_ALLFLAGS, DX_VEX|DX_IGNOREL|DX_NOVREG,
        1, 0x000000FF, 0x0000002F, 0x00,
        B_SREGF64L,
        B_SSEF64L,
        B_NONE,
        B_NONE },

  { T("COMISS"),
        D_SSE|D_MUSTNONE|D_ALLFLAGS, 0,
        2, 0x0000FFFF, 0x00002F0F, 0x00,
        B_SREGF32L,
        B_SSEF32L,
        B_NONE,
        B_NONE },

  { T("VCOMISS"),
        D_AVX|D_MUSTNONE|D_ALLFLAGS, DX_VEX|DX_IGNOREL|DX_NOVREG,
        1, 0x000000FF, 0x0000002F, 0x00,
        B_SREGF32L,
        B_SSEF32L,
        B_NONE,
        B_NONE },

  { T("CVTDQ2PD"),
        D_SSE|D_MUSTF3, 0,
        2, 0x0000FFFF, 0x0000E60F, 0x00,
        B_SREGF64x2|B_CHG,
        B_SSEI32x2L,
        B_NONE,
        B_NONE },

  { T("VCVTDQ2PD"),
        D_AVX|D_MUSTF3|D_REGISTER, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x000000E6, 0x00,
        B_SREGF64x2|B_CHG,
        B_SSEI32x2L|B_REGISTER|B_NOVEXSIZE|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("VCVTDQ2PD"),
        D_AVX|D_MUSTF3|D_MEMORY, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x000000E6, 0x00,
        B_SREGF64x2|B_CHG,
        B_SSEI32x2L|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("CVTDQ2PS"),
        D_SSE|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x00005B0F, 0x00,
        B_SREGF32x4|B_CHG,
        B_SSEI32x4,
        B_NONE,
        B_NONE },

  { T("VCVTDQ2PS"),
        D_AVX|D_MUSTNONE, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x0000005B, 0x00,
        B_SREGF32x4|B_CHG,
        B_SSEI32x4,
        B_NONE,
        B_NONE },

  { T("CVTPD2DQ"),
        D_SSE|D_MUSTF2, 0,
        2, 0x0000FFFF, 0x0000E60F, 0x00,
        B_SREGI32x2L|B_CHG,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("VCVTPD2DQ"),
        D_AVX|D_MUSTF2, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x000000E6, 0x00,
        B_SREGI32x2L|B_NOVEXSIZE|B_CHG,
        B_SSEF64x2|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("CVTPD2PI"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x00002D0F, 0x00,
        B_MREG32x2|B_SIGNED|B_CHG,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("CVTPD2PS"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x00005A0F, 0x00,
        B_SREGF32x2L|B_CHG,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("VCVTPD2PS"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x0000005A, 0x00,
        B_SREGF32x2L|B_NOVEXSIZE|B_CHG,
        B_SSEF64x2|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("CVTPI2PD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x00002A0F, 0x00,
        B_SREGF64x2|B_CHG,
        B_MMX32x2|B_SIGNED,
        B_NONE,
        B_NONE },

  { T("CVTPI2PS"),
        D_SSE|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x00002A0F, 0x00,
        B_SREGF32x2L|B_CHG,
        B_MMX32x2|B_SIGNED,
        B_NONE,
        B_NONE },

  { T("CVTPS2DQ"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x00005B0F, 0x00,
        B_SREGI32x4|B_CHG,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("VCVTPS2DQ"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x0000005B, 0x00,
        B_SREGI32x4|B_CHG,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("CVTPS2PD"),
        D_SSE|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x00005A0F, 0x00,
        B_SREGF64x2|B_CHG,
        B_SSEF32x2L,
        B_NONE,
        B_NONE },

  { T("VCVTPS2PD"),
        D_AVX|D_MUSTNONE|D_REGISTER, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x0000005A, 0x00,
        B_SREGF64x2|B_CHG,
        B_SSEF32x2L|B_REGISTER|B_NOVEXSIZE|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("VCVTPS2PD"),
        D_AVX|D_MUSTNONE|D_MEMORY, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x0000005A, 0x00,
        B_SREGF64x2|B_CHG,
        B_SSEF32x2L|B_MEMORY|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("CVTPS2PI"),
        D_SSE|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x00002D0F, 0x00,
        B_MREG32x2|B_CHG,
        B_SSEF32x2L,
        B_NONE,
        B_NONE },

  { T("CVTSD2SI"),
        D_SSE|D_MUSTF2, 0,
        2, 0x0000FFFF, 0x00002D0F, 0x00,
        B_REG32|B_CHG,
        B_SSEF64L,
        B_NONE,
        B_NONE },

  { T("VCVTSD2SI"),
        D_AVX|D_MUSTF2, DX_VEX|DX_IGNOREL|DX_NOVREG,
        1, 0x000000FF, 0x0000002D, 0x00,
        B_REG32|B_CHG,
        B_SSEF64L,
        B_NONE,
        B_NONE },

  { T("CVTSD2SS"),
        D_SSE|D_MUSTF2, 0,
        2, 0x0000FFFF, 0x00005A0F, 0x00,
        B_SREGF32L|B_CHG,
        B_SSEF64L,
        B_NONE,
        B_NONE },

  { T("VCVTSD2SS"),
        D_AVX|D_MUSTF2, DX_VEX|DX_IGNOREL,
        1, 0x000000FF, 0x0000005A, 0x00,
        B_SREGF32L|B_CHG,
        B_SVEXF32L,
        B_SSEF64L,
        B_NONE },

  { T("CVTSI2SD"),
        D_SSE|D_MUSTF2, 0,
        2, 0x0000FFFF, 0x00002A0F, 0x00,
        B_SREGF64L|B_CHG,
        B_INT32|B_SIGNED,
        B_NONE,
        B_NONE },

  { T("VCVTSI2SD"),
        D_AVX|D_MUSTF2, DX_VEX|DX_IGNOREL,
        1, 0x000000FF, 0x0000002A, 0x00,
        B_SREGF64L|B_CHG,
        B_SVEXF64L,
        B_INT32|B_SIGNED,
        B_NONE },

  { T("CVTSI2SS"),
        D_SSE|D_MUSTF3, 0,
        2, 0x0000FFFF, 0x00002A0F, 0x00,
        B_SREGF32L|B_CHG,
        B_INT32|B_SIGNED,
        B_NONE,
        B_NONE },

  { T("VCVTSI2SS"),
        D_AVX|D_MUSTF3, DX_VEX|DX_IGNOREL,
        1, 0x000000FF, 0x0000002A, 0x00,
        B_SREGF32L|B_CHG,
        B_SVEXF32L,
        B_INT32|B_SIGNED,
        B_NONE },

  { T("CVTSS2SD"),
        D_SSE|D_MUSTF3, 0,
        2, 0x0000FFFF, 0x00005A0F, 0x00,
        B_SREGF64L|B_CHG,
        B_SSEF32L,
        B_NONE,
        B_NONE },

  { T("VCVTSS2SD"),
        D_AVX|D_MUSTF3, DX_VEX|DX_IGNOREL,
        1, 0x000000FF, 0x0000005A, 0x00,
        B_SREGF64L|B_CHG,
        B_SVEXF64L,
        B_SSEF32L,
        B_NONE },

  { T("CVTSS2SI"),
        D_SSE|D_MUSTF3, 0,
        2, 0x0000FFFF, 0x00002D0F, 0x00,
        B_REG32|B_CHG,
        B_SSEF32L,
        B_NONE,
        B_NONE },

  { T("VCVTSS2SI"),
        D_AVX|D_MUSTF3, DX_VEX|DX_IGNOREL|DX_NOVREG,
        1, 0x000000FF, 0x0000002D, 0x00,
        B_REG32|B_CHG,
        B_SSEF32L,
        B_NONE,
        B_NONE },

  { T("CVTTPD2PI"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x00002C0F, 0x00,
        B_MREG32x2|B_CHG,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("CVTTPD2DQ"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000E60F, 0x00,
        B_SREGI32x2L|B_CHG,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("VCVTTPD2DQ"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x000000E6, 0x00,
        B_SREGI32x2L|B_NOVEXSIZE|B_CHG,
        B_SSEF64x2|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("CVTTPS2DQ"),
        D_SSE|D_MUSTF3, 0,
        2, 0x0000FFFF, 0x00005B0F, 0x00,
        B_SREGI32x4|B_CHG,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("VCVTTPS2DQ"),
        D_AVX|D_MUSTF3, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x0000005B, 0x00,
        B_SREGI32x4|B_CHG,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("CVTTPS2PI"),
        D_SSE|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x00002C0F, 0x00,
        B_MREG32x2|B_CHG,
        B_SSEF32x2L,
        B_NONE,
        B_NONE },

  { T("CVTTSD2SI"),
        D_SSE|D_MUSTF2, 0,
        2, 0x0000FFFF, 0x00002C0F, 0x00,
        B_REG32|B_CHG,
        B_SSEF64L,
        B_NONE,
        B_NONE },

  { T("VCVTTSD2SI"),
        D_AVX|D_MUSTF2, DX_VEX|DX_IGNOREL|DX_NOVREG,
        1, 0x000000FF, 0x0000002C, 0x00,
        B_REG32|B_CHG,
        B_SSEF64L,
        B_NONE,
        B_NONE },

  { T("CVTTSS2SI"),
        D_SSE|D_MUSTF3, 0,
        2, 0x0000FFFF, 0x00002C0F, 0x00,
        B_REG32|B_CHG,
        B_SSEF32L,
        B_NONE,
        B_NONE },

  { T("VCVTTSS2SI"),
        D_AVX|D_MUSTF3, DX_VEX|DX_IGNOREL|DX_NOVREG,
        1, 0x000000FF, 0x0000002C, 0x00,
        B_REG32|B_CHG,
        B_SSEF32L,
        B_NONE,
        B_NONE },

  { T("DIVPD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x00005E0F, 0x00,
        B_SREGF64x2|B_UPD,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("VDIVPD"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x0000005E, 0x00,
        B_SREGF64x2|B_CHG,
        B_SVEXF64x2,
        B_SSEF64x2,
        B_NONE },

  { T("DIVPS"),
        D_SSE|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x00005E0F, 0x00,
        B_SREGF32x4|B_UPD,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("VDIVPS"),
        D_AVX|D_MUSTNONE, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x0000005E, 0x00,
        B_SREGF32x4|B_CHG,
        B_SVEXF32x4,
        B_SSEF32x4,
        B_NONE },

  { T("DIVSD"),
        D_SSE|D_MUSTF2, 0,
        2, 0x0000FFFF, 0x00005E0F, 0x00,
        B_SREGF64L|B_UPD,
        B_SSEF64L,
        B_NONE,
        B_NONE },

  { T("VDIVSD"),
        D_AVX|D_MUSTF2, DX_VEX|DX_IGNOREL,
        1, 0x000000FF, 0x0000005E, 0x00,
        B_SREGF64L|B_CHG,
        B_SVEXF64L,
        B_SSEF64L,
        B_NONE },

  { T("DIVSS"),
        D_SSE|D_MUSTF3, 0,
        2, 0x0000FFFF, 0x00005E0F, 0x00,
        B_SREGF32L|B_UPD,
        B_SSEF32L,
        B_NONE,
        B_NONE },

  { T("VDIVSS"),
        D_AVX|D_MUSTF3, DX_VEX|DX_IGNOREL,
        1, 0x000000FF, 0x0000005E, 0x00,
        B_SREGF32L|B_CHG,
        B_SVEXF32L,
        B_SSEF32L,
        B_NONE },

  { T("HADDPD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x00007C0F, 0x00,
        B_SREGF64x2|B_UPD,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("VHADDPD"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x0000007C, 0x00,
        B_SREGF64x2|B_CHG,
        B_SVEXF64x2,
        B_SSEF64x2,
        B_NONE },

  { T("HADDPS"),
        D_SSE|D_MUSTF2, 0,
        2, 0x0000FFFF, 0x00007C0F, 0x00,
        B_SREGF32x4|B_UPD,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("VHADDPS"),
        D_AVX|D_MUSTF2, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x0000007C, 0x00,
        B_SREGF32x4|B_CHG,
        B_SVEXF32x4,
        B_SSEF32x4,
        B_NONE },

  { T("HSUBPD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x00007D0F, 0x00,
        B_SREGF64x2|B_UPD,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("VHSUBPD"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x0000007D, 0x00,
        B_SREGF64x2|B_CHG,
        B_SVEXF64x2,
        B_SSEF64x2,
        B_NONE },

  { T("HSUBPS"),
        D_SSE|D_MUSTF2, 0,
        2, 0x0000FFFF, 0x00007D0F, 0x00,
        B_SREGF32x4|B_UPD,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("VHSUBPS"),
        D_AVX|D_MUSTF2, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x0000007D, 0x00,
        B_SREGF32x4|B_CHG,
        B_SVEXF32x4,
        B_SSEF32x4,
        B_NONE },

  { T("LDDQU"),
        D_SSE|D_MUSTF2, 0,
        2, 0x0000FFFF, 0x0000F00F, 0x00,
        B_SREGF64x2|B_CHG,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("VLDDQU"),
        D_AVX|D_MUSTF2, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x000000F0, 0x00,
        B_SREGF64x2|B_CHG,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("LDMXCSR"),
        D_CMD|D_MEMORY, 0,
        2, 0x0038FFFF, 0x0010AE0F, 0x00,
        B_MXCSR|B_CHG|B_PSEUDO,
        B_INT32|B_BINARY|B_MEMORY,
        B_NONE,
        B_NONE },

  { T("VLDMXCSR"),
        D_CMD|D_MEMORY, DX_VEX|DX_LSHORT|DX_NOVREG,
        1, 0x000038FF, 0x000010AE, 0x00,
        B_MXCSR|B_CHG|B_PSEUDO,
        B_INT32|B_BINARY|B_MEMORY,
        B_NONE,
        B_NONE },

  { T("VSTMXCSR"),
        D_CMD|D_MEMORY, DX_VEX|DX_LSHORT|DX_NOVREG,
        1, 0x000038FF, 0x000018AE, 0x00,
        B_INT32|B_BINARY|B_MEMONLY|B_NOESP|B_SHOWSIZE|B_CHG,
        B_MXCSR|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("MASKMOVDQU"),
        D_SSE|D_MUST66|D_REGISTER, 0,
        2, 0x00C0FFFF, 0x00C0F70F, 0x00,
        B_SSEI8x16DI|B_UPD|B_PSEUDO,
        B_SREGI8x16|B_BINARY,
        B_SSEI8x16|B_REGISTER,
        B_NONE },

  { T("VMASKMOVDQU"),
        D_AVX|D_MUST66|D_REGISTER, DX_VEX|DX_LSHORT|DX_NOVREG,
        1, 0x0000C0FF, 0x0000C0F7, 0x00,
        B_SSEI8x16DI|B_UPD|B_PSEUDO,
        B_SREGI8x16|B_BINARY,
        B_SSEI8x16|B_REGISTER,
        B_NONE },

  { T("MASKMOVQ"),
        D_MMX|D_MUSTNONE|D_REGISTER, 0,
        2, 0x00C0FFFF, 0x00C0F70F, 0x00,
        B_MMX8x8DI|B_UPD|B_PSEUDO,
        B_MREG8x8,
        B_MMX8x8|B_REGISTER,
        B_NONE },

  { T("MAXPD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x00005F0F, 0x00,
        B_SREGF64x2|B_UPD,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("VMAXPD"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x0000005F, 0x00,
        B_SREGF64x2|B_CHG,
        B_SVEXF64x2,
        B_SSEF64x2,
        B_NONE },

  { T("MAXPS"),
        D_SSE|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x00005F0F, 0x00,
        B_SREGF32x4|B_UPD,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("VMAXPS"),
        D_AVX|D_MUSTNONE, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x0000005F, 0x00,
        B_SREGF32x4|B_CHG,
        B_SVEXF32x4,
        B_SSEF32x4,
        B_NONE },

  { T("MAXSD"),
        D_SSE|D_MUSTF2, 0,
        2, 0x0000FFFF, 0x00005F0F, 0x00,
        B_SREGF64L|B_UPD,
        B_SSEF64L,
        B_NONE,
        B_NONE },

  { T("VMAXSD"),
        D_AVX|D_MUSTF2, DX_VEX|DX_IGNOREL,
        1, 0x000000FF, 0x0000005F, 0x00,
        B_SREGF64L|B_CHG,
        B_SVEXF64L,
        B_SSEF64L,
        B_NONE },

  { T("MAXSS"),
        D_SSE|D_MUSTF3, 0,
        2, 0x0000FFFF, 0x00005F0F, 0x00,
        B_SREGF32L|B_UPD,
        B_SSEF32L,
        B_NONE,
        B_NONE },

  { T("VMAXSS"),
        D_AVX|D_MUSTF3, DX_VEX|DX_IGNOREL,
        1, 0x000000FF, 0x0000005F, 0x00,
        B_SREGF32L|B_CHG,
        B_SVEXF32L,
        B_SSEF32L,
        B_NONE },

  { T("MFENCE"),
        D_SSE, 0,
        3, 0x00FFFFFF, 0x00F0AE0F, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("MINPD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x00005D0F, 0x00,
        B_SREGF64x2|B_UPD,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("VMINPD"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x0000005D, 0x00,
        B_SREGF64x2|B_CHG,
        B_SVEXF64x2,
        B_SSEF64x2,
        B_NONE },

  { T("MINPS"),
        D_SSE|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x00005D0F, 0x00,
        B_SREGF32x4|B_UPD,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("VMINPS"),
        D_AVX|D_MUSTNONE, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x0000005D, 0x00,
        B_SREGF32x4|B_CHG,
        B_SVEXF32x4,
        B_SSEF32x4,
        B_NONE },

  { T("MINSD"),
        D_SSE|D_MUSTF2, 0,
        2, 0x0000FFFF, 0x00005D0F, 0x00,
        B_SREGF64L|B_UPD,
        B_SSEF64L,
        B_NONE,
        B_NONE },

  { T("VMINSD"),
        D_AVX|D_MUSTF2, DX_VEX|DX_IGNOREL,
        1, 0x000000FF, 0x0000005D, 0x00,
        B_SREGF64L|B_CHG,
        B_SVEXF64L,
        B_SSEF64L,
        B_NONE },

  { T("MINSS"),
        D_SSE|D_MUSTF3, 0,
        2, 0x0000FFFF, 0x00005D0F, 0x00,
        B_SREGF32L|B_UPD,
        B_SSEF32L,
        B_NONE,
        B_NONE },

  { T("VMINSS"),
        D_AVX|D_MUSTF3, DX_VEX|DX_IGNOREL,
        1, 0x000000FF, 0x0000005D, 0x00,
        B_SREGF32L|B_CHG,
        B_SVEXF32L,
        B_SSEF32L,
        B_NONE },

  { T("MOVAPD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000280F, 0x00,
        B_SREGF64x2|B_CHG,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("VMOVAPD"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x00000028, 0x00,
        B_SREGF64x2|B_CHG,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("MOVAPD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000290F, 0x00,
        B_SSEF64x2|B_CHG,
        B_SREGF64x2,
        B_NONE,
        B_NONE },

  { T("VMOVAPD"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x00000029, 0x00,
        B_SSEF64x2|B_CHG,
        B_SREGF64x2,
        B_NONE,
        B_NONE },

  { T("MOVAPS"),
        D_SSE|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000280F, 0x00,
        B_SREGF32x4|B_CHG,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("VMOVAPS"),
        D_AVX|D_MUSTNONE, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x00000028, 0x00,
        B_SREGF32x4|B_CHG,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("MOVAPS"),
        D_SSE|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000290F, 0x00,
        B_SSEF32x4|B_CHG,
        B_SREGF32x4,
        B_NONE,
        B_NONE },

  { T("VMOVAPS"),
        D_AVX|D_MUSTNONE, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x00000029, 0x00,
        B_SSEF32x4|B_CHG,
        B_SREGF32x4,
        B_NONE,
        B_NONE },

  { T("MOVD"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x00006E0F, 0x00,
        B_MREG32x2|B_CHG,
        B_INT32,
        B_NONE,
        B_NONE },

  { T("MOVD"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x00007E0F, 0x00,
        B_INT32|B_CHG,
        B_MREG32x2,
        B_NONE,
        B_NONE },

  { T("MOVD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x00006E0F, 0x00,
        B_SREGI32x2L|B_CHG,
        B_INT32,
        B_NONE,
        B_NONE },

  { T("VMOVD"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_NOVREG,
        1, 0x000000FF, 0x0000006E, 0x00,
        B_SREGI32x2L|B_CHG,
        B_INT32,
        B_NONE,
        B_NONE },

  { T("MOVD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x00007E0F, 0x00,
        B_INT32|B_CHG,
        B_SREGI32L,
        B_NONE,
        B_NONE },

  { T("VMOVD"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_NOVREG,
        1, 0x000000FF, 0x0000007E, 0x00,
        B_INT32|B_CHG,
        B_SREGI32L,
        B_NONE,
        B_NONE },

  { T("MOVDDUP"),
        D_SSE|D_MUSTF2, 0,
        2, 0x0000FFFF, 0x0000120F, 0x00,
        B_SREGF64x2|B_CHG,
        B_SSEF64L,
        B_NONE,
        B_NONE },

  { T("VMOVDDUP"),
        D_AVX|D_MUSTF2, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x00000012, 0x00,
        B_SREGF64x2|B_CHG,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("MOVDQA"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x00006F0F, 0x00,
        B_SREGF64x2|B_CHG,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("VMOVDQA"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x0000006F, 0x00,
        B_SREGF64x2|B_CHG,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("MOVDQA"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x00007F0F, 0x00,
        B_SSEF64x2|B_CHG,
        B_SREGF64x2,
        B_NONE,
        B_NONE },

  { T("VMOVDQA"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x0000007F, 0x00,
        B_SSEF64x2|B_CHG,
        B_SREGF64x2,
        B_NONE,
        B_NONE },

  { T("MOVDQU"),
        D_SSE|D_MUSTF3, 0,
        2, 0x0000FFFF, 0x00006F0F, 0x00,
        B_SREGF64x2|B_CHG,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("VMOVDQU"),
        D_AVX|D_MUSTF3, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x0000006F, 0x00,
        B_SREGF64x2|B_CHG,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("MOVDQU"),
        D_SSE|D_MUSTF3, 0,
        2, 0x0000FFFF, 0x00007F0F, 0x00,
        B_SSEF64x2|B_CHG,
        B_SREGF64x2,
        B_NONE,
        B_NONE },

  { T("VMOVDQU"),
        D_AVX|D_MUSTF3, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x0000007F, 0x00,
        B_SSEF64x2|B_CHG,
        B_SREGF64x2,
        B_NONE,
        B_NONE },

  { T("MOVDQ2Q"),
        D_MMX|D_MUSTF2|D_REGISTER, 0,
        2, 0x00C0FFFF, 0x00C0D60F, 0x00,
        B_MREG32x2|B_CHG,
        B_SSEI32x2L|B_REGISTER,
        B_NONE,
        B_NONE },

  { T("MOVHLPS"),
        D_SSE|D_MUSTNONE|D_REGISTER, 0,
        2, 0x00C0FFFF, 0x00C0120F, 0x00,
        B_SREGF32x4|B_UPD,
        B_SSEF32x4|B_REGISTER,
        B_NONE,
        B_NONE },

  { T("VMOVHLPS"),
        D_AVX|D_MUSTNONE|D_REGISTER, DX_VEX|DX_LSHORT,
        1, 0x0000C0FF, 0x0000C012, 0x00,
        B_SREGF32x4|B_CHG,
        B_SVEXF32x4,
        B_SSEF32x4|B_REGISTER,
        B_NONE },

  { T("MOVHPD"),
        D_SSE|D_MUST66|D_MEMORY, 0,
        2, 0x0000FFFF, 0x0000160F, 0x00,
        B_SREGF64x2|B_UPD,
        B_SSEF64L|B_MEMORY,
        B_NONE,
        B_NONE },

  { T("VMOVHPD"),
        D_AVX|D_MUST66|D_MEMORY, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x00000016, 0x00,
        B_SREGF64x2|B_CHG,
        B_SVEXF64x2,
        B_SSEF64L|B_MEMORY,
        B_NONE },

  { T("MOVHPD"),
        D_SSE|D_MUST66|D_MEMORY, 0,
        2, 0x0000FFFF, 0x0000170F, 0x00,
        B_SSEF64L|B_MEMORY|B_UPD,
        B_SREGF64x2,
        B_NONE,
        B_NONE },

  { T("VMOVHPD"),
        D_AVX|D_MUST66|D_MEMORY, DX_VEX|DX_LSHORT|DX_NOVREG,
        1, 0x000000FF, 0x00000017, 0x00,
        B_SSEF64L|B_MEMORY|B_UPD,
        B_SREGF64x2,
        B_NONE,
        B_NONE },

  { T("MOVHPS"),
        D_SSE|D_MUSTNONE|D_MEMORY, 0,
        2, 0x0000FFFF, 0x0000160F, 0x00,
        B_SREGF32x4|B_UPD,
        B_SSEF32x2L|B_MEMORY,
        B_NONE,
        B_NONE },

  { T("VMOVHPS"),
        D_AVX|D_MUSTNONE|D_MEMORY, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x00000016, 0x00,
        B_SREGF32x4|B_CHG,
        B_SVEXF32x4,
        B_SSEF32x2L|B_MEMORY,
        B_NONE },

  { T("MOVHPS"),
        D_SSE|D_MUSTNONE|D_MEMORY, 0,
        2, 0x0000FFFF, 0x0000170F, 0x00,
        B_SSEF32x2L|B_MEMORY|B_UPD,
        B_SREGF32x4,
        B_NONE,
        B_NONE },

  { T("VMOVHPS"),
        D_AVX|D_MUSTNONE|D_MEMORY, DX_VEX|DX_LSHORT|DX_NOVREG,
        1, 0x000000FF, 0x00000017, 0x00,
        B_SSEF32x2L|B_MEMORY|B_UPD,
        B_SREGF32x4,
        B_NONE,
        B_NONE },

  { T("MOVLHPS"),
        D_SSE|D_MUSTNONE|D_REGISTER, 0,
        2, 0x00C0FFFF, 0x00C0160F, 0x00,
        B_SREGF32x4|B_UPD,
        B_SSEF32x2L|B_REGISTER,
        B_NONE,
        B_NONE },

  { T("VMOVLHPS"),
        D_AVX|D_MUSTNONE|D_REGISTER, DX_VEX|DX_LSHORT,
        1, 0x0000C0FF, 0x0000C016, 0x00,
        B_SREGF32x4|B_UPD,
        B_SVEXF32x4,
        B_SSEF32x2L|B_REGISTER,
        B_NONE },

  { T("MOVLPD"),
        D_SSE|D_MUST66|D_MEMORY, 0,
        2, 0x0000FFFF, 0x0000120F, 0x00,
        B_SREGF64L|B_UPD,
        B_SSEF64L|B_MEMORY,
        B_NONE,
        B_NONE },

  { T("VMOVLPD"),
        D_AVX|D_MUST66|D_MEMORY, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x00000012, 0x00,
        B_SREGF64x2|B_UPD,
        B_SVEXF64x2,
        B_SSEF64L|B_MEMORY,
        B_NONE },

  { T("MOVLPD"),
        D_SSE|D_MUST66|D_MEMORY, 0,
        2, 0x0000FFFF, 0x0000130F, 0x00,
        B_SSEF64L|B_MEMORY|B_UPD,
        B_SREGF64L,
        B_NONE,
        B_NONE },

  { T("VMOVLPD"),
        D_AVX|D_MUST66|D_MEMORY, DX_VEX|DX_LSHORT|DX_NOVREG,
        1, 0x000000FF, 0x00000013, 0x00,
        B_SSEF64L|B_MEMORY|B_UPD,
        B_SREGF64L,
        B_NONE,
        B_NONE },

  { T("MOVLPS"),
        D_SSE|D_MUSTNONE|D_MEMORY, 0,
        2, 0x0000FFFF, 0x0000120F, 0x00,
        B_SREGF32x2L|B_UPD,
        B_SSEF32x2L|B_MEMORY,
        B_NONE,
        B_NONE },

  { T("VMOVLPS"),
        D_AVX|D_MUSTNONE|D_MEMORY, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x00000012, 0x00,
        B_SREGF32x4|B_UPD,
        B_SVEXF32x4,
        B_SSEF32x2L|B_MEMORY,
        B_NONE },

  { T("MOVLPS"),
        D_SSE|D_MUSTNONE|D_MEMORY, 0,
        2, 0x0000FFFF, 0x0000130F, 0x00,
        B_SSEF32x2L|B_MEMORY|B_UPD,
        B_SREGF32x2L,
        B_NONE,
        B_NONE },

  { T("VMOVLPS"),
        D_AVX|D_MUSTNONE|D_MEMORY, DX_VEX|DX_LSHORT|DX_NOVREG,
        1, 0x000000FF, 0x00000013, 0x00,
        B_SSEF32x2L|B_MEMORY|B_UPD,
        B_SREGF32x2L,
        B_NONE,
        B_NONE },

  { T("MOVMSKPD"),
        D_SSE|D_MUST66|D_REGISTER, 0,
        2, 0x00C0FFFF, 0x00C0500F, 0x00,
        B_REG32|B_CHG,
        B_SSEF64x2|B_REGONLY,
        B_NONE,
        B_NONE },

  { T("VMOVMSKPD"),
        D_AVX|D_MUST66|D_REGISTER, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x0000C0FF, 0x0000C050, 0x00,
        B_REG32|B_CHG,
        B_SSEF64x2|B_REGONLY,
        B_NONE,
        B_NONE },

  { T("MOVMSKPS"),
        D_SSE|D_MUSTNONE|D_REGISTER, 0,
        2, 0x00C0FFFF, 0x00C0500F, 0x00,
        B_REG32|B_CHG,
        B_SSEF32x4|B_REGONLY,
        B_NONE,
        B_NONE },

  { T("VMOVMSKPS"),
        D_AVX|D_MUSTNONE|D_REGISTER, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x0000C0FF, 0x0000C050, 0x00,
        B_REG32|B_CHG,
        B_SSEF32x4|B_REGONLY,
        B_NONE,
        B_NONE },

  { T("MOVNTDQ"),
        D_SSE|D_MUST66|D_MEMORY, 0,
        2, 0x0000FFFF, 0x0000E70F, 0x00,
        B_SSEI8x16|B_MEMORY|B_CHG,
        B_SREGI8x16|B_BINARY,
        B_NONE,
        B_NONE },

  { T("VMOVNTDQ"),
        D_AVX|D_MUST66|D_MEMORY, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x000000E7, 0x00,
        B_SSEI8x16|B_MEMORY|B_CHG,
        B_SREGI8x16|B_BINARY,
        B_NONE,
        B_NONE },

  { T("MOVNTI"),
        D_SSE|D_MUSTNONE|D_MEMORY, 0,
        2, 0x0000FFFF, 0x0000C30F, 0x00,
        B_INT32|B_MEMORY|B_CHG,
        B_REG32,
        B_NONE,
        B_NONE },

  { T("MOVNTPD"),
        D_SSE|D_MUST66|D_MEMORY, 0,
        2, 0x0000FFFF, 0x00002B0F, 0x00,
        B_SSEF64x2|B_MEMORY|B_CHG,
        B_SREGF64x2,
        B_NONE,
        B_NONE },

  { T("VMOVNTPD"),
        D_AVX|D_MUST66|D_MEMORY, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x0000002B, 0x00,
        B_SSEF64x2|B_MEMORY|B_CHG,
        B_SREGF64x2,
        B_NONE,
        B_NONE },

  { T("MOVNTPS"),
        D_SSE|D_MUSTNONE|D_MEMORY, 0,
        2, 0x0000FFFF, 0x00002B0F, 0x00,
        B_SSEF32x4|B_MEMORY|B_CHG,
        B_SREGF32x4,
        B_NONE,
        B_NONE },

  { T("VMOVNTPS"),
        D_AVX|D_MUSTNONE|D_MEMORY, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x0000002B, 0x00,
        B_SSEF32x4|B_MEMORY|B_CHG,
        B_SREGF32x4,
        B_NONE,
        B_NONE },

  { T("MOVNTQ"),
        D_MMX|D_MUSTNONE|D_MEMORY, 0,
        2, 0x0000FFFF, 0x0000E70F, 0x00,
        B_MMX64|B_MEMORY|B_CHG,
        B_MREG64,
        B_NONE,
        B_NONE },

  { T("MOVQ"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x00006F0F, 0x00,
        B_MREG64|B_CHG,
        B_MMX64,
        B_NONE,
        B_NONE },

  { T("MOVQ"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x00007F0F, 0x00,
        B_MMX64|B_CHG,
        B_MREG64,
        B_NONE,
        B_NONE },

  { T("MOVQ"),
        D_SSE|D_MUSTF3, 0,
        2, 0x0000FFFF, 0x00007E0F, 0x00,
        B_SREGF64L|B_CHG,
        B_SSEF64L,
        B_NONE,
        B_NONE },

  { T("VMOVQ"),
        D_AVX|D_MUSTF3, DX_VEX|DX_LSHORT|DX_NOVREG,
        1, 0x000000FF, 0x0000007E, 0x00,
        B_SREGF64L|B_CHG,
        B_SSEF64L,
        B_NONE,
        B_NONE },

  { T("MOVQ"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000D60F, 0x00,
        B_SSEF64L|B_CHG,
        B_SREGF64L,
        B_NONE,
        B_NONE },

  { T("VMOVQ"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_NOVREG,
        1, 0x000000FF, 0x000000D6, 0x00,
        B_SSEF64L|B_CHG,
        B_SREGF64L,
        B_NONE,
        B_NONE },

  { T("MOVQ2DQ"),
        D_MMX|D_MUSTF3|D_REGISTER, 0,
        2, 0x00C0FFFF, 0x00C0D60F, 0x00,
        B_SREGF64L|B_UPD,
        B_MMX8x8|B_REGISTER,
        B_NONE,
        B_NONE },

  { T("MOVSD"),
        D_SSE|D_MUSTF2, 0,
        2, 0x0000FFFF, 0x0000100F, 0x00,
        B_SREGF64L|B_UPD,
        B_SSEF64L,
        B_NONE,
        B_NONE },

  { T("VMOVSD"),
        D_AVX|D_MUSTF2|D_MEMORY, DX_VEX|DX_IGNOREL|DX_NOVREG,
        1, 0x000000FF, 0x00000010, 0x00,
        B_SREGF64L|B_UPD,
        B_SSEF64L,
        B_NONE,
        B_NONE },

  { T("VMOVSD"),
        D_AVX|D_MUSTF2|D_REGISTER, DX_VEX|DX_IGNOREL,
        1, 0x000000FF, 0x00000010, 0x00,
        B_SREGF64L|B_UPD,
        B_SVEXF64x2,
        B_SSEF64L,
        B_NONE },

  { T("MOVSD"),
        D_SSE|D_MUSTF2, 0,
        2, 0x0000FFFF, 0x0000110F, 0x00,
        B_SSEF64L|B_UPD,
        B_SREGF64L,
        B_NONE,
        B_NONE },

  { T("VMOVSD"),
        D_AVX|D_MUSTF2|D_MEMORY, DX_VEX|DX_IGNOREL|DX_NOVREG,
        1, 0x000000FF, 0x00000011, 0x00,
        B_SSEF64L|B_UPD,
        B_SREGF64L,
        B_NONE,
        B_NONE },

  { T("VMOVSD"),
        D_AVX|D_MUSTF2|D_REGISTER, DX_VEX|DX_IGNOREL,
        1, 0x000000FF, 0x00000011, 0x00,
        B_SSEF64L|B_UPD,
        B_SVEXF64x2,
        B_SREGF64L,
        B_NONE },

  { T("MOVSS"),
        D_SSE|D_MUSTF3, 0,
        2, 0x0000FFFF, 0x0000100F, 0x00,
        B_SREGF32L|B_UPD,
        B_SSEF32L,
        B_NONE,
        B_NONE },

  { T("VMOVSS"),
        D_AVX|D_MUSTF3|D_MEMORY, DX_VEX|DX_IGNOREL|DX_NOVREG,
        1, 0x000000FF, 0x00000010, 0x00,
        B_SREGF32L|B_UPD,
        B_SSEF32L,
        B_NONE,
        B_NONE },

  { T("VMOVSS"),
        D_AVX|D_MUSTF3|D_REGISTER, DX_VEX|DX_IGNOREL,
        1, 0x000000FF, 0x00000010, 0x00,
        B_SREGF32L|B_UPD,
        B_SVEXF32x4,
        B_SSEF32L,
        B_NONE },

  { T("MOVSS"),
        D_SSE|D_MUSTF3, 0,
        2, 0x0000FFFF, 0x0000110F, 0x00,
        B_SSEF32L|B_UPD,
        B_SREGF32L,
        B_NONE,
        B_NONE },

  { T("VMOVSS"),
        D_AVX|D_MUSTF3|D_MEMORY, DX_VEX|DX_IGNOREL|DX_NOVREG,
        1, 0x000000FF, 0x00000011, 0x00,
        B_SSEF32L|B_UPD,
        B_SREGF32L,
        B_NONE,
        B_NONE },

  { T("VMOVSS"),
        D_AVX|D_MUSTF3|D_REGISTER, DX_VEX|DX_IGNOREL,
        1, 0x000000FF, 0x00000011, 0x00,
        B_SSEF32L|B_UPD,
        B_SVEXF32x4,
        B_SREGF32L,
        B_NONE },

  { T("MOVSHDUP"),
        D_SSE|D_MUSTF3, 0,
        2, 0x0000FFFF, 0x0000160F, 0x00,
        B_SREGF32x4|B_CHG,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("VMOVSHDUP"),
        D_AVX|D_MUSTF3, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x00000016, 0x00,
        B_SREGF32x4|B_CHG,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("MOVSLDUP"),
        D_SSE|D_MUSTF3, 0,
        2, 0x0000FFFF, 0x0000120F, 0x00,
        B_SREGF32x4|B_CHG,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("VMOVSLDUP"),
        D_AVX|D_MUSTF3, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x00000012, 0x00,
        B_SREGF32x4|B_CHG,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("MOVUPD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000100F, 0x00,
        B_SREGF64x2|B_CHG,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("VMOVUPD"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x00000010, 0x00,
        B_SREGF64x2|B_CHG,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("MOVUPD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000110F, 0x00,
        B_SSEF64x2|B_CHG,
        B_SREGF64x2,
        B_NONE,
        B_NONE },

  { T("VMOVUPD"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x00000011, 0x00,
        B_SSEF64x2|B_CHG,
        B_SREGF64x2,
        B_NONE,
        B_NONE },

  { T("MOVUPS"),
        D_SSE|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000100F, 0x00,
        B_SREGF32x4|B_CHG,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("VMOVUPS"),
        D_AVX|D_MUSTNONE, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x00000010, 0x00,
        B_SREGF32x4|B_CHG,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("MOVUPS"),
        D_SSE|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000110F, 0x00,
        B_SSEF32x4|B_CHG,
        B_SREGF32x4,
        B_NONE,
        B_NONE },

  { T("VMOVUPS"),
        D_AVX|D_MUSTNONE, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x00000011, 0x00,
        B_SSEF32x4|B_CHG,
        B_SREGF32x4,
        B_NONE,
        B_NONE },

  { T("MULPD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000590F, 0x00,
        B_SREGF64x2|B_UPD,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("VMULPD"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x00000059, 0x00,
        B_SREGF64x2|B_UPD,
        B_SVEXF64x2,
        B_SSEF64x2,
        B_NONE },

  { T("MULPS"),
        D_SSE|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000590F, 0x00,
        B_SREGF32x4|B_UPD,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("VMULPS"),
        D_AVX|D_MUSTNONE, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x00000059, 0x00,
        B_SREGF32x4|B_UPD,
        B_SVEXF32x4,
        B_SSEF32x4,
        B_NONE },

  { T("MULSD"),
        D_SSE|D_MUSTF2, 0,
        2, 0x0000FFFF, 0x0000590F, 0x00,
        B_SREGF64L|B_UPD,
        B_SSEF64L,
        B_NONE,
        B_NONE },

  { T("VMULSD"),
        D_AVX|D_MUSTF2, DX_VEX|DX_IGNOREL,
        1, 0x000000FF, 0x00000059, 0x00,
        B_SREGF64L|B_UPD,
        B_SVEXF64L,
        B_SSEF64L,
        B_NONE },

  { T("MULSS"),
        D_SSE|D_MUSTF3, 0,
        2, 0x0000FFFF, 0x0000590F, 0x00,
        B_SREGF32L|B_UPD,
        B_SSEF32L,
        B_NONE,
        B_NONE },

  { T("VMULSS"),
        D_AVX|D_MUSTF3, DX_VEX|DX_IGNOREL,
        1, 0x000000FF, 0x00000059, 0x00,
        B_SREGF32L|B_UPD,
        B_SVEXF32L,
        B_SSEF32L,
        B_NONE },

  { T("ORPD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000560F, 0x00,
        B_SREGF64x2|B_UPD,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("VORPD"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x00000056, 0x00,
        B_SREGF64x2|B_UPD,
        B_SVEXF64x2,
        B_SSEF64x2,
        B_NONE },

  { T("ORPS"),
        D_SSE|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000560F, 0x00,
        B_SREGF32x4|B_UPD,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("VORPS"),
        D_AVX|D_MUSTNONE, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x00000056, 0x00,
        B_SREGF32x4|B_UPD,
        B_SVEXF32x4,
        B_SSEF32x4,
        B_NONE },

  { T("PACKSSWB"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000630F, 0x00,
        B_MREG16x4|B_UPD,
        B_MMX16x4,
        B_NONE,
        B_NONE },

  { T("PACKSSWB"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000630F, 0x00,
        B_SREGI16x8|B_UPD,
        B_SSEI16x8,
        B_NONE,
        B_NONE },

  { T("VPACKSSWB"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x00000063, 0x00,
        B_SREGI8x16|B_UPD,
        B_SVEXI16x8,
        B_SSEI16x8,
        B_NONE },

  { T("PACKSSDW"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x00006B0F, 0x00,
        B_MREG32x2|B_UPD,
        B_MMX32x2,
        B_NONE,
        B_NONE },

  { T("PACKSSDW"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x00006B0F, 0x00,
        B_SREGI32x4|B_UPD,
        B_SSEI32x4,
        B_NONE,
        B_NONE },

  { T("VPACKSSDW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x0000006B, 0x00,
        B_SREGI16x8|B_UPD,
        B_SVEXI32x4,
        B_SSEI32x4,
        B_NONE },

  { T("PACKUSWB"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000670F, 0x00,
        B_MREG16x4|B_UPD,
        B_MMX16x4,
        B_NONE,
        B_NONE },

  { T("PACKUSWB"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000670F, 0x00,
        B_SREGI16x8|B_UPD,
        B_SSEI16x8,
        B_NONE,
        B_NONE },

  { T("VPACKUSWB"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x00000067, 0x00,
        B_SREGI8x16|B_UPD,
        B_SVEXI16x8,
        B_SSEI16x8,
        B_NONE },

  { T("PADDB"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000FC0F, 0x00,
        B_MREG8x8|B_UPD,
        B_MMX8x8,
        B_NONE,
        B_NONE },

  { T("PADDW"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000FD0F, 0x00,
        B_MREG16x4|B_UPD,
        B_MMX16x4,
        B_NONE,
        B_NONE },

  { T("PADDD"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000FE0F, 0x00,
        B_MREG32x2|B_UPD,
        B_MMX32x2,
        B_NONE,
        B_NONE },

  { T("PADDB"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000FC0F, 0x00,
        B_SREGI8x16|B_UPD,
        B_SSEI8x16,
        B_NONE,
        B_NONE },

  { T("VPADDB"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000FC, 0x00,
        B_SREGI8x16|B_UPD,
        B_SVEXI8x16,
        B_SSEI8x16,
        B_NONE },

  { T("PADDW"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000FD0F, 0x00,
        B_SREGI16x8|B_UPD,
        B_SSEI16x8,
        B_NONE,
        B_NONE },

  { T("VPADDW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000FD, 0x00,
        B_SREGI16x8|B_UPD,
        B_SVEXI16x8,
        B_SSEI16x8,
        B_NONE },

  { T("PADDD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000FE0F, 0x00,
        B_SREGI32x4|B_UPD,
        B_SSEI32x4,
        B_NONE,
        B_NONE },

  { T("VPADDD"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000FE, 0x00,
        B_SREGI32x4|B_UPD,
        B_SVEXI32x4,
        B_SSEI32x4,
        B_NONE },

  { T("PADDQ"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000D40F, 0x00,
        B_MREG64|B_UPD,
        B_MMX64,
        B_NONE,
        B_NONE },

  { T("PADDQ"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000D40F, 0x00,
        B_SREGI64x2|B_UPD,
        B_SSEI64x2,
        B_NONE,
        B_NONE },

  { T("VPADDQ"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000D4, 0x00,
        B_SREGI64x2|B_UPD,
        B_SVEXI64x2,
        B_SSEI64x2,
        B_NONE },

  { T("PADDSB"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000EC0F, 0x00,
        B_MREG8x8|B_UPD,
        B_MMX8x8,
        B_NONE,
        B_NONE },

  { T("PADDSW"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000ED0F, 0x00,
        B_MREG16x4|B_UPD,
        B_MMX16x4,
        B_NONE,
        B_NONE },

  { T("PADDSB"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000EC0F, 0x00,
        B_SREGI8x16|B_UPD,
        B_SSEI8x16,
        B_NONE,
        B_NONE },

  { T("VPADDSB"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000EC, 0x00,
        B_SREGI8x16|B_UPD,
        B_SVEXI8x16,
        B_SSEI8x16,
        B_NONE },

  { T("PADDSW"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000ED0F, 0x00,
        B_SREGI16x8|B_UPD,
        B_SSEI16x8,
        B_NONE,
        B_NONE },

  { T("VPADDSW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000ED, 0x00,
        B_SREGI16x8|B_UPD,
        B_SVEXI16x8,
        B_SSEI16x8,
        B_NONE },

  { T("PADDUSB"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000DC0F, 0x00,
        B_MREG8x8|B_UPD,
        B_MMX8x8,
        B_NONE,
        B_NONE },

  { T("PADDUSW"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000DD0F, 0x00,
        B_MREG16x4|B_UPD,
        B_MMX16x4,
        B_NONE,
        B_NONE },

  { T("PADDUSB"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000DC0F, 0x00,
        B_SREGI8x16|B_UPD,
        B_SSEI8x16,
        B_NONE,
        B_NONE },

  { T("VPADDUSB"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000DC, 0x00,
        B_SREGI8x16|B_UPD,
        B_SVEXI8x16,
        B_SSEI8x16,
        B_NONE },

  { T("PADDUSW"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000DD0F, 0x00,
        B_SREGI16x8|B_UPD,
        B_SSEI16x8,
        B_NONE,
        B_NONE },

  { T("VPADDUSW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000DD, 0x00,
        B_SREGI16x8|B_UPD,
        B_SVEXI16x8,
        B_SSEI16x8,
        B_NONE },

  { T("PAND"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000DB0F, 0x00,
        B_MREG8x8|B_UPD,
        B_MMX8x8,
        B_NONE,
        B_NONE },

  { T("PAND"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000DB0F, 0x00,
        B_SREGI8x16|B_UPD,
        B_SSEI8x16,
        B_NONE,
        B_NONE },

  { T("VPAND"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000DB, 0x00,
        B_SREGI8x16|B_UPD,
        B_SVEXI8x16,
        B_SSEI8x16,
        B_NONE },

  { T("PANDN"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000DF0F, 0x00,
        B_MREG8x8|B_UPD,
        B_MMX8x8,
        B_NONE,
        B_NONE },

  { T("PANDN"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000DF0F, 0x00,
        B_SREGI8x16|B_UPD,
        B_SSEI8x16,
        B_NONE,
        B_NONE },

  { T("VPANDN"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000DF, 0x00,
        B_SREGI8x16|B_UPD,
        B_SVEXI8x16,
        B_SSEI8x16,
        B_NONE },

  { T("PAVGB"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000E00F, 0x00,
        B_MREG8x8|B_UPD,
        B_MMX8x8,
        B_NONE,
        B_NONE },

  { T("PAVGW"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000E30F, 0x00,
        B_MREG16x4|B_UPD,
        B_MMX16x4,
        B_NONE,
        B_NONE },

  { T("PAVGB"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000E00F, 0x00,
        B_SREGI8x16|B_UPD,
        B_SSEI8x16,
        B_NONE,
        B_NONE },

  { T("VPAVGB"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000E0, 0x00,
        B_SREGI8x16|B_UPD,
        B_SVEXI8x16,
        B_SSEI8x16,
        B_NONE },

  { T("PAVGW"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000E30F, 0x00,
        B_SREGI16x8|B_UPD,
        B_SSEI16x8,
        B_NONE,
        B_NONE },

  { T("VPAVGW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000E3, 0x00,
        B_SREGI16x8|B_UPD,
        B_SVEXI16x8,
        B_SSEI16x8,
        B_NONE },

  { T("PCMPEQB"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000740F, 0x00,
        B_MREG8x8|B_UPD,
        B_MMX8x8,
        B_NONE,
        B_NONE },

  { T("PCMPEQW"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000750F, 0x00,
        B_MREG16x4|B_UPD,
        B_MMX16x4,
        B_NONE,
        B_NONE },

  { T("PCMPEQD"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000760F, 0x00,
        B_MREG32x2|B_UPD,
        B_MMX32x2,
        B_NONE,
        B_NONE },

  { T("PCMPEQB"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000740F, 0x00,
        B_SREGI8x16|B_UPD,
        B_SSEI8x16,
        B_NONE,
        B_NONE },

  { T("VPCMPEQB"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x00000074, 0x00,
        B_SREGI8x16|B_UPD,
        B_SVEXI8x16,
        B_SSEI8x16,
        B_NONE },

  { T("PCMPEQW"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000750F, 0x00,
        B_SREGI16x8|B_UPD,
        B_SSEI16x8,
        B_NONE,
        B_NONE },

  { T("VPCMPEQW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x00000075, 0x00,
        B_SREGI16x8|B_UPD,
        B_SVEXI16x8,
        B_SSEI16x8,
        B_NONE },

  { T("PCMPEQD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000760F, 0x00,
        B_SREGI32x4|B_UPD,
        B_SSEI32x4,
        B_NONE,
        B_NONE },

  { T("VPCMPEQD"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x00000076, 0x00,
        B_SREGI32x4|B_UPD,
        B_SVEXI32x4,
        B_SSEI32x4,
        B_NONE },

  { T("PCMPGTB"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000640F, 0x00,
        B_MREG8x8|B_UPD,
        B_MMX8x8,
        B_NONE,
        B_NONE },

  { T("PCMPGTW"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000650F, 0x00,
        B_MREG16x4|B_UPD,
        B_MMX16x4,
        B_NONE,
        B_NONE },

  { T("PCMPGTD"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000660F, 0x00,
        B_MREG32x2|B_UPD,
        B_MMX32x2,
        B_NONE,
        B_NONE },

  { T("PCMPGTB"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000640F, 0x00,
        B_SREGI8x16|B_UPD,
        B_SSEI8x16,
        B_NONE,
        B_NONE },

  { T("VPCMPGTB"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x00000064, 0x00,
        B_SREGI8x16|B_UPD,
        B_SVEXI8x16,
        B_SSEI8x16,
        B_NONE },

  { T("PCMPGTW"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000650F, 0x00,
        B_SREGI16x8|B_UPD,
        B_SSEI16x8,
        B_NONE,
        B_NONE },

  { T("VPCMPGTW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x00000065, 0x00,
        B_SREGI16x8|B_UPD,
        B_SVEXI16x8,
        B_SSEI16x8,
        B_NONE },

  { T("PCMPGTD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000660F, 0x00,
        B_SREGI32x4|B_UPD,
        B_SSEI32x4,
        B_NONE,
        B_NONE },

  { T("VPCMPGTD"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x00000066, 0x00,
        B_SREGI32x4|B_UPD,
        B_SVEXI32x4,
        B_SSEI32x4,
        B_NONE },

  { T("PEXTRW"),
        D_MMX|D_MUSTNONE|D_REGISTER, 0,
        2, 0x00C0FFFF, 0x00C0C50F, 0x00,
        B_REG32|B_CHG,
        B_MMX16x4|B_REGISTER,
        B_CONST8|B_COUNT,
        B_NONE },

  { T("PEXTRW"),
        D_SSE|D_MUST66|D_REGISTER, 0,
        2, 0x00C0FFFF, 0x00C0C50F, 0x00,
        B_REG32|B_CHG,
        B_SSEI16x8|B_REGISTER,
        B_CONST8|B_COUNT,
        B_NONE },

  { T("VPEXTRW"),
        D_AVX|D_MUST66|D_REGISTER, DX_VEX|DX_LSHORT|DX_NOVREG,
        1, 0x0000C0FF, 0x0000C0C5, 0x00,
        B_REG32|B_CHG,
        B_SSEI16x8|B_REGISTER,
        B_CONST8|B_COUNT,
        B_NONE },

  { T("PINSRW"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000C40F, 0x00,
        B_MREG16x4|B_UPD,
        B_INT16,
        B_CONST8|B_COUNT,
        B_NONE },

  { T("PINSRW"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x00C0FFFF, 0x00C0C40F, 0x00,
        B_MREG16x4|B_UPD,
        B_INT32|B_REGISTER,
        B_CONST8|B_COUNT,
        B_NONE },

  { T("PINSRW"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000C40F, 0x00,
        B_SREGI16x8|B_UPD,
        B_INT16,
        B_CONST8|B_COUNT,
        B_NONE },

  { T("VPINSRW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000C4, 0x00,
        B_SREGI16x8|B_UPD,
        B_SVEXI16x8,
        B_INT16,
        B_CONST8|B_COUNT },

  { T("PINSRW"),
        D_SSE|D_MUST66, 0,
        2, 0x00C0FFFF, 0x00C0C40F, 0x00,
        B_SREGI16x8|B_UPD,
        B_INT32|B_REGISTER,
        B_CONST8|B_COUNT,
        B_NONE },

  { T("VPINSRW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x0000C0FF, 0x0000C0C4, 0x00,
        B_SREGI16x8|B_UPD,
        B_SVEXI16x8,
        B_INT32|B_REGISTER,
        B_CONST8|B_COUNT },

  { T("PMADDWD"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000F50F, 0x00,
        B_MREG16x4|B_UPD,
        B_MMX16x4,
        B_NONE,
        B_NONE },

  { T("PMADDWD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000F50F, 0x00,
        B_SREGI16x8|B_UPD,
        B_SSEI16x8,
        B_NONE,
        B_NONE },

  { T("VPMADDWD"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000F5, 0x00,
        B_SREGI16x8|B_UPD,
        B_SVEXI16x8,
        B_SSEI16x8,
        B_NONE },

  { T("PMAXSW"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000EE0F, 0x00,
        B_MREG16x4|B_UPD,
        B_MMX16x4,
        B_NONE,
        B_NONE },

  { T("PMAXSW"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000EE0F, 0x00,
        B_SREGI16x8|B_UPD,
        B_SSEI16x8,
        B_NONE,
        B_NONE },

  { T("VPMAXSW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000EE, 0x00,
        B_SREGI16x8|B_UPD,
        B_SVEXI16x8,
        B_SSEI16x8,
        B_NONE },

  { T("PMAXUB"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000DE0F, 0x00,
        B_MREG8x8|B_UPD,
        B_MMX8x8,
        B_NONE,
        B_NONE },

  { T("PMAXUB"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000DE0F, 0x00,
        B_SREGI8x16|B_UPD,
        B_SSEI8x16,
        B_NONE,
        B_NONE },

  { T("VPMAXUB"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000DE, 0x00,
        B_SREGI8x16|B_UPD,
        B_SVEXI8x16,
        B_SSEI8x16,
        B_NONE },

  { T("PMINSW"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000EA0F, 0x00,
        B_MREG16x4|B_UPD,
        B_MMX16x4,
        B_NONE,
        B_NONE },

  { T("PMINSW"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000EA0F, 0x00,
        B_SREGI16x8|B_UPD,
        B_SSEI16x8,
        B_NONE,
        B_NONE },

  { T("VPMINSW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000EA, 0x00,
        B_SREGI16x8|B_UPD,
        B_SVEXI16x8,
        B_SSEI16x8,
        B_NONE },

  { T("PMINUB"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000DA0F, 0x00,
        B_MREG8x8|B_UPD,
        B_MMX8x8,
        B_NONE,
        B_NONE },

  { T("PMINUB"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000DA0F, 0x00,
        B_SREGI8x16|B_UPD,
        B_SSEI8x16,
        B_NONE,
        B_NONE },

  { T("VPMINUB"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000DA, 0x00,
        B_SREGI8x16|B_UPD,
        B_SVEXI8x16,
        B_SSEI8x16,
        B_NONE },

  { T("PMOVMSKB"),
        D_MMX|D_MUSTNONE|D_REGISTER, 0,
        2, 0x00C0FFFF, 0x00C0D70F, 0x00,
        B_REG32|B_CHG,
        B_MMX8x8|B_REGISTER,
        B_NONE,
        B_NONE },

  { T("PMOVMSKB"),
        D_SSE|D_MUST66|D_REGISTER, 0,
        2, 0x00C0FFFF, 0x00C0D70F, 0x00,
        B_REG32|B_CHG,
        B_SSEI8x16|B_REGISTER,
        B_NONE,
        B_NONE },

  { T("VPMOVMSKB"),
        D_AVX|D_MUST66|D_REGISTER, DX_VEX|DX_LSHORT|DX_NOVREG,
        1, 0x0000C0FF, 0x0000C0D7, 0x00,
        B_REG32|B_CHG,
        B_SSEI8x16|B_REGISTER,
        B_NONE,
        B_NONE },

  { T("PMULHUW"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000E40F, 0x00,
        B_MREG16x4|B_UPD,
        B_MMX16x4,
        B_NONE,
        B_NONE },

  { T("PMULHUW"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000E40F, 0x00,
        B_SREGI16x8|B_UPD,
        B_SSEI16x8,
        B_NONE,
        B_NONE },

  { T("VPMULHUW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000E4, 0x00,
        B_SREGI16x8|B_UPD,
        B_SVEXI16x8,
        B_SSEI16x8,
        B_NONE },

  { T("PMULHW"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000E50F, 0x00,
        B_MREG16x4|B_UPD,
        B_MMX16x4,
        B_NONE,
        B_NONE },

  { T("PMULHW"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000E50F, 0x00,
        B_SREGI16x8|B_UPD,
        B_SSEI16x8,
        B_NONE,
        B_NONE },

  { T("VPMULHW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000E5, 0x00,
        B_SREGI16x8|B_UPD,
        B_SVEXI16x8,
        B_SSEI16x8,
        B_NONE },

  { T("PMULLW"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000D50F, 0x00,
        B_MREG16x4|B_UPD,
        B_MMX16x4,
        B_NONE,
        B_NONE },

  { T("PMULLW"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000D50F, 0x00,
        B_SREGI16x8|B_UPD,
        B_SSEI16x8,
        B_NONE,
        B_NONE },

  { T("VPMULLW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000D5, 0x00,
        B_SREGI16x8|B_UPD,
        B_SVEXI16x8,
        B_SSEI16x8,
        B_NONE },

  { T("PMULUDQ"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000F40F, 0x00,
        B_MREG32x2|B_UPD,
        B_MMX32x2,
        B_NONE,
        B_NONE },

  { T("PMULUDQ"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000F40F, 0x00,
        B_SREGI32x4|B_UPD,
        B_SSEI32x4,
        B_NONE,
        B_NONE },

  { T("VPMULUDQ"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000F4, 0x00,
        B_SREGI32x4|B_UPD,
        B_SVEXI32x4,
        B_SSEI32x4,
        B_NONE },

  { T("POR"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000EB0F, 0x00,
        B_MREG8x8|B_UPD,
        B_MMX8x8,
        B_NONE,
        B_NONE },

  { T("POR"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000EB0F, 0x00,
        B_SREGI8x16|B_BINARY|B_UPD,
        B_SSEI8x16,
        B_NONE,
        B_NONE },

  { T("VPOR"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000EB, 0x00,
        B_SREGI8x16|B_BINARY|B_UPD,
        B_SVEXI8x16|B_BINARY,
        B_SSEI8x16,
        B_NONE },

  { T("PSADBW"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000F60F, 0x00,
        B_MREG8x8|B_UPD,
        B_MMX8x8,
        B_NONE,
        B_NONE },

  { T("PSADBW"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000F60F, 0x00,
        B_SREGI8x16|B_UPD,
        B_SSEI8x16,
        B_NONE,
        B_NONE },

  { T("VPSADBW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000F6, 0x00,
        B_SREGI8x16|B_UPD,
        B_SVEXI8x16,
        B_SSEI8x16,
        B_NONE },

  { T("PSHUFD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000700F, 0x00,
        B_SREGI32x4|B_CHG,
        B_SSEI32x4,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("VPSHUFD"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_NOVREG,
        1, 0x000000FF, 0x00000070, 0x00,
        B_SREGI32x4|B_CHG,
        B_SSEI32x4,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("PSHUFHW"),
        D_SSE|D_MUSTF3, 0,
        2, 0x0000FFFF, 0x0000700F, 0x00,
        B_SREGI16x8|B_CHG,
        B_SSEI16x8,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("VPSHUFHW"),
        D_AVX|D_MUSTF3, DX_VEX|DX_LSHORT|DX_NOVREG,
        1, 0x000000FF, 0x00000070, 0x00,
        B_SREGI16x8|B_CHG,
        B_SSEI16x8,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("PSHUFLW"),
        D_SSE|D_MUSTF2, 0,
        2, 0x0000FFFF, 0x0000700F, 0x00,
        B_SREGI16x8|B_CHG,
        B_SSEI16x8,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("VPSHUFLW"),
        D_AVX|D_MUSTF2, DX_VEX|DX_LSHORT|DX_NOVREG,
        1, 0x000000FF, 0x00000070, 0x00,
        B_SREGI16x8|B_CHG,
        B_SSEI16x8,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("PSHUFW"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000700F, 0x00,
        B_MREG16x4|B_CHG,
        B_MMX16x4,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("PSLLDQ"),
        D_SSE|D_MUST66|D_REGISTER, 0,
        2, 0x00F8FFFF, 0x00F8730F, 0x00,
        B_SSEI8x16|B_REGISTER|B_UPD,
        B_CONST8|B_COUNT,
        B_NONE,
        B_NONE },

  { T("VPSLLDQ"),
        D_AVX|D_MUST66|D_REGISTER, DX_VEX|DX_LSHORT,
        1, 0x0000F8FF, 0x0000F873, 0x00,
        B_SVEXI8x16|B_UPD,
        B_SSEI8x16|B_REGISTER,
        B_CONST8|B_COUNT,
        B_NONE },

  { T("PSLLW"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000F10F, 0x00,
        B_MREG16x4|B_UPD,
        B_MMX16x4,
        B_NONE,
        B_NONE },

  { T("PSLLW"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000F10F, 0x00,
        B_SREGI16x8|B_UPD,
        B_SSEI16x8,
        B_NONE,
        B_NONE },

  { T("VPSLLW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000F1, 0x00,
        B_SREGI16x8|B_UPD,
        B_SVEXI16x8,
        B_SSEI16x8,
        B_NONE },

  { T("PSLLW"),
        D_MMX|D_MUSTNONE|D_REGISTER, 0,
        2, 0x00F8FFFF, 0x00F0710F, 0x00,
        B_MMX16x4|B_REGISTER|B_UPD,
        B_CONST8|B_COUNT,
        B_NONE,
        B_NONE },

  { T("PSLLW"),
        D_SSE|D_MUST66|D_REGISTER, 0,
        2, 0x00F8FFFF, 0x00F0710F, 0x00,
        B_SSEI16x8|B_REGISTER|B_UPD,
        B_CONST8|B_COUNT,
        B_NONE,
        B_NONE },

  { T("VPSLLW"),
        D_AVX|D_MUST66|D_REGISTER, DX_VEX|DX_LSHORT,
        1, 0x0000F8FF, 0x0000F071, 0x00,
        B_SVEXI16x8|B_UPD,
        B_SSEI16x8|B_REGISTER,
        B_CONST8|B_COUNT,
        B_NONE },

  { T("PSLLD"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000F20F, 0x00,
        B_MREG32x2|B_UPD,
        B_MMX32x2,
        B_NONE,
        B_NONE },

  { T("PSLLD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000F20F, 0x00,
        B_SREGI32x4|B_UPD,
        B_SSEI32x4,
        B_NONE,
        B_NONE },

  { T("VPSLLD"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000F2, 0x00,
        B_SREGI32x4|B_UPD,
        B_SVEXI32x4,
        B_SSEI32x4,
        B_NONE },

  { T("PSLLD"),
        D_MMX|D_MUSTNONE|D_REGISTER, 0,
        2, 0x00F8FFFF, 0x00F0720F, 0x00,
        B_MMX32x2|B_REGISTER|B_UPD,
        B_CONST8|B_COUNT,
        B_NONE,
        B_NONE },

  { T("PSLLD"),
        D_SSE|D_MUST66|D_REGISTER, 0,
        2, 0x00F8FFFF, 0x00F0720F, 0x00,
        B_SSEI32x4|B_REGISTER|B_UPD,
        B_CONST8|B_COUNT,
        B_NONE,
        B_NONE },

  { T("VPSLLD"),
        D_AVX|D_MUST66|D_REGISTER, DX_VEX|DX_LSHORT,
        1, 0x0000F8FF, 0x0000F072, 0x00,
        B_SVEXI32x4|B_UPD,
        B_SSEI32x4|B_REGISTER,
        B_CONST8|B_COUNT,
        B_NONE },

  { T("PSLLQ"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000F30F, 0x00,
        B_MREG64|B_UPD,
        B_MMX64,
        B_NONE,
        B_NONE },

  { T("PSLLQ"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000F30F, 0x00,
        B_SREGI64x2|B_UPD,
        B_SSEI64x2,
        B_NONE,
        B_NONE },

  { T("VPSLLQ"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000F3, 0x00,
        B_SREGI64x2|B_UPD,
        B_SVEXI64x2,
        B_SSEI64x2,
        B_NONE },

  { T("PSLLQ"),
        D_MMX|D_MUSTNONE|D_REGISTER, 0,
        2, 0x00F8FFFF, 0x00F0730F, 0x00,
        B_MMX64|B_REGISTER|B_UPD,
        B_CONST8|B_COUNT,
        B_NONE,
        B_NONE },

  { T("PSLLQ"),
        D_SSE|D_MUST66|D_REGISTER, 0,
        2, 0x00F8FFFF, 0x00F0730F, 0x00,
        B_SSEI64x2|B_REGISTER|B_UPD,
        B_CONST8|B_COUNT,
        B_NONE,
        B_NONE },

  { T("VPSLLQ"),
        D_AVX|D_MUST66|D_REGISTER, DX_VEX|DX_LSHORT,
        1, 0x0000F8FF, 0x0000F073, 0x00,
        B_SVEXI64x2|B_UPD,
        B_SSEI64x2|B_REGISTER,
        B_CONST8|B_COUNT,
        B_NONE },

  { T("PSRAW"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000E10F, 0x00,
        B_MREG16x4|B_UPD,
        B_MMX16x4,
        B_NONE,
        B_NONE },

  { T("PSRAW"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000E10F, 0x00,
        B_SREGI16x8|B_UPD,
        B_SSEI16x8,
        B_NONE,
        B_NONE },

  { T("VPSRAW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000E1, 0x00,
        B_SREGI16x8|B_UPD,
        B_SVEXI16x8,
        B_SSEI16x8,
        B_NONE },

  { T("PSRAW"),
        D_MMX|D_MUSTNONE|D_REGISTER, 0,
        2, 0x00F8FFFF, 0x00E0710F, 0x00,
        B_MMX16x4|B_REGISTER|B_UPD,
        B_CONST8|B_COUNT,
        B_NONE,
        B_NONE },

  { T("PSRAW"),
        D_SSE|D_MUST66|D_REGISTER, 0,
        2, 0x00F8FFFF, 0x00E0710F, 0x00,
        B_SSEI16x8|B_REGISTER|B_UPD,
        B_CONST8|B_COUNT,
        B_NONE,
        B_NONE },

  { T("VPSRAW"),
        D_AVX|D_MUST66|D_REGISTER, DX_VEX|DX_LSHORT,
        1, 0x0000F8FF, 0x0000E071, 0x00,
        B_SVEXI16x8|B_UPD,
        B_SSEI16x8|B_REGISTER,
        B_CONST8|B_COUNT,
        B_NONE },

  { T("PSRAD"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000E20F, 0x00,
        B_MREG32x2|B_UPD,
        B_MMX32x2,
        B_NONE,
        B_NONE },

  { T("PSRAD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000E20F, 0x00,
        B_SREGI32x4|B_UPD,
        B_SSEI32x4,
        B_NONE,
        B_NONE },

  { T("VPSRAD"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000E2, 0x00,
        B_SREGI32x4|B_UPD,
        B_SVEXI32x4,
        B_SSEI32x4,
        B_NONE },

  { T("PSRAD"),
        D_MMX|D_MUSTNONE|D_REGISTER, 0,
        2, 0x00F8FFFF, 0x00E0720F, 0x00,
        B_MMX32x2|B_REGISTER|B_UPD,
        B_CONST8|B_COUNT,
        B_NONE,
        B_NONE },

  { T("PSRAD"),
        D_SSE|D_MUST66|D_REGISTER, 0,
        2, 0x00F8FFFF, 0x00E0720F, 0x00,
        B_SSEI32x4|B_REGISTER|B_UPD,
        B_CONST8|B_COUNT,
        B_NONE,
        B_NONE },

  { T("VPSRAD"),
        D_AVX|D_MUST66|D_REGISTER, DX_VEX|DX_LSHORT,
        1, 0x0000F8FF, 0x0000E072, 0x00,
        B_SVEXI32x4|B_UPD,
        B_SSEI32x4|B_REGISTER,
        B_CONST8|B_COUNT,
        B_NONE },

  { T("PSRLDQ"),
        D_SSE|D_MUST66|D_REGISTER, 0,
        2, 0x00F8FFFF, 0x00D8730F, 0x00,
        B_SSEI8x16|B_REGISTER|B_UPD,
        B_CONST8|B_COUNT,
        B_NONE,
        B_NONE },

  { T("VPSRLDQ"),
        D_AVX|D_MUST66|D_REGISTER, DX_VEX|DX_LSHORT,
        1, 0x0000F8FF, 0x0000D873, 0x00,
        B_SVEXI8x16|B_UPD,
        B_SSEI8x16|B_REGISTER,
        B_CONST8|B_COUNT,
        B_NONE },

  { T("PSRLW"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000D10F, 0x00,
        B_MREG16x4|B_UPD,
        B_MMX16x4,
        B_NONE,
        B_NONE },

  { T("PSRLW"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000D10F, 0x00,
        B_SREGI16x8|B_UPD,
        B_SSEI16x8,
        B_NONE,
        B_NONE },

  { T("VPSRLW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000D1, 0x00,
        B_SREGI16x8|B_UPD,
        B_SVEXI16x8,
        B_SSEI16x8,
        B_NONE },

  { T("PSRLW"),
        D_MMX|D_MUSTNONE|D_REGISTER, 0,
        2, 0x00F8FFFF, 0x00D0710F, 0x00,
        B_MMX16x4|B_REGISTER|B_UPD,
        B_CONST8|B_COUNT,
        B_NONE,
        B_NONE },

  { T("PSRLW"),
        D_SSE|D_MUST66|D_REGISTER, 0,
        2, 0x00F8FFFF, 0x00D0710F, 0x00,
        B_SSEI16x8|B_REGISTER|B_UPD,
        B_CONST8|B_COUNT,
        B_NONE,
        B_NONE },

  { T("VPSRLW"),
        D_AVX|D_MUST66|D_REGISTER, DX_VEX|DX_LSHORT,
        1, 0x0000F8FF, 0x0000D071, 0x00,
        B_SVEXI16x8|B_UPD,
        B_SSEI16x8|B_REGISTER,
        B_CONST8|B_COUNT,
        B_NONE },

  { T("PSRLD"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000D20F, 0x00,
        B_MREG32x2|B_UPD,
        B_MMX32x2,
        B_NONE,
        B_NONE },

  { T("PSRLD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000D20F, 0x00,
        B_SREGI32x4|B_UPD,
        B_SSEI32x4,
        B_NONE,
        B_NONE },

  { T("VPSRLD"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000D2, 0x00,
        B_SREGI32x4|B_UPD,
        B_SVEXI32x4,
        B_SSEI32x4,
        B_NONE },

  { T("PSRLD"),
        D_MMX|D_MUSTNONE|D_REGISTER, 0,
        2, 0x00F8FFFF, 0x00D0720F, 0x00,
        B_MMX32x2|B_REGISTER|B_UPD,
        B_CONST8|B_COUNT,
        B_NONE,
        B_NONE },

  { T("PSRLD"),
        D_SSE|D_MUST66|D_REGISTER, 0,
        2, 0x00F8FFFF, 0x00D0720F, 0x00,
        B_SSEI32x4|B_REGISTER|B_UPD,
        B_CONST8|B_COUNT,
        B_NONE,
        B_NONE },

  { T("VPSRLD"),
        D_AVX|D_MUST66|D_REGISTER, DX_VEX|DX_LSHORT,
        1, 0x0000F8FF, 0x0000D072, 0x00,
        B_SVEXI32x4|B_UPD,
        B_SSEI32x4|B_REGISTER,
        B_CONST8|B_COUNT,
        B_NONE },

  { T("PSRLQ"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000D30F, 0x00,
        B_MREG64|B_UPD,
        B_MMX64,
        B_NONE,
        B_NONE },

  { T("PSRLQ"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000D30F, 0x00,
        B_SREGI64x2|B_UPD,
        B_SSEI64x2,
        B_NONE,
        B_NONE },

  { T("VPSRLQ"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000D3, 0x00,
        B_SREGI64x2|B_UPD,
        B_SVEXI64x2,
        B_SSEI64x2,
        B_NONE },

  { T("PSRLQ"),
        D_MMX|D_MUSTNONE|D_REGISTER, 0,
        2, 0x00F8FFFF, 0x00D0730F, 0x00,
        B_MMX64|B_REGISTER|B_UPD,
        B_CONST8|B_COUNT,
        B_NONE,
        B_NONE },

  { T("PSRLQ"),
        D_SSE|D_MUST66|D_REGISTER, 0,
        2, 0x00F8FFFF, 0x00D0730F, 0x00,
        B_SSEI64x2|B_REGISTER|B_UPD,
        B_CONST8|B_COUNT,
        B_NONE,
        B_NONE },

  { T("VPSRLQ"),
        D_AVX|D_MUST66|D_REGISTER, DX_VEX|DX_LSHORT,
        1, 0x0000F8FF, 0x0000D073, 0x00,
        B_SVEXI64x2|B_UPD,
        B_SSEI64x2|B_REGISTER,
        B_CONST8|B_COUNT,
        B_NONE },

  { T("PSUBB"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000F80F, 0x00,
        B_MREG8x8|B_UPD,
        B_MMX8x8,
        B_NONE,
        B_NONE },

  { T("PSUBW"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000F90F, 0x00,
        B_MREG16x4|B_UPD,
        B_MMX16x4,
        B_NONE,
        B_NONE },

  { T("PSUBD"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000FA0F, 0x00,
        B_MREG32x2|B_UPD,
        B_MMX32x2,
        B_NONE,
        B_NONE },

  { T("PSUBB"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000F80F, 0x00,
        B_SREGI8x16|B_UPD,
        B_SSEI8x16,
        B_NONE,
        B_NONE },

  { T("VPSUBB"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000F8, 0x00,
        B_SREGI8x16|B_UPD,
        B_SVEXI8x16,
        B_SSEI8x16,
        B_NONE },

  { T("PSUBW"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000F90F, 0x00,
        B_SREGI16x8|B_UPD,
        B_SSEI16x8,
        B_NONE,
        B_NONE },

  { T("VPSUBW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000F9, 0x00,
        B_SREGI16x8|B_UPD,
        B_SVEXI16x8,
        B_SSEI16x8,
        B_NONE },

  { T("PSUBD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000FA0F, 0x00,
        B_SREGI32x4|B_UPD,
        B_SSEI32x4,
        B_NONE,
        B_NONE },

  { T("VPSUBD"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000FA, 0x00,
        B_SREGI32x4|B_UPD,
        B_SVEXI32x4,
        B_SSEI32x4,
        B_NONE },

  { T("PSUBQ"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000FB0F, 0x00,
        B_MREG64|B_UPD,
        B_MMX64,
        B_NONE,
        B_NONE },

  { T("PSUBQ"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000FB0F, 0x00,
        B_SREGI64x2|B_UPD,
        B_SSEI64x2,
        B_NONE,
        B_NONE },

  { T("VPSUBQ"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000FB, 0x00,
        B_SREGI64x2|B_UPD,
        B_SVEXI64x2,
        B_SSEI64x2,
        B_NONE },

  { T("PSUBSB"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000E80F, 0x00,
        B_MREG8x8|B_UPD,
        B_MMX8x8,
        B_NONE,
        B_NONE },

  { T("PSUBSW"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000E90F, 0x00,
        B_MREG16x4|B_UPD,
        B_MMX16x4,
        B_NONE,
        B_NONE },

  { T("PSUBSB"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000E80F, 0x00,
        B_SREGI8x16|B_UPD,
        B_SSEI8x16,
        B_NONE,
        B_NONE },

  { T("VPSUBSB"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000E8, 0x00,
        B_SREGI8x16|B_UPD,
        B_SVEXI8x16,
        B_SSEI8x16,
        B_NONE },

  { T("PSUBSW"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000E90F, 0x00,
        B_SREGI16x8|B_UPD,
        B_SSEI16x8,
        B_NONE,
        B_NONE },

  { T("VPSUBSW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000E9, 0x00,
        B_SREGI16x8|B_UPD,
        B_SVEXI16x8,
        B_SSEI16x8,
        B_NONE },

  { T("PSUBUSB"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000D80F, 0x00,
        B_MREG8x8|B_UPD,
        B_MMX8x8,
        B_NONE,
        B_NONE },

  { T("PSUBUSW"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000D90F, 0x00,
        B_MREG16x4|B_UPD,
        B_MMX16x4,
        B_NONE,
        B_NONE },

  { T("PSUBUSB"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000D80F, 0x00,
        B_SREGI8x16|B_UPD,
        B_SSEI8x16,
        B_NONE,
        B_NONE },

  { T("VPSUBUSB"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000D8, 0x00,
        B_SREGI8x16|B_UPD,
        B_SVEXI8x16,
        B_SSEI8x16,
        B_NONE },

  { T("PSUBUSW"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000D90F, 0x00,
        B_SREGI16x8|B_UPD,
        B_SSEI16x8,
        B_NONE,
        B_NONE },

  { T("VPSUBUSW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000D9, 0x00,
        B_SREGI16x8|B_UPD,
        B_SVEXI16x8,
        B_SSEI16x8,
        B_NONE },

  { T("PUNPCKHBW"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000680F, 0x00,
        B_MREG8x8|B_UPD,
        B_MMX8x8,
        B_NONE,
        B_NONE },

  { T("PUNPCKHBW"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000680F, 0x00,
        B_SREGI8x16|B_UPD,
        B_SSEI8x16,
        B_NONE,
        B_NONE },

  { T("VPUNPCKHBW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x00000068, 0x00,
        B_SREGI8x16|B_UPD,
        B_SVEXI8x16,
        B_SSEI8x16,
        B_NONE },

  { T("PUNPCKHWD"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000690F, 0x00,
        B_MREG16x4|B_UPD,
        B_MMX16x4,
        B_NONE,
        B_NONE },

  { T("PUNPCKHWD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000690F, 0x00,
        B_SREGI16x8|B_UPD,
        B_SSEI16x8,
        B_NONE,
        B_NONE },

  { T("VPUNPCKHWD"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x00000069, 0x00,
        B_SREGI16x8|B_UPD,
        B_SVEXI16x8,
        B_SSEI16x8,
        B_NONE },

  { T("PUNPCKHDQ"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x00006A0F, 0x00,
        B_MREG32x2|B_UPD,
        B_MMX32x2,
        B_NONE,
        B_NONE },

  { T("PUNPCKHDQ"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x00006A0F, 0x00,
        B_SREGI32x4|B_UPD,
        B_SSEI32x4,
        B_NONE,
        B_NONE },

  { T("VPUNPCKHDQ"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x0000006A, 0x00,
        B_SREGI32x4|B_UPD,
        B_SVEXI32x4,
        B_SSEI32x4,
        B_NONE },

  { T("PUNPCKHQDQ"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x00006D0F, 0x00,
        B_SREGI32x4|B_UPD,
        B_SSEI32x4,
        B_NONE,
        B_NONE },

  { T("VPUNPCKHQDQ"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x0000006D, 0x00,
        B_SREGI32x4|B_UPD,
        B_SVEXI32x4,
        B_SSEI32x4,
        B_NONE },

  { T("PUNPCKLBW"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000600F, 0x00,
        B_MREG8x8|B_UPD,
        B_MMX8x8,
        B_NONE,
        B_NONE },

  { T("PUNPCKLBW"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000600F, 0x00,
        B_SREGI8x16|B_UPD,
        B_SSEI8x16,
        B_NONE,
        B_NONE },

  { T("VPUNPCKLBW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x00000060, 0x00,
        B_SREGI8x16|B_UPD,
        B_SVEXI8x16,
        B_SSEI8x16,
        B_NONE },

  { T("PUNPCKLWD"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000610F, 0x00,
        B_MREG16x4|B_UPD,
        B_MMX16x4,
        B_NONE,
        B_NONE },

  { T("PUNPCKLWD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000610F, 0x00,
        B_SREGI16x8|B_UPD,
        B_SSEI16x8,
        B_NONE,
        B_NONE },

  { T("VPUNPCKLWD"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x00000061, 0x00,
        B_SREGI16x8|B_UPD,
        B_SVEXI16x8,
        B_SSEI16x8,
        B_NONE },

  { T("PUNPCKLDQ"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000620F, 0x00,
        B_MREG32x2|B_UPD,
        B_MMX32x2,
        B_NONE,
        B_NONE },

  { T("PUNPCKLDQ"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000620F, 0x00,
        B_SREGI32x4|B_UPD,
        B_SSEI32x4,
        B_NONE,
        B_NONE },

  { T("VPUNPCKLDQ"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x00000062, 0x00,
        B_SREGI32x4|B_UPD,
        B_SVEXI32x4,
        B_SSEI32x4,
        B_NONE },

  { T("PUNPCKLQDQ"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x00006C0F, 0x00,
        B_SREGI32x4|B_UPD,
        B_SSEI32x4,
        B_NONE,
        B_NONE },

  { T("VPUNPCKLQDQ"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x0000006C, 0x00,
        B_SREGI32x4|B_UPD,
        B_SVEXI32x4,
        B_SSEI32x4,
        B_NONE },

  { T("PXOR"),
        D_MMX|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000EF0F, 0x00,
        B_MREG8x8|B_UPD,
        B_MMX8x8,
        B_NONE,
        B_NONE },

  { T("PXOR"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000EF0F, 0x00,
        B_SREGI8x16|B_BINARY|B_UPD,
        B_SSEI8x16,
        B_NONE,
        B_NONE },

  { T("VPXOR"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT,
        1, 0x000000FF, 0x000000EF, 0x00,
        B_SREGI8x16|B_BINARY|B_UPD,
        B_SVEXI8x16|B_BINARY,
        B_SSEI8x16,
        B_NONE },

  { T("RCPPS"),
        D_SSE|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000530F, 0x00,
        B_SREGF32x4|B_CHG,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("VRCPPS"),
        D_AVX|D_MUSTNONE, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x00000053, 0x00,
        B_SREGF32x4|B_CHG,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("RCPSS"),
        D_SSE|D_MUSTF3, 0,
        2, 0x0000FFFF, 0x0000530F, 0x00,
        B_SREGF32L|B_CHG,
        B_SSEF32L,
        B_NONE,
        B_NONE },

  { T("VRCPSS"),
        D_AVX|D_MUSTF3, DX_VEX|DX_IGNOREL,
        1, 0x000000FF, 0x00000053, 0x00,
        B_SREGF32L|B_CHG,
        B_SVEXF32L|B_CHG,
        B_SSEF32L,
        B_NONE },

  { T("RSQRTPS"),
        D_SSE|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000520F, 0x00,
        B_SREGF32x4|B_CHG,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("VRSQRTPS"),
        D_AVX|D_MUSTNONE, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x00000052, 0x00,
        B_SREGF32x4|B_CHG,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("RSQRTSS"),
        D_SSE|D_MUSTF3, 0,
        2, 0x0000FFFF, 0x0000520F, 0x00,
        B_SREGF32L|B_CHG,
        B_SSEF32L,
        B_NONE,
        B_NONE },

  { T("VRSQRTSS"),
        D_AVX|D_MUSTF3, DX_VEX|DX_IGNOREL,
        1, 0x000000FF, 0x00000052, 0x00,
        B_SREGF32L|B_CHG,
        B_SVEXF32L,
        B_SSEF32L,
        B_NONE },

  { T("SHUFPD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000C60F, 0x00,
        B_SREGF64x2|B_UPD,
        B_SSEF64x2,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("VSHUFPD"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x000000C6, 0x00,
        B_SREGF64x2|B_UPD,
        B_SVEXF64x2,
        B_SSEF64x2,
        B_CONST8|B_BINARY },

  { T("SHUFPS"),
        D_SSE|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000C60F, 0x00,
        B_SREGF32x4|B_UPD,
        B_SSEF32x4,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("VSHUFPS"),
        D_AVX|D_MUSTNONE, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x000000C6, 0x00,
        B_SREGF32x4|B_UPD,
        B_SVEXF32x4,
        B_SSEF32x4,
        B_CONST8|B_BINARY },

  { T("SQRTPD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000510F, 0x00,
        B_SREGF64x2|B_CHG,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("VSQRTPD"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x00000051, 0x00,
        B_SREGF64x2|B_CHG,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("SQRTPS"),
        D_SSE|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000510F, 0x00,
        B_SREGF32x4|B_CHG,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("VSQRTPS"),
        D_AVX|D_MUSTNONE, DX_VEX|DX_LBOTH|DX_NOVREG,
        1, 0x000000FF, 0x00000051, 0x00,
        B_SREGF32x4|B_CHG,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("SQRTSD"),
        D_SSE|D_MUSTF2, 0,
        2, 0x0000FFFF, 0x0000510F, 0x00,
        B_SREGF64L|B_CHG,
        B_SSEF64L,
        B_NONE,
        B_NONE },

  { T("VSQRTSD"),
        D_AVX|D_MUSTF2, DX_VEX|DX_IGNOREL,
        1, 0x000000FF, 0x00000051, 0x00,
        B_SREGF64L|B_CHG,
        B_SVEXF64L,
        B_SSEF64L,
        B_NONE },

  { T("SQRTSS"),
        D_SSE|D_MUSTF3, 0,
        2, 0x0000FFFF, 0x0000510F, 0x00,
        B_SREGF32L|B_CHG,
        B_SSEF32L,
        B_NONE,
        B_NONE },

  { T("VSQRTSS"),
        D_AVX|D_MUSTF3, DX_VEX|DX_IGNOREL,
        1, 0x000000FF, 0x00000051, 0x00,
        B_SREGF32L|B_CHG,
        B_SVEXF32L,
        B_SSEF32L,
        B_NONE },

  { T("SUBPD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x00005C0F, 0x00,
        B_SREGF64x2|B_UPD,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("VSUBPD"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x0000005C, 0x00,
        B_SREGF64x2|B_UPD,
        B_SVEXF64x2,
        B_SSEF64x2,
        B_NONE },

  { T("SUBPS"),
        D_SSE|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x00005C0F, 0x00,
        B_SREGF32x4|B_UPD,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("VSUBPS"),
        D_AVX|D_MUSTNONE, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x0000005C, 0x00,
        B_SREGF32x4|B_UPD,
        B_SVEXF32x4,
        B_SSEF32x4,
        B_NONE },

  { T("SUBSD"),
        D_SSE|D_MUSTF2, 0,
        2, 0x0000FFFF, 0x00005C0F, 0x00,
        B_SREGF64L|B_UPD,
        B_SSEF64L,
        B_NONE,
        B_NONE },

  { T("VSUBSD"),
        D_AVX|D_MUSTF2, DX_VEX|DX_IGNOREL,
        1, 0x000000FF, 0x0000005C, 0x00,
        B_SREGF64L|B_UPD,
        B_SVEXF64L,
        B_SSEF64L,
        B_NONE },

  { T("SUBSS"),
        D_SSE|D_MUSTF3, 0,
        2, 0x0000FFFF, 0x00005C0F, 0x00,
        B_SREGF32L|B_UPD,
        B_SSEF32L,
        B_NONE,
        B_NONE },

  { T("VSUBSS"),
        D_AVX|D_MUSTF3, DX_VEX|DX_IGNOREL,
        1, 0x000000FF, 0x0000005C, 0x00,
        B_SREGF32L|B_UPD,
        B_SVEXF32L,
        B_SSEF32L,
        B_NONE },

  { T("UNPCKHPD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000150F, 0x00,
        B_SREGF64x2|B_UPD,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("VUNPCKHPD"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x00000015, 0x00,
        B_SREGF64x2|B_UPD,
        B_SVEXF64x2,
        B_SSEF64x2,
        B_NONE },

  { T("UNPCKHPS"),
        D_SSE|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000150F, 0x00,
        B_SREGF32x4|B_UPD,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("VUNPCKHPS"),
        D_AVX|D_MUSTNONE, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x00000015, 0x00,
        B_SREGF32x4|B_UPD,
        B_SVEXF32x4,
        B_SSEF32x4,
        B_NONE },

  { T("UNPCKLPD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000140F, 0x00,
        B_SREGF64x2|B_UPD,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("VUNPCKLPD"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x00000014, 0x00,
        B_SREGF64x2|B_UPD,
        B_SVEXF64x2,
        B_SSEF64x2,
        B_NONE },

  { T("UNPCKLPS"),
        D_SSE|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000140F, 0x00,
        B_SREGF32x4|B_UPD,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("VUNPCKLPS"),
        D_AVX|D_MUSTNONE, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x00000014, 0x00,
        B_SREGF32x4|B_UPD,
        B_SVEXF32x4,
        B_SSEF32x4,
        B_NONE },

  { T("UCOMISD"),
        D_SSE|D_MUST66|D_ALLFLAGS, 0,
        2, 0x0000FFFF, 0x00002E0F, 0x00,
        B_SREGF64L,
        B_SSEF64L,
        B_NONE,
        B_NONE },

  { T("VUCOMISD"),
        D_AVX|D_MUST66|D_ALLFLAGS, DX_VEX|DX_IGNOREL|DX_NOVREG,
        1, 0x000000FF, 0x0000002E, 0x00,
        B_SREGF64L,
        B_SSEF64L,
        B_NONE,
        B_NONE },

  { T("UCOMISS"),
        D_SSE|D_MUSTNONE|D_ALLFLAGS, 0,
        2, 0x0000FFFF, 0x00002E0F, 0x00,
        B_SREGF32L,
        B_SSEF32L,
        B_NONE,
        B_NONE },

  { T("VUCOMISS"),
        D_AVX|D_MUSTNONE|D_ALLFLAGS, DX_VEX|DX_IGNOREL|DX_NOVREG,
        1, 0x000000FF, 0x0000002E, 0x00,
        B_SREGF32L,
        B_SSEF32L,
        B_NONE,
        B_NONE },

  { T("XORPD"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000570F, 0x00,
        B_SREGF64x2|B_UPD,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("VXORPD"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x00000057, 0x00,
        B_SREGF64x2|B_UPD,
        B_SVEXF64x2,
        B_SSEF64x2,
        B_NONE },

  { T("XORPS"),
        D_SSE|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x0000570F, 0x00,
        B_SREGF32x4|B_UPD,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("VXORPS"),
        D_AVX|D_MUSTNONE, DX_VEX|DX_LBOTH,
        1, 0x000000FF, 0x00000057, 0x00,
        B_SREGF32x4|B_UPD,
        B_SVEXF32x4,
        B_SSEF32x4,
        B_NONE },

  { T("FXRSTOR"),
        D_SSE|D_MEMORY, 0,
        2, 0x0038FFFF, 0x0008AE0F, 0x00,
        B_LONGDATA|B_MEMORY,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FXSAVE"),
        D_SSE|D_MEMORY, 0,
        2, 0x0038FFFF, 0x0000AE0F, 0x00,
        B_LONGDATA|B_MEMORY|B_CHG,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("LFENCE"),
        D_SSE, 0,
        3, 0x00FFFFFF, 0x00E8AE0F, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("PREFETCHT0"),
        D_SSE|D_MUSTNONE|D_MEMORY, 0,
        2, 0x0038FFFF, 0x0008180F, 0x00,
        B_ANYMEM|B_MEMORY,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("PREFETCHT1"),
        D_SSE|D_MUSTNONE|D_MEMORY, 0,
        2, 0x0038FFFF, 0x0010180F, 0x00,
        B_ANYMEM|B_MEMORY,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("PREFETCHT2"),
        D_SSE|D_MUSTNONE|D_MEMORY, 0,
        2, 0x0038FFFF, 0x0018180F, 0x00,
        B_ANYMEM|B_MEMORY,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("PREFETCHNTA"),
        D_SSE|D_MUSTNONE|D_MEMORY, 0,
        2, 0x0038FFFF, 0x0000180F, 0x00,
        B_ANYMEM|B_MEMORY,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("NOP"),
        D_SSE|D_MUSTNONE|D_MEMORY|D_UNDOC, DX_NOP,
        2, 0x0020FFFF, 0x0020180F, 0x00,
        B_ANYMEM|B_MEMORY,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("PREFETCH"),
        D_SSE|D_MUSTNONE|D_MEMORY, 0,
        2, 0x0038FFFF, 0x00000D0F, 0x00,
        B_ANYMEM|B_MEMORY,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("PREFETCHW"),
        D_SSE|D_MUSTNONE|D_MEMORY, 0,
        2, 0x0038FFFF, 0x00080D0F, 0x00,
        B_ANYMEM|B_MEMORY,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("SFENCE"),
        D_SSE, 0,
        3, 0x00FFFFFF, 0x00F8AE0F, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("BLENDPD"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x000D3A0F, 0x00,
        B_SREGF64x2|B_UPD,
        B_SSEF64x2,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("VBLENDPD"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH|DX_LEAD3A,
        1, 0x000000FF, 0x0000000D, 0x00,
        B_SREGF64x2|B_UPD,
        B_SVEXF64x2,
        B_SSEF64x2,
        B_CONST8|B_BINARY },

  { T("BLENDPS"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x000C3A0F, 0x00,
        B_SREGF32x4|B_UPD,
        B_SSEF32x4,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("VBLENDPS"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH|DX_LEAD3A,
        1, 0x000000FF, 0x0000000C, 0x00,
        B_SREGF32x4|B_UPD,
        B_SVEXF32x4,
        B_SSEF32x4,
        B_CONST8|B_BINARY },

  { T("BLENDVPD"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0015380F, 0x00,
        B_SREGF64x2|B_UPD,
        B_SSEF64x2,
        B_XMM0I64x2,
        B_NONE },

  { T("BLENDVPD"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0015380F, 0x00,
        B_SREGF64x2|B_UPD,
        B_SSEF64x2,
        B_XMM0I64x2|B_PSEUDO,
        B_NONE },

  { T("VBLENDVPD"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH|DX_W0|DX_LEAD3A,
        1, 0x000000FF, 0x0000004B, 0x00,
        B_SREGF64x2|B_UPD,
        B_SVEXF64x2,
        B_SSEF64x2,
        B_SIMMI8x16 },

  { T("BLENDVPS"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0014380F, 0x00,
        B_SREGF32x4|B_UPD,
        B_SSEF32x4,
        B_XMM0I32x4,
        B_NONE },

  { T("BLENDVPS"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0014380F, 0x00,
        B_SREGF32x4|B_UPD,
        B_SSEF32x4,
        B_XMM0I32x4|B_PSEUDO,
        B_NONE },

  { T("VBLENDVPS"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH|DX_W0|DX_LEAD3A,
        1, 0x000000FF, 0x0000004A, 0x00,
        B_SREGF32x4|B_UPD,
        B_SVEXF32x4,
        B_SSEF32x4,
        B_SIMMI8x16 },

  { T("CRC32"),
        D_CMD|D_NEEDF2, 0,
        3, 0x00FFFFFF, 0x00F0380F, 0x00,
        B_REG32|B_NOADDR|B_UPD,
        B_INT8|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("CRC32"),
        D_CMD|D_NEEDF2, 0,
        3, 0x00FFFFFF, 0x00F1380F, 0x00,
        B_REG32|B_NOADDR|B_UPD,
        B_INT1632|B_NOADDR|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("DPPD"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x00413A0F, 0x00,
        B_SREGF64x2|B_UPD,
        B_SSEF64x2,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("VDPPD"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD3A,
        1, 0x000000FF, 0x00000041, 0x00,
        B_SREGF64x2|B_UPD,
        B_SVEXF64x2,
        B_SSEF64x2,
        B_CONST8|B_BINARY },

  { T("DPPS"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x00403A0F, 0x00,
        B_SREGF32x4|B_UPD,
        B_SSEF32x4,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("VDPPS"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH|DX_LEAD3A,
        1, 0x000000FF, 0x00000040, 0x00,
        B_SREGF32x4|B_UPD,
        B_SVEXF32x4,
        B_SSEF32x4,
        B_CONST8|B_BINARY },

  { T("EXTRACTPS"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x00173A0F, 0x00,
        B_INT32|B_CHG,
        B_SREGF32x4,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("VEXTRACTPS"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_NOVREG|DX_LEAD3A,
        1, 0x000000FF, 0x00000017, 0x00,
        B_INT32|B_CHG,
        B_SREGF32x4,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("INSERTPS"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x00213A0F, 0x00,
        B_SREGF32x4|B_UPD,
        B_SSEF32L,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("VINSERTPS"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD3A,
        1, 0x000000FF, 0x00000021, 0x00,
        B_SREGF32x4|B_UPD,
        B_SVEXF32x4,
        B_SSEF32L,
        B_CONST8|B_BINARY },

  { T("MOVNTDQA"),
        D_SSE|D_MUST66|D_MEMORY, 0,
        3, 0x00FFFFFF, 0x002A380F, 0x00,
        B_SREGI8x16|B_BINARY|B_CHG,
        B_SSEI8x16|B_MEMORY,
        B_NONE,
        B_NONE },

  { T("VMOVNTDQA"),
        D_AVX|D_MUST66|D_MEMORY, DX_VEX|DX_LSHORT|DX_NOVREG|DX_LEAD38,
        1, 0x000000FF, 0x0000002A, 0x00,
        B_SREGI8x16|B_BINARY|B_CHG,
        B_SSEI8x16|B_MEMORY,
        B_NONE,
        B_NONE },

  { T("MPSADBW"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x00423A0F, 0x00,
        B_SREGI8x16|B_UPD,
        B_SSEI8x16,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("VMPSADBW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD3A,
        1, 0x000000FF, 0x00000042, 0x00,
        B_SREGI8x16|B_UPD,
        B_SVEXI8x16,
        B_SSEI8x16,
        B_CONST8|B_BINARY },

  { T("PACKUSDW"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x002B380F, 0x00,
        B_SREGI32x4|B_UPD,
        B_SSEI32x4,
        B_NONE,
        B_NONE },

  { T("VPACKUSDW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD38,
        1, 0x000000FF, 0x0000002B, 0x00,
        B_SREGI32x4|B_UPD,
        B_SVEXI32x4|B_UPD,
        B_SSEI32x4,
        B_NONE },

  { T("PBLENDVB"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0010380F, 0x00,
        B_SREGI8x16|B_UPD,
        B_SSEI8x16,
        B_XMM0I8x16,
        B_NONE },

  { T("PBLENDVB"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0010380F, 0x00,
        B_SREGI8x16|B_UPD,
        B_SSEI8x16,
        B_XMM0I8x16|B_PSEUDO,
        B_NONE },

  { T("VPBLENDVB"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_W0|DX_LEAD3A,
        1, 0x000000FF, 0x0000004C, 0x00,
        B_SREGI8x16|B_UPD,
        B_SVEXI8x16,
        B_SSEI8x16,
        B_SIMMI8x16 },

  { T("PBLENDW"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x000E3A0F, 0x00,
        B_SREGI16x8|B_UPD,
        B_SSEI16x8,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("VPBLENDW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD3A,
        1, 0x000000FF, 0x0000000E, 0x00,
        B_SREGI16x8|B_UPD,
        B_SVEXI16x8,
        B_SSEI16x8,
        B_CONST8|B_BINARY },

  { T("PCLMULLQLQDQ"),
        D_SSE|D_POSTBYTE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x00443A0F, 0x00,
        B_SREGI64x2|B_UPD,
        B_SSEI64x2,
        B_NONE,
        B_NONE },

  { T("VPCLMULLQLQDQ"),
        D_AVX|D_POSTBYTE|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD3A,
        1, 0x000000FF, 0x00000044, 0x00,
        B_SREGI64x2|B_UPD,
        B_SVEXI64x2,
        B_SSEI64x2,
        B_NONE },

  { T("PCLMULHQLQDQ"),
        D_SSE|D_POSTBYTE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x00443A0F, 0x01,
        B_SREGI64x2|B_UPD,
        B_SSEI64x2,
        B_NONE,
        B_NONE },

  { T("VPCLMULHQLQDQ"),
        D_AVX|D_POSTBYTE|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD3A,
        1, 0x000000FF, 0x00000044, 0x01,
        B_SREGI64x2|B_UPD,
        B_SVEXI64x2,
        B_SSEI64x2,
        B_NONE },

  { T("PCLMULLQHDQ"),
        D_SSE|D_POSTBYTE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x00443A0F, 0x10,
        B_SREGI64x2|B_UPD,
        B_SSEI64x2,
        B_NONE,
        B_NONE },

  { T("VPCLMULLQHDQ"),
        D_AVX|D_POSTBYTE|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD3A,
        1, 0x000000FF, 0x00000044, 0x10,
        B_SREGI64x2|B_UPD,
        B_SVEXI64x2,
        B_SSEI64x2,
        B_NONE },

  { T("PCLMULHQHDQ"),
        D_SSE|D_POSTBYTE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x00443A0F, 0x11,
        B_SREGI64x2|B_UPD,
        B_SSEI64x2,
        B_NONE,
        B_NONE },

  { T("VPCLMULHQHDQ"),
        D_AVX|D_POSTBYTE|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD3A,
        1, 0x000000FF, 0x00000044, 0x11,
        B_SREGI64x2|B_UPD,
        B_SVEXI64x2,
        B_SSEI64x2,
        B_NONE },

  { T("PCLMULQDQ"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x00443A0F, 0x00,
        B_SREGI64x2|B_UPD,
        B_SSEI64x2,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("VPCLMULQDQ"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD3A,
        1, 0x000000FF, 0x00000044, 0x00,
        B_SREGI64x2|B_UPD,
        B_SVEXI64x2,
        B_SSEI64x2,
        B_CONST8|B_BINARY },

  { T("PCMPEQQ"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0029380F, 0x00,
        B_SREGI64x2|B_UPD,
        B_SSEI64x2,
        B_NONE,
        B_NONE },

  { T("VPCMPEQQ"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD38,
        1, 0x000000FF, 0x00000029, 0x00,
        B_SREGI64x2|B_UPD,
        B_SVEXI64x2,
        B_SSEI64x2,
        B_NONE },

  { T("PCMPESTRI"),
        D_SSE|D_MUST66|D_ALLFLAGS, 0,
        3, 0x00FFFFFF, 0x00613A0F, 0x00,
        B_SREGI8x16,
        B_SSEI8x16,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("VPCMPESTRI"),
        D_AVX|D_MUST66|D_ALLFLAGS, DX_VEX|DX_LSHORT|DX_NOVREG|DX_LEAD3A,
        1, 0x000000FF, 0x00000061, 0x00,
        B_SREGI8x16,
        B_SSEI8x16,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("PCMPESTRM"),
        D_SSE|D_MUST66|D_ALLFLAGS, 0,
        3, 0x00FFFFFF, 0x00603A0F, 0x00,
        B_SREGI8x16,
        B_SSEI8x16,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("VPCMPESTRM"),
        D_AVX|D_MUST66|D_ALLFLAGS, DX_VEX|DX_LSHORT|DX_NOVREG|DX_LEAD3A,
        1, 0x000000FF, 0x00000060, 0x00,
        B_SREGI8x16,
        B_SSEI8x16,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("PCMPISTRI"),
        D_SSE|D_MUST66|D_ALLFLAGS, 0,
        3, 0x00FFFFFF, 0x00633A0F, 0x00,
        B_SREGI8x16,
        B_SSEI8x16,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("VPCMPISTRI"),
        D_AVX|D_MUST66|D_ALLFLAGS, DX_VEX|DX_LSHORT|DX_NOVREG|DX_LEAD3A,
        1, 0x000000FF, 0x00000063, 0x00,
        B_SREGI8x16,
        B_SSEI8x16,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("PCMPISTRM"),
        D_SSE|D_MUST66|D_ALLFLAGS, 0,
        3, 0x00FFFFFF, 0x00623A0F, 0x00,
        B_SREGI8x16,
        B_SSEI8x16,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("VPCMPISTRM"),
        D_AVX|D_MUST66|D_ALLFLAGS, DX_VEX|DX_LSHORT|DX_NOVREG|DX_LEAD3A,
        1, 0x000000FF, 0x00000062, 0x00,
        B_SREGI8x16,
        B_SSEI8x16,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("PCMPGTQ"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0037380F, 0x00,
        B_SREGI64x2|B_UPD,
        B_SSEI64x2,
        B_NONE,
        B_NONE },

  { T("VPCMPGTQ"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD38,
        1, 0x000000FF, 0x00000037, 0x00,
        B_SREGI64x2|B_UPD,
        B_SVEXI64x2,
        B_SSEI64x2,
        B_NONE },

  { T("PEXTRB"),
        D_SSE|D_MUST66|D_MEMORY, 0,
        3, 0x00FFFFFF, 0x00143A0F, 0x00,
        B_INT8|B_MEMORY|B_CHG,
        B_SREGI8x16,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("PEXTRB"),
        D_SSE|D_MUST66|D_REGISTER, 0,
        3, 0x00FFFFFF, 0x00143A0F, 0x00,
        B_INT32|B_REGISTER|B_CHG,
        B_SREGI8x16,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("VPEXTRB"),
        D_AVX|D_MUST66|D_MEMORY, DX_VEX|DX_LSHORT|DX_NOVREG|DX_W0|DX_LEAD3A,
        1, 0x000000FF, 0x00000014, 0x00,
        B_INT8|B_MEMORY|B_CHG,
        B_SREGI8x16,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("VPEXTRB"),
        D_AVX|D_MUST66|D_REGISTER, DX_VEX|DX_LSHORT|DX_NOVREG|DX_W0|DX_LEAD3A,
        1, 0x000000FF, 0x00000014, 0x00,
        B_INT32|B_REGISTER|B_CHG,
        B_SREGI8x16,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("PEXTRD"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x00163A0F, 0x00,
        B_INT32|B_CHG,
        B_SREGI32x4,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("VPEXTRD"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_NOVREG|DX_LEAD3A,
        1, 0x000000FF, 0x00000016, 0x00,
        B_INT32|B_CHG,
        B_SREGI32x4,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("PEXTRW"),
        D_SSE|D_MUST66|D_MEMORY, 0,
        3, 0x00FFFFFF, 0x00153A0F, 0x00,
        B_INT16|B_MEMORY|B_CHG,
        B_SREGI16x8,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("PEXTRW"),
        D_SSE|D_MUST66|D_REGISTER, 0,
        3, 0x00FFFFFF, 0x00153A0F, 0x00,
        B_INT32|B_REGISTER|B_CHG,
        B_SREGI16x8,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("VPEXTRW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_NOVREG|DX_LEAD3A,
        1, 0x000000FF, 0x00000015, 0x00,
        B_INT16|B_CHG,
        B_SREGI16x8,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("PHMINPOSUW"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0041380F, 0x00,
        B_SREGI16x8|B_CHG,
        B_SSEI16x8,
        B_NONE,
        B_NONE },

  { T("VPHMINPOSUW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_NOVREG|DX_LEAD38,
        1, 0x000000FF, 0x00000041, 0x00,
        B_SREGI16x8|B_CHG,
        B_SSEI16x8,
        B_NONE,
        B_NONE },

  { T("PINSRB"),
        D_SSE|D_MUST66|D_MEMORY, 0,
        3, 0x00FFFFFF, 0x00203A0F, 0x00,
        B_SREGI8x16|B_UPD,
        B_INT8|B_MEMORY,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("VPINSRB"),
        D_AVX|D_MUST66|D_MEMORY, DX_VEX|DX_LSHORT|DX_W0|DX_LEAD3A,
        1, 0x000000FF, 0x00000020, 0x00,
        B_SREGI8x16|B_UPD,
        B_SVEXI8x16,
        B_INT8|B_MEMORY,
        B_CONST8|B_BINARY },

  { T("PINSRB"),
        D_SSE|D_MUST66|D_REGISTER, 0,
        3, 0x00FFFFFF, 0x00203A0F, 0x00,
        B_SREGI8x16|B_UPD,
        B_INT32|B_REGISTER,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("VPINSRB"),
        D_AVX|D_MUST66|D_REGISTER, DX_VEX|DX_LSHORT|DX_W0|DX_LEAD3A,
        1, 0x000000FF, 0x00000020, 0x00,
        B_SREGI8x16|B_UPD,
        B_SVEXI8x16,
        B_INT32|B_REGISTER,
        B_CONST8|B_BINARY },

  { T("PINSRD"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x00223A0F, 0x00,
        B_SREGI32x4|B_UPD,
        B_INT32,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("VPINSRD"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_W0|DX_LEAD3A,
        1, 0x000000FF, 0x00000022, 0x00,
        B_SREGI32x4|B_UPD,
        B_SVEXI32x4,
        B_INT32,
        B_CONST8|B_BINARY },

  { T("PMAXSB"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x003C380F, 0x00,
        B_SREGI8x16|B_SIGNED|B_UPD,
        B_SSEI8x16|B_SIGNED,
        B_NONE,
        B_NONE },

  { T("VPMAXSB"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD38,
        1, 0x000000FF, 0x0000003C, 0x00,
        B_SREGI8x16|B_SIGNED|B_UPD,
        B_SVEXI8x16|B_SIGNED,
        B_SSEI8x16|B_SIGNED,
        B_NONE },

  { T("PMAXSD"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x003D380F, 0x00,
        B_SREGI32x4|B_SIGNED|B_UPD,
        B_SSEI32x4|B_SIGNED,
        B_NONE,
        B_NONE },

  { T("VPMAXSD"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD38,
        1, 0x000000FF, 0x0000003D, 0x00,
        B_SREGI32x4|B_SIGNED|B_UPD,
        B_SVEXI32x4|B_SIGNED,
        B_SSEI32x4|B_SIGNED,
        B_NONE },

  { T("PMAXUD"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x003F380F, 0x00,
        B_SREGI32x4|B_UNSIGNED|B_UPD,
        B_SSEI32x4|B_UNSIGNED,
        B_NONE,
        B_NONE },

  { T("VPMAXUD"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD38,
        1, 0x000000FF, 0x0000003F, 0x00,
        B_SREGI32x4|B_UNSIGNED|B_UPD,
        B_SVEXI32x4|B_UNSIGNED,
        B_SSEI32x4|B_UNSIGNED,
        B_NONE },

  { T("PMAXUW"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x003E380F, 0x00,
        B_SREGI16x8|B_UNSIGNED|B_UPD,
        B_SSEI16x8|B_UNSIGNED,
        B_NONE,
        B_NONE },

  { T("VPMAXUW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD38,
        1, 0x000000FF, 0x0000003E, 0x00,
        B_SREGI16x8|B_UNSIGNED|B_UPD,
        B_SVEXI16x8|B_UNSIGNED,
        B_SSEI16x8|B_UNSIGNED,
        B_NONE },

  { T("PMINSB"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0038380F, 0x00,
        B_SREGI8x16|B_SIGNED|B_UPD,
        B_SSEI8x16|B_SIGNED,
        B_NONE,
        B_NONE },

  { T("VPMINSB"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD38,
        1, 0x000000FF, 0x00000038, 0x00,
        B_SREGI8x16|B_SIGNED|B_UPD,
        B_SVEXI8x16|B_SIGNED,
        B_SSEI8x16|B_SIGNED,
        B_NONE },

  { T("PMINSD"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0039380F, 0x00,
        B_SREGI32x4|B_SIGNED|B_UPD,
        B_SSEI32x4|B_SIGNED,
        B_NONE,
        B_NONE },

  { T("VPMINSD"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD38,
        1, 0x000000FF, 0x00000039, 0x00,
        B_SREGI32x4|B_SIGNED|B_UPD,
        B_SVEXI32x4|B_SIGNED,
        B_SSEI32x4|B_SIGNED,
        B_NONE },

  { T("PMINUD"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x003B380F, 0x00,
        B_SREGI32x4|B_UNSIGNED|B_UPD,
        B_SSEI32x4|B_UNSIGNED,
        B_NONE,
        B_NONE },

  { T("VPMINUD"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD38,
        1, 0x000000FF, 0x0000003B, 0x00,
        B_SREGI32x4|B_UNSIGNED|B_UPD,
        B_SVEXI32x4|B_UNSIGNED,
        B_SSEI32x4|B_UNSIGNED,
        B_NONE },

  { T("PMINUW"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x003A380F, 0x00,
        B_SREGI16x8|B_UNSIGNED|B_UPD,
        B_SSEI16x8|B_UNSIGNED,
        B_NONE,
        B_NONE },

  { T("VPMINUW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD38,
        1, 0x000000FF, 0x0000003A, 0x00,
        B_SREGI16x8|B_UNSIGNED|B_UPD,
        B_SVEXI16x8|B_UNSIGNED,
        B_SSEI16x8|B_UNSIGNED,
        B_NONE },

  { T("PMOVSXBW"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0020380F, 0x00,
        B_SREGI16x8|B_SIGNED|B_CHG,
        B_SSEI8x8L,
        B_NONE,
        B_NONE },

  { T("VPMOVSXBW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_NOVREG|DX_LEAD38,
        1, 0x000000FF, 0x00000020, 0x00,
        B_SREGI16x8|B_SIGNED|B_CHG,
        B_SSEI8x8L,
        B_NONE,
        B_NONE },

  { T("PMOVSXBD"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0021380F, 0x00,
        B_SREGI32x4|B_SIGNED|B_CHG,
        B_SSEI8x4L,
        B_NONE,
        B_NONE },

  { T("VPMOVSXBD"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_NOVREG|DX_LEAD38,
        1, 0x000000FF, 0x00000021, 0x00,
        B_SREGI32x4|B_SIGNED|B_CHG,
        B_SSEI8x4L,
        B_NONE,
        B_NONE },

  { T("PMOVSXBQ"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0022380F, 0x00,
        B_SREGI64x2|B_SIGNED|B_CHG,
        B_SSEI8x2L,
        B_NONE,
        B_NONE },

  { T("VPMOVSXBQ"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_NOVREG|DX_LEAD38,
        1, 0x000000FF, 0x00000022, 0x00,
        B_SREGI64x2|B_SIGNED|B_CHG,
        B_SSEI8x2L,
        B_NONE,
        B_NONE },

  { T("PMOVSXWD"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0023380F, 0x00,
        B_SREGI32x4|B_SIGNED|B_CHG,
        B_SSEI16x4L,
        B_NONE,
        B_NONE },

  { T("VPMOVSXWD"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_NOVREG|DX_LEAD38,
        1, 0x000000FF, 0x00000023, 0x00,
        B_SREGI32x4|B_SIGNED|B_CHG,
        B_SSEI16x4L,
        B_NONE,
        B_NONE },

  { T("PMOVSXWQ"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0024380F, 0x00,
        B_SREGI64x2|B_SIGNED|B_CHG,
        B_SSEI16x2L,
        B_NONE,
        B_NONE },

  { T("VPMOVSXWQ"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_NOVREG|DX_LEAD38,
        1, 0x000000FF, 0x00000024, 0x00,
        B_SREGI64x2|B_SIGNED|B_CHG,
        B_SSEI16x2L,
        B_NONE,
        B_NONE },

  { T("PMOVSXDQ"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0025380F, 0x00,
        B_SREGI64x2|B_SIGNED|B_CHG,
        B_SSEI32x2L,
        B_NONE,
        B_NONE },

  { T("VPMOVSXDQ"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_NOVREG|DX_LEAD38,
        1, 0x000000FF, 0x00000025, 0x00,
        B_SREGI64x2|B_SIGNED|B_CHG,
        B_SSEI32x2L,
        B_NONE,
        B_NONE },

  { T("PMOVZXBW"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0030380F, 0x00,
        B_SREGI16x8|B_CHG,
        B_SSEI8x8L,
        B_NONE,
        B_NONE },

  { T("VPMOVZXBW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_NOVREG|DX_LEAD38,
        1, 0x000000FF, 0x00000030, 0x00,
        B_SREGI16x8|B_CHG,
        B_SSEI8x8L,
        B_NONE,
        B_NONE },

  { T("PMOVZXBD"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0031380F, 0x00,
        B_SREGI32x4|B_CHG,
        B_SSEI8x4L,
        B_NONE,
        B_NONE },

  { T("VPMOVZXBD"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_NOVREG|DX_LEAD38,
        1, 0x000000FF, 0x00000031, 0x00,
        B_SREGI32x4|B_CHG,
        B_SSEI8x4L,
        B_NONE,
        B_NONE },

  { T("PMOVZXBQ"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0032380F, 0x00,
        B_SREGI64x2|B_CHG,
        B_SSEI8x2L,
        B_NONE,
        B_NONE },

  { T("VPMOVZXBQ"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_NOVREG|DX_LEAD38,
        1, 0x000000FF, 0x00000032, 0x00,
        B_SREGI64x2|B_CHG,
        B_SSEI8x2L,
        B_NONE,
        B_NONE },

  { T("PMOVZXWD"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0033380F, 0x00,
        B_SREGI32x4|B_CHG,
        B_SSEI16x4L,
        B_NONE,
        B_NONE },

  { T("VPMOVZXWD"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_NOVREG|DX_LEAD38,
        1, 0x000000FF, 0x00000033, 0x00,
        B_SREGI32x4|B_CHG,
        B_SSEI16x4L,
        B_NONE,
        B_NONE },

  { T("PMOVZXWQ"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0034380F, 0x00,
        B_SREGI64x2|B_CHG,
        B_SSEI16x2L,
        B_NONE,
        B_NONE },

  { T("VPMOVZXWQ"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_NOVREG|DX_LEAD38,
        1, 0x000000FF, 0x00000034, 0x00,
        B_SREGI64x2|B_CHG,
        B_SSEI16x2L,
        B_NONE,
        B_NONE },

  { T("PMOVZXDQ"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0035380F, 0x00,
        B_SREGI64x2|B_CHG,
        B_SSEI32x2L,
        B_NONE,
        B_NONE },

  { T("VPMOVZXDQ"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_NOVREG|DX_LEAD38,
        1, 0x000000FF, 0x00000035, 0x00,
        B_SREGI64x2|B_CHG,
        B_SSEI32x2L,
        B_NONE,
        B_NONE },

  { T("PMULDQ"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0028380F, 0x00,
        B_SREGI32x4|B_SIGNED|B_UPD,
        B_SSEI32x4|B_SIGNED,
        B_NONE,
        B_NONE },

  { T("VPMULDQ"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD38,
        1, 0x000000FF, 0x00000028, 0x00,
        B_SREGI32x4|B_SIGNED|B_UPD,
        B_SVEXI32x4|B_SIGNED,
        B_SSEI32x4|B_SIGNED,
        B_NONE },

  { T("PMULLD"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0040380F, 0x00,
        B_SREGI32x4|B_SIGNED|B_UPD,
        B_SSEI32x4|B_SIGNED,
        B_NONE,
        B_NONE },

  { T("VPMULLD"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD38,
        1, 0x000000FF, 0x00000040, 0x00,
        B_SREGI32x4|B_SIGNED|B_UPD,
        B_SVEXI32x4|B_SIGNED,
        B_SSEI32x4|B_SIGNED,
        B_NONE },

  { T("PTEST"),
        D_SSE|D_MUST66|D_ALLFLAGS, 0,
        3, 0x00FFFFFF, 0x0017380F, 0x00,
        B_SREGI32x4,
        B_SSEI32x4,
        B_NONE,
        B_NONE },

  { T("VPTEST"),
        D_AVX|D_MUST66|D_ALLFLAGS, DX_VEX|DX_LBOTH|DX_NOVREG|DX_LEAD38,
        1, 0x000000FF, 0x00000017, 0x00,
        B_SREGI32x4,
        B_SSEI32x4,
        B_NONE,
        B_NONE },

  { T("ROUNDPD"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x00093A0F, 0x00,
        B_SREGF64x2|B_CHG,
        B_SSEF64x2,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("VROUNDPD"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH|DX_NOVREG|DX_LEAD3A,
        1, 0x000000FF, 0x00000009, 0x00,
        B_SREGF64x2|B_CHG,
        B_SSEF64x2,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("ROUNDPS"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x00083A0F, 0x00,
        B_SREGF32x4|B_CHG,
        B_SSEF32x4,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("VROUNDPS"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH|DX_NOVREG|DX_LEAD3A,
        1, 0x000000FF, 0x00000008, 0x00,
        B_SREGF32x4|B_CHG,
        B_SSEF32x4,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("ROUNDSD"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x000B3A0F, 0x00,
        B_SREGF64L|B_CHG,
        B_SSEF64L,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("VROUNDSD"),
        D_AVX|D_MUST66, DX_VEX|DX_IGNOREL|DX_LEAD3A,
        1, 0x000000FF, 0x0000000B, 0x00,
        B_SREGF64L|B_CHG,
        B_SVEXF64L,
        B_SSEF64L,
        B_CONST8|B_BINARY },

  { T("ROUNDSS"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x000A3A0F, 0x00,
        B_SREGF32L|B_CHG,
        B_SSEF32L,
        B_CONST8|B_BINARY,
        B_NONE },

  { T("VROUNDSS"),
        D_AVX|D_MUST66, DX_VEX|DX_IGNOREL|DX_LEAD3A,
        1, 0x000000FF, 0x0000000A, 0x00,
        B_SREGF32L|B_CHG,
        B_SVEXF32L,
        B_SSEF32L,
        B_CONST8|B_BINARY },

  { T("PABSB"),
        D_MMX|D_MUSTNONE, 0,
        3, 0x00FFFFFF, 0x001C380F, 0x00,
        B_MREG8x8|B_UNSIGNED|B_CHG,
        B_MMX8x8|B_SIGNED,
        B_NONE,
        B_NONE },

  { T("PABSB"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x001C380F, 0x00,
        B_SREGI8x16|B_UNSIGNED|B_CHG,
        B_SSEI8x16|B_SIGNED,
        B_NONE,
        B_NONE },

  { T("VPABSB"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_NOVREG|DX_LEAD38,
        1, 0x000000FF, 0x0000001C, 0x00,
        B_SREGI8x16|B_UNSIGNED|B_CHG,
        B_SSEI8x16|B_SIGNED,
        B_NONE,
        B_NONE },

  { T("PABSW"),
        D_MMX|D_MUSTNONE, 0,
        3, 0x00FFFFFF, 0x001D380F, 0x00,
        B_MREG16x4|B_UNSIGNED|B_CHG,
        B_MMX16x4|B_SIGNED,
        B_NONE,
        B_NONE },

  { T("PABSW"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x001D380F, 0x00,
        B_SREGI16x8|B_UNSIGNED|B_CHG,
        B_SSEI16x8|B_SIGNED,
        B_NONE,
        B_NONE },

  { T("VPABSW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_NOVREG|DX_LEAD38,
        1, 0x000000FF, 0x0000001D, 0x00,
        B_SREGI16x8|B_UNSIGNED|B_CHG,
        B_SSEI16x8|B_SIGNED,
        B_NONE,
        B_NONE },

  { T("PABSD"),
        D_MMX|D_MUSTNONE, 0,
        3, 0x00FFFFFF, 0x001E380F, 0x00,
        B_MREG32x2|B_UNSIGNED|B_CHG,
        B_MMX32x2|B_SIGNED,
        B_NONE,
        B_NONE },

  { T("PABSD"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x001E380F, 0x00,
        B_SREGI32x4|B_UNSIGNED|B_CHG,
        B_SSEI32x4|B_SIGNED,
        B_NONE,
        B_NONE },

  { T("VPABSD"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_NOVREG|DX_LEAD38,
        1, 0x000000FF, 0x0000001E, 0x00,
        B_SREGI32x4|B_UNSIGNED|B_CHG,
        B_SSEI32x4|B_SIGNED,
        B_NONE,
        B_NONE },

  { T("PALIGNR"),
        D_MMX|D_MUSTNONE, 0,
        3, 0x00FFFFFF, 0x000F3A0F, 0x00,
        B_MREG8x8|B_BINARY|B_UPD,
        B_MMX8x8|B_BINARY,
        B_CONST8|B_UNSIGNED,
        B_NONE },

  { T("PALIGNR"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x000F3A0F, 0x00,
        B_SREGI8x16|B_BINARY|B_UPD,
        B_SSEI8x16|B_BINARY,
        B_CONST8|B_UNSIGNED,
        B_NONE },

  { T("VPALIGNR"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD3A,
        1, 0x000000FF, 0x0000000F, 0x00,
        B_SREGI8x16|B_BINARY|B_UPD,
        B_SVEXI8x16|B_BINARY,
        B_SSEI8x16|B_BINARY,
        B_CONST8|B_UNSIGNED },

  { T("PHADDW"),
        D_MMX|D_MUSTNONE, 0,
        3, 0x00FFFFFF, 0x0001380F, 0x00,
        B_MREG16x4|B_UPD,
        B_MMX16x4,
        B_NONE,
        B_NONE },

  { T("PHADDW"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0001380F, 0x00,
        B_SREGI16x8|B_UPD,
        B_SSEI16x8,
        B_NONE,
        B_NONE },

  { T("VPHADDW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD38,
        1, 0x000000FF, 0x00000001, 0x00,
        B_SREGI16x8|B_UPD,
        B_SVEXI16x8,
        B_SSEI16x8,
        B_NONE },

  { T("PHADDD"),
        D_MMX|D_MUSTNONE, 0,
        3, 0x00FFFFFF, 0x0002380F, 0x00,
        B_MREG32x2|B_UPD,
        B_MMX32x2,
        B_NONE,
        B_NONE },

  { T("PHADDD"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0002380F, 0x00,
        B_SREGI32x4|B_UPD,
        B_SSEI32x4,
        B_NONE,
        B_NONE },

  { T("VPHADDD"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD38,
        1, 0x000000FF, 0x00000002, 0x00,
        B_SREGI32x4|B_UPD,
        B_SVEXI32x4,
        B_SSEI32x4,
        B_NONE },

  { T("PHSUBW"),
        D_MMX|D_MUSTNONE, 0,
        3, 0x00FFFFFF, 0x0005380F, 0x00,
        B_MREG16x4|B_UPD,
        B_MMX16x4,
        B_NONE,
        B_NONE },

  { T("PHSUBW"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0005380F, 0x00,
        B_SREGI16x8|B_UPD,
        B_SSEI16x8,
        B_NONE,
        B_NONE },

  { T("VPHSUBW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD38,
        1, 0x000000FF, 0x00000005, 0x00,
        B_SREGI16x8|B_UPD,
        B_SVEXI16x8,
        B_SSEI16x8,
        B_NONE },

  { T("PHSUBD"),
        D_MMX|D_MUSTNONE, 0,
        3, 0x00FFFFFF, 0x0006380F, 0x00,
        B_MREG32x2|B_UPD,
        B_MMX32x2,
        B_NONE,
        B_NONE },

  { T("PHSUBD"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0006380F, 0x00,
        B_SREGI32x4|B_UPD,
        B_SSEI32x4,
        B_NONE,
        B_NONE },

  { T("VPHSUBD"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD38,
        1, 0x000000FF, 0x00000006, 0x00,
        B_SREGI32x4|B_UPD,
        B_SVEXI32x4,
        B_SSEI32x4,
        B_NONE },

  { T("PHADDSW"),
        D_MMX|D_MUSTNONE, 0,
        3, 0x00FFFFFF, 0x0003380F, 0x00,
        B_MREG16x4|B_UPD,
        B_MMX16x4,
        B_NONE,
        B_NONE },

  { T("PHADDSW"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0003380F, 0x00,
        B_SREGI16x8|B_UPD,
        B_SSEI16x8,
        B_NONE,
        B_NONE },

  { T("VPHADDSW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD38,
        1, 0x000000FF, 0x00000003, 0x00,
        B_SREGI16x8|B_UPD,
        B_SVEXI16x8,
        B_SSEI16x8,
        B_NONE },

  { T("PHSUBSW"),
        D_MMX|D_MUSTNONE, 0,
        3, 0x00FFFFFF, 0x0007380F, 0x00,
        B_MREG16x4|B_UPD,
        B_MMX16x4,
        B_NONE,
        B_NONE },

  { T("PHSUBSW"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0007380F, 0x00,
        B_SREGI16x8|B_UPD,
        B_SSEI16x8,
        B_NONE,
        B_NONE },

  { T("VPHSUBSW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD38,
        1, 0x000000FF, 0x00000007, 0x00,
        B_SREGI16x8|B_UPD,
        B_SVEXI16x8,
        B_SSEI16x8,
        B_NONE },

  { T("PMADDUBSW"),
        D_MMX|D_MUSTNONE, 0,
        3, 0x00FFFFFF, 0x0004380F, 0x00,
        B_MREG8x8|B_UNSIGNED|B_UPD,
        B_MMX8x8|B_SIGNED,
        B_NONE,
        B_NONE },

  { T("PMADDUBSW"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0004380F, 0x00,
        B_SREGI8x16|B_UNSIGNED|B_UPD,
        B_SSEI8x16|B_SIGNED,
        B_NONE,
        B_NONE },

  { T("VPMADDUBSW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD38,
        1, 0x000000FF, 0x00000004, 0x00,
        B_SREGI8x16|B_UNSIGNED|B_UPD,
        B_SVEXI8x16|B_UNSIGNED,
        B_SSEI8x16|B_SIGNED,
        B_NONE },

  { T("PMULHRSW"),
        D_MMX|D_MUSTNONE, 0,
        3, 0x00FFFFFF, 0x000B380F, 0x00,
        B_MREG16x4|B_UPD,
        B_MMX16x4,
        B_NONE,
        B_NONE },

  { T("PMULHRSW"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x000B380F, 0x00,
        B_SREGI16x8|B_UPD,
        B_SSEI16x8,
        B_NONE,
        B_NONE },

  { T("VPMULHRSW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD38,
        1, 0x000000FF, 0x0000000B, 0x00,
        B_SREGI16x8|B_UPD,
        B_SVEXI16x8,
        B_SSEI16x8,
        B_NONE },

  { T("PSHUFB"),
        D_MMX|D_MUSTNONE, 0,
        3, 0x00FFFFFF, 0x0000380F, 0x00,
        B_MREG8x8|B_UPD,
        B_MMX8x8|B_BINARY,
        B_NONE,
        B_NONE },

  { T("PSHUFB"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0000380F, 0x00,
        B_SREGI8x16|B_UPD,
        B_SSEI8x16|B_BINARY,
        B_NONE,
        B_NONE },

  { T("VPSHUFB"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD38,
        1, 0x000000FF, 0x00000000, 0x00,
        B_SREGI8x16|B_UPD,
        B_SVEXI8x16,
        B_SSEI8x16|B_BINARY,
        B_NONE },

  { T("PSIGNB"),
        D_MMX|D_MUSTNONE, 0,
        3, 0x00FFFFFF, 0x0008380F, 0x00,
        B_MREG8x8|B_SIGNED|B_UPD,
        B_MMX8x8|B_SIGNED,
        B_NONE,
        B_NONE },

  { T("PSIGNB"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0008380F, 0x00,
        B_SREGI8x16|B_SIGNED|B_UPD,
        B_SSEI8x16|B_SIGNED,
        B_NONE,
        B_NONE },

  { T("VPSIGNB"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD38,
        1, 0x000000FF, 0x00000008, 0x00,
        B_SREGI8x16|B_SIGNED|B_UPD,
        B_SVEXI8x16|B_SIGNED,
        B_SSEI8x16|B_SIGNED,
        B_NONE },

  { T("PSIGNW"),
        D_MMX|D_MUSTNONE, 0,
        3, 0x00FFFFFF, 0x0009380F, 0x00,
        B_MREG16x4|B_SIGNED|B_UPD,
        B_MMX16x4|B_SIGNED,
        B_NONE,
        B_NONE },

  { T("PSIGNW"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x0009380F, 0x00,
        B_SREGI16x8|B_SIGNED|B_UPD,
        B_SSEI16x8|B_SIGNED,
        B_NONE,
        B_NONE },

  { T("VPSIGNW"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD38,
        1, 0x000000FF, 0x00000009, 0x00,
        B_SREGI16x8|B_SIGNED|B_UPD,
        B_SVEXI16x8|B_SIGNED,
        B_SSEI16x8|B_SIGNED,
        B_NONE },

  { T("PSIGND"),
        D_MMX|D_MUSTNONE, 0,
        3, 0x00FFFFFF, 0x000A380F, 0x00,
        B_MREG32x2|B_SIGNED|B_UPD,
        B_MMX32x2|B_SIGNED,
        B_NONE,
        B_NONE },

  { T("PSIGND"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x000A380F, 0x00,
        B_SREGI32x4|B_SIGNED|B_UPD,
        B_SSEI32x4|B_SIGNED,
        B_NONE,
        B_NONE },

  { T("VPSIGND"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD38,
        1, 0x000000FF, 0x0000000A, 0x00,
        B_SREGI32x4|B_SIGNED|B_UPD,
        B_SVEXI32x4|B_SIGNED,
        B_SSEI32x4|B_SIGNED,
        B_NONE },

  { T("VBROADCASTSS"),
        D_AVX|D_MUST66|D_MEMORY, DX_VEX|DX_LBOTH|DX_NOVREG|DX_W0|DX_LEAD38,
        1, 0x000000FF, 0x00000018, 0x00,
        B_SREGF32x4|B_CHG,
        B_SSEF32L|B_MEMORY,
        B_NONE,
        B_NONE },

  { T("VBROADCASTSD"),
        D_AVX|D_MUST66|D_MEMORY, DX_VEX|DX_LLONG|DX_NOVREG|DX_W0|DX_LEAD38,
        1, 0x000000FF, 0x00000019, 0x00,
        B_SREGF64x2|B_CHG,
        B_SSEF64L|B_MEMORY,
        B_NONE,
        B_NONE },

  { T("VBROADCASTF128"),
        D_AVX|D_MUST66|D_MEMORY, DX_VEX|DX_LLONG|DX_NOVREG|DX_W0|DX_LEAD38,
        1, 0x000000FF, 0x0000001A, 0x00,
        B_SREGF64x2|B_CHG,
        B_SSEF64x2|B_MEMORY|B_NOVEXSIZE|B_SHOWSIZE,
        B_NONE,
        B_NONE },

  { T("VEXTRACTF128"),
        D_AVX|D_MUST66, DX_VEX|DX_LLONG|DX_NOVREG|DX_W0|DX_LEAD3A,
        1, 0x000000FF, 0x00000019, 0x00,
        B_SSEF64x2|B_NOVEXSIZE|B_SHOWSIZE|B_CHG,
        B_SREGF64x2,
        B_CONST8,
        B_NONE },

  { T("VINSERTF128"),
        D_AVX|D_MUST66, DX_VEX|DX_LLONG|DX_W0|DX_LEAD3A,
        1, 0x000000FF, 0x00000018, 0x00,
        B_SREGF64x2|B_CHG,
        B_SVEXF64x2,
        B_SSEF64x2|B_NOVEXSIZE|B_SHOWSIZE,
        B_CONST8 },

  { T("VMASKMOVPS"),
        D_AVX|D_MUST66|D_MEMORY, DX_VEX|DX_LBOTH|DX_W0|DX_LEAD38,
        1, 0x000000FF, 0x0000002C, 0x00,
        B_SREGF32x4|B_CHG,
        B_SVEXF32x4,
        B_SSEF32x4|B_MEMORY,
        B_NONE },

  { T("VMASKMOVPS"),
        D_AVX|D_MUST66|D_MEMORY, DX_VEX|DX_LBOTH|DX_W0|DX_LEAD38,
        1, 0x000000FF, 0x0000002E, 0x00,
        B_SSEF32x4|B_MEMORY|B_CHG,
        B_SVEXF32x4,
        B_SREGF32x4,
        B_NONE },

  { T("VMASKMOVPD"),
        D_AVX|D_MUST66|D_MEMORY, DX_VEX|DX_LBOTH|DX_W0|DX_LEAD38,
        1, 0x000000FF, 0x0000002D, 0x00,
        B_SREGF64x2|B_CHG,
        B_SVEXF64x2,
        B_SSEF64x2|B_MEMORY,
        B_NONE },

  { T("VMASKMOVPD"),
        D_AVX|D_MUST66|D_MEMORY, DX_VEX|DX_LBOTH|DX_W0|DX_LEAD38,
        1, 0x000000FF, 0x0000002F, 0x00,
        B_SSEF64x2|B_MEMORY|B_CHG,
        B_SVEXF64x2,
        B_SREGF64x2,
        B_NONE },

  { T("VPERMILPD"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH|DX_W0|DX_LEAD38,
        1, 0x000000FF, 0x0000000D, 0x00,
        B_SREGF64x2|B_CHG,
        B_SVEXF64x2,
        B_SSEI64x2,
        B_NONE },

  { T("VPERMILPD"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH|DX_NOVREG|DX_W0|DX_LEAD3A,
        1, 0x000000FF, 0x00000005, 0x00,
        B_SREGF64x2|B_CHG,
        B_SSEF64x2,
        B_CONST8,
        B_NONE },

  { T("VPERMILPS"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH|DX_W0|DX_LEAD38,
        1, 0x000000FF, 0x0000000C, 0x00,
        B_SREGF32x4|B_CHG,
        B_SVEXF32x4,
        B_SSEI32x4,
        B_NONE },

  { T("VPERMILPS"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH|DX_NOVREG|DX_W0|DX_LEAD3A,
        1, 0x000000FF, 0x00000004, 0x00,
        B_SREGF32x4|B_CHG,
        B_SSEF32x4,
        B_CONST8,
        B_NONE },

  { T("VPERM2F128"),
        D_AVX|D_MUST66, DX_VEX|DX_LLONG|DX_W0|DX_LEAD3A,
        1, 0x000000FF, 0x00000006, 0x00,
        B_SREGF64x2|B_CHG,
        B_SVEXF64x2,
        B_SSEF64x2,
        B_CONST8 },

  { T("VTESTPS"),
        D_AVX|D_MUST66|D_ALLFLAGS, DX_VEX|DX_LBOTH|DX_NOVREG|DX_W0|DX_LEAD38,
        1, 0x000000FF, 0x0000000E, 0x00,
        B_SREGF32x4,
        B_SSEF32x4,
        B_NONE,
        B_NONE },

  { T("VTESTPD"),
        D_AVX|D_MUST66|D_ALLFLAGS, DX_VEX|DX_LBOTH|DX_NOVREG|DX_W0|DX_LEAD38,
        1, 0x000000FF, 0x0000000F, 0x00,
        B_SREGF64x2,
        B_SSEF64x2,
        B_NONE,
        B_NONE },

  { T("VZEROALL"),
        D_AVX|D_MUSTNONE, DX_VEX|DX_LLONG|DX_NOVREG,
        1, 0x000000FF, 0x00000077, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("VZEROUPPER"),
        D_AVX|D_MUSTNONE, DX_VEX|DX_LSHORT|DX_NOVREG,
        1, 0x000000FF, 0x00000077, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("AESDEC"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x00DE380F, 0x00,
        B_SREGI8x16|B_UPD,
        B_SSEI8x16,
        B_NONE,
        B_NONE },

  { T("VAESDEC"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD38,
        1, 0x000000FF, 0x000000DE, 0x00,
        B_SREGI8x16|B_CHG,
        B_SVEXI8x16,
        B_SSEI8x16,
        B_NONE },

  { T("AESDECLAST"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x00DF380F, 0x00,
        B_SREGI8x16|B_UPD,
        B_SSEI8x16,
        B_NONE,
        B_NONE },

  { T("VAESDECLAST"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD38,
        1, 0x000000FF, 0x000000DF, 0x00,
        B_SREGI8x16|B_CHG,
        B_SVEXI8x16,
        B_SSEI8x16,
        B_NONE },

  { T("AESENC"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x00DC380F, 0x00,
        B_SREGI8x16|B_UPD,
        B_SSEI8x16,
        B_NONE,
        B_NONE },

  { T("VAESENC"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD38,
        1, 0x000000FF, 0x000000DC, 0x00,
        B_SREGI8x16|B_CHG,
        B_SVEXI8x16,
        B_SSEI8x16,
        B_NONE },

  { T("AESENCLAST"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x00DD380F, 0x00,
        B_SREGI8x16|B_UPD,
        B_SSEI8x16,
        B_NONE,
        B_NONE },

  { T("VAESENCLAST"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_LEAD38,
        1, 0x000000FF, 0x000000DD, 0x00,
        B_SREGI8x16|B_CHG,
        B_SVEXI8x16,
        B_SSEI8x16,
        B_NONE },

  { T("AESIMC"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x00DB380F, 0x00,
        B_SREGI8x16|B_CHG,
        B_SSEI8x16,
        B_NONE,
        B_NONE },

  { T("VAESIMC"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_NOVREG|DX_LEAD38,
        1, 0x000000FF, 0x000000DB, 0x00,
        B_SREGI8x16|B_CHG,
        B_SSEI8x16,
        B_NONE,
        B_NONE },

  { T("AESKEYGENASSIST"),
        D_SSE|D_MUST66, 0,
        3, 0x00FFFFFF, 0x00DF3A0F, 0x00,
        B_SREGI8x16|B_CHG,
        B_SSEI8x16,
        B_CONST8|B_COUNT,
        B_NONE },

  { T("VAESKEYGENASSIST"),
        D_AVX|D_MUST66, DX_VEX|DX_LSHORT|DX_NOVREG|DX_LEAD3A,
        1, 0x000000FF, 0x000000DF, 0x00,
        B_SREGI8x16|B_CHG,
        B_SSEI8x16,
        B_CONST8|B_COUNT,
        B_NONE },

  { T("VCVTPH2PS"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH|DX_NOVREG|DX_W0|DX_LEAD38,
        1, 0x000000FF, 0x00000013, 0x00,
        B_SREGF32x4|B_CHG,
        B_SSEI16x4L,
        B_NONE,
        B_NONE },

  { T("VCVTPS2PH"),
        D_AVX|D_MUST66, DX_VEX|DX_LBOTH|DX_NOVREG|DX_W0|DX_LEAD3A,
        1, 0x000000FF, 0x0000001D, 0x00,
        B_SSEI16x4L|B_CHG,
        B_SREGF32x4,
        B_NONE,
        B_NONE },

  { T("LZCNT"),
        D_CMD|D_NEEDF3|D_ALLFLAGS, 0,
        2, 0x0000FFFF, 0x0000BD0F, 0x00,
        B_REG|B_CHG,
        B_INT|B_BINARY,
        B_NONE,
        B_NONE },

  { T("POPCNT"),
        D_CMD|D_NEEDF3|D_ALLFLAGS, 0,
        2, 0x0000FFFF, 0x0000B80F, 0x00,
        B_REG|B_CHG,
        B_INT|B_NOADDR,
        B_NONE,
        B_NONE },

  { T("EXTRQ"),
        D_SSE|D_MUST66, 0,
        2, 0x0038FFFF, 0x0000780F, 0x00,
        B_SSEI8x16|B_REGONLY|B_UPD,
        B_CONST8|B_COUNT,
        B_CONST8_2|B_COUNT,
        B_NONE },

  { T("EXTRQ"),
        D_SSE|D_MUST66, 0,
        2, 0x0000FFFF, 0x0000790F, 0x00,
        B_SREGI8x16|B_UPD,
        B_SSEI8x2L|B_REGONLY,
        B_NONE,
        B_NONE },

  { T("INSERTQ"),
        D_SSE|D_MUSTF2, 0,
        2, 0x0000FFFF, 0x0000780F, 0x00,
        B_SREGI8x16|B_REGONLY|B_UPD,
        B_SSEI8x8L,
        B_CONST8|B_COUNT,
        B_CONST8_2|B_COUNT },

  { T("INSERTQ"),
        D_SSE|D_MUSTF2, 0,
        2, 0x0000FFFF, 0x0000790F, 0x00,
        B_SREGI8x16|B_REGONLY|B_UPD,
        B_SSEI8x16,
        B_NONE,
        B_NONE },

  { T("MOVNTSD"),
        D_SSE|D_MUSTF2, 0,
        2, 0x0000FFFF, 0x00002B0F, 0x00,
        B_SSEF64L|B_MEMONLY|B_CHG,
        B_SREGF64L,
        B_NONE,
        B_NONE },

  { T("MOVNTSS"),
        D_SSE|D_MUSTF3, 0,
        2, 0x0000FFFF, 0x00002B0F, 0x00,
        B_SSEF32L|B_MEMONLY|B_CHG,
        B_SREGF32L,
        B_NONE,
        B_NONE },

  { T("INVEPT"),
        D_PRIVILEGED|D_MUST66|D_MEMORY|D_RARE, 0,
        3, 0x00FFFFFF, 0x0080380F, 0x00,
        B_REG32,
        B_INT128,
        B_NONE,
        B_NONE },

  { T("INVVPID"),
        D_PRIVILEGED|D_MUST66|D_MEMORY|D_RARE, 0,
        3, 0x00FFFFFF, 0x0081380F, 0x00,
        B_REG32,
        B_INT128,
        B_NONE,
        B_NONE },

  { T("VMCALL"),
        D_PRIVILEGED|D_RARE, 0,
        3, 0x00FFFFFF, 0x00C1010F, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("VMCLEAR"),
        D_PRIVILEGED|D_MUST66|D_MEMORY|D_RARE, 0,
        2, 0x0038FFFF, 0x0030C70F, 0x00,
        B_INT64|B_MEMONLY,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("VMLAUNCH"),
        D_PRIVILEGED|D_RARE, 0,
        3, 0x00FFFFFF, 0x00C2010F, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("VMFUNC"),
        D_PRIVILEGED|D_RARE, 0,
        3, 0x00FFFFFF, 0x00D4010F, 0x00,
        B_EAX|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("XEND"),
        D_PRIVILEGED|D_RARE, 0,
        3, 0x00FFFFFF, 0x00D5010F, 0x00,
        B_EAX|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("XTEST"),
        D_PRIVILEGED|D_RARE, 0,
        3, 0x00FFFFFF, 0x00D6010F, 0x00,
        B_EAX|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("VMRESUME"),
        D_PRIVILEGED|D_RARE, 0,
        3, 0x00FFFFFF, 0x00C3010F, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("VMPTRLD"),
        D_PRIVILEGED|D_MUSTNONE|D_MEMORY|D_RARE, 0,
        2, 0x0038FFFF, 0x0030C70F, 0x00,
        B_INT64|B_MEMONLY,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("VMPTRST"),
        D_PRIVILEGED|D_MUSTNONE|D_MEMORY|D_RARE, 0,
        2, 0x0038FFFF, 0x0038C70F, 0x00,
        B_INT64|B_MEMONLY|B_CHG,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("VMREAD"),
        D_PRIVILEGED|D_MUSTNONE|D_RARE, 0,
        2, 0x0000FFFF, 0x0000780F, 0x00,
        B_INT32|B_CHG,
        B_REG32,
        B_NONE,
        B_NONE },

  { T("VMWRITE"),
        D_PRIVILEGED|D_RARE, 0,
        2, 0x0000FFFF, 0x0000790F, 0x00,
        B_REG32,
        B_INT32,
        B_NONE,
        B_NONE },

  { T("VMXOFF"),
        D_PRIVILEGED|D_RARE, 0,
        3, 0x00FFFFFF, 0x00C4010F, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("VMXON"),
        D_PRIVILEGED|D_MUSTF3|D_MEMORY|D_RARE, 0,
        2, 0x0038FFFF, 0x0030C70F, 0x00,
        B_INT64,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("GETSEC"),
        D_PRIVILEGED|D_RARE, 0,
        2, 0x0000FFFF, 0x0000370F, 0x00,
        B_EAX|B_UPD|B_PSEUDO,
        B_EBX|B_PSEUDO,
        B_ECX|B_PSEUDO,
        B_NONE },

  { T("FEMMS"),
        D_CMD, 0,
        2, 0x0000FFFF, 0x00000E0F, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("PAVGUSB"),
        D_MMX|D_POSTBYTE|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x00000F0F, 0xBF,
        B_MREG8x8|B_UPD,
        B_MMX8x8,
        B_NONE,
        B_NONE },

  { T("PF2ID"),
        D_3DNOW|D_POSTBYTE, 0,
        2, 0x0000FFFF, 0x00000F0F, 0x1D,
        B_MREG32x2|B_CHG,
        B_3DNOW,
        B_NONE,
        B_NONE },

  { T("PFACC"),
        D_3DNOW|D_POSTBYTE, 0,
        2, 0x0000FFFF, 0x00000F0F, 0xAE,
        B_3DREG|B_UPD,
        B_3DNOW,
        B_NONE,
        B_NONE },

  { T("PFADD"),
        D_3DNOW|D_POSTBYTE, 0,
        2, 0x0000FFFF, 0x00000F0F, 0x9E,
        B_3DREG|B_UPD,
        B_3DNOW,
        B_NONE,
        B_NONE },

  { T("PFCMPEQ"),
        D_3DNOW|D_POSTBYTE, 0,
        2, 0x0000FFFF, 0x00000F0F, 0xB0,
        B_3DREG|B_UPD,
        B_3DNOW,
        B_NONE,
        B_NONE },

  { T("PFCMPGE"),
        D_3DNOW|D_POSTBYTE, 0,
        2, 0x0000FFFF, 0x00000F0F, 0x90,
        B_3DREG|B_UPD,
        B_3DNOW,
        B_NONE,
        B_NONE },

  { T("PFCMPGT"),
        D_3DNOW|D_POSTBYTE, 0,
        2, 0x0000FFFF, 0x00000F0F, 0xA0,
        B_3DREG|B_UPD,
        B_3DNOW,
        B_NONE,
        B_NONE },

  { T("PFMAX"),
        D_3DNOW|D_POSTBYTE, 0,
        2, 0x0000FFFF, 0x00000F0F, 0xA4,
        B_3DREG|B_UPD,
        B_3DNOW,
        B_NONE,
        B_NONE },

  { T("PFMIN"),
        D_3DNOW|D_POSTBYTE, 0,
        2, 0x0000FFFF, 0x00000F0F, 0x94,
        B_3DREG|B_UPD,
        B_3DNOW,
        B_NONE,
        B_NONE },

  { T("PFMUL"),
        D_3DNOW|D_POSTBYTE, 0,
        2, 0x0000FFFF, 0x00000F0F, 0xB4,
        B_3DREG|B_UPD,
        B_3DNOW,
        B_NONE,
        B_NONE },

  { T("PFRCP"),
        D_3DNOW|D_POSTBYTE, 0,
        2, 0x0000FFFF, 0x00000F0F, 0x96,
        B_3DREG|B_UPD,
        B_3DNOW,
        B_NONE,
        B_NONE },

  { T("PFRCPIT1"),
        D_3DNOW|D_POSTBYTE, 0,
        2, 0x0000FFFF, 0x00000F0F, 0xA6,
        B_3DREG|B_UPD,
        B_3DNOW,
        B_NONE,
        B_NONE },

  { T("PFRCPIT2"),
        D_3DNOW|D_POSTBYTE, 0,
        2, 0x0000FFFF, 0x00000F0F, 0xB6,
        B_3DREG|B_UPD,
        B_3DNOW,
        B_NONE,
        B_NONE },

  { T("PFRSQIT1"),
        D_3DNOW|D_POSTBYTE, 0,
        2, 0x0000FFFF, 0x00000F0F, 0xA7,
        B_3DREG|B_UPD,
        B_3DNOW,
        B_NONE,
        B_NONE },

  { T("PFRSQRT"),
        D_3DNOW|D_POSTBYTE, 0,
        2, 0x0000FFFF, 0x00000F0F, 0x97,
        B_3DREG|B_UPD,
        B_3DNOW,
        B_NONE,
        B_NONE },

  { T("PFSUB"),
        D_3DNOW|D_POSTBYTE, 0,
        2, 0x0000FFFF, 0x00000F0F, 0x9A,
        B_3DREG|B_UPD,
        B_3DNOW,
        B_NONE,
        B_NONE },

  { T("PFSUBR"),
        D_3DNOW|D_POSTBYTE, 0,
        2, 0x0000FFFF, 0x00000F0F, 0xAA,
        B_3DREG|B_UPD,
        B_3DNOW,
        B_NONE,
        B_NONE },

  { T("PI2FD"),
        D_3DNOW|D_POSTBYTE, 0,
        2, 0x0000FFFF, 0x00000F0F, 0x0D,
        B_3DREG|B_UPD,
        B_MMX32x2|B_SIGNED,
        B_NONE,
        B_NONE },

  { T("PMULHRW"),
        D_MMX|D_POSTBYTE|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x00000F0F, 0xB7,
        B_MREG16x4|B_UPD,
        B_MMX16x4,
        B_NONE,
        B_NONE },

  { T("PF2IW"),
        D_3DNOW|D_POSTBYTE, 0,
        2, 0x0000FFFF, 0x00000F0F, 0x1C,
        B_MREG32x2|B_UPD,
        B_3DNOW,
        B_NONE,
        B_NONE },

  { T("PFNACC"),
        D_3DNOW|D_POSTBYTE, 0,
        2, 0x0000FFFF, 0x00000F0F, 0x8A,
        B_3DREG|B_UPD,
        B_3DNOW,
        B_NONE,
        B_NONE },

  { T("PFPNACC"),
        D_3DNOW|D_POSTBYTE, 0,
        2, 0x0000FFFF, 0x00000F0F, 0x8E,
        B_3DREG|B_UPD,
        B_3DNOW,
        B_NONE,
        B_NONE },

  { T("PI2FW"),
        D_3DNOW|D_POSTBYTE, 0,
        2, 0x0000FFFF, 0x00000F0F, 0x0C,
        B_3DREG|B_UPD,
        B_MMX16x4,
        B_NONE,
        B_NONE },

  { T("PSWAPD"),
        D_MMX|D_POSTBYTE|D_MUSTNONE, 0,
        2, 0x0000FFFF, 0x00000F0F, 0xBB,
        B_MREG32x2|B_UPD,
        B_MMX32x2,
        B_NONE,
        B_NONE },

  { T("SYSCALL"),
        D_SYS|D_RARE, 0,
        2, 0x0000FFFF, 0x0000050F, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("SYSRET"),
        D_SYS|D_ALLFLAGS|D_SUSPICIOUS, 0,
        2, 0x0000FFFF, 0x0000070F, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("CLGI"),
        D_PRIVILEGED, 0,
        3, 0x00FFFFFF, 0x00DD010F, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("STGI"),
        D_PRIVILEGED, 0,
        3, 0x00FFFFFF, 0x00DC010F, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("INVLPGA"),
        D_PRIVILEGED|D_RARE, 0,
        3, 0x00FFFFFF, 0x00DF010F, 0x00,
        B_EAX|B_PSEUDO,
        B_ECX|B_PSEUDO,
        B_NONE,
        B_NONE },

  { T("SKINIT"),
        D_PRIVILEGED, 0,
        3, 0x00FFFFFF, 0x00DE010F, 0x00,
        B_EAX|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("VMLOAD"),
        D_PRIVILEGED, 0,
        3, 0x00FFFFFF, 0x00DA010F, 0x00,
        B_EAX|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("VMMCALL"),
        D_SYS|D_SUSPICIOUS, 0,
        3, 0x00FFFFFF, 0x00D9010F, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("VMRUN"),
        D_PRIVILEGED, 0,
        3, 0x00FFFFFF, 0x00D8010F, 0x00,
        B_EAX|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("VMSAVE"),
        D_PRIVILEGED, 0,
        3, 0x00FFFFFF, 0x00DB010F, 0x00,
        B_EAX|B_PSEUDO,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("ES:"),
        D_PREFIX|D_SUSPICIOUS, 0,
        1, 0x000000FF, 0x00000026, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("CS:"),
        D_PREFIX|D_SUSPICIOUS, 0,
        1, 0x000000FF, 0x0000002E, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("SS:"),
        D_PREFIX|D_SUSPICIOUS, 0,
        1, 0x000000FF, 0x00000036, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("DS:"),
        D_PREFIX|D_SUSPICIOUS, 0,
        1, 0x000000FF, 0x0000003E, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("FS:"),
        D_PREFIX|D_SUSPICIOUS, 0,
        1, 0x000000FF, 0x00000064, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("GS:"),
        D_PREFIX|D_SUSPICIOUS, 0,
        1, 0x000000FF, 0x00000065, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("DATASIZE:"),
        D_PREFIX|D_SUSPICIOUS, 0,
        1, 0x000000FF, 0x00000066, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("ADDRSIZE:"),
        D_PREFIX|D_SUSPICIOUS, 0,
        1, 0x000000FF, 0x00000067, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("LOCK"),
        D_PREFIX|D_SUSPICIOUS, 0,
        1, 0x000000FF, 0x000000F0, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("REPNE"),
        D_PREFIX|D_SUSPICIOUS, 0,
        1, 0x000000FF, 0x000000F2, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("REPNZ"),
        D_PREFIX|D_SUSPICIOUS, 0,
        1, 0x000000FF, 0x000000F2, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("REP"),
        D_PREFIX|D_SUSPICIOUS, 0,
        1, 0x000000FF, 0x000000F3, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("REPE"),
        D_PREFIX|D_SUSPICIOUS, 0,
        1, 0x000000FF, 0x000000F3, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("REPZ"),
        D_PREFIX|D_SUSPICIOUS, 0,
        1, 0x000000FF, 0x000000F3, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JCC"),
        D_PSEUDO|D_BHINT|D_COND, 0,
        1, 0x000000F0, 0x00000070, 0x00,
        B_BYTEOFFS|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("JCC"),
        D_PSEUDO|D_BHINT|D_COND, 0,
        2, 0x0000F0FF, 0x0000800F, 0x00,
        B_OFFSET|B_JMPCALL,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("SETCC"),
        D_PSEUDO|D_COND, 0,
        2, 0x0038F0FF, 0x0000900F, 0x00,
        B_INT8|B_SHOWSIZE|B_CHG,
        B_NONE,
        B_NONE,
        B_NONE },

  { T("CMOVCC"),
        D_PSEUDO|D_COND, 0,
        2, 0x0000F0FF, 0x0000400F, 0x00,
        B_REG|B_UPD,
        B_INT,
        B_NONE,
        B_NONE },

  { T("FCMOVCC"),
        D_PSEUDO|D_COND, 0,
        2, 0x0000E0FF, 0x0000C0DA, 0x00,
        B_ST0|B_CHG,
        B_ST,
        B_NONE,
        B_NONE },

  { T("FCMOVCC"),
        D_PSEUDO|D_COND, 0,
        2, 0x0000E0FF, 0x0000C0DB, 0x00,
        B_ST0|B_CHG,
        B_ST,
        B_NONE,
        B_NONE },

  { NULL,
        0, 0,
        0, 0x00000000, 0x00000000, 0x00,
        B_NONE,
        B_NONE,
        B_NONE,
        B_NONE
  }

};


////////////////////////////////////////////////////////////////////////////////
//////////////////////////////// SYMBOLIC NAMES ////////////////////////////////

// 8-bit register names, sorted by 'natural' index (as understood by CPU, not
// in the alphabetical order as some 'programmers' prefer).
const tchar *regname8[NREG] = {
  T("AL"),       T("CL"),       T("DL"),       T("BL"),
  T("AH"),       T("CH"),       T("DH"),       T("BH")  };

// 16-bit register names.
const tchar *regname16[NREG] = {
  T("AX"),       T("CX"),       T("DX"),       T("BX"),
  T("SP"),       T("BP"),       T("SI"),       T("DI")  };

// 32-bit register names.
const tchar *regname32[NREG] = {
  T("EAX"),      T("ECX"),      T("EDX"),      T("EBX"),
  T("ESP"),      T("EBP"),      T("ESI"),      T("EDI") };

// Names of segment registers.
const tchar *segname[NREG] = {
  T("ES"),       T("CS"),       T("SS"),       T("DS"),
  T("FS"),       T("GS"),       T("SEG6:"),    T("SEG7:") };

// Names of FPU registers, classical form.
const tchar *fpulong[NREG] = {
  T("ST(0)"),    T("ST(1)"),    T("ST(2)"),    T("ST(3)"),
  T("ST(4)"),    T("ST(5)"),    T("ST(6)"),    T("ST(7)") };

// Names of FPU registers, short form.
const tchar *fpushort[NREG] = {
  T("ST0"),      T("ST1"),      T("ST2"),      T("ST3"),
  T("ST4"),      T("ST5"),      T("ST6"),      T("ST7") };

// Names of MMX/3DNow! registers.
const tchar *mmxname[NREG] = {
  T("MM0"),      T("MM1"),      T("MM2"),      T("MM3"),
  T("MM4"),      T("MM5"),      T("MM6"),      T("MM7") };

// Names of 128-bit SSE registers.
const tchar *sse128[NREG] = {
  T("XMM0"),     T("XMM1"),     T("XMM2"),     T("XMM3"),
  T("XMM4"),     T("XMM5"),     T("XMM6"),     T("XMM7") };

// Names of 256-bit SSE registers.
const tchar *sse256[NREG] = {
  T("YMM0"),     T("YMM1"),     T("YMM2"),     T("YMM3"),
  T("YMM4"),     T("YMM5"),     T("YMM6"),     T("YMM7") };

// Names of control registers.
const tchar *crname[NREG] = {
  T("CR0"),      T("CR1"),      T("CR2"),      T("CR3"),
  T("CR4"),      T("CR5"),      T("CR6"),      T("CR7") };

// Names of debug registers.
const tchar *drname[NREG] = {
  T("DR0"),      T("DR1"),      T("DR2"),      T("DR3"),
  T("DR4"),      T("DR5"),      T("DR6"),      T("DR7") };

// Declarations for data types. Depending on ssesizemode, name of 16-byte data
// type (DQWORD) may be changed to XMMWORD and that of 32-bit type (QQWORD) to
// YMMWORD.
const tchar *sizename[33] = {
  NULL,          T("BYTE"),     T("WORD"),     NULL,
  T("DWORD"),    NULL,          T("FWORD"),    NULL,
  T("QWORD"),    NULL,          T("TBYTE"),    NULL,
  NULL,          NULL,          NULL,          NULL,
  T("DQWORD"),   NULL,          NULL,          NULL,
  NULL,          NULL,          NULL,          NULL,
  NULL,          NULL,          NULL,          NULL,
  NULL,          NULL,          NULL,          NULL,
  T("QQWORD") };

// Keywords for immediate data. HLA uses sizename[] instead of sizekey[].
const tchar *sizekey[33] = {
  NULL,          T("DB"),       T("DW"),       NULL,
  T("DD"),       NULL,          T("DF"),       NULL,
  T("DQ"),       NULL,          T("DT"),       NULL,
  NULL,          NULL,          NULL,          NULL,
  T("DDQ"),      NULL,          NULL,          NULL,
  NULL,          NULL,          NULL,          NULL,
  NULL,          NULL,          NULL,          NULL,
  NULL,          NULL,          NULL,          NULL,
  T("DQQ") };

// Keywords for immediate data in AT&T format.
const tchar *sizeatt[33] = {
  NULL,          T(".BYTE"),    T(".WORD"),    NULL,
  T(".LONG"),    NULL,          T(".FWORD"),   NULL,
  T(".QUAD"),    NULL,          T(".TBYTE"),   NULL,
  NULL,          NULL,          NULL,          NULL,
  T(".DQUAD"),   NULL,          NULL,          NULL,
  NULL,          NULL,          NULL,          NULL,
  NULL,          NULL,          NULL,          NULL,
  NULL,          NULL,          NULL,          NULL,
  T(".QQUAD") };

// Comparison predicates in SSE [0..7] and VEX commands [0..31].
const tchar *ssepredicate[32] = {
  T("EQ"),       T("LT"),       T("LE"),       T("UNORD"),
  T("NEQ"),      T("NLT"),      T("NLE"),      T("ORD"),
  T("EQ_UQ"),    T("NGE"),      T("NGT"),      T("FALSE"),
  T("NEQ_OQ"),   T("GE"),       T("GT"),       T("TRUE"),
  T("EQ_OS"),    T("LT_OQ"),    T("LE_OQ"),    T("UNORD_S"),
  T("NEQ_US"),   T("NLT_UQ"),   T("NLE_UQ"),   T("ORD_S"),
  T("EQ_US"),    T("NGE_UQ"),   T("NGT_UQ"),   T("FALSE_OS"),
  T("NEQ_OS"),   T("GE_OQ"),    T("GT_OQ"),    T("TRUE_US") };


////////////////////////////////////////////////////////////////////////////////
///////////////////////////////// DISASSEMBLER /////////////////////////////////

typedef struct t_imdata {              // Intermediate disassembler data
  t_disasm       *da;                  // Result of disassembly
  ulong          damode;               // Disassembling mode, set of DA_xxx
  t_config       *config;              // Disassembler configuration
  int            (*decodeaddress)(tchar *s,ulong addr);
  ulong          prefixlist;           // List of command's prefixes, PF_xxx
  int            ssesize;              // Size of SSE operands (16/32 bytes)
  ulong          immsize1;             // Size of first immediate constant
  ulong          immsize2;             // Size of second immediate constant
  ulong          mainsize;             // Size of command with prefixes
  ulong          modsize;              // Size of ModRegRM/SIB bytes
  ulong          dispsize;             // Size of address offset
  int            usesdatasize;         // May have data size prefix
  int            usesaddrsize;         // May have address size prefix
  int            usessegment;          // May have segment override prefix
} t_imdata;

static t_config  defconfig = {         // Default disassembler configuration
  DAMODE_MASM,                         // Main style, one of DAMODE_xxx
  NUM_STD|NUM_DECIMAL,                 // Constant part of address, NUM_xxx
  NUM_STD|NUM_LONG,                    // Jump/call destination, NUM_xxx
  NUM_STD|NUM_LONG,                    // Binary constants, NUM_xxx
  NUM_STD|NUM_DECIMAL,                 // Numeric constants, NUM_xxx
  0,                                   // Force lowercase display
  0,                                   // Tab between mnemonic and arguments
  0,                                   // Extra space between arguments
  0,                                   // Use RET instead of RETN
  1,                                   // Use short form of string commands
  0,                                   // Display default segments in listing
  1,                                   // Always show memory size
  0,                                   // Show NEAR modifiers
  1,                                   // How to decode size of SSE operands
  0,                                   // How to decode jump hints
  0,                                   // How to decode size-sensitive mnemonics
  0,                                   // How to decode top of FPU stack
  0                                    // Highlight operands
};

static t_config  attconfig = {         // AT&T disassembler configuration
  DAMODE_ATT,                          // Main style, one of DAMODE_xxx
  NUM_X|NUM_DECIMAL,                   // Constant part of address, NUM_xxx
  NUM_X|NUM_LONG,                      // Jump/call destination, NUM_xxx
  NUM_X|NUM_LONG,                      // Binary constants, NUM_xxx
  NUM_X|NUM_DECIMAL,                   // Numeric constants, NUM_xxx
  1,                                   // Force lowercase display
  1,                                   // Tab between mnemonic and arguments
  1,                                   // Extra space between arguments
  0,                                   // Use RET instead of RETN
  1,                                   // Use short form of string commands
  0,                                   // Display default segments in listing
  0,                                   // Always show memory size
  0,                                   // Show NEAR modifiers
  1,                                   // How to decode size of SSE operands
  0,                                   // How to decode jump hints
  0,                                   // How to decode size-sensitive mnemonics
  0,                                   // How to decode top of FPU stack
  0                                    // Highlight operands
};

////////////////////////////////////////////////////////////////////////////////
////////////////////////////// SERVICE FUNCTIONS ///////////////////////////////

static tchar hexcharu[16] = {          // Nibble-to-hexdigit table, uppercase
  T('0'), T('1'), T('2'), T('3'), T('4'), T('5'), T('6'), T('7'),
  T('8'), T('9'), T('A'), T('B'), T('C'), T('D'), T('E'), T('F') };

static tchar hexcharl[16] = {          // Nibble-to-hexdigit table, lowercase
  T('0'), T('1'), T('2'), T('3'), T('4'), T('5'), T('6'), T('7'),
  T('8'), T('9'), T('a'), T('b'), T('c'), T('d'), T('e'), T('f') };

static tchar cvtlower[256];

// Copies at most n-1 wide characters from src to dest and assures that dest is
// null-terminated. Slow but reliable. Returns number of copied characters, not
// including the terminal null. Attention, does not check that input parameters
// are correct!
static int Tstrcopy(tchar *dest,int n,const tchar *src) {
  int i;
  if (n<=0)
    return 0;
  for (i=0; i<n-1; i++) {
    if (*src==T('\0')) break;
    *dest++=*src++; };
  *dest=T('\0');
  return i;
};

// Copies at most n-1 wide characters from src to dest and assures that dest is
// null-terminated. If lowercase is 1, simultaneously converts it to lower
// case. Slow but reliable. Returns number of copied characters, not including
// the terminal null. Attention, does not check that input parameters are
// correct!
static int Tcopycase(tchar *dest,int n,const tchar *src,int lowercase) {
  int i;
  if (n<=0)
    return 0;
  for (i=0; i<n-1; i++) {
    if (*src==T('\0')) break;
    if (lowercase)
      *dest++=cvtlower[*src++];        // Much faster than call to tolower()
    else
      *dest++=*src++;
    ;
  };
  *dest=T('\0');
  return i;
};

// Dumps ncode bytes of code to the string s. Returns length of resulting text,
// characters, not including terminal zero. Attention, does not check that
// input parameters are correct or that s has sufficient length!
static int Thexdump(tchar *s,uchar *code,int ncode,int lowercase) {
  int d,n;
  static tchar *hexchar;
  hexchar=(lowercase?hexcharl:hexcharu);
  n=0;
  while (ncode>0) {
    d=*code++;
    s[n++]=hexchar[(d>>4) & 0x0F];
    s[n++]=hexchar[d & 0x0F];
    ncode--;
  };
  s[n]=T('\0');
  return n;
};

// Converts unsigned 1-, 2- or 4-byte number to hexadecimal text, according to
// the specified mode and type of argument. String s must be at least SHORTNAME
// characters long. Returns length of resulting text in characters, not
// including the terminal zero.
static int Hexprint(int size,tchar *s,ulong u,const t_imdata *im,ulong arg) {
  int i,k,ndigit,lastdigit = 0; //[bytepointer.com]
  ulong nummode,mod = 0; //[bytepointer.com]
  tchar buf[SHORTNAME];
  static tchar *hexchar;
  if (size==1)
    u&=0x000000FF;                     // 8-bit number
  else if (size==2)
    u&=0x0000FFFF;                     // 16-bit number
  else
    size=4;                            // Correct possible errors
  mod=arg & B_MODMASK;
  if (mod==B_ADDR)
    nummode=im->config->memmode;
  else if (mod==B_JMPCALL || mod==B_JMPCALLFAR)
    nummode=im->config->jmpmode;
  else if (mod==B_BINARY)
    nummode=im->config->binconstmode;
  else
    nummode=im->config->constmode;
  hexchar=(im->config->lowercase?hexcharl:hexcharu);
  buf[SHORTNAME-1]=T('\0');
  k=SHORTNAME-1;
  if ((nummode & NUM_DECIMAL)!=0 && (mod==B_SIGNED || mod==B_UNSIGNED ||
    (u<DECLIMIT && mod!=B_BINARY && mod!=B_JMPCALL && mod!=B_JMPCALLFAR))
  ) {
    // Decode as decimal unsigned number.
    if ((nummode & NUM_STYLE)==NUM_OLLY && u>=10)
      buf[--k]=T('.');                 // Period marks decimals in OllyDbg
    do {
      buf[--k]=hexchar[u%10];
      u/=10;
    } while (u!=0); }
  else {
    // Decode as hexadecimal number.
    if (nummode & NUM_LONG)            // 2, 4 or 8 significant digits
      ndigit=size*2;
    else
      ndigit=1;
    if ((nummode & NUM_STYLE)==NUM_STD)
      buf[--k]=T('h');
    for (i=0; i<ndigit || u!=0; i++) {
      lastdigit=u & 0x0F;
      buf[--k]=hexchar[lastdigit];
      u=(u>>4) & 0x0FFFFFFF; };
    if ((nummode & NUM_STYLE)==NUM_X) {
      buf[--k]=T('x');
      buf[--k]=T('0'); }
    else if (lastdigit>=10 &&
      ((nummode & NUM_STYLE)!=NUM_OLLY || i<(mod==B_BINARY?size*2:8)))
      buf[--k]=T('0');
    ;
  };
  return Tstrcopy(s,SHORTNAME,buf+k);
};


////////////////////////////////////////////////////////////////////////////////
///////////////////////// INTERNAL DISASSEMBLER TABLES /////////////////////////

t_chain          *cmdchain;            // Commands sorted by first CMDMASK bits
t_modrm          modrm16[256];         // 16-bit ModRM decodings
t_modrm          modrm32[256];         // 32-bit ModRM decodings without SIB
t_modrm          sib0[256];            // ModRM-SIB decodings with Mod=00
t_modrm          sib1[256];            // ModRM-SIB decodings with Mod=01
t_modrm          sib2[256];            // ModRM-SIB decodings with Mod=10

// Initializes disassembler tables. Call this function once during startup.
// Returns 0 on success and -1 if initialization was unsuccessful. In the last
// case, continuation is not possible and program must terminate.
int Preparedisasm(void) {
  int n,c,reg,sreg,scale,nchain;
  ulong u,code,mask;
  const t_bincmd *pcmd;
  t_chain *pchain;
  t_modrm *pmrm,*psib;
  if (cmdchain!=NULL)
    return 0;                          // Already initialized
  // Sort command descriptors into command chains by first CMDMASK bits.
  cmdchain=(t_chain *)malloc(NCHAIN*sizeof(t_chain));
  if (cmdchain==NULL)                  // Low memory
    return -1;
  memset(cmdchain,0,NCHAIN*sizeof(t_chain));
  nchain=CMDMASK+1;                    // Number of command chains
  for (pcmd=bincmd; pcmd->length!=0; pcmd++) {
    if ((pcmd->cmdtype & D_CMDTYPE)==D_PSEUDO)
      continue;                        // Pseudocommand, for search models only
    code=pcmd->code;
    mask=pcmd->mask & CMDMASK;
    for (u=0; u<CMDMASK+1; u++) {
      if (((u ^ code) & mask)!=0)
        continue;                      // Command has different first bytes
      pchain=cmdchain+u;
      while (pchain->pcmd!=NULL && pchain->pnext!=NULL)
        pchain=pchain->pnext;          // Walk chain to the end
      if (pchain->pcmd==NULL)
        pchain->pcmd=pcmd;
      else if (nchain>=NCHAIN)
        return -1;                     // Too many commands
      else {
        pchain->pnext=cmdchain+nchain; // Prolongate chain
        pchain=pchain->pnext;
        pchain->pcmd=pcmd;
        nchain++;
      };
    };
  };
  // Prepare 16-bit ModRM decodings.
  memset(modrm16,0,sizeof(modrm16));
  for (c=0x00,pmrm=modrm16; c<=0xFF; c++,pmrm++) {
    reg=c & 0x07;
    if ((c & 0xC0)==0xC0) {
      // Register in ModRM.
      pmrm->size=1;
      pmrm->features=0;                // Register, its type as yet unknown
      pmrm->reg=reg;
      pmrm->defseg=SEG_UNDEF;
      pmrm->basereg=REG_UNDEF; }
    else if ((c & 0xC7)==0x06) {
      // Special case of immediate address.
      pmrm->size=3;
      pmrm->dispsize=2;
      pmrm->features=OP_MEMORY|OP_OPCONST|OP_ADDR16;
      pmrm->reg=REG_UNDEF;
      pmrm->defseg=SEG_DS;
      pmrm->basereg=REG_UNDEF; }
    else {
      pmrm->features=OP_MEMORY|OP_INDEXED|OP_ADDR16;
      if ((c & 0xC0)==0x40) {
        pmrm->dispsize=1; pmrm->features|=OP_OPCONST; }
      else if ((c & 0xC0)==0x80) {
        pmrm->dispsize=2; pmrm->features|=OP_OPCONST; };
      pmrm->size=pmrm->dispsize+1;
      pmrm->reg=REG_UNDEF;
      switch (reg) {
        case 0:
          pmrm->scale[REG_EBX]=1; pmrm->scale[REG_ESI]=1;
          pmrm->defseg=SEG_DS;
          tstrcpy(pmrm->ardec,T("BX+SI"));
          tstrcpy(pmrm->aratt,T("%BX,%SI"));
          pmrm->aregs=(1<<REG_EBX)|(1<<REG_ESI);
          pmrm->basereg=REG_ESI; break;
        case 1:
          pmrm->scale[REG_EBX]=1; pmrm->scale[REG_EDI]=1;
          pmrm->defseg=SEG_DS;
          tstrcpy(pmrm->ardec,T("BX+DI"));
          tstrcpy(pmrm->aratt,T("%BX,%DI"));
          pmrm->aregs=(1<<REG_EBX)|(1<<REG_EDI);
          pmrm->basereg=REG_EDI; break;
        case 2:
          pmrm->scale[REG_EBP]=1; pmrm->scale[REG_ESI]=1;
          pmrm->defseg=SEG_SS;
          tstrcpy(pmrm->ardec,T("BP+SI"));
          tstrcpy(pmrm->aratt,T("%BP,%SI"));
          pmrm->aregs=(1<<REG_EBP)|(1<<REG_ESI);
          pmrm->basereg=REG_ESI; break;
        case 3:
          pmrm->scale[REG_EBP]=1; pmrm->scale[REG_EDI]=1;
          pmrm->defseg=SEG_SS;
          tstrcpy(pmrm->ardec,T("BP+DI"));
          tstrcpy(pmrm->aratt,T("%BP,%DI"));
          pmrm->aregs=(1<<REG_EBP)|(1<<REG_EDI);
          pmrm->basereg=REG_EDI; break;
        case 4:
          pmrm->scale[REG_ESI]=1;
          pmrm->defseg=SEG_DS;
          tstrcpy(pmrm->ardec,T("SI"));
          tstrcpy(pmrm->aratt,T("%SI"));
          pmrm->aregs=(1<<REG_ESI);
          pmrm->basereg=REG_ESI; break;
        case 5:
          pmrm->scale[REG_EDI]=1;
          pmrm->defseg=SEG_DS;
          tstrcpy(pmrm->ardec,T("DI"));
          tstrcpy(pmrm->aratt,T("%DI"));
          pmrm->aregs=(1<<REG_EDI);
          pmrm->basereg=REG_EDI; break;
        case 6:
          pmrm->scale[REG_EBP]=1;
          pmrm->defseg=SEG_SS;
          tstrcpy(pmrm->ardec,T("BP"));
          tstrcpy(pmrm->aratt,T("%BP"));
          pmrm->aregs=(1<<REG_EBP);
          pmrm->basereg=REG_EBP; break;
        case 7:
          pmrm->scale[REG_EBX]=1;
          pmrm->defseg=SEG_DS;
          tstrcpy(pmrm->ardec,T("BX"));
          tstrcpy(pmrm->aratt,T("%BX"));
          pmrm->aregs=(1<<REG_EBX);
          pmrm->basereg=REG_EBX;
        break;
      };
    };
  };
  // Prepare 32-bit ModRM decodings without SIB.
  memset(modrm32,0,sizeof(modrm32));
  for (c=0x00,pmrm=modrm32; c<=0xFF; c++,pmrm++) {
    reg=c & 0x07;
    if ((c & 0xC0)==0xC0) {
      // Register in ModRM.
      pmrm->size=1;
      pmrm->features=0;                // Register, its type as yet unknown
      pmrm->reg=reg;
      pmrm->defseg=SEG_UNDEF;
      pmrm->basereg=REG_UNDEF; }
    else if ((c & 0xC7)==0x05) {
      // Special case of 32-bit immediate address.
      pmrm->size=5;
      pmrm->dispsize=4;
      pmrm->features=OP_MEMORY|OP_OPCONST;
      pmrm->reg=REG_UNDEF;
      pmrm->defseg=SEG_DS;
      pmrm->basereg=REG_UNDEF; }
    else {
      // Regular memory address.
      pmrm->features=OP_MEMORY;
      pmrm->reg=REG_UNDEF;
      if ((c & 0xC0)==0x40) {
        pmrm->dispsize=1;              // 8-bit sign-extended displacement
        pmrm->features|=OP_OPCONST; }
      else if ((c & 0xC0)==0x80) {
        pmrm->dispsize=4;              // 32-bit displacement
        pmrm->features|=OP_OPCONST; };
      if (reg==REG_ESP) {
        // SIB byte follows, decode with sib32.
        if ((c & 0xC0)==0x00) pmrm->psib=sib0;
        else if ((c & 0xC0)==0x40) pmrm->psib=sib1;
        else pmrm->psib=sib2;
        pmrm->basereg=REG_UNDEF; }
      else {
        pmrm->size=1+pmrm->dispsize;
        pmrm->features|=OP_INDEXED;
        pmrm->defseg=(reg==REG_EBP?SEG_SS:SEG_DS);
        pmrm->scale[reg]=1;
        tstrcpy(pmrm->ardec,regname32[reg]);
        pmrm->aratt[0]=T('%');
        Tstrcopy(pmrm->aratt+1,SHORTNAME-1,regname32[reg]);
        pmrm->aregs=(1<<reg);
        pmrm->basereg=reg;
      };
    };
  };
  // Prepare 32-bit ModRM decodings with SIB, case Mod=00: usually no disp.
  memset(sib0,0,sizeof(sib0));
  for (c=0x00,psib=sib0; c<=0xFF; c++,psib++) {
    psib->features=OP_MEMORY;
    psib->reg=REG_UNDEF;
    reg=c & 0x07;
    sreg=(c>>3) & 0x07;
    if ((c & 0xC0)==0) scale=1;
    else if ((c & 0xC0)==0x40) scale=2;
    else if ((c & 0xC0)==0x80) scale=4;
    else scale=8;
    if (sreg!=REG_ESP) {
      psib->scale[sreg]=(uchar)scale;
      n=Tstrcopy(psib->ardec,SHORTNAME,regname32[sreg]);
      psib->aregs=(1<<sreg);
      psib->features|=OP_INDEXED;
      if (scale>1) {
        psib->ardec[n++]=T('*');
        psib->ardec[n++]=(tchar)(T('0')+scale);
        psib->ardec[n]=T('\0');
      }; }
    else
      n=0;
    if (reg==REG_EBP) {
      psib->size=6;
      psib->dispsize=4;
      psib->features|=OP_OPCONST;
      psib->defseg=SEG_DS;
      psib->basereg=REG_UNDEF; }
    else {
      psib->size=2;
      psib->defseg=((reg==REG_ESP || reg==REG_EBP)?SEG_SS:SEG_DS);
      psib->scale[reg]++;
      psib->features|=OP_INDEXED;
      if (n!=0) psib->ardec[n++]=T('+');
      Tstrcopy(psib->ardec+n,SHORTNAME-n,regname32[reg]);
      psib->aregs|=(1<<reg);
      psib->basereg=reg; };
    if (reg!=REG_EBP) {
      psib->aratt[0]=T('%'); n=1;
      n+=Tstrcopy(psib->aratt+n,SHORTNAME-n,regname32[reg]); }
    else
      n=0;
    if (sreg!=REG_ESP) {
      n+=Tstrcopy(psib->aratt+n,SHORTNAME-n,T(",%"));
      n+=Tstrcopy(psib->aratt+n,SHORTNAME-n,regname32[sreg]);
      if (scale>1) {
        psib->aratt[n++]=T(',');
        psib->aratt[n++]=(tchar)(T('0')+scale);
        psib->aratt[n]=T('\0');
      };
    };
  };
  // Prepare 32-bit ModRM decodings with SIB, case Mod=01: 8-bit displacement.
  memset(sib1,0,sizeof(sib1));
  for (c=0x00,psib=sib1; c<=0xFF; c++,psib++) {
    psib->features=OP_MEMORY|OP_INDEXED|OP_OPCONST;
    psib->reg=REG_UNDEF;
    reg=c & 0x07;
    sreg=(c>>3) & 0x07;
    if ((c & 0xC0)==0) scale=1;
    else if ((c & 0xC0)==0x40) scale=2;
    else if ((c & 0xC0)==0x80) scale=4;
    else scale=8;
    psib->size=3;
    psib->dispsize=1;
    psib->defseg=((reg==REG_ESP || reg==REG_EBP)?SEG_SS:SEG_DS);
    psib->scale[reg]=1;
    psib->basereg=reg;
    psib->aregs=(1<<reg);
    if (sreg!=REG_ESP) {
      psib->scale[sreg]+=(uchar)scale;
      n=Tstrcopy(psib->ardec,SHORTNAME,regname32[sreg]);
      psib->aregs|=(1<<sreg);
      if (scale>1) {
        psib->ardec[n++]=T('*');
        psib->ardec[n++]=(tchar)(T('0')+scale);
      }; }
    else
      n=0;
    if (n!=0) psib->ardec[n++]=T('+');
    Tstrcopy(psib->ardec+n,SHORTNAME-n,regname32[reg]);
    psib->aratt[0]=T('%'); n=1;
    n+=Tstrcopy(psib->aratt+n,SHORTNAME-n,regname32[reg]);
    if (sreg!=REG_ESP) {
      n+=Tstrcopy(psib->aratt+n,SHORTNAME-n,T(",%"));
      n+=Tstrcopy(psib->aratt+n,SHORTNAME-n,regname32[sreg]);
      if (scale>1) {
        psib->aratt[n++]=T(',');
        psib->aratt[n++]=(tchar)(T('0')+scale);
        psib->aratt[n]=T('\0');
      };
    };
  };
  // Prepare 32-bit ModRM decodings with SIB, case Mod=10: 32-bit displacement.
  memset(sib2,0,sizeof(sib2));
  for (c=0x00,psib=sib2; c<=0xFF; c++,psib++) {
    psib->features=OP_MEMORY|OP_INDEXED|OP_OPCONST;
    psib->reg=REG_UNDEF;
    reg=c & 0x07;
    sreg=(c>>3) & 0x07;
    if ((c & 0xC0)==0) scale=1;
    else if ((c & 0xC0)==0x40) scale=2;
    else if ((c & 0xC0)==0x80) scale=4;
    else scale=8;
    psib->size=6;
    psib->dispsize=4;
    psib->defseg=((reg==REG_ESP || reg==REG_EBP)?SEG_SS:SEG_DS);
    psib->scale[reg]=1;
    psib->basereg=reg;
    psib->aregs=(1<<reg);
    if (sreg!=REG_ESP) {
      psib->scale[sreg]+=(uchar)scale;
      n=Tstrcopy(psib->ardec,SHORTNAME,regname32[sreg]);
      psib->aregs|=(1<<sreg);
      if (scale>1) {
        psib->ardec[n++]=T('*');
        psib->ardec[n++]=(tchar)(T('0')+scale);
      }; }
    else
      n=0;
    if (n!=0) psib->ardec[n++]=T('+');
    Tstrcopy(psib->ardec+n,SHORTNAME-n,regname32[reg]);
    psib->aratt[0]=T('%'); n=1;
    n+=Tstrcopy(psib->aratt+n,SHORTNAME-n,regname32[reg]);
    if (sreg!=REG_ESP) {
      n+=Tstrcopy(psib->aratt+n,SHORTNAME-n,T(",%"));
      n+=Tstrcopy(psib->aratt+n,SHORTNAME-n,regname32[sreg]);
      if (scale>1) {
        psib->aratt[n++]=T(',');
        psib->aratt[n++]=(tchar)(T('0')+scale);
        psib->aratt[n]=T('\0');
      };
    };
  };
  // Fill lowercase conversion table. This table replaces tolower(). When
  // compiled with Borland C++ Builder, spares significant time.
  for (c=0; c<256; c++)
    cvtlower[c]=(tchar)ttolower(c);
  // Report success.
  return 0;
};

// Frees resources allocated by Preparedisasm(). Call this function once
// during shutdown after disassembling service is no longer necessary.
void Finishdisasm(void) {
  if (cmdchain!=NULL) {
    free(cmdchain);
    cmdchain=NULL;
  };
};


////////////////////////////////////////////////////////////////////////////////
////////////////////////////// AUXILIARY ROUTINES //////////////////////////////

// Given index of byte register, returns index of 32-bit container.
static int Byteregtodwordreg(int bytereg) {
  if (bytereg<0 || bytereg>=NREG)
    return REG_UNDEF;
  if (bytereg>=4)
    return bytereg-4;
  return bytereg;
};

// Checks prefix override flags and generates warnings if prefix is superfluous.
// Returns index of segment register. Note that Disasm() assures that two
// segment override bits in im->prefixlist can't be set simultaneously.
static int Getsegment(t_imdata *im,int arg,int defseg) {
  if ((im->prefixlist & PF_SEGMASK)==0)
    return defseg;                     // Optimization for most frequent case
  switch (im->prefixlist & PF_SEGMASK) {
    case PF_ES:
      if (defseg==SEG_ES) im->da->warnings|=DAW_DEFSEG;
      if (arg & B_NOSEG) im->da->warnings|=DAW_SEGPREFIX;
      return SEG_ES;
    case PF_CS:
      if (defseg==SEG_CS) im->da->warnings|=DAW_DEFSEG;
      if (arg & B_NOSEG) im->da->warnings|=DAW_SEGPREFIX;
      return SEG_CS;
    case PF_SS:
      if (defseg==SEG_SS) im->da->warnings|=DAW_DEFSEG;
      if (arg & B_NOSEG) im->da->warnings|=DAW_SEGPREFIX;
      return SEG_SS;
    case PF_DS:
      if (defseg==SEG_DS) im->da->warnings|=DAW_DEFSEG;
      if (arg & B_NOSEG) im->da->warnings|=DAW_SEGPREFIX;
      return SEG_DS;
    case PF_FS:
      if (defseg==SEG_FS) im->da->warnings|=DAW_DEFSEG;
      if (arg & B_NOSEG) im->da->warnings|=DAW_SEGPREFIX;
      return SEG_FS;
    case PF_GS:
      if (defseg==SEG_GS) im->da->warnings|=DAW_DEFSEG;
      if (arg & B_NOSEG) im->da->warnings|=DAW_SEGPREFIX;
      return SEG_GS;
    default: return defseg;            // Most frequent case of default segment
  };
};

// Decodes generalized memory address to text.
static void Memaddrtotext(t_imdata *im,int arg,int datasize,int seg,
  const tchar *regpart,long constpart,tchar *s) {
  int n;
  tchar label[TEXTLEN];
  if (im->config->disasmmode==DAMODE_ATT) {
    // AT&T memory address syntax is so different from Intel that I process it
    // separately from the rest.
    n=0;
    if ((arg & B_MODMASK)==B_JMPCALL)
      s[n++]=T('*');
    // On request, I show only explicit segments.
    if ((im->config->putdefseg && (arg & B_NOSEG)==0) ||
      (im->prefixlist & PF_SEGMASK)!=0
    ) {
      s[n++]=T('%');
      n+=Tcopycase(s+n,TEXTLEN-n,segname[seg],im->config->lowercase);
      s[n++]=T(':'); };
    // Add constant part (offset).
    if (constpart<0 && constpart>NEGLIMIT) {
      s[n++]=T('-');
      n+=Hexprint((im->prefixlist & PF_ASIZE?2:4),s+n,-constpart,im,B_ADDR); }
    else if (constpart!=0) {
      if (seg!=SEG_FS && seg!=SEG_GS &&
        im->decodeaddress!=NULL &&
        im->decodeaddress(label,constpart)!=0)
        n+=Tstrcopy(s+n,TEXTLEN-n,label);
      else
        n+=Hexprint((im->prefixlist & PF_ASIZE?2:4),s+n,constpart,im,B_ADDR);
      ;
    };
    // Add register part of address, may be absent.
    if (regpart[0]!=T('\0')) {
      n+=Tstrcopy(s+n,TEXTLEN-n,T("("));
      n+=Tcopycase(s+n,TEXTLEN-n,regpart,im->config->lowercase);
      n+=Tstrcopy(s+n,TEXTLEN-n,T(")"));
    }; }
  else {
    // Mark far and near jump/call addresses.
    if ((arg & B_MODMASK)==B_JMPCALLFAR)
      n=Tcopycase(s,TEXTLEN,T("FAR "),im->config->lowercase);
    else if (im->config->shownear && (arg & B_MODMASK)==B_JMPCALL)
      n=Tcopycase(s,TEXTLEN,T("NEAR "),im->config->lowercase);
    else
      n=0;
    if (im->config->disasmmode!=DAMODE_MASM) {
      s[n++]=T('[');
      if ((im->prefixlist & PF_ASIZE)!=0 && regpart[0]==T('\0'))
        n+=Tcopycase(s+n,TEXTLEN-n,T("SMALL "),im->config->lowercase);
      ;
    };
    // If operand is longer than 32 bytes or of type B_ANYMEM (memory contents
    // unimportant), its size is not displayed. Otherwise, bit B_SHOWSIZE
    // indicates that explicit operand's size can't be omitted.
    if (datasize<=32 && (arg & B_ARGMASK)!=B_ANYMEM &&
      (im->config->showmemsize!=0 || (arg & B_SHOWSIZE)!=0)
    ) {
      if (im->config->disasmmode==DAMODE_HLA)
        n+=Tcopycase(s+n,TEXTLEN-n,T("TYPE "),im->config->lowercase);
      if ((arg & B_ARGMASK)==B_INTPAIR && im->config->disasmmode==DAMODE_IDEAL){
        // If operand is a pair of integers (BOUND), Borland in IDEAL mode
        // expects size of single integer, whereas MASM requires size of the
        // whole pair.
        n+=Tcopycase(s+n,TEXTLEN-n,sizename[datasize/2],im->config->lowercase);
        s[n++]=T(' '); }
      else if (datasize==16 && im->config->ssesizemode==1)
        n+=Tcopycase(s+n,TEXTLEN-n,T("XMMWORD "),im->config->lowercase);
      else if (datasize==32 && im->config->ssesizemode==1)
        n+=Tcopycase(s+n,TEXTLEN-n,T("YMMWORD "),im->config->lowercase);
      else {
        n+=Tcopycase(s+n,TEXTLEN-n,sizename[datasize],im->config->lowercase);
        s[n++]=T(' '); };
      if (im->config->disasmmode==DAMODE_MASM)
        n+=Tcopycase(s+n,TEXTLEN-n,T("PTR "),im->config->lowercase);
      ;
    };
    // On request, I show only explicit segments.
    if ((im->config->putdefseg && (arg & B_NOSEG)==0) ||
      (im->prefixlist & PF_SEGMASK)!=0
    ) {
      n+=Tcopycase(s+n,TEXTLEN-n,segname[seg],im->config->lowercase);
      s[n++]=T(':'); };
    if (im->config->disasmmode==DAMODE_MASM) {
      s[n++]=T('[');
      if ((im->prefixlist & PF_ASIZE)!=0 && regpart[0]==T('\0'))
        n+=Tcopycase(s+n,TEXTLEN-n,T("SMALL "),im->config->lowercase);
      ;
    };
    // Add register part of address, may be absent.
    if (regpart[0]!=T('\0'))
      n+=Tcopycase(s+n,TEXTLEN-n,regpart,im->config->lowercase);
    if (regpart[0]!=T('\0') && constpart<0 && constpart>NEGLIMIT) {
      s[n++]=T('-');
      n+=Hexprint((im->prefixlist & PF_ASIZE?2:4),s+n,-constpart,im,B_ADDR); }
    else if (constpart!=0 || regpart[0]==T('\0')) {
      if (regpart[0]!=T('\0')) s[n++]=T('+');
      if (seg!=SEG_FS && seg!=SEG_GS &&
        im->decodeaddress!=NULL &&
        im->decodeaddress(label,constpart)!=0)
        n+=Tstrcopy(s+n,TEXTLEN-n,label);
      else
        n+=Hexprint((im->prefixlist & PF_ASIZE?2:4),s+n,constpart,im,B_ADDR);
      ;
    };
    n+=Tstrcopy(s+n,TEXTLEN-n,T("]"));
  };
  s[n]=T('\0');
};

// Service function, returns granularity of MMX, 3DNow! and SSE operands.
static int Getgranularity(ulong arg) {
  int granularity;
  switch (arg & B_ARGMASK) {
    case B_MREG8x8:                    // MMX register as 8 8-bit integers
    case B_MMX8x8:                     // MMX reg/memory as 8 8-bit integers
    case B_MMX8x8DI:                   // MMX 8 8-bit integers at [DS:(E)DI]
    case B_XMM0I8x16:                  // XMM0 as 16 8-bit integers
    case B_SREGI8x16:                  // SSE register as 16 8-bit sigints
    case B_SVEXI8x16:                  // SSE reg in VEX as 16 8-bit sigints
    case B_SIMMI8x16:                  // SSE reg in immediate 8-bit constant
    case B_SSEI8x16:                   // SSE reg/memory as 16 8-bit sigints
    case B_SSEI8x16DI:                 // SSE 16 8-bit sigints at [DS:(E)DI]
    case B_SSEI8x8L:                   // Low 8 8-bit ints in SSE reg/memory
    case B_SSEI8x4L:                   // Low 4 8-bit ints in SSE reg/memory
    case B_SSEI8x2L:                   // Low 2 8-bit ints in SSE reg/memory
      granularity=1; break;
    case B_MREG16x4:                   // MMX register as 4 16-bit integers
    case B_MMX16x4:                    // MMX reg/memory as 4 16-bit integers
    case B_SREGI16x8:                  // SSE register as 8 16-bit sigints
    case B_SVEXI16x8:                  // SSE reg in VEX as 8 16-bit sigints
    case B_SSEI16x8:                   // SSE reg/memory as 8 16-bit sigints
    case B_SSEI16x4L:                  // Low 4 16-bit ints in SSE reg/memory
    case B_SSEI16x2L:                  // Low 2 16-bit ints in SSE reg/memory
      granularity=2; break;
    case B_MREG32x2:                   // MMX register as 2 32-bit integers
    case B_MMX32x2:                    // MMX reg/memory as 2 32-bit integers
    case B_3DREG:                      // 3DNow! register as 2 32-bit floats
    case B_3DNOW:                      // 3DNow! reg/memory as 2 32-bit floats
    case B_SREGF32x4:                  // SSE register as 4 32-bit floats
    case B_SVEXF32x4:                  // SSE reg in VEX as 4 32-bit floats
    case B_SREGF32L:                   // Low 32-bit float in SSE register
    case B_SVEXF32L:                   // Low 32-bit float in SSE in VEX
    case B_SREGF32x2L:                 // Low 2 32-bit floats in SSE register
    case B_SSEF32x4:                   // SSE reg/memory as 4 32-bit floats
    case B_SSEF32L:                    // Low 32-bit float in SSE reg/memory
    case B_SSEF32x2L:                  // Low 2 32-bit floats in SSE reg/memory
      granularity=4; break;
    case B_XMM0I32x4:                  // XMM0 as 4 32-bit integers
    case B_SREGI32x4:                  // SSE register as 4 32-bit sigints
    case B_SVEXI32x4:                  // SSE reg in VEX as 4 32-bit sigints
    case B_SREGI32L:                   // Low 32-bit sigint in SSE register
    case B_SREGI32x2L:                 // Low 2 32-bit sigints in SSE register
    case B_SSEI32x4:                   // SSE reg/memory as 4 32-bit sigints
    case B_SSEI32x2L:                  // Low 2 32-bit sigints in SSE reg/memory
      granularity=4; break;
    case B_MREG64:                     // MMX register as 1 64-bit integer
    case B_MMX64:                      // MMX reg/memory as 1 64-bit integer
    case B_XMM0I64x2:                  // XMM0 as 2 64-bit integers
    case B_SREGF64x2:                  // SSE register as 2 64-bit floats
    case B_SVEXF64x2:                  // SSE reg in VEX as 2 64-bit floats
    case B_SREGF64L:                   // Low 64-bit float in SSE register
    case B_SVEXF64L:                   // Low 64-bit float in SSE in VEX
    case B_SSEF64x2:                   // SSE reg/memory as 2 64-bit floats
    case B_SSEF64L:                    // Low 64-bit float in SSE reg/memory
      granularity=8; break;
    case B_SREGI64x2:                  // SSE register as 2 64-bit sigints
    case B_SVEXI64x2:                  // SSE reg in VEX as 2 64-bit sigints
    case B_SSEI64x2:                   // SSE reg/memory as 2 64-bit sigints
    case B_SREGI64L:                   // Low 64-bit sigint in SSE register
      granularity=8; break;
    default:
      granularity=1;                   // Treat unknown ops as string of bytes
    break; };
  return granularity;
};


////////////////////////////////////////////////////////////////////////////////
////////////////////////// OPERAND DECODING ROUTINES ///////////////////////////

// Decodes 8/16/32-bit integer register operand. ATTENTION, calling routine
// must set usesdatasize and usesaddrsize by itself!
static void Operandintreg(t_imdata *im,ulong datasize,int index,t_operand *op) {
  int n,reg32;
  op->features=OP_REGISTER;
  op->opsize=op->granularity=datasize;
  op->reg=index;
  op->seg=SEG_UNDEF;
  // Add container register to lists of used and modified registers.
  if (datasize==1)
    reg32=Byteregtodwordreg(index);
  else
    reg32=index;
  if ((op->arg & B_CHG)==0) {
    op->uses=(1<<reg32);
    im->da->uses|=(1<<reg32); };
  if (op->arg & (B_CHG|B_UPD)) {
    op->modifies=(1<<reg32);
    im->da->modifies|=(1<<reg32); };
  // Warn if ESP is misused.
  if ((op->arg & B_NOESP)!=0 && reg32==REG_ESP)
    im->da->warnings|=DAW_NOESP;
  // Decode name of integer register.
  if (im->damode & DA_TEXT) {
    n=0;
    if (im->config->disasmmode==DAMODE_ATT) {
      if ((op->arg & B_MODMASK)==B_JMPCALL) op->text[n++]=T('*');
      op->text[n++]=T('%'); };
    if (datasize==4)                   // Most frequent case first
      Tcopycase(op->text+n,TEXTLEN-n,regname32[index],im->config->lowercase);
    else if (datasize==1)
      Tcopycase(op->text+n,TEXTLEN-n,regname8[index],im->config->lowercase);
    else                               // 16-bit registers are seldom
      Tcopycase(op->text+n,TEXTLEN-n,regname16[index],im->config->lowercase);
    ;
  };
};

// Decodes 16/32-bit memory address in ModRM/SIB bytes. Returns full length of
// address (ModRM+SIB+displacement) in bytes, 0 if ModRM indicates register
// operand and -1 on error. ATTENTION, calling routine must set usesdatasize,
// granularity (preset to datasize) and reg together with OP_MODREG by itself!
static int Operandmodrm(t_imdata *im,ulong datasize,uchar *cmd,ulong cmdsize,
  t_operand *op) {
  tchar *ardec;
  t_modrm *pmrm;
  if (cmdsize==0) {
    im->da->errors|=DAE_CROSS;         // Command crosses end of memory block
    return -1; };
  // Decode ModRM/SIB. Most of the work is already done in Preparedisasm(), we
  // only need to find corresponding t_modrm.
  if (im->prefixlist & PF_ASIZE) {
    pmrm=modrm16+cmd[0];               // 16-bit address
    im->modsize=1; }
  else {
    pmrm=modrm32+cmd[0];
    if (pmrm->psib==NULL)
      im->modsize=1;                   // No SIB byte
    else {
      if (cmdsize<2) {
        im->da->errors|=DAE_CROSS;     // Command crosses end of memory block
        return -1; };
      pmrm=pmrm->psib+cmd[1];
      im->modsize=2;                   // Both ModRM and SIB
    };
  };
  // Check whether ModRM indicates register operand and immediately return if
  // true. As a side effect, modsize is already set.
  if ((cmd[0] & 0xC0)==0xC0)
    return 0;
  // Operand in memory.
  op->opsize=datasize;
  op->granularity=datasize;            // Default, may be overriden later
  op->reg=REG_UNDEF;
  im->usesaddrsize=1;                  // Address size prefix is meaningful
  im->usessegment=1;                   // Segment override prefix is meaningful
  // Fetch precalculated t_modrm fields.
  op->features=pmrm->features;
  memcpy(op->scale,pmrm->scale,8);
  op->aregs=pmrm->aregs;
  im->da->uses|=pmrm->aregs;           // Mark registers used to form address
  // Get displacement, if any.
  im->dispsize=pmrm->dispsize;
  if (pmrm->dispsize!=0) {
    if (cmdsize<pmrm->size) {
      im->da->errors|=DAE_CROSS;       // Command crosses end of memory block
      return -1; };
    if (pmrm->dispsize==1)             // 8-bit displacement is sign-extended
      op->opconst=im->da->memconst=(signed char)cmd[im->modsize];
    else if (pmrm->dispsize==4) {      // 32-bit full displacement
      im->da->memfixup=
        im->mainsize+im->modsize;      // Possible 32-bit fixup
      op->opconst=im->da->memconst=*(ulong *)(cmd+im->modsize); }
    else                               // 16-bit displacement, very rare
      op->opconst=im->da->memconst=*(ushort *)(cmd+im->modsize);
    ;
  };
  // Get segment.
  op->seg=Getsegment(im,op->arg,pmrm->defseg);
  // Warn if memory contents is 16-bit jump/call destination.
  if (datasize==2 && (op->arg & B_MODMASK)==B_JMPCALL)
    im->da->warnings|=DAW_JMP16;
  // Decode memory operand to text, if requested.
  if (im->damode & DA_TEXT) {
    ardec=(im->config->disasmmode==DAMODE_ATT?pmrm->aratt:pmrm->ardec);
    Memaddrtotext(im,op->arg,datasize,op->seg,ardec,op->opconst,op->text);
  };
  return pmrm->size;
};

// Decodes 16/32-bit immediate address (used only for 8/16/32-bit memory-
// accumulator moves). ATTENTION, calling routine must set usesdatasize by
// itself!
static void Operandimmaddr(t_imdata *im,ulong datasize,uchar *cmd,
  ulong cmdsize,t_operand *op) {
  if (im->prefixlist & PF_ASIZE)
    im->dispsize=2;
  else
    im->dispsize=4;
  if (cmdsize<im->dispsize) {
    im->da->errors|=DAE_CROSS;         // Command crosses end of memory block
    return; };
  op->features=OP_MEMORY|OP_OPCONST;
  op->opsize=op->granularity=datasize;
  op->reg=REG_UNDEF;
  im->usesaddrsize=1;                  // Address size prefix is meaningful
  im->usessegment=1;                   // Segment override prefix is meaningful
  if (im->dispsize==4) {               // 32-bit immediate address
    // 32-bit address means possible fixup, calculate offset.
    im->da->memfixup=im->mainsize;
    op->opconst=im->da->memconst=*(ulong *)cmd; }
  else {                               // 16-bit immediate address, very rare
    op->opconst=im->da->memconst=*(ushort *)cmd;
    op->features|=OP_ADDR16; };
  // Get segment.
  op->seg=Getsegment(im,op->arg,SEG_DS);
  // Decode memory operand to text, if requested.
  if (im->damode & DA_TEXT)
    Memaddrtotext(im,op->arg,datasize,op->seg,T(""),op->opconst,op->text);
  ;
};

// Decodes simple register address ([reg16] or [reg32]). Flag changesreg must
// be 0 if register remains unchanged and 1 if it changes. If fixseg is set to
// SEG_UNDEF, assumes overridable DS:, otherwise assumes fixsegment that cannot
// be overriden with segment prefix. If fixaddrsize is 2 or 4, assumes 16- or
// 32-bit addressing only, otherwise uses default. ATTENTION, calling routine
// must set usesdatasize by itself!
static void Operandindirect(t_imdata *im,int index,int changesreg,int fixseg,
  int fixaddrsize,ulong datasize,t_operand *op) {
  int n;
  ulong originallist;
  tchar ardec[SHORTNAME];
  op->features=OP_MEMORY|OP_INDEXED;
  if (changesreg) {
    op->features|=OP_MODREG;
    op->reg=index;
    im->da->modifies|=(1<<index); }
  else
    op->reg=REG_UNDEF;
  if (fixaddrsize==2)
    op->features|=OP_ADDR16;
  else if (fixaddrsize==0) {
    im->usesaddrsize=1;                // Address size prefix is meaningful
    if (im->prefixlist & PF_ASIZE) {
      op->features|=OP_ADDR16;
      fixaddrsize=2;
    };
  };
  // Get segment.
  if (fixseg==SEG_UNDEF) {
    op->seg=Getsegment(im,op->arg,SEG_DS);
    im->usessegment=1; }               // Segment override prefix is meaningful
  else
    op->seg=fixseg;
  op->opsize=datasize;
  op->granularity=datasize;            // Default, may be overriden later
  op->scale[index]=1;
  op->aregs=(1<<index);
  im->da->uses|=(1<<index);
  // Warn if memory contents is 16-bit jump/call destination.
  if (datasize==2 && (op->arg & B_MODMASK)==B_JMPCALL)
    im->da->warnings|=DAW_JMP16;
  // Decode source operand to text, if requested.
  if (im->damode & DA_TEXT) {
    if (im->config->disasmmode==DAMODE_ATT) {
      ardec[0]=T('%'); n=1; }
    else
      n=0;
    if (fixaddrsize==2)
      Tstrcopy(ardec+n,SHORTNAME-n,regname16[index]);
    else
      Tstrcopy(ardec+n,SHORTNAME-n,regname32[index]);
    if (fixseg==SEG_UNDEF)
      Memaddrtotext(im,op->arg,datasize,op->seg,ardec,0,op->text);
    else {
      originallist=im->prefixlist;
      im->prefixlist&=~PF_SEGMASK;
      Memaddrtotext(im,op->arg,datasize,op->seg,ardec,0,op->text);
      im->prefixlist=originallist;
    };
  };
};

// Decodes XLAT source address ([(E)BX+AL]). Note that I set scale of EAX to 1,
// which is not exactly true. ATTENTION, calling routine must set usesdatasize
// by itself!
static void Operandxlat(t_imdata *im,t_operand *op) {
  tchar *ardec;
  op->features=OP_MEMORY|OP_INDEXED;
  if (im->prefixlist & PF_ASIZE)
    op->features|=OP_ADDR16;
  im->usesaddrsize=1;                  // Address size prefix is meaningful
  im->usessegment=1;                   // Segment override prefix is meaningful
  op->opsize=1;
  op->granularity=1;
  op->reg=REG_UNDEF;
  // Get segment.
  op->seg=Getsegment(im,op->arg,SEG_DS);
  op->scale[REG_EAX]=1;                // This is not correct!
  op->scale[REG_EBX]=1;
  op->aregs=(1<<REG_EAX)|(1<<REG_EBX);
  im->da->uses|=op->aregs;
  // Decode address to text, if requested.
  if (im->damode & DA_TEXT) {
    if (im->config->disasmmode==DAMODE_ATT)
      ardec=(im->prefixlist & PF_ASIZE?T("%BX,%AL"):T("%EBX,%AL"));
    else
      ardec=(im->prefixlist & PF_ASIZE?T("BX+AL"):T("EBX+AL"));
    Memaddrtotext(im,op->arg,1,op->seg,ardec,0,op->text);
  };
};

// Decodes stack pushes of any size, including implicit return address in
// CALLs. ATTENTION, calling routine must set usesdatasize by itself!
static void Operandpush(t_imdata *im,ulong datasize,t_operand *op) {
  int n,addrsize;
  ulong originallist;
  tchar ardec[SHORTNAME];
  op->features=OP_MEMORY|OP_INDEXED|OP_MODREG;
  op->reg=REG_ESP;
  op->aregs=(1<<REG_ESP);
  im->da->modifies|=op->aregs;
  im->usesaddrsize=1;                  // Address size prefix is meaningful
  if (im->prefixlist & PF_ASIZE) {
    op->features|=OP_ADDR16;
    addrsize=2; }
  else
    addrsize=4;                        // Flat model!
  op->seg=SEG_SS;
  if ((op->arg & B_ARGMASK)==B_PUSHA) {
    im->da->uses=0xFF;                 // Uses all general registers
    op->opsize=datasize*8; }
  else if ((op->arg & B_ARGMASK)==B_PUSHRETF) {
    im->da->uses|=op->aregs;
    op->opsize=datasize*2; }
  else {
    im->da->uses|=op->aregs;
    // Warn if memory contents is 16-bit jump/call destination.
    if (datasize==2 && (op->arg & B_MODMASK)==B_JMPCALL)
      im->da->warnings|=DAW_JMP16;
    op->opsize=datasize;
  };
  op->opconst=-(long)op->opsize;       // ESP is predecremented
  op->granularity=datasize;            // Default, may be overriden later
  op->scale[REG_ESP]=1;
  // Decode source operand to text, if requested.
  if (im->damode & DA_TEXT) {
    if (im->config->disasmmode==DAMODE_ATT) {
      ardec[0]=T('%'); n=1; }
    else
      n=0;
    if (addrsize==2)
      Tstrcopy(ardec+n,SHORTNAME-n,regname16[REG_ESP]);
    else
      Tstrcopy(ardec+n,SHORTNAME-n,regname32[REG_ESP]);
    originallist=im->prefixlist;
    im->prefixlist&=~PF_SEGMASK;
    Memaddrtotext(im,op->arg,datasize,op->seg,ardec,0,op->text);
    im->prefixlist=originallist;
  };
};

// Decodes segment register.
static void Operandsegreg(t_imdata *im,int index,t_operand *op) {
  int n;
  op->features=OP_SEGREG;
  if (index>=NSEG) {
    op->features|=OP_INVALID;          // Invalid segment register
    im->da->errors|=DAE_BADSEG; };
  op->opsize=op->granularity=2;
  op->reg=index;
  op->seg=SEG_UNDEF;                   // Because this is not a memory address
  if (op->arg & (B_CHG|B_UPD))
    im->da->warnings|=DAW_SEGMOD;      // Modifies segment register
  // Decode name of segment register.
  if (im->damode & DA_TEXT) {
    n=0;
    if (im->config->disasmmode==DAMODE_ATT) op->text[n++]=T('%');
    Tcopycase(op->text+n,TEXTLEN-n,segname[index],im->config->lowercase);
  };
};

// Decodes FPU register operand.
static void Operandfpureg(t_imdata *im,int index,t_operand *op) {
  op->features=OP_FPUREG;
  op->opsize=op->granularity=10;
  op->reg=index;
  op->seg=SEG_UNDEF;                   // Because this is not a memory address
  // Decode name of FPU register.
  if (im->damode & DA_TEXT) {
    if (im->config->disasmmode==DAMODE_ATT) {
      if (im->config->simplifiedst && index==0)
        Tcopycase(op->text,TEXTLEN,T("%ST"),im->config->lowercase);
      else {
        op->text[0]=T('%');
        Tcopycase(op->text+1,TEXTLEN-1,fpushort[index],im->config->lowercase);
      }; }
    else if (im->config->simplifiedst && index==0)
      Tcopycase(op->text,TEXTLEN,T("ST"),im->config->lowercase);
    else if (im->config->disasmmode!=DAMODE_HLA)
      Tcopycase(op->text,TEXTLEN,fpulong[index],im->config->lowercase);
    else
      Tcopycase(op->text,TEXTLEN,fpushort[index],im->config->lowercase);
    ;
  };
};

// Decodes MMX register operands. ATTENTION, does not set correct granularity!
static void Operandmmxreg(t_imdata *im,int index,t_operand *op) {
  int n;
  op->features=OP_MMXREG;
  op->opsize=8;
  op->granularity=4;                   // Default, correct it later!
  op->reg=index;
  op->seg=SEG_UNDEF;
  // Decode name of MMX/3DNow! register.
  if (im->damode & DA_TEXT) {
    n=0;
    if (im->config->disasmmode==DAMODE_ATT) op->text[n++]=T('%');
    Tcopycase(op->text+n,TEXTLEN-n,mmxname[index],im->config->lowercase);
  };
};

// Decodes 3DNow! register operands. ATTENTION, does not set correct
// granularity!
static void Operandnowreg(t_imdata *im,int index,t_operand *op) {
  int n;
  op->features=OP_3DNOWREG;
  op->opsize=8;
  op->granularity=4;                   // Default, correct it later!
  op->reg=index;
  op->seg=SEG_UNDEF;
  // Decode name of MMX/3DNow! register.
  if (im->damode & DA_TEXT) {
    n=0;
    if (im->config->disasmmode==DAMODE_ATT) op->text[n++]=T('%');
    Tcopycase(op->text+n,TEXTLEN-n,mmxname[index],im->config->lowercase);
  };
};

// Decodes SSE register operands. ATTENTION, does not set correct granularity!
static void Operandssereg(t_imdata *im,int index,t_operand *op) {
  int n;
  op->features=OP_SSEREG;
  if (op->arg & B_NOVEXSIZE)
    op->opsize=16;
  else
    op->opsize=im->ssesize;
  op->granularity=4;                   // Default, correct it later!
  op->reg=index;
  op->seg=SEG_UNDEF;
  // Note that some rare SSE commands may use Reg without ModRM.
  if (im->modsize==0)
    im->modsize=1;
  // Decode name of SSE register.
  if (im->damode & DA_TEXT) {
    n=0;
    if (im->config->disasmmode==DAMODE_ATT) op->text[n++]=T('%');
    if (op->opsize==32)
      Tcopycase(op->text+n,TEXTLEN-n,sse256[index],im->config->lowercase);
    else
      Tcopycase(op->text+n,TEXTLEN-n,sse128[index],im->config->lowercase);
    ;
  };
};

// Decodes flag register EFL.
static void Operandefl(t_imdata *im,ulong datasize,t_operand *op) {
  op->features=OP_OTHERREG;
  op->opsize=op->granularity=datasize;
  op->reg=REG_UNDEF;
  op->seg=SEG_UNDEF;
  // Decode name of register.
  if (im->damode & DA_TEXT) {
    if (im->config->disasmmode==DAMODE_ATT)
      Tcopycase(op->text,TEXTLEN,T("%EFL"),im->config->lowercase);
    else
      Tcopycase(op->text,TEXTLEN,T("EFL"),im->config->lowercase);
    ;
  };
};

// Decodes 8/16/32-bit immediate jump/call offset relative to EIP of next
// command.
static void Operandoffset(t_imdata *im,ulong offsetsize,ulong datasize,
  uchar *cmd,ulong cmdsize,ulong offsaddr,t_operand *op) {
  int n;
  tchar label[TEXTLEN];
  if (cmdsize<offsetsize) {
    im->da->errors|=DAE_CROSS;         // Command crosses end of memory block
    return; };
  op->features=OP_CONST;
  op->opsize=op->granularity=datasize; // NOT offsetsize!
  im->immsize1=offsetsize;
  op->reg=REG_UNDEF;
  op->seg=SEG_UNDEF;
  offsaddr+=offsetsize;
  if (offsetsize==1)                   // Sign-extandable constant
    op->opconst=*(signed char *)cmd+offsaddr;
  else if (offsetsize==2)              // 16-bit immediate offset, rare
    op->opconst=*(ushort *)cmd+offsaddr;
  else                                 // 32-bit immediate offset
    op->opconst=*(ulong *)cmd+offsaddr;
  if (datasize==2) {
    op->opconst&=0x0000FFFF;
    im->da->warnings|=DAW_JMP16; };    // Practically unused in Win32 code
  im->usesdatasize=1;
  // Decode address of destination to text, if requested.
  if (im->damode & DA_TEXT) {
    if (offsetsize==1 && im->config->disasmmode!=DAMODE_HLA &&
      im->config->disasmmode!=DAMODE_ATT)
      n=Tcopycase(op->text,TEXTLEN,T("SHORT "),im->config->lowercase);
    else
      n=0;
    if (datasize==4) {
      if (im->decodeaddress!=NULL &&
        im->decodeaddress(label,op->opconst)!=0)
        Tstrcopy(op->text+n,TEXTLEN-n,label);
      else {
        if (im->config->disasmmode==DAMODE_ATT)
          op->text[n++]=T('$');
        Hexprint(4,op->text+n,op->opconst,im,op->arg);
      }; }
    else {
      if (im->config->disasmmode==DAMODE_ATT)
        op->text[n++]=T('$');
      Hexprint(2,op->text+n,op->opconst,im,op->arg);
    };
  };
};

// Decodes 16:16/16:32-bit immediate absolute far jump/call address.
static void Operandimmfaraddr(t_imdata *im,ulong datasize,uchar *cmd,
  ulong cmdsize,t_operand *op) {
  int n;
  if (cmdsize<datasize+2) {
    im->da->errors|=DAE_CROSS;         // Command crosses end of memory block
    return; };
  op->features=OP_CONST|OP_SELECTOR;
  op->opsize=datasize+2;
  op->granularity=datasize;            // Attention, non-standard case!
  op->reg=REG_UNDEF;
  op->seg=SEG_UNDEF;
  im->immsize1=datasize;
  im->immsize2=2;
  if (datasize==2) {
    op->opconst=*(ushort *)cmd;
    im->da->warnings|=DAW_JMP16; }     // Practically unused in Win32 code
  else {
    op->opconst=*(ulong *)cmd;
    im->da->immfixup=im->mainsize; };
  op->selector=*(ushort *)(cmd+datasize);
  im->usesdatasize=1;
  // Decode address of destination to text, if requested.
  if (im->damode & DA_TEXT) {
    if (im->config->disasmmode==DAMODE_ATT) {
      op->text[0]=T('$'); n=1; }
    else
      n=Tcopycase(op->text,TEXTLEN,T("FAR "),im->config->lowercase);
    n+=Hexprint(2,op->text+n,op->selector,im,op->arg);
    op->text[n++]=T(':');
    if (im->config->disasmmode==DAMODE_ATT)
      op->text[n++]=T('$');
    Hexprint(4,op->text+n,op->opconst,im,op->arg);
  };
};

// Decodes immediate constant 1 used in shift operations.
static void Operandone(t_imdata *im,t_operand *op) {
  op->features=OP_CONST;
  op->opsize=op->granularity=1;        // Just to make it defined
  op->reg=REG_UNDEF;
  op->seg=SEG_UNDEF;
  op->opconst=1;
  if (im->damode & DA_TEXT) {
    if (im->config->disasmmode==DAMODE_ATT)
      Tstrcopy(op->text,TEXTLEN,T("$1"));
    else
      Tstrcopy(op->text,TEXTLEN,T("1"));
    ;
  };
};

// Decodes 8/16/32-bit immediate constant (possibly placed after ModRegRM-SIB-
// Disp combination). Constant is nbytes long in the command and extends to
// constsize bytes. If constant is a count, it deals with data of size datasize.
// ATTENTION, calling routine must set usesdatasize by itself!
static void Operandimmconst(t_imdata *im,ulong nbytes,ulong constsize,
  ulong datasize,uchar *cmd,ulong cmdsize,int issecond,t_operand *op) {
  int n;
  ulong u,mod;
  tchar label[TEXTLEN];
  if (cmdsize<im->modsize+im->dispsize+nbytes+(issecond?im->immsize1:0)) {
    im->da->errors|=DAE_CROSS;         // Command crosses end of memory block
    return; };
  op->features=OP_CONST;
  op->opsize=op->granularity=constsize;
  cmd+=im->modsize+im->dispsize;
  if (issecond==0)
    im->immsize1=nbytes;               // First constant
  else {
    im->immsize2=nbytes;               // Second constant (ENTER only)
    cmd+=im->immsize1; };
  op->reg=REG_UNDEF;
  op->seg=SEG_UNDEF;
  if (nbytes==4) {                     // 32-bit immediate constant
    op->opconst=*(ulong *)cmd;
    im->da->immfixup=im->mainsize+im->modsize+im->dispsize+
      (issecond?im->immsize1:0);
    ; }
  else if (nbytes==1)                  // 8-byte constant, maybe sign-extendable
    op->opconst=*(signed char *)cmd;
  else                                 // 16-bite immediate constant, rare
    op->opconst=*(ushort *)cmd;
  if (constsize==1)
    op->opconst&=0x000000FF;
  else if (constsize==2)
    op->opconst&=0x0000FFFF;
  switch (op->arg & B_MODMASK) {
    case B_BITCNT:                     // Constant is a bit count
      if ((datasize==4 && op->opconst>31) ||
        (datasize==1 && op->opconst>7) ||
        (datasize==2 && op->opconst>15)) im->da->warnings|=DAW_SHIFT;
      break;
    case B_SHIFTCNT:                   // Constant is a shift count
      if (op->opconst==0 ||
        (datasize==4 && op->opconst>31) || (datasize==1 && op->opconst>7) ||
        (datasize==2 && op->opconst>15)) im->da->warnings|=DAW_SHIFT;
      break;
    case B_STACKINC:                   // Stack increment must be DWORD-aligned
      if ((op->opconst & 0x3)!=0)
        im->da->warnings|=DAW_STACK;
      im->da->stackinc=op->opconst;
      break;
    default: break; };
  if (im->damode & DA_TEXT) {
    mod=op->arg & B_MODMASK;
    n=0;
    if (constsize==1) {                // 8-bit constant
      if (im->config->disasmmode==DAMODE_ATT)
        op->text[n++]=T('$');
      Hexprint(1,op->text+n,op->opconst,im,op->arg); }
    else if (constsize==4) {           // 32-bit constant
      if (im->decodeaddress!=NULL &&
        (mod==B_NONSPEC || mod==B_JMPCALL || mod==B_JMPCALLFAR) &&
        im->decodeaddress(label,op->opconst)!=0)
        Tstrcopy(op->text+n,TEXTLEN-n,label);
      else {
        if (im->config->disasmmode==DAMODE_ATT)
          op->text[n++]=T('$');
        if (mod!=B_UNSIGNED && mod!=B_BINARY && mod!=B_PORT &&
          (long)op->opconst<0 &&
          (mod==B_SIGNED || (long)op->opconst>NEGLIMIT)
        ) {
          op->text[n++]=T('-'); u=-(long)op->opconst; }
        else
          u=op->opconst;
        Hexprint(4,op->text+n,u,im,op->arg);
      }; }
    else {                             // 16-bit constant
      if (im->config->disasmmode==DAMODE_ATT)
        op->text[n++]=T('$');
      else if ((op->arg & B_SHOWSIZE)!=0) {
        n+=Tcopycase(op->text+n,TEXTLEN-n,sizename[constsize],
          im->config->lowercase);
        n+=Tstrcopy(op->text+n,TEXTLEN-n,T(" ")); };
      Hexprint(2,op->text+n,op->opconst,im,op->arg);
    };
  };
  return;
};

// Decodes contrtol register operands.
static void Operandcreg(t_imdata *im,int index,t_operand *op) {
  int n;
  op->features=OP_CREG;
  op->opsize=op->granularity=4;
  op->reg=index;
  op->seg=SEG_UNDEF;
  // Decode name of control register.
  if (im->damode & DA_TEXT) {
    n=0;
    if (im->config->disasmmode==DAMODE_ATT) op->text[n++]=T('%');
    Tcopycase(op->text+n,TEXTLEN-n,crname[index],im->config->lowercase); };
  // Some control registers are physically absent.
  if (index!=0 && index!=2 && index!=3 && index!=4)
    im->da->errors|=DAE_BADCR;
  ;
};

// Decodes debug register operands.
static void Operanddreg(t_imdata *im,int index,t_operand *op) {
  int n;
  op->features=OP_DREG;
  op->opsize=op->granularity=4;
  op->reg=index;
  op->seg=SEG_UNDEF;
  // Decode name of debug register.
  if (im->damode & DA_TEXT) {
    n=0;
    if (im->config->disasmmode==DAMODE_ATT) op->text[n++]=T('%');
    Tcopycase(op->text+n,TEXTLEN-n,drname[index],im->config->lowercase);
  };
};

// Decodes FPU status register FST.
static void Operandfst(t_imdata *im,t_operand *op) {
  op->features=OP_OTHERREG;
  op->opsize=op->granularity=2;
  op->reg=REG_UNDEF;
  op->seg=SEG_UNDEF;
  // Decode name of register.
  if (im->damode & DA_TEXT) {
    if (im->config->disasmmode==DAMODE_ATT)
      Tcopycase(op->text,TEXTLEN,T("%FST"),im->config->lowercase);
    else
      Tcopycase(op->text,TEXTLEN,T("FST"),im->config->lowercase);
    ;
  };
};

// Decodes FPU control register FCW.
static void Operandfcw(t_imdata *im,t_operand *op) {
  op->features=OP_OTHERREG;
  op->opsize=op->granularity=2;
  op->reg=REG_UNDEF;
  op->seg=SEG_UNDEF;
  // Decode name of register.
  if (im->damode & DA_TEXT) {
    if (im->config->disasmmode==DAMODE_ATT)
      Tcopycase(op->text,TEXTLEN,T("%FCW"),im->config->lowercase);
    else
      Tcopycase(op->text,TEXTLEN,T("FCW"),im->config->lowercase);
    ;
  };
};

// Decodes SSE control register MXCSR.
static void Operandmxcsr(t_imdata *im,t_operand *op) {
  op->features=OP_OTHERREG;
  op->opsize=op->granularity=4;
  op->reg=REG_UNDEF;
  op->seg=SEG_UNDEF;
  // Decode name of register.
  if (im->damode & DA_TEXT) {
    if (im->config->disasmmode==DAMODE_ATT)
      Tcopycase(op->text,TEXTLEN,T("%MXCSR"),im->config->lowercase);
    else
      Tcopycase(op->text,TEXTLEN,T("MXCSR"),im->config->lowercase);
    ;
  };
};


////////////////////////////////////////////////////////////////////////////////
///////////////////////////////// DISASSEMBLER /////////////////////////////////

// Disassembles first command in the binary code of given length at given
// address. Assumes that address and data size attributes of all participating
// segments are 32 bit (flat model). Returns length of the command or 0 in case
// of severe error.

//[bytepointer.com] - removed const qualifier from first argument in function list
ulong Disasm(uchar *cmd,ulong cmdsize,ulong ip,t_disasm *da,
  int damode,t_config *config,int (*decodeaddress)(tchar *s,ulong addr)) {
  //[bytepointer.com] added zero initialization to relevant members to silence compiler warning: C4701: local variable 'xxx' may be used without having been initialized
  int i,j,k,q,noperand = 0,nout,enclose = 0,vexreg = 0,success,cfill = 0,ofill;
  ulong m,n,u,prefix,prefixmask = 0,code,arg,cmdtype = 0,datasize = 0;
  ulong type,vex,vexlead;
  t_imdata im;
  const t_chain *pchain = 0;
  const t_bincmd *pcmd = 0;
  const t_modrm *pmrm = 0;
  t_operand *op,pseudoop;

  //[bytepointer.com] added memset() calls to init im struct
  memset(&im,0,sizeof(im));
  memset(&im,0,sizeof(im));

  // Verify input parameters.
  if (cmd==NULL || cmdsize==0 || da==NULL || cmdchain==NULL)
    return 0;                          // Error in parameters or uninitialized
  // Initialize t_disasm structure that receives results of disassembly. This
  // structure is very large, memset() or several memset()'s would take much,
  // much longer.
  da->ip=ip;
  da->memfixup=da->immfixup=(ulong)-1; //[bytepointer.com] added ulong cast
  da->errors=DAE_NOERR;
  da->warnings=DAW_NOWARN;
  da->uses=0;
  da->modifies=0;
  da->memconst=0;
  da->stackinc=0;
  for (i=0,op=da->op; i<NOPERAND; i++,op++) {
    op->features=0;
    op->arg=0;
    op->opsize=op->granularity=0;
    op->reg=REG_UNDEF;
    op->uses=0;
    op->modifies=0;
    op->seg=SEG_UNDEF;
    ((ulong *)op->scale)[0]=0;
    ((ulong *)op->scale)[1]=0;
    op->aregs=0;
    op->opconst=0;
    op->selector=0;
    op->text[0]=T('\0'); };
  da->dump[0]=T('\0');
  da->result[0]=T('\0');
  da->masksize=0;
  // Prepare intermediate data. This data allows to keep Disasm() reentrant
  // (thread-safe).
  im.da=da;
  im.damode=damode;
  if (config==NULL)
    im.config=config=&defconfig;       // Use default configuration
  else
    im.config=config;
  im.decodeaddress=decodeaddress;
  im.prefixlist=0;
  im.ssesize=16;                       // Default
  im.immsize1=im.immsize2=0;
  // Correct 80x86 command may contain up to 4 prefixes belonging to different
  // prefix groups. If Disasm() detects second prefix from the same group, it
  // flushes first prefix in the sequence as a pseudocommand. (This is not
  // quite true; all CPUs that I have tested accept repeating prefixes. Still,
  // who will place superfluous and possibly nonportable prefixes into the
  // code?)
  for (n=0; ; n++) {
    if (n>=cmdsize) {                  // Command crosses end of memory block
      n=0; im.prefixlist=0;            // Decode as standalone prefix
      break; };
    // Note that some CPUs treat REPx and LOCK as belonging to the same group.
    switch (cmd[n]) {
      case 0x26: prefix=PF_ES; prefixmask=PF_SEGMASK; break;
      case 0x2E: prefix=PF_CS; prefixmask=PF_SEGMASK; break;
      case 0x36: prefix=PF_SS; prefixmask=PF_SEGMASK; break;
      case 0x3E: prefix=PF_DS; prefixmask=PF_SEGMASK; break;
      case 0x64: prefix=PF_FS; prefixmask=PF_SEGMASK; break;
      case 0x65: prefix=PF_GS; prefixmask=PF_SEGMASK; break;
      case 0x66: prefix=prefixmask=PF_DSIZE; break;
      case 0x67: prefix=prefixmask=PF_ASIZE; break;
      case 0xF0: prefix=prefixmask=PF_LOCK; break;
      case 0xF2: prefix=PF_REPNE; prefixmask=PF_REPMASK; break;
      case 0xF3: prefix=PF_REP; prefixmask=PF_REPMASK; break;
      default: prefix=0; break; };
    if (prefix==0)
      break;
    if (im.prefixlist & prefixmask) {
      da->errors|=DAE_SAMEPREF;        // Two prefixes from the same group
      break; };
    im.prefixlist|=prefix;
  };
  // There may be VEX prefix preceding command body. Yes, VEX is supported in
  // the 32-bit mode! And even in the 16-bit, but who cares?
  vex=0; vexlead=0;
  if (cmdsize>=n+3 && (*(ushort *)(cmd+n) & 0xC0FE)==0xC0C4) {
    // VEX is not compatible with LOCK, 66, F2 and F3 prefixes. VEX is not
    // compatible with REX, too, but REX prefixes are missing in 32-bit mode.
    if (im.prefixlist & (PF_LOCK|PF_66|PF_F2|PF_F3))
      da->errors|=DAE_SAMEPREF;        // Incompatible prefixes
    else {
      if (cmd[n]==0xC5) {
        // 2-byte VEX prefix.
        im.prefixlist|=PF_VEX2;
        vex=cmd[n+1];
        vexlead=DX_VEX|DX_LEAD0F; n+=2; }
      else {
        // 3-byte VEX prefix.
        im.prefixlist|=PF_VEX3;
        vex=cmd[n+2]+(cmd[n+1]<<8);    // Note the order of the bytes!
        switch (vex & 0x1F00) {
          case 0x0100: vexlead=DX_VEX|DX_LEAD0F; n+=3; break;
          case 0x0200: vexlead=DX_VEX|DX_LEAD38; n+=3; break;
          case 0x0300: vexlead=DX_VEX|DX_LEAD3A; n+=3; break;
          default: vex=0; break;       // Unsupported VEX, decode as LES
        };
      };
      if (vex!=0) {
        // Get size of operands.
        if (vex & 0x0004)
          im.ssesize=32;               // 256-bit SSE operands
        // Get register encoded in VEX prefix.
        vexreg=(~vex>>3) & 0x07;
        // Check for SIMD prefix.
        switch (vex & 0x3) {
          case 0x0001: im.prefixlist|=PF_66; break;
          case 0x0002: im.prefixlist|=PF_F3; break;
          case 0x0003: im.prefixlist|=PF_F2; break;
        };
      };
    };
    if (n>=cmdsize) {                  // Command crosses end of memory block
      n=0; vex=0; im.prefixlist=0;     // Decode as LES
    };
  };
  // We have gathered all prefixes, including those that are integral part of
  // the SSE command.
  if (n>4 || (da->errors & DAE_SAMEPREF)!=0) {
    if (n>4) da->errors|=DAE_MANYPREF;
    n=0; im.prefixlist=0; };           // Decode as standalone prefix
  da->prefixes=im.prefixlist;
  da->nprefix=n;
  // Fetch first 4 bytes of the command and find start of command chain in the
  // command table.
  if (cmdsize>=n+sizeof(ulong))
    code=*(ulong *)(cmd+n);            // Optimization for most frequent case
  else {
    code=cmd[n];
    if (cmdsize>n+1) ((uchar *)&code)[1]=cmd[n+1];
    if (cmdsize>n+2) ((uchar *)&code)[2]=cmd[n+2];
    if (cmdsize>n+3) ((uchar *)&code)[3]=cmd[n+3]; };
  // Walk chain and search for matching command. Command is matched if:
  // (1) code bits allowed in mask coincide in command and descriptor;
  // (2) when command type contains D_MEMORY, ModRegRM byte must indicate
  //     memory, and when type contains D_REGISTER, Mod must indicate register;
  // (3) when bits D_DATAxx or D_ADDRxx are set, size of data and/or code must
  //     match these bits;
  // (4) field D_MUSTMASK must match gathered prefixes;
  // (5) presence or absence of VEX prefix must be matched by DX_VEX. If VEX
  //     is present, implied leading bytes must match vexlead and bit L must
  //     match DX_VLMASK;
  // (6) if short form of string commands is requested, bit D_LONGFORM must be
  //     cleared, or segment override prefix other that DS:, or address size
  //     prefix must be present;
  // (7) when bit D_POSTBYTE is set, byte after ModRegRM/SIB/offset must match
  //     postbyte. Note that all postbyted commands include memory address in
  //     ModRegRM form and do not include immediate constants;
  // (8) if alternative forms of conditional commands are requested, command
  //     is conditional, and it is marked as DX_ZEROMASK or DX_CARRYMASK,
  //     check whether these bits match damode. (Conditional branch on flag
  //     Z!=0 can be disassembled either as JZ or JE. First form is preferrable
  //     after SUB or DEC; second form is more natural after CMP);
  // (9) if command has mnemonics RETN but alternative form RET is expected,
  //     skip it - RET will follow.
  success=0;
  for (pchain=cmdchain+(code & CMDMASK); ; pchain=pchain->pnext) {
    if (pchain==NULL || pchain->pcmd==NULL)
      break;                           // End of chain, no match
    pcmd=pchain->pcmd;
    if (((code ^ pcmd->code) & pcmd->mask)!=0)
      continue;                        // (1) Different code bits
    cmdtype=pcmd->cmdtype;
    if ((damode & DA_TEXT)!=0) {
      if ((pcmd->exttype & DX_RETN)!=0 && config->useretform!=0)
        continue;                      // (9) RET form of near return expected
      if ((cmdtype & D_COND)!=0 &&
        (pcmd->exttype & (DX_ZEROMASK|DX_CARRYMASK))!=0
      ) {
        if ((damode & DA_JZ)!=0 && (pcmd->exttype & DX_ZEROMASK)==DX_JE)
          continue;                    // (8) Wait for DX_JZ
        if ((damode & DA_JC)!=0 && (pcmd->exttype & DX_CARRYMASK)==DX_JB)
          continue;                    // (8) Wait for DX_JC
        ;
      };
    };
    if ((pcmd->exttype & (DX_VEX|DX_LEADMASK))!=vexlead)
      continue;                        // (5) Unmatched VEX prefix
    if (pcmd->exttype & DX_VEX) {
      if (((pcmd->exttype & DX_VLMASK)==DX_LSHORT && (vex & 0x04)!=0) ||
        ((pcmd->exttype & DX_VLMASK)==DX_LLONG && (vex & 0x04)==0))
        continue;                      // (5) Unmatched VEX.L
      ;
    };
    if ((cmdtype &
      (D_MEMORY|D_REGISTER|D_LONGFORM|D_SIZEMASK|D_MUSTMASK|D_POSTBYTE))==0
    ) {
      success=1; break; };             // Optimization for most frequent case
    switch (cmdtype & D_MUSTMASK) {
      case D_MUST66:                   // (4) (SSE) Requires 66, no F2 or F3
        if ((im.prefixlist & (PF_66|PF_F2|PF_F3))!=PF_66) continue;
        break;
      case D_MUSTF2:                   // (4) (SSE) Requires F2, no 66 or F3
        if ((im.prefixlist & (PF_66|PF_F2|PF_F3))!=PF_F2) continue;
        break;
      case D_MUSTF3:                   // (4) (SSE) Requires F3, no 66 or F2
        if ((im.prefixlist & (PF_66|PF_F2|PF_F3))!=PF_F3) continue;
        break;
      case D_MUSTNONE:                 // (4) (MMX,SSE) Requires no 66, F2, F3
        if ((im.prefixlist & (PF_66|PF_F2|PF_F3))!=0) continue;
        break;
      case D_NEEDF2:                   // (4) (SSE) Requires F2, no F3
        if ((im.prefixlist & (PF_F2|PF_F3))!=PF_F2) continue;
        break;
      case D_NEEDF3:                   // (4) (SSE) Requires F3, no F2
        if ((im.prefixlist & (PF_F2|PF_F3))!=PF_F3) continue;
        break;
      case D_NOREP:                    // (4) Must not include F2 or F3
        if ((im.prefixlist & (PF_REP|PF_REPNE))!=0) continue;
        break;
      case D_MUSTREP:                  // (4) Must include F3 (REP)
      case D_MUSTREPE:                 // (4) Must include F3 (REPE)
        if ((im.prefixlist & PF_REP)==0) continue;
        break;
      case D_MUSTREPNE:                // (4) Must include F2 (REPNE)
        if ((im.prefixlist & PF_REPNE)==0) continue;
        break;
      default: break; };
    if ((cmdtype & D_DATA16)!=0 && (im.prefixlist & PF_DSIZE)==0)
      continue;                        // (3) 16-bit data expected
    if ((cmdtype & D_DATA32)!=0 && (im.prefixlist & PF_DSIZE)!=0)
      continue;                        // (3) 32-bit data expected
    if ((cmdtype & D_ADDR16)!=0 && (im.prefixlist & PF_ASIZE)==0)
      continue;                        // (3) 16-bit address expected
    if ((cmdtype & D_ADDR32)!=0 && (im.prefixlist & PF_ASIZE)!=0)
      continue;                        // (3) 32-bit address expected
    if ((cmdtype & D_LONGFORM)!=0 && config->shortstringcmds!=0 &&
      (im.prefixlist & (PF_ES|PF_CS|PF_SS|PF_FS|PF_GS|PF_ASIZE))==0)
      continue;                        // (6) Short form of string cmd expected
    if (cmdtype & D_MEMORY) {
      // (2) Command expects operand in memory (Mod in ModRegRM is not 11b).
      if (n+pcmd->length>=cmdsize)
        break;                         // Command longer than available code
      if ((cmd[n+pcmd->length] & 0xC0)==0xC0) continue; }
    else if (cmdtype & D_REGISTER) {
      // (2) Command expects operand in register (Mod in ModRegRM is 11b).
      if (n+pcmd->length>=cmdsize)
        break;                         // Command longer than available code
      if ((cmd[n+pcmd->length] & 0xC0)!=0xC0) continue; }
    if (cmdtype & D_POSTBYTE) {
      // Command expects postbyte after ModRegRM/SIB/offset as part of the
      // code. If command is longer than available code, immediately report
      // match - error will be reported elsewhere.
      m=n+pcmd->length;                // Offset to ModRegRM byte
      if (m>=cmdsize)
        break;                         // Command longer than available code
      if (im.prefixlist & PF_ASIZE)
        m+=modrm16[cmd[m]].size;       // 16-bit address
      else {
        pmrm=modrm32+cmd[m];
        if (pmrm->psib==NULL)          // 32-bit address without SIB
          m+=pmrm->size;
        else if (m+1>=cmdsize)         // Command longer than available code
          break;
        else                           // 32-bit address with SIB
          m+=pmrm->psib[cmd[m+1]].size;
        ;
      };
      if (m>=cmdsize)
        break;                         // Command longer than available code
      // Asterisk in SSE and AVX commands means comparison predicate. Check
      // for predefined range.
      if (cmd[m]==(uchar)pcmd->postbyte ||
        ((cmdtype & D_WILDCARD)!=0 && cmd[m]<(pcmd->exttype & DX_VEX?32:8)))
        im.immsize1=1;                 // (7) Interprete postbyte as imm const
      else
        continue;                      // (7)
      ;
    };
    success=1;
    break;                             // Perfect match, command found
  };
  // If command is bad but preceded with prefixes, decode first prefix as
  // standalone. In this case, list of command's prefixes is empty.
  if (success==0) {
    pcmd=NULL;
    if (im.prefixlist!=0) {
      n=0; da->nprefix=0; da->prefixes=im.prefixlist=0;
      code=cmd[n] & CMDMASK;
      for (pchain=cmdchain+code; ; pchain=pchain->pnext) {
        if (pchain==NULL || pchain->pcmd==NULL) {
          pcmd=NULL; break; };         // End of chain, no match
        pcmd=pchain->pcmd;
        if ((pcmd->cmdtype & D_CMDTYPE)!=D_PREFIX)
          continue;
        if (((code ^ pcmd->code) & pcmd->mask)==0) {
          cmdtype=pcmd->cmdtype;
          da->errors|=DAE_BADCMD;
          break;
        };
      };
    };
    // If matching command is still not found, report error and return one byte
    // as a command length.
    if (pcmd==NULL) {
      if (damode & DA_DUMP)
        Thexdump(da->dump,cmd,1,config->lowercase);
      if (damode & DA_TEXT) {
        if (config->disasmmode==DAMODE_HLA)
          j=Tcopycase(da->result,TEXTLEN,sizename[1],config->lowercase);
        else if (config->disasmmode==DAMODE_ATT)
          j=Tcopycase(da->result,TEXTLEN,sizeatt[1],config->lowercase);
        else
          j=Tcopycase(da->result,TEXTLEN,sizekey[1],config->lowercase);
        j+=Tstrcopy(da->result+j,TEXTLEN-j,T(" "));
        Thexdump(da->result+j,cmd,1,config->lowercase); };
      da->size=1;
      da->nprefix=0;
      da->prefixes=0;
      da->cmdtype=D_BAD;
      da->exttype=0;
      da->errors|=DAE_BADCMD;          // Unrecognized command
      if (damode & DA_HILITE) {
        da->masksize=(int)tstrlen(da->result); //[bytepointer.com] added int cast
        memset(da->mask,DRAW_SUSPECT,da->masksize); };
      return 1;
    };
  };
  // Exclude prefixes that are integral part of the command from the list of
  // prefixes. First comparison optimizes for the most frequent case of no
  // obligatory prefixes.
  if (cmdtype & (D_SIZEMASK|D_MUSTMASK)) {
    switch (cmdtype & D_MUSTMASK) {
      case D_MUST66:                   // (SSE) Requires 66, no F2 or F3
      case D_MUSTF2:                   // (SSE) Requires F2, no 66 or F3
      case D_MUSTF3:                   // (SSE) Requires F3, no 66 or F2
        im.prefixlist&=~(PF_66|PF_F2|PF_F3); break;
      case D_NEEDF2:                   // (SSE) Requires F2, no F3
      case D_NEEDF3:                   // (SSE) Requires F3, no F2
        im.prefixlist&=~(PF_F2|PF_F3); break;
      default: break; };
    if (cmdtype & D_DATA16)            // Must include data size prefix
      im.prefixlist&=~PF_DSIZE;
    if (cmdtype & D_ADDR16)            // Must include address size prefix
      im.prefixlist&=~PF_ASIZE;
    ;
  };
  // Prepare for disassembling.
  im.modsize=0;                        // Size of ModRegRM/SIB bytes
  im.dispsize=0;                       // Size of address offset
  im.usesdatasize=0;
  im.usesaddrsize=0;
  im.usessegment=0;
  da->cmdtype=cmdtype;
  da->exttype=pcmd->exttype;
  n+=pcmd->length;                     // Offset of ModRegRM or imm constant
  if (n>cmdsize) {
    da->errors|=DAE_CROSS;             // Command crosses end of memory block
    goto error; };
  im.mainsize=n;                       // Size of command with prefixes
  // Set default data size (note that many commands and operands override it).
  if ((cmdtype & D_SIZE01)!=0 && (cmd[n-1] & 0x01)==0) {
    if (im.prefixlist & PF_DSIZE)
      da->warnings|=DAW_DATASIZE;      // Superfluous data size prefix
    datasize=1; }
  else if (im.prefixlist & PF_DSIZE)
    datasize=2;
  else
    datasize=4;
  // Process operands.
  noperand=0;
  for (i=0; i<NOPERAND; i++) {
    arg=pcmd->arg[i];
    if ((arg & B_ARGMASK)==B_NONE)
      break;                           // Optimization for most frequent case
    // If pseudooperands to be skipped, I process them nevertheless. Such
    // operands may contain important information.
    if ((arg & B_PSEUDO)!=0 && (damode & DA_PSEUDO)==0)
      op=&pseudoop;                    // Request to skip pseudooperands
    else
      op=da->op+noperand++;
    op->arg=arg;
    switch (arg & B_ARGMASK) {
      case B_AL:                       // Register AL
        Operandintreg(&im,1,REG_AL,op);
        break;
      case B_AH:                       // Register AH
        Operandintreg(&im,1,REG_AH,op);
        break;
      case B_AX:                       // Register AX
        Operandintreg(&im,2,REG_EAX,op);
        break;
      case B_CL:                       // Register CL
        Operandintreg(&im,1,REG_CL,op);
        break;
      case B_CX:                       // Register CX
        Operandintreg(&im,2,REG_ECX,op);
        break;
      case B_DX:                       // Register DX
        Operandintreg(&im,2,REG_EDX,op);
        break;
      case B_DXPORT:                   // Register DX as I/O port address
        Operandintreg(&im,2,REG_EDX,op);
        op->features|=OP_PORT;
        break;
      case B_EAX:                      // Register EAX
        Operandintreg(&im,4,REG_EAX,op);
        break;
      case B_EBX:                      // Register EBX
        Operandintreg(&im,4,REG_EBX,op);
        break;
      case B_ECX:                      // Register ECX
        Operandintreg(&im,4,REG_ECX,op);
        break;
      case B_EDX:                      // Register EDX
        Operandintreg(&im,4,REG_EDX,op);
        break;
      case B_ACC:                      // Accumulator (AL/AX/EAX)
        Operandintreg(&im,datasize,REG_EAX,op);
        im.usesdatasize=1;
        break;
      case B_STRCNT:                   // Register CX or ECX as REPxx counter
        Operandintreg(&im,(im.prefixlist & PF_ASIZE?2:4),REG_ECX,op);
        im.usesaddrsize=1;
        break;
      case B_DXEDX:                    // Register DX or EDX in DIV/MUL
        Operandintreg(&im,datasize,REG_EDX,op);
        im.usesdatasize=1;
        break;
      case B_BPEBP:                    // Register BP or EBP in ENTER/LEAVE
        Operandintreg(&im,datasize,REG_EBP,op);
        im.usesdatasize=1;
        break;
      case B_REG:                      // 8/16/32-bit register in Reg
        // Note that all commands that use B_REG have also another operand
        // that requires ModRM, so we don't need to set modsize here.
        if (n>=cmdsize)
          da->errors|=DAE_CROSS;       // Command crosses end of memory block
        else {
          Operandintreg(&im,datasize,(cmd[n]>>3) & 0x07,op);
          im.usesdatasize=1; };
        break;
      case B_REG16:                    // 16-bit register in Reg
        if (n>=cmdsize)
          da->errors|=DAE_CROSS;       // Command crosses end of memory block
        else
          Operandintreg(&im,2,(cmd[n]>>3) & 0x07,op);
        break;
      case B_REG32:                    // 32-bit register in Reg
        if (n>=cmdsize)
          da->errors|=DAE_CROSS;       // Command crosses end of memory block
        else
          Operandintreg(&im,4,(cmd[n]>>3) & 0x07,op);
        break;
      case B_REGCMD:                   // 16/32-bit register in last cmd byte
        Operandintreg(&im,datasize,cmd[n-1] & 0x07,op);
        im.usesdatasize=1;
        break;
      case B_REGCMD8:                  // 8-bit register in last cmd byte
        Operandintreg(&im,1,cmd[n-1] & 0x07,op);
        break;
      case B_ANYREG:                   // Reg field is unused, any allowed
        break;
      case B_INT:                      // 8/16/32-bit register/memory in ModRM
      case B_INT1632:                  // 16/32-bit register/memory in ModRM
        k=Operandmodrm(&im,datasize,cmd+n,cmdsize-n,op);
        if (k<0) break;                // Error in address
        if (k==0) Operandintreg(&im,datasize,cmd[n] & 0x07,op);
        im.usesdatasize=1;
        break;
      case B_INT8:                     // 8-bit register/memory in ModRM
        k=Operandmodrm(&im,1,cmd+n,cmdsize-n,op);
        if (k<0) break;                // Error in address
        if (k==0) Operandintreg(&im,1,cmd[n] & 0x07,op);
        break;
      case B_INT16:                    // 16-bit register/memory in ModRM
        k=Operandmodrm(&im,2,cmd+n,cmdsize-n,op);
        if (k<0) break;                // Error in address
        if (k==0) Operandintreg(&im,2,cmd[n] & 0x07,op);
        break;
      case B_INT32:                    // 32-bit register/memory in ModRM
        k=Operandmodrm(&im,4,cmd+n,cmdsize-n,op);
        if (k<0) break;                // Error in address
        if (k==0) Operandintreg(&im,4,cmd[n] & 0x07,op);
        break;
      case B_INT64:                    // 64-bit integer in ModRM, memory only
        k=Operandmodrm(&im,8,cmd+n,cmdsize-n,op);
        if (k<0) break;                // Error in address
        if (k==0) {
          // Register is not allowed, decode as 32-bit register and set error.
          Operandintreg(&im,4,cmd[n] & 0x07,op);
          op->features|=OP_INVALID;
          da->errors|=DAE_MEMORY; break; };
        break;
      case B_INT128:                   // 128-bit integer in ModRM, memory only
        k=Operandmodrm(&im,16,cmd+n,cmdsize-n,op);
        if (k<0) break;                // Error in address
        if (k==0) {
          // Register is not allowed, decode as 32-bit register and set error.
          Operandintreg(&im,4,cmd[n] & 0x07,op);
          op->features|=OP_INVALID;
          da->errors|=DAE_MEMORY; break; };
        break;
      case B_IMMINT:                   // 8/16/32-bit int at immediate addr
        Operandimmaddr(&im,datasize,cmd+n,cmdsize-n,op);
        im.usesdatasize=1;
        break;
      case B_INTPAIR:                  // Two signed 16/32 in ModRM, memory only
        k=Operandmodrm(&im,2*datasize,cmd+n,cmdsize-n,op);
        if (k<0) break;                // Error in address
        op->granularity=datasize;
        if (k==0) {
          // Register is not allowed, decode as register and set error.
          Operandintreg(&im,datasize,cmd[n] & 0x07,op);
          op->features|=OP_INVALID;
          da->errors|=DAE_MEMORY; break; };
        im.usesdatasize=1;
        break;
      case B_SEGOFFS:                  // 16:16/16:32 absolute address in memory
        k=Operandmodrm(&im,datasize+2,cmd+n,cmdsize-n,op);
        if (k<0) break;                // Error in address
        if (k==0) {
          // Register is not allowed, decode and set error.
          Operandintreg(&im,datasize,cmd[n] & 0x07,op);
          op->features|=OP_INVALID;
          da->errors|=DAE_MEMORY; break; };
        im.usesdatasize=1;
        break;
      case B_STRDEST:                  // 8/16/32-bit string dest, [ES:(E)DI]
        Operandindirect(&im,REG_EDI,1,SEG_ES,0,datasize,op);
        im.usesdatasize=1;
        break;
      case B_STRDEST8:                 // 8-bit string destination, [ES:(E)DI]
        Operandindirect(&im,REG_EDI,1,SEG_ES,0,1,op);
        break;
      case B_STRSRC:                   // 8/16/32-bit string source, [(E)SI]
        Operandindirect(&im,REG_ESI,1,SEG_UNDEF,0,datasize,op);
        im.usesdatasize=1;
        break;
      case B_STRSRC8:                  // 8-bit string source, [(E)SI]
        Operandindirect(&im,REG_ESI,1,SEG_UNDEF,0,1,op);
        break;
      case B_XLATMEM:                  // 8-bit memory in XLAT, [(E)BX+AL]
        Operandxlat(&im,op);
        break;
      case B_EAXMEM:                   // Reference to memory addressed by [EAX]
        Operandindirect(&im,REG_EAX,0,SEG_UNDEF,4,1,op);
        break;
      case B_LONGDATA:                 // Long data in ModRM, mem only
        k=Operandmodrm(&im,256,cmd+n,cmdsize-n,op);
        if (k<0) break;                // Error in address
        op->granularity=1;             // Just a trick
        if (k==0) {
          // Register is not allowed, decode and set error.
          Operandintreg(&im,4,cmd[n] & 0x07,op);
          op->features|=OP_INVALID;
          da->errors|=DAE_MEMORY; break; };
        im.usesdatasize=1;             // Caveat user
        break;
      case B_ANYMEM:                   // Reference to memory, data unimportant
        k=Operandmodrm(&im,1,cmd+n,cmdsize-n,op);
        if (k<0) break;                // Error in address
        if (k==0) {
          // Register is not allowed, decode and set error.
          Operandintreg(&im,4,cmd[n] & 0x07,op);
          op->features|=OP_INVALID;
          da->errors|=DAE_MEMORY; };
        break;
      case B_STKTOP:                   // 16/32-bit int top of stack
        Operandindirect(&im,REG_ESP,1,SEG_SS,0,datasize,op);
        im.usesdatasize=1;
        break;
      case B_STKTOPFAR:                // Top of stack (16:16/16:32 far addr)
        Operandindirect(&im,REG_ESP,1,SEG_SS,0,datasize*2,op);
        op->granularity=datasize;
        im.usesdatasize=1;
        break;
      case B_STKTOPEFL:                // 16/32-bit flags on top of stack
        Operandindirect(&im,REG_ESP,1,SEG_SS,0,datasize,op);
        im.usesdatasize=1;
        break;
      case B_STKTOPA:                  // 16/32-bit top of stack all registers
        Operandindirect(&im,REG_ESP,1,SEG_SS,0,datasize*8,op);
        op->granularity=datasize;
        op->modifies=da->modifies=0xFF;
        im.usesdatasize=1;
        break;
      case B_PUSH:                     // 16/32-bit int push to stack
      case B_PUSHRET:                  // 16/32-bit push of return address
      case B_PUSHRETF:                 // 16:16/16:32-bit push of far retaddr
      case B_PUSHA:                    // 16/32-bit push all registers
        Operandpush(&im,datasize,op);
        im.usesdatasize=1;
        break;
      case B_EBPMEM:                   // 16/32-bit int at [EBP]
        Operandindirect(&im,REG_EBP,1,SEG_SS,0,datasize,op);
        im.usesdatasize=1;
        break;
      case B_SEG:                      // Segment register in Reg
        if (n>=cmdsize)
          da->errors|=DAE_CROSS;       // Command crosses end of memory block
        else
          Operandsegreg(&im,(cmd[n]>>3) & 0x07,op);
        break;
      case B_SEGNOCS:                  // Segment register in Reg, but not CS
        if (n>=cmdsize)
          da->errors|=DAE_CROSS;       // Command crosses end of memory block
        else {
          k=(cmd[n]>>3) & 0x07;
          Operandsegreg(&im,k,op);
          if (k==SEG_SS)
            da->exttype|=DX_WONKYTRAP;
          if (k==SEG_CS) {
            op->features|=OP_INVALID;
            da->errors|=DAE_BADSEG;
          };
        };
        break;
      case B_SEGCS:                    // Segment register CS
        Operandsegreg(&im,SEG_CS,op);
        break;
      case B_SEGDS:                    // Segment register DS
        Operandsegreg(&im,SEG_DS,op);
        break;
      case B_SEGES:                    // Segment register ES
        Operandsegreg(&im,SEG_ES,op);
        break;
      case B_SEGFS:                    // Segment register FS
        Operandsegreg(&im,SEG_FS,op);
        break;
      case B_SEGGS:                    // Segment register GS
        Operandsegreg(&im,SEG_GS,op);
        break;
      case B_SEGSS:                    // Segment register SS
        Operandsegreg(&im,SEG_SS,op);
        break;
      case B_ST:                       // 80-bit FPU register in last cmd byte
        Operandfpureg(&im,cmd[n-1] & 0x07,op);
        break;
      case B_ST0:                      // 80-bit FPU register ST0
        Operandfpureg(&im,0,op);
        break;
      case B_ST1:                      // 80-bit FPU register ST1
        Operandfpureg(&im,1,op);
        break;
      case B_FLOAT32:                  // 32-bit float in ModRM, memory only
        k=Operandmodrm(&im,4,cmd+n,cmdsize-n,op);
        if (k<0) break;                // Error in address
        if (k==0) {
          // Register is not allowed, decode as FPU register and set error.
          Operandfpureg(&im,cmd[n] & 0x07,op);
          op->features|=OP_INVALID;
          da->errors|=DAE_MEMORY; };
        break;
      case B_FLOAT64:                  // 64-bit float in ModRM, memory only
        k=Operandmodrm(&im,8,cmd+n,cmdsize-n,op);
        if (k<0) break;                // Error in address
        if (k==0) {
          // Register is not allowed, decode as FPU register and set error.
          Operandfpureg(&im,cmd[n] & 0x07,op);
          op->features|=OP_INVALID;
          da->errors|=DAE_MEMORY; };
        break;
      case B_FLOAT80:                  // 80-bit float in ModRM, memory only
        k=Operandmodrm(&im,10,cmd+n,cmdsize-n,op);
        if (k<0) break;                // Error in address
        if (k==0) {
          // Register is not allowed, decode as FPU register and set error.
          Operandfpureg(&im,cmd[n] & 0x07,op);
          op->features|=OP_INVALID;
          da->errors|=DAE_MEMORY; };
        break;
      case B_BCD:                      // 80-bit BCD in ModRM, memory only
        k=Operandmodrm(&im,10,cmd+n,cmdsize-n,op);
        if (k<0) break;                // Error in address
        if (k==0) {
          // Register is not allowed, decode as FPU register and set error.
          Operandfpureg(&im,cmd[n] & 0x07,op);
          op->features|=OP_INVALID;
          da->errors|=DAE_MEMORY; };
        break;
      case B_MREG8x8:                  // MMX register as 8 8-bit integers
      case B_MREG16x4:                 // MMX register as 4 16-bit integers
      case B_MREG32x2:                 // MMX register as 2 32-bit integers
      case B_MREG64:                   // MMX register as 1 64-bit integer
        if (n>=cmdsize)
          da->errors|=DAE_CROSS;       // Command crosses end of memory block
        else {
          Operandmmxreg(&im,(cmd[n]>>3) & 0x07,op);
          op->granularity=Getgranularity(arg); };
        break;
      case B_MMX8x8:                   // MMX reg/memory as 8 8-bit integers
      case B_MMX16x4:                  // MMX reg/memory as 4 16-bit integers
      case B_MMX32x2:                  // MMX reg/memory as 2 32-bit integers
      case B_MMX64:                    // MMX reg/memory as 1 64-bit integer
        k=Operandmodrm(&im,8,cmd+n,cmdsize-n,op);
        if (k<0) break;                // Error in address
        if (k==0) Operandmmxreg(&im,cmd[n] & 0x07,op);
        op->granularity=Getgranularity(arg);
        break;
      case B_MMX8x8DI:                 // MMX 8 8-bit integers at [DS:(E)DI]
        Operandindirect(&im,REG_EDI,0,SEG_UNDEF,0,8,op);
        op->granularity=Getgranularity(arg);
        break;
      case B_3DREG:                    // 3DNow! register as 2 32-bit floats
        if (n>=cmdsize)
          da->errors|=DAE_CROSS;       // Command crosses end of memory block
        else {
          Operandnowreg(&im,(cmd[n]>>3) & 0x07,op);
          op->granularity=4; };
        break;
      case B_3DNOW:                    // 3DNow! reg/memory as 2 32-bit floats
        k=Operandmodrm(&im,8,cmd+n,cmdsize-n,op);
        if (k<0) break;                // Error in address
        if (k==0) Operandnowreg(&im,cmd[n] & 0x07,op);
        op->granularity=4;
        break;
      case B_SREGF32x4:                // SSE register as 4 32-bit floats
      case B_SREGF32L:                 // Low 32-bit float in SSE register
      case B_SREGF32x2L:               // Low 2 32-bit floats in SSE register
      case B_SREGF64x2:                // SSE register as 2 64-bit floats
      case B_SREGF64L:                 // Low 64-bit float in SSE register
        if (n>=cmdsize)
          da->errors|=DAE_CROSS;       // Command crosses end of memory block
        else {
          Operandssereg(&im,(cmd[n]>>3) & 0x07,op);
          op->granularity=Getgranularity(arg); };
        break;
      case B_SVEXF32x4:                // SSE reg in VEX as 4 32-bit floats
      case B_SVEXF32L:                 // Low 32-bit float in SSE in VEX
      case B_SVEXF64x2:                // SSE reg in VEX as 2 64-bit floats
      case B_SVEXF64L:                 // Low 64-bit float in SSE in VEX
        Operandssereg(&im,vexreg,op);
        op->granularity=Getgranularity(arg);
        break;
      case B_SSEF32x4:                 // SSE reg/memory as 4 32-bit floats
      case B_SSEF64x2:                 // SSE reg/memory as 2 64-bit floats
        k=Operandmodrm(&im,
          (arg & B_NOVEXSIZE?16:im.ssesize),cmd+n,cmdsize-n,op);
        if (k<0) break;                // Error in address
        if (k==0) Operandssereg(&im,cmd[n] & 0x07,op);
        op->granularity=Getgranularity(arg);
        break;
      case B_SSEF32L:                  // Low 32-bit float in SSE reg/memory
        k=Operandmodrm(&im,4,cmd+n,cmdsize-n,op);
        if (k<0) break;                // Error in address
        if (k==0)                      // Operand in SSE register
          Operandssereg(&im,cmd[n] & 0x07,op);
        op->granularity=4;
        break;
      case B_SSEF32x2L:                // Low 2 32-bit floats in SSE reg/memory
        k=Operandmodrm(&im,
          (arg & B_NOVEXSIZE?16:im.ssesize)/2,cmd+n,cmdsize-n,op);
        if (k<0) break;                // Error in address
        if (k==0)                      // Operand in SSE register
          Operandssereg(&im,cmd[n] & 0x07,op);
        op->granularity=4;
        break;
      case B_SSEF64L:                  // Low 64-bit float in SSE reg/memory
        k=Operandmodrm(&im,8,cmd+n,cmdsize-n,op);
        if (k<0) break;                // Error in address
        if (k==0)                      // Operand in SSE register
          Operandssereg(&im,cmd[n] & 0x07,op);
        op->granularity=8;
        break;
      case B_XMM0I32x4:                // XMM0 as 4 32-bit integers
      case B_XMM0I64x2:                // XMM0 as 2 64-bit integers
      case B_XMM0I8x16:                // XMM0 as 16 8-bit integers
        Operandssereg(&im,0,op);
        op->granularity=Getgranularity(arg);
        break;
      case B_SREGI8x16:                // SSE register as 16 8-bit sigints
      case B_SREGI16x8:                // SSE register as 8 16-bit sigints
      case B_SREGI32x4:                // SSE register as 4 32-bit sigints
      case B_SREGI64x2:                // SSE register as 2 64-bit sigints
      case B_SREGI32L:                 // Low 32-bit sigint in SSE register
      case B_SREGI32x2L:               // Low 2 32-bit sigints in SSE register
      case B_SREGI64L:                 // Low 64-bit sigint in SSE register
        if (n>=cmdsize)
          da->errors|=DAE_CROSS;       // Command crosses end of memory block
        else {
          Operandssereg(&im,(cmd[n]>>3) & 0x07,op);
          op->granularity=Getgranularity(arg); };
        break;
      case B_SVEXI8x16:                // SSE reg in VEX as 16 8-bit sigints
      case B_SVEXI16x8:                // SSE reg in VEX as 8 16-bit sigints
      case B_SVEXI32x4:                // SSE reg in VEX as 4 32-bit sigints
      case B_SVEXI64x2:                // SSE reg in VEX as 2 64-bit sigints
        Operandssereg(&im,vexreg,op);
        op->granularity=Getgranularity(arg);
        break;
      case B_SSEI8x16:                 // SSE reg/memory as 16 8-bit sigints
      case B_SSEI16x8:                 // SSE reg/memory as 8 16-bit sigints
      case B_SSEI32x4:                 // SSE reg/memory as 4 32-bit sigints
      case B_SSEI64x2:                 // SSE reg/memory as 2 64-bit sigints
        k=Operandmodrm(&im,
          (arg & B_NOVEXSIZE?16:im.ssesize),cmd+n,cmdsize-n,op);
        if (k<0) break;                // Error in address
        if (k==0)
          Operandssereg(&im,cmd[n] & 0x07,op);
        op->granularity=Getgranularity(arg);
        break;
      case B_SSEI8x8L:                 // Low 8 8-bit ints in SSE reg/memory
      case B_SSEI16x4L:                // Low 4 16-bit ints in SSE reg/memory
      case B_SSEI32x2L:                // Low 2 32-bit sigints in SSE reg/memory
        k=Operandmodrm(&im,
          (arg & B_NOVEXSIZE?16:im.ssesize)/2,cmd+n,cmdsize-n,op);
        if (k<0) break;                // Error in address
        if (k==0)
          Operandssereg(&im,cmd[n] & 0x07,op);
        op->granularity=Getgranularity(arg);
        break;
      case B_SSEI8x4L:                 // Low 4 8-bit ints in SSE reg/memory
      case B_SSEI16x2L:                // Low 2 16-bit ints in SSE reg/memory
        k=Operandmodrm(&im,4,cmd+n,cmdsize-n,op);
        if (k<0) break;                // Error in address
        if (k==0)
          Operandssereg(&im,cmd[n] & 0x07,op);
        op->granularity=Getgranularity(arg);
        break;
      case B_SSEI8x2L:                 // Low 2 8-bit ints in SSE reg/memory
        k=Operandmodrm(&im,2,cmd+n,cmdsize-n,op);
        if (k<0) break;                // Error in address
        if (k==0)
          Operandssereg(&im,cmd[n] & 0x07,op);
        op->granularity=Getgranularity(arg);
        break;
      case B_SSEI8x16DI:               // SSE 16 8-bit sigints at [DS:(E)DI]
        Operandindirect(&im,REG_EDI,0,SEG_UNDEF,0,
          (arg & B_NOVEXSIZE?16:im.ssesize),op);
        op->granularity=1;
        break;
      case B_EFL:                      // Flags register EFL
        Operandefl(&im,4,op);
        break;
      case B_FLAGS8:                   // Flags (low byte)
        Operandefl(&im,1,op);
        break;
      case B_OFFSET:                   // 16/32 const offset from next command
        Operandoffset(&im,datasize,datasize,cmd+n,cmdsize-n,da->ip+n,op);
        break;
      case B_BYTEOFFS:                 // 8-bit sxt const offset from next cmd
        Operandoffset(&im,1,datasize,cmd+n,cmdsize-n,da->ip+n,op);
        break;
      case B_FARCONST:                 // 16:16/16:32 absolute address constant
        Operandimmfaraddr(&im,datasize,cmd+n,cmdsize-n,op);
        break;
      case B_DESCR:                    // 16:32 descriptor in ModRM
        k=Operandmodrm(&im,6,cmd+n,cmdsize-n,op);
        if (k<0) break;                // Error in address
        if (k==0) {
          // Register is not allowed, decode as 32-bit register and set error.
          Operandintreg(&im,4,cmd[n] & 0x07,op);
          op->features|=OP_INVALID;
          da->errors|=DAE_MEMORY; };
        break;
      case B_1:                        // Immediate constant 1
        Operandone(&im,op);
        break;
      case B_CONST8:                   // Immediate 8-bit constant
        Operandimmconst(&im,1,1,datasize,cmd+n,cmdsize-n,0,op);
        if (arg & B_PORT) op->features|=OP_PORT;
        break;
      case B_SIMMI8x16:                // SSE reg in immediate 8-bit constant
        if (cmdsize-n<im.modsize+im.dispsize+1) {
          da->errors|=DAE_CROSS;       // Command crosses end of memory block
          break; };
        im.immsize1=1;
        Operandssereg(&im,(cmd[n+im.modsize+im.dispsize]>>4) & 0x07,op);
        op->granularity=Getgranularity(arg);
        break;
      case B_CONST8_2:                 // Immediate 8-bit const, second in cmd
        Operandimmconst(&im,1,1,datasize,cmd+n,cmdsize-n,1,op);
        break;
      case B_CONST16:                  // Immediate 16-bit constant
        Operandimmconst(&im,2,2,datasize,cmd+n,cmdsize-n,0,op);
        break;
      case B_CONST:                    // Immediate 8/16/32-bit constant
      case B_CONSTL:                   // Immediate 16/32-bit constant
        Operandimmconst(&im,datasize,datasize,datasize,cmd+n,cmdsize-n,0,op);
        im.usesdatasize=1;
        break;
      case B_SXTCONST:                 // Immediate 8-bit sign-extended to size
        Operandimmconst(&im,1,datasize,datasize,cmd+n,cmdsize-n,0,op);
        im.usesdatasize=1;
        break;
      case B_CR:                       // Control register in Reg
        Operandcreg(&im,(cmd[n]>>3) & 0x07,op);
        break;
      case B_CR0:                      // Control register CR0
        Operandcreg(&im,0,op);
        break;
      case B_DR:                       // Debug register in Reg
        Operanddreg(&im,(cmd[n]>>3) & 0x07,op);
        break;
      case B_FST:                      // FPU status register
        Operandfst(&im,op);
        break;
      case B_FCW:                      // FPU control register
        Operandfcw(&im,op);
        break;
      case B_MXCSR:                    // SSE media control and status register
        Operandmxcsr(&im,op);
        break;
      default:                         // Internal error
        da->errors|=DAE_INTERN;
      break; };
    if ((arg & B_32BITONLY)!=0 && op->opsize!=4)
      da->warnings|=DAW_NONCLASS;
    if ((arg & B_MODMASK)==B_JMPCALLFAR)
      da->warnings|=DAW_FARADDR;
    if (arg & B_PSEUDO) op->features|=OP_PSEUDO;
    if (arg & (B_CHG|B_UPD)) op->features|=OP_MOD;
  };
  if (im.prefixlist!=0) {                 // Optimization for most frequent case
    // If LOCK prefix is present, report error if prefix is not allowed by
    // command and warning otherwise. Application code usually doesn't need
    // atomic bus access.
    if ((im.prefixlist & PF_LOCK)!=0) {
      if ((cmdtype & D_LOCKABLE)==0) da->errors|=DAE_LOCK;
      else da->warnings|=DAW_LOCK; };
    // Warn if data size prefix is present but not used by command.
    if ((im.prefixlist & PF_DSIZE)!=0 && im.usesdatasize==0 &&
      (pcmd->exttype & DX_TYPEMASK)!=DX_NOP)
      da->warnings|=DAW_DATASIZE;
    // Warn if address size prefix is present but not used by command.
    if ((im.prefixlist & PF_ASIZE)!=0 && im.usesaddrsize==0)
      da->warnings|=DAW_ADDRSIZE;
    // Warn if segment override prefix is present but command doesn't access
    // memory. Prefixes CS: and DS: are also used as branch hints in
    // conditional branches.
    if ((im.prefixlist & PF_SEGMASK)!=0 && im.usessegment==0) {
      if ((cmdtype & D_BHINT)==0 || (im.prefixlist & PF_HINT)==0)
      da->warnings|=DAW_SEGPREFIX; };
    // Warn if REPxx prefix is present but not used by command. Attention,
    // Intel frequently uses these prefixes for different means!
    if (im.prefixlist & PF_REPMASK) {
      if (((im.prefixlist & PF_REP)!=0 && (cmdtype & D_MUSTMASK)!=D_MUSTREP &&
      (cmdtype & D_MUSTMASK)!=D_MUSTREPE) ||
      ((im.prefixlist & PF_REPNE)!=0 && (cmdtype & D_MUSTMASK)!=D_MUSTREPNE))
      da->warnings|=DAW_REPPREFIX;
    };
  };
  // Warn on unaligned stack, I/O and privileged commands.
  switch (cmdtype & D_CMDTYPE) {
    case D_PUSH:
      if (datasize==2) da->warnings|=DAW_STACK; break;
    case D_INT:
      da->warnings|=DAW_INTERRUPT; break;
    case D_IO:
      da->warnings|=DAW_IO; break;
    case D_PRIVILEGED:
      da->warnings|=DAW_PRIV;
    break; };
  // Warn on system, privileged  and undocumented commands.
  if ((cmdtype & D_USEMASK)!=0) {
    if ((cmdtype & D_USEMASK)==D_RARE || (cmdtype & D_USEMASK)==D_SUSPICIOUS)
      da->warnings|=DAW_RARE;
    if ((cmdtype & D_USEMASK)==D_UNDOC) da->warnings|=DAW_NONCLASS; };
  // If command implicitly changes ESP, it uses and modifies this register.
  if (cmdtype & D_CHGESP) {
    da->uses|=(1<<REG_ESP);
    da->modifies|=(1<<REG_ESP);
  };
error:
  // Prepare hex dump, if requested. As maximal size of command is limited to
  // MAXCMDSIZE=16 bytes, string can't overflow.
  if (damode & DA_DUMP) {
    if (da->errors & DAE_CROSS)        // Incomplete command
      Thexdump(da->dump,cmd,cmdsize,config->lowercase);
    else {
      j=0;
      // Dump prefixes. REPxx is treated as prefix and separated from command
      // with semicolon; prefixes 66, F2 and F3 that are part of SSE command
      // are glued with command's body - well, at least if there are no other
      // prefixes inbetween.
      for (u=0; u<da->nprefix; u++) {
        j+=Thexdump(da->dump+j,cmd+u,1,config->lowercase);
        if (cmd[u]==0x66 && (cmdtype & D_MUSTMASK)==D_MUST66) continue;
        if (cmd[u]==0xF2 && ((cmdtype & D_MUSTMASK)==D_MUSTF2 ||
          (cmdtype & D_MUSTMASK)==D_NEEDF2)) continue;
        if (cmd[u]==0xF3 && ((cmdtype & D_MUSTMASK)==D_MUSTF3 ||
          (cmdtype & D_MUSTMASK)==D_NEEDF3)) continue;
        if ((im.prefixlist & (PF_VEX2|PF_VEX3))!=0 && u==da->nprefix-2)
          continue;
        if ((im.prefixlist & PF_VEX3)!=0 && u==da->nprefix-3)
          continue;
        da->dump[j++]=T(':'); };
      // Dump body of the command, including ModRegRM and SIB bytes.
      j+=Thexdump(da->dump+j,cmd+u,im.mainsize+im.modsize-u,
        config->lowercase);
      // Dump displacement, if any, separated with space from command's body.
      if (im.dispsize>0) {
        da->dump[j++]=T(' ');
        j+=Thexdump(da->dump+j,cmd+im.mainsize+im.modsize,im.dispsize,
          config->lowercase);
        ;
      };
      // Dump immediate constants, if any.
      if (im.immsize1>0) {
        da->dump[j++]=T(' ');
        j+=Thexdump(da->dump+j,cmd+im.mainsize+im.modsize+im.dispsize,
          im.immsize1,config->lowercase)
        ;
      };
      if (im.immsize2>0) {
        da->dump[j++]=T(' ');
        Thexdump(da->dump+j,cmd+im.mainsize+im.modsize+im.dispsize+im.immsize1,
          im.immsize2,config->lowercase);
        ;
      };
    };
  };
  // Prepare disassembled command. There are many options that control look
  // and feel of disassembly, so the procedure is a bit, errr, boring.
  if (damode & DA_TEXT) {
    if (da->errors & DAE_CROSS) {      // Incomplete command
      q=Tstrcopy(da->result,TEXTLEN,T("???"));
      if (damode & DA_HILITE) {
        memset(da->mask,DRAW_SUSPECT,q);
        da->masksize=q;
      }; }
    else {
      j=0;
      // If LOCK and/or REPxx prefix is present, prepend it to the command.
      // Such cases are rare, first comparison makes small optimization.
      if (im.prefixlist & (PF_LOCK|PF_REPMASK)) {
        if (im.prefixlist & PF_LOCK)
          j=Tcopycase(da->result,TEXTLEN,T("LOCK "),config->lowercase);
        if (im.prefixlist & PF_REPNE)
          j+=Tcopycase(da->result+j,TEXTLEN-j,T("REPNE "),config->lowercase);
        else if (im.prefixlist & PF_REP) {
          if ((cmdtype & D_MUSTMASK)==D_MUSTREPE)
            j+=Tcopycase(da->result+j,TEXTLEN-j,T("REPE "),config->lowercase);
          else
            j+=Tcopycase(da->result+j,TEXTLEN-j,T("REP "),config->lowercase);
          ;
        };
      };
      // If there is a branch hint, prefix jump mnemonics with '+' (taken) or
      // '-' (not taken), or use pseudoprefixes BHT/BHNT. I don't know how MASM
      // indicates hints.
      if (cmdtype & D_BHINT) {
        if (config->jumphintmode==0) {
          if (im.prefixlist & PF_TAKEN)
            da->result[j++]=T('+');
          else if (im.prefixlist & PF_NOTTAKEN)
            da->result[j++]=T('-');
          ; }
        else {
          if (im.prefixlist & PF_TAKEN)
            j+=Tcopycase(da->result+j,TEXTLEN-j,T("BHT "),config->lowercase);
          else if (im.prefixlist & PF_NOTTAKEN)
            j+=Tcopycase(da->result+j,TEXTLEN-j,T("BHNT "),config->lowercase);
          ;
        };
      };
      // Get command mnemonics. If mnemonics contains asterisk, it must be
      // replaced by W, D or none according to sizesens. Asterisk in SSE and
      // AVX commands means comparison predicate.
      if (cmdtype & D_WILDCARD) {
        for (i=0; ; i++) {
          if (pcmd->name[i]==T('\0'))
            break;
          else if (pcmd->name[i]!=T('*'))
            da->result[j++]=pcmd->name[i];
          else if (cmdtype & D_POSTBYTE)
            j+=Tstrcopy(da->result+j,TEXTLEN-j,
            ssepredicate[cmd[im.mainsize+im.modsize+im.dispsize]]);
          else if (datasize==4 &&
            (config->sizesens==0 || config->sizesens==1))
            da->result[j++]=T('D');
          else if (datasize==2 &&
            (config->sizesens==1 || config->sizesens==2))
            da->result[j++]=T('W');
          ; };
        da->result[j]=T('\0');
        if (config->lowercase) tstrlwr(da->result); }
      else {
        j+=Tcopycase(da->result+j,TEXTLEN-j,pcmd->name,config->lowercase);
        if (config->disasmmode==DAMODE_ATT && im.usesdatasize!=0) {
          // AT&T mnemonics are suffixed with the operand's size.
          if ((cmdtype & D_CMDTYPE)!=D_CMD &&
            (cmdtype & D_CMDTYPE)!=D_MOV &&
            (cmdtype & D_CMDTYPE)!=D_MOVC &&
            (cmdtype & D_CMDTYPE)!=D_TEST &&
            (cmdtype & D_CMDTYPE)!=D_STRING &&
            (cmdtype & D_CMDTYPE)!=D_PUSH &&
            (cmdtype & D_CMDTYPE)!=D_POP) ;
          else if (datasize==1) j+=Tcopycase(da->result+j,TEXTLEN-j,
            T("B"),config->lowercase);
          else if (datasize==2) j+=Tcopycase(da->result+j,TEXTLEN-j,
            T("W"),config->lowercase);
          else if (datasize==4) j+=Tcopycase(da->result+j,TEXTLEN-j,
            T("L"),config->lowercase);
          else if (datasize==8) j+=Tcopycase(da->result+j,TEXTLEN-j,
            T("Q"),config->lowercase);
          ;
        };
      };
      if (damode & DA_HILITE) {
        type=cmdtype & D_CMDTYPE;
        if (da->errors!=0)
          cfill=DRAW_SUSPECT;
        else switch (cmdtype & D_CMDTYPE) {
          case D_JMP:                  // Unconditional near jump
          case D_JMPFAR:               // Unconditional far jump
            cfill=DRAW_JUMP; break;
          case D_JMC:                  // Conditional jump on flags
          case D_JMCX:                 // Conditional jump on (E)CX (and flags)
            cfill=DRAW_CJMP; break;
          case D_PUSH:                 // PUSH exactly 1 (d)word of data
          case D_POP:                  // POP exactly 1 (d)word of data
            cfill=DRAW_PUSHPOP; break;
          case D_CALL:                 // Plain near call
          case D_CALLFAR:              // Far call
          case D_INT:                  // Interrupt
            cfill=DRAW_CALL; break;
          case D_RET:                  // Plain near return from call
          case D_RETFAR:               // Far return or IRET
            cfill=DRAW_RET; break;
          case D_FPU:                  // FPU command
          case D_MMX:                  // MMX instruction, incl. SSE extensions
          case D_3DNOW:                // 3DNow! instruction
          case D_SSE:                  // SSE instruction
          case D_AVX:                  // AVX instruction
            cfill=DRAW_FPU; break;
          case D_IO:                   // Accesses I/O ports
          case D_SYS:                  // Legal but useful in system code only
          case D_PRIVILEGED:           // Privileged (non-Ring3) command
            cfill=DRAW_SUSPECT; break;
          default:
            cfill=DRAW_PLAIN;
          break; };
        memset(da->mask,cfill,j);
        da->masksize=j;
      };
      // Add decoded operands. In HLA mode, order of operands is inverted
      // except for comparison commands (marked with bit D_HLADIR) and
      // arguments are enclosed in parenthesis (except for immediate jumps).
      // In AT&T mode, order of operands is always inverted. Operands of type
      // B_PSEUDO are implicit and don't appear in text.
      if (config->disasmmode==DAMODE_HLA &&
        (pcmd->arg[0] & B_ARGMASK)!=B_OFFSET &&
        (pcmd->arg[0] & B_ARGMASK)!=B_BYTEOFFS &&
        (pcmd->arg[0] & B_ARGMASK)!=B_FARCONST)
        enclose=1;                     // Enclose operand list in parenthesis
      else
        enclose=0;
      if ((damode & DA_HILITE)!=0 && config->hiliteoperands!=0)
        cfill=DRAW_PLAIN;
      nout=0;
      for (i=0; i<noperand; i++) {
        if ((config->disasmmode==DAMODE_HLA && (cmdtype & D_HLADIR)==0) ||
          config->disasmmode==DAMODE_ATT)
          k=noperand-1-i;              // Inverted (HLA/AT&T) order of operands
        else
          k=i;                         // Direct (Intel's) order of operands
        arg=da->op[k].arg;
        if ((arg & B_ARGMASK)==B_NONE || (arg & B_PSEUDO)!=0)
          continue;                    // Empty or implicit operand
        q=j;
        if (nout==0) {
          // Spaces between mnemonic and first operand.
          da->result[j++]=T(' ');
          if (config->tabarguments) {
            for ( ; j<8; j++) da->result[j]=T(' '); };
          if (enclose) {
            da->result[j++]=T('(');
            if (config->extraspace) da->result[j++]=(' ');
          }; }
        else {
          // Comma and optional space between operands.
          da->result[j++]=T(',');
          if (config->extraspace) da->result[j++]=T(' ');
        };
        if (damode & DA_HILITE) {
          memset(da->mask+q,cfill,j-q);
          da->masksize=j; };
        // Operand itself.
        q=j;
        op=da->op+k;
        j+=Tstrcopy(da->result+j,TEXTLEN-j-10,op->text);
        if (damode & DA_HILITE) {
          if (config->hiliteoperands==0)
            ofill=cfill;
          else if (op->features & OP_REGISTER)
            ofill=DRAW_IREG;
          else if (op->features & (OP_FPUREG|OP_MMXREG|OP_3DNOWREG|OP_SSEREG))
            ofill=DRAW_FREG;
          else if (op->features & (OP_SEGREG|OP_CREG|OP_DREG))
            ofill=DRAW_SYSREG;
          else if (op->features & OP_MEMORY) {
            if (op->scale[REG_ESP]!=0 || op->scale[REG_EBP]!=0)
              ofill=DRAW_STKMEM;
            else
              ofill=DRAW_MEM;
            ; }
          else if (op->features & OP_CONST)
            ofill=DRAW_CONST;
          else
            ofill=cfill;
          memset(da->mask+q,ofill,j-q);
          da->masksize=j;
        };
        nout++;
      };
      // All arguments added, close list.
      if (enclose && nout!=0) {
        q=j;
        if (config->extraspace) da->result[j++]=T(' ');
        da->result[j++]=T(')');
        if (damode & DA_HILITE) {
          memset(da->mask+q,cfill,j-q);
          da->masksize=j;
        };
      };
      da->result[j]=T('\0');
    };
  };
  // Calculate total size of command.
  if (da->errors & DAE_CROSS)          // Incomplete command
    n=cmdsize;
  else
    n+=im.modsize+im.dispsize+im.immsize1+im.immsize2;
  da->size=n;
  return n;
};

// Given error and warning lists, returns pointer to the string describing
// relatively most severe error or warning, or NULL if there are no errors or
// warnings.
tchar *Geterrwarnmessage(ulong errors,ulong warnings) {
  tchar *ps;
  if (errors==0 && warnings==0)
    ps=NULL;
  else if (errors & DAE_BADCMD)
    ps=T("Unknown command");
  else if (errors & DAE_CROSS)
    ps=T("Command crosses end of memory block");
  else if (errors & DAE_MEMORY)
    ps=T("Illegal use of register");
  else if (errors & DAE_REGISTER)
    ps=T("Memory address is not allowed");
  else if (errors & DAE_LOCK)
    ps=T("LOCK prefix is not allowed");
  else if (errors & DAE_BADSEG)
    ps=T("Invalid segment register");
  else if (errors & DAE_SAMEPREF)
    ps=T("Two prefixes from the same group");
  else if (errors & DAE_MANYPREF)
    ps=T("More than 4 prefixes");
  else if (errors & DAE_BADCR)
    ps=T("Invalid CR register");
  else if (errors & DAE_INTERN)
    ps=T("Internal OllyDbg error");
  else if (warnings & DAW_DATASIZE)
    ps=T("Superfluous operand size prefix");
  else if (warnings & DAW_ADDRSIZE)
    ps=T("Superfluous address size prefix");
  else if (warnings & DAW_SEGPREFIX)
    ps=T("Superfluous segment override prefix");
  else if (warnings & DAW_REPPREFIX)
    ps=T("Superfluous REPxx prefix");
  else if (warnings & DAW_DEFSEG)
    ps=T("Explicit default segment register");
  else if (warnings & DAW_JMP16)
    ps=T("16-bit jump, call or return");
  else if (warnings & DAW_FARADDR)
    ps=T("Far jump or call");
  else if (warnings & DAW_SEGMOD)
    ps=T("Modification of segment register");
  else if (warnings & DAW_PRIV)
    ps=T("Privileged instruction");
  else if (warnings & DAW_IO)
    ps=T("I/O command");
  else if (warnings & DAW_SHIFT)
    ps=T("Shift out of range");
  else if (warnings & DAW_LOCK)
    ps=T("Command uses (valid) LOCK prefix");
  else if (warnings & DAW_STACK)
    ps=T("Unaligned stack operation");
  else if (warnings & DAW_NOESP)
    ps=T("Suspicious use of stack pointer");
  else if (warnings & DAW_NONCLASS)
    ps=T("Undocumented instruction or encoding");
  else
    ps=NULL;
  return ps;
};

//[bytepointer.com] this callback moved here from original main.c and modified to always return zero
//  and original code left commented; for non-interpreted disassembly, we don't care about any named labels
int Decodeaddress(tchar *s,ulong addr)
{
    return(0);
    /*
    if (addr<1024 || addr>50000) return 0;
    return sprintf(s,"LABEL_%x",addr); 
    */
};
